Package: fabletools
Title: Core Tools for Packages in the 'fable' Framework
Version: 0.7.0
Authors@R: 
    c(person(given = "Mitchell",
             family = "O'Hara-Wild",
             role = c("aut", "cre"),
             email = "mail@mitchelloharawild.com",
             comment = c(ORCID = "0000-0001-6729-7695")),
      person(given = "Rob",
             family = "Hyndman",
             role = "aut"),
      person(given = "Earo",
             family = "Wang",
             role = "aut",
             comment = c(ORCID = "0000-0001-6448-5260")),
      person(given = "Di",
             family = "Cook",
             role = "ctb"),
      person(given = "George",
             family = "Athanasopoulos",
             role = "ctb"),
      person(given = "David",
             family = "Holt",
             role = "ctb"))
Description: Provides tools, helpers and data structures for
    developing models and time series functions for 'fable' and extension
    packages. These tools support a consistent and tidy interface for time
    series modelling and analysis.
License: GPL-3
URL: https://fabletools.tidyverts.org/,
        https://github.com/tidyverts/fabletools
BugReports: https://github.com/tidyverts/fabletools/issues
Depends: R (>= 3.1.3)
Imports: tsibble (>= 0.9.0), tibble (>= 1.4.1), ggplot2 (>= 3.0.0),
        tidyselect, rlang (>= 0.4.5), stats, dplyr (>= 1.0.0), tidyr
        (>= 1.1.0), generics, R6, utils, vctrs (>= 0.2.2),
        distributional (>= 0.3.1), progressr, lifecycle, ggdist,
        scales, cli
Suggests: covr, crayon, fable (>= 0.2.0), future, future.apply, knitr,
        pillar (>= 1.0.1), feasts (>= 0.1.2), rmarkdown, spelling,
        testthat, tsibbledata (>= 0.2.0), lubridate, urca, mvtnorm,
        Matrix, ggtime
VignetteBuilder: knitr
RdMacros: lifecycle
ByteCompile: true
Encoding: UTF-8
Language: en-GB
Config/roxygen2/version: 7.3.3.9000
NeedsCompilation: no
Packaged: 2026-04-29 19:45:30 UTC; mitchell
Author: Mitchell O'Hara-Wild [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-6729-7695>),
  Rob Hyndman [aut],
  Earo Wang [aut] (ORCID: <https://orcid.org/0000-0001-6448-5260>),
  Di Cook [ctb],
  George Athanasopoulos [ctb],
  David Holt [ctb]
Maintainer: Mitchell O'Hara-Wild <mail@mitchelloharawild.com>
Repository: CRAN
Date/Publication: 2026-04-30 05:10:17 UTC
Built: R 4.6.0; ; 2026-05-01 21:50:39 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method("!=",agg_vec)
S3method("$<-",mdl_df)
S3method("==",agg_vec)
S3method("[",dcmp_ts)
S3method("[",fbl_ts)
S3method("[",mdl_df)
S3method("names<-",mdl_df)
S3method(IRF,mdl_df)
S3method(IRF,mdl_ts)
S3method(Ops,lst_mdl)
S3method(Ops,mdl_defn)
S3method(Ops,mdl_lst)
S3method(Ops,mdl_ts)
S3method(accuracy,fbl_ts)
S3method(accuracy,mdl_df)
S3method(accuracy,mdl_ts)
S3method(aggregate_key,tbl_ts)
S3method(as_dable,tbl_df)
S3method(as_dable,tbl_ts)
S3method(as_fable,fbl_ts)
S3method(as_fable,forecast)
S3method(as_fable,grouped_df)
S3method(as_fable,grouped_ts)
S3method(as_fable,tbl_df)
S3method(as_fable,tbl_ts)
S3method(as_mable,data.frame)
S3method(as_tibble,fbl_ts)
S3method(as_tibble,grouped_fbl)
S3method(as_tibble,mdl_df)
S3method(as_tsibble,dcmp_ts)
S3method(as_tsibble,fbl_ts)
S3method(as_tsibble,grouped_fbl)
S3method(augment,lst_mdl)
S3method(augment,mdl_df)
S3method(augment,mdl_lst)
S3method(augment,mdl_ts)
S3method(coef,lst_mdl)
S3method(coef,mdl_df)
S3method(coef,mdl_lst)
S3method(coef,mdl_ts)
S3method(coherent_cmat,default)
S3method(coherent_smat,default)
S3method(coherent_smat,tbl)
S3method(common_periods,default)
S3method(common_periods,interval)
S3method(common_periods,tbl_ts)
S3method(components,mdl_df)
S3method(components,mdl_ts)
S3method(distribution_var,fbl_ts)
S3method(dplyr_col_modify,dcmp_ts)
S3method(dplyr_col_modify,fbl_ts)
S3method(dplyr_col_modify,grouped_fbl)
S3method(dplyr_col_modify,mdl_df)
S3method(dplyr_reconstruct,dcmp_ts)
S3method(dplyr_reconstruct,fbl_ts)
S3method(dplyr_reconstruct,grouped_fbl)
S3method(dplyr_reconstruct,mdl_df)
S3method(dplyr_row_slice,dcmp_ts)
S3method(dplyr_row_slice,fbl_ts)
S3method(dplyr_row_slice,grouped_fbl)
S3method(dplyr_row_slice,mdl_df)
S3method(equation,mdl_df)
S3method(equation,mdl_ts)
S3method(estimate,tbl_ts)
S3method(features,tbl_ts)
S3method(features_all,tbl_ts)
S3method(features_at,tbl_ts)
S3method(features_if,tbl_ts)
S3method(fill_gaps,fbl_ts)
S3method(fitted,"NULL")
S3method(fitted,mdl_df)
S3method(fitted,mdl_ts)
S3method(fitted,model_combination)
S3method(fitted,null_mdl)
S3method(forecast,"NULL")
S3method(forecast,fbl_ts)
S3method(forecast,lst_btmup_mdl)
S3method(forecast,lst_mdl)
S3method(forecast,lst_midout_mdl)
S3method(forecast,lst_mint_mdl)
S3method(forecast,lst_topdwn_mdl)
S3method(forecast,mdl_df)
S3method(forecast,mdl_lst)
S3method(forecast,mdl_ts)
S3method(forecast,model_combination)
S3method(forecast,null_mdl)
S3method(format,agg_vec)
S3method(format,lst_mdl)
S3method(format,mdl_lst)
S3method(format,mdl_ts)
S3method(gather,mdl_df)
S3method(generate,"NULL")
S3method(generate,lst_mdl)
S3method(generate,mdl_df)
S3method(generate,mdl_lst)
S3method(generate,mdl_ts)
S3method(generate,model_combination)
S3method(generate,null_mdl)
S3method(get_frequencies,"NULL")
S3method(get_frequencies,Period)
S3method(get_frequencies,character)
S3method(get_frequencies,numeric)
S3method(glance,lst_mdl)
S3method(glance,mdl_df)
S3method(glance,mdl_lst)
S3method(glance,mdl_ts)
S3method(glance,null_mdl)
S3method(group_by,fbl_ts)
S3method(group_by,grouped_fbl)
S3method(group_data,mdl_df)
S3method(hfitted,mdl_ts)
S3method(hilo,fbl_ts)
S3method(hypothesize,mdl_df)
S3method(hypothesize,mdl_ts)
S3method(interpolate,mdl_df)
S3method(interpolate,mdl_ts)
S3method(invert_transformation,transformation)
S3method(is.na,agg_vec)
S3method(key,mdl_df)
S3method(key_data,mdl_df)
S3method(key_vars,mdl_df)
S3method(mable_vars,mdl_df)
S3method(model,tbl_ts)
S3method(model_sum,decomposition_model)
S3method(model_sum,default)
S3method(model_sum,mdl_ts)
S3method(model_sum,model_combination)
S3method(model_sum,null_mdl)
S3method(outliers,mdl_df)
S3method(outliers,mdl_ts)
S3method(pivot_longer,mdl_df)
S3method(print,transformation)
S3method(rbind,dcmp_ts)
S3method(rbind,fbl_ts)
S3method(recode,agg_vec)
S3method(reconcile,mdl_df)
S3method(refit,"NULL")
S3method(refit,lst_mdl)
S3method(refit,mdl_df)
S3method(refit,mdl_lst)
S3method(refit,mdl_ts)
S3method(refit,null_mdl)
S3method(report,"NULL")
S3method(report,mdl_df)
S3method(report,mdl_ts)
S3method(report,model_combination)
S3method(report,null_mdl)
S3method(residuals,"NULL")
S3method(residuals,mdl_df)
S3method(residuals,mdl_ts)
S3method(residuals,model_combination)
S3method(residuals,null_mdl)
S3method(response,mdl_df)
S3method(response,mdl_ts)
S3method(response_vars,dcmp_ts)
S3method(response_vars,fbl_ts)
S3method(response_vars,mdl_df)
S3method(response_vars,mdl_ts)
S3method(select,fbl_ts)
S3method(select,grouped_fbl)
S3method(select,mdl_df)
S3method(stream,"NULL")
S3method(stream,lst_mdl)
S3method(stream,mdl_df)
S3method(stream,mdl_lst)
S3method(stream,mdl_ts)
S3method(stream,null_mdl)
S3method(summarise,fbl_ts)
S3method(summarise,grouped_fbl)
S3method(tidy,lst_mdl)
S3method(tidy,mdl_df)
S3method(tidy,mdl_lst)
S3method(tidy,mdl_ts)
S3method(tidy,null_mdl)
S3method(transmute,fbl_ts)
S3method(transmute,grouped_fbl)
S3method(transmute,mdl_df)
S3method(ungroup,fbl_ts)
S3method(ungroup,grouped_fbl)
S3method(vec_cast,agg_vec)
S3method(vec_cast,agg_vec.agg_vec)
S3method(vec_cast,agg_vec.character)
S3method(vec_cast,agg_vec.default)
S3method(vec_cast,character.agg_vec)
S3method(vec_cast,character.lst_mdl)
S3method(vec_cast,character.mdl_lst)
S3method(vec_cast,data.frame.dcmp_ts)
S3method(vec_cast,data.frame.fbl_ts)
S3method(vec_cast,data.frame.mdl_df)
S3method(vec_cast,dcmp_ts)
S3method(vec_cast,dcmp_ts.data.frame)
S3method(vec_cast,dcmp_ts.dcmp_ts)
S3method(vec_cast,dcmp_ts.tbl_df)
S3method(vec_cast,fbl_ts)
S3method(vec_cast,fbl_ts.data.frame)
S3method(vec_cast,fbl_ts.fbl_ts)
S3method(vec_cast,fbl_ts.tbl_df)
S3method(vec_cast,mdl_df)
S3method(vec_cast,mdl_df.data.frame)
S3method(vec_cast,mdl_df.mdl_df)
S3method(vec_cast,mdl_df.tbl_df)
S3method(vec_cast,mdl_lst.mdl_lst)
S3method(vec_cast,tbl_df.dcmp_ts)
S3method(vec_cast,tbl_df.fbl_ts)
S3method(vec_cast,tbl_df.mdl_df)
S3method(vec_cast,tbl_ts.dcmp_ts)
S3method(vec_cast,tbl_ts.fbl_ts)
S3method(vec_proxy_compare,agg_vec)
S3method(vec_ptype2,agg_vec)
S3method(vec_ptype2,agg_vec.agg_vec)
S3method(vec_ptype2,agg_vec.character)
S3method(vec_ptype2,agg_vec.default)
S3method(vec_ptype2,character.agg_vec)
S3method(vec_ptype2,data.frame.dcmp_ts)
S3method(vec_ptype2,data.frame.fbl_ts)
S3method(vec_ptype2,data.frame.mdl_df)
S3method(vec_ptype2,dcmp_ts)
S3method(vec_ptype2,dcmp_ts.data.frame)
S3method(vec_ptype2,dcmp_ts.dcmp_ts)
S3method(vec_ptype2,dcmp_ts.tbl_df)
S3method(vec_ptype2,fbl_ts)
S3method(vec_ptype2,fbl_ts.data.frame)
S3method(vec_ptype2,fbl_ts.fbl_ts)
S3method(vec_ptype2,fbl_ts.tbl_df)
S3method(vec_ptype2,mdl_df)
S3method(vec_ptype2,mdl_df.data.frame)
S3method(vec_ptype2,mdl_df.mdl_df)
S3method(vec_ptype2,mdl_df.tbl_df)
S3method(vec_ptype2,mdl_lst.mdl_lst)
S3method(vec_ptype2,tbl_df.dcmp_ts)
S3method(vec_ptype2,tbl_df.fbl_ts)
S3method(vec_ptype2,tbl_df.mdl_df)
S3method(vec_ptype_abbr,agg_vec)
S3method(vec_restore,fbl_ts)
export("%>%")
export(ACF1)
export(CRPS)
export(IRF)
export(MAAPE)
export(MAE)
export(MAPE)
export(MASE)
export(MDA)
export(MDPV)
export(MDV)
export(ME)
export(MPE)
export(MSE)
export(RMSE)
export(RMSSE)
export(accuracy)
export(agg_vec)
export(aggregate_key)
export(as_dable)
export(as_fable)
export(as_mable)
export(as_tsibble)
export(augment)
export(autolayer)
export(autoplot)
export(bias_adjust)
export(bottom_up)
export(box_cox)
export(coherent_cmat)
export(coherent_smat)
export(combination_ensemble)
export(combination_model)
export(combination_weighted)
export(common_periods)
export(common_xregs)
export(components)
export(construct_fc)
export(dable)
export(decomposition_model)
export(directional_accuracy_measures)
export(distribution_accuracy_measures)
export(distribution_var)
export(equation)
export(estimate)
export(fable)
export(feature_set)
export(features)
export(features_all)
export(features_at)
export(features_if)
export(forecast)
export(generate)
export(get_frequencies)
export(glance)
export(hfitted)
export(hilo)
export(hypothesize)
export(interpolate)
export(interval_accuracy_measures)
export(inv_box_cox)
export(invert_transformation)
export(is_aggregated)
export(is_dable)
export(is_fable)
export(is_mable)
export(is_model)
export(is_null_model)
export(mable)
export(mable_vars)
export(middle_out)
export(min_trace)
export(model)
export(model_lhs)
export(model_rhs)
export(model_sum)
export(new_model_class)
export(new_model_definition)
export(new_specials)
export(new_transformation)
export(null_model)
export(outliers)
export(percentile_score)
export(pinball_loss)
export(point_accuracy_measures)
export(quantile_score)
export(reconcile)
export(refit)
export(register_feature)
export(report)
export(response)
export(response_vars)
export(scaled_pinball_loss)
export(scenarios)
export(skill_score)
export(special_xreg)
export(stream)
export(tidy)
export(top_down)
export(unpack_hilo)
export(vars)
export(winkler_score)
import(rlang)
import(tsibble)
import(vctrs)
importFrom(distributional,hilo)
importFrom(dplyr,"%>%")
importFrom(dplyr,anti_join)
importFrom(dplyr,bind_cols)
importFrom(dplyr,bind_rows)
importFrom(dplyr,dplyr_col_modify)
importFrom(dplyr,dplyr_reconstruct)
importFrom(dplyr,dplyr_row_slice)
importFrom(dplyr,filter)
importFrom(dplyr,full_join)
importFrom(dplyr,group_by)
importFrom(dplyr,group_data)
importFrom(dplyr,groups)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,recode)
importFrom(dplyr,rename)
importFrom(dplyr,select)
importFrom(dplyr,semi_join)
importFrom(dplyr,summarise)
importFrom(dplyr,transmute)
importFrom(dplyr,ungroup)
importFrom(dplyr,vars)
importFrom(generics,accuracy)
importFrom(generics,augment)
importFrom(generics,components)
importFrom(generics,equation)
importFrom(generics,forecast)
importFrom(generics,generate)
importFrom(generics,glance)
importFrom(generics,hypothesize)
importFrom(generics,interpolate)
importFrom(generics,refit)
importFrom(generics,tidy)
importFrom(ggplot2,aes)
importFrom(ggplot2,after_stat)
importFrom(ggplot2,autolayer)
importFrom(ggplot2,autoplot)
importFrom(ggplot2,facet_grid)
importFrom(ggplot2,facet_wrap)
importFrom(ggplot2,fortify)
importFrom(ggplot2,geom_line)
importFrom(ggplot2,geom_rect)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,vars)
importFrom(ggplot2,xlab)
importFrom(ggplot2,ylab)
importFrom(lifecycle,deprecate_warn)
importFrom(lifecycle,deprecated)
importFrom(stats,fitted)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,quantile)
importFrom(stats,residuals)
importFrom(stats,terms)
importFrom(stats,var)
importFrom(tibble,new_tibble)
importFrom(tidyr,gather)
importFrom(tidyr,nest)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,spread)
importFrom(tidyr,unnest)
importFrom(tidyselect,all_of)
importFrom(tsibble,as_tsibble)
importFrom(vctrs,vec_ptype2)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# fabletools

<!-- badges: start -->

[![R build
status](https://github.com/tidyverts/fabletools/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/tidyverts/fabletools/actions/workflows/R-CMD-check.yaml)
[![Code test
coverage](https://codecov.io/gh/tidyverts/fabletools/branch/master/graph/badge.svg)](https://app.codecov.io/gh/tidyverts/fabletools?branch=master)
[![lifecycle](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html)
[![CRAN
status](https://www.r-pkg.org/badges/version/fabletools)](https://CRAN.R-project.org/package=fabletools)
<!-- badges: end -->

The R package *fabletools* provides tools for building modelling
packages, with a focus on time series forecasting. This package allows
package developers to extend *fable* with additional models, without
needing to depend on the models supported by *fable*.

## Installation

You could install the **stable** version on
[CRAN](https://cran.r-project.org/package=fabletools):

``` r
install.packages("fabletools")
```

You can install the **development** version from
[GitHub](https://github.com/tidyverts/fabletools):

``` r
# install.packages("remotes")
remotes::install_github("tidyverts/fabletools")
```

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----setup--------------------------------------------------------------------
library(fabletools)

## -----------------------------------------------------------------------------
#' Seasonal mean models
#' 
#' Add the rest of your documentation here.
#' Typically this includes a "Specials" section
#' 
#' @export
SMEAN <- function(formula, ...) {
  # Create a model class which combines the training method, specials, and data checks
  model_smean <- new_model_class("smean",
    # The training method (more on this later)
    train = train_smean,
    # The formula specials (the next section)
    specials = specials_smean,
    # Any checks of the unprocessed data, like gaps, ordered, regular, etc.
    check = function(.data) { 
      if (!tsibble::is_regular(.data)) stop("Data must be regular") 
    }
  )
  
  # Return a model definition which stores the user's model specification
  new_model_definition(model_smean, {{formula}}, ...)
}

## -----------------------------------------------------------------------------
specials_smean <- new_specials(
  season = function(period = NULL) {
    # Your input handling code here.
    get_frequencies(period, self$data, .auto = "smallest")
  },
  xreg = function(...) {
    # This model doesn't support exogenous regressors, time to error.
    stop("Exogenous regressors aren't supported by `SMEAN()`")
  },
  # This model requires `season()`
  # Adding this allows `SMEAN(y)` to automatically include the `season()` special
  .required_specials = "season"
)

## -----------------------------------------------------------------------------
train_smean <- function(.data, specials, ...){
  # Extract a vector of response data
  mv <- tsibble::measured_vars(.data)
  if(length(mv) > 1) stop("SMEAN() is a univariate model.")
  y <- .data[[mv]]
  
  # Pull out inputs from the specials
  if(length(specials$season) > 1) stop("The `season()` special of `SMEAN()` should only be used once.")
  m <- specials$season[[1]]
  
  # Compute the seasonal averages
  season_id <- seq(0, length(y) - 1) %% m
  season_y <- split(y, season_id)
  season_avg <- vapply(season_y, FUN = mean, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE)
  
  # Compute fitted values and residuals
  fit <- season_avg[season_id+1]
  e <- y - fit
  
  # Create S3 model object
  # It should be small, but contain everything needed for methods below
  structure(
    list(
      coef = season_avg,
      n = length(y),
      y_name = mv,
      fitted = fit,
      residuals = e,
      sigma2 = var(e, na.rm = TRUE)
    ),
    class = "model_smean"
  )
}

## ----first-mable--------------------------------------------------------------
fit <- tsibbledata::aus_production %>%
  model(SMEAN(Beer))
fit

## ----echo = FALSE-------------------------------------------------------------
tibble::tribble(
  ~ Method, ~ Value, ~ Description,
  "`model_sum()`", "`character(1L)`", "A short summary of the model to display in the mable",
  "`report()`", "console output", "A detailed summary of the model, similar to `summary()`",
  "`equation()`", "character(1L)", "The mathematical equation for the fitted model",
  "`forecast()`", "distribution", "Produce forecasts from the model",
  "`stream()`", "updated model", "Extend the fit of the model with additional data",
  "`generate()`", "tsibble", "Generate potential reponse values at certain times from the model",
  "`interpolate()`", "tsibble", "Interpolate missing values using the model",
  "`refit()`", "refitted model", "Apply the model to a new dataset",
  "`tidy()`", "tibble of coefficients", "Extract coefficients from the model",
  "`glance()`", "tibble of statistics", "Extract summary statistics from the model",
  "`augment()`", "tibble of data", "Augment a dataset with information from the model",
  "`components()`", "dable of components", "Extract decomposed elements from the model",
  "`fitted()`", "numeric", "Extract fitted values from the model",
  "`residuals()`", "numeric", "Extract residuals from the model"
) %>% 
  knitr::kable()

## -----------------------------------------------------------------------------
#' @importFrom fabletools model_sum
#' @export
model_sum.model_smean <- function(x){
  sprintf("SMEAN[%i]", length(x$coef))
}

fit

## -----------------------------------------------------------------------------
#' @importFrom fabletools report
#' @export
report.model_smean <- function(x){
  m <- length(x$coef)
  
  cat("\n")
  cat(paste("Seasonal period:", m))
  cat("\n\n")
  cat("Seasonal averages:\n")
  
  print.default(
    setNames(x$coef, paste0("s", seq_len(m))),
    print.gap = 2
  )
  cat(paste("\nsigma^2:", round(x$sigma2, 4), "\n"))
}

report(fit)

## -----------------------------------------------------------------------------
#' @importFrom fabletools tidy
#' @export
tidy.model_smean <- function(x){
  tibble::tibble(
    term = paste0("season_", seq_along(x$coef)), 
    estimate = x$coef
  )
}

tidy(fit)

## -----------------------------------------------------------------------------
#' @importFrom fabletools glance
#' @export
glance.model_smean <- function(x){
  tibble::tibble(
    sigma2 = x$sigma2
  )
}

glance(fit)

## -----------------------------------------------------------------------------
#' @importFrom fabletools forecast
#' @export
forecast.model_smean <- function(object, new_data, ...){
  # Extract required parameters
  h <- NROW(new_data)
  n <- object$n
  m <- length(object$coef)
  coef <- object$coef
  
  # Compute forecast variance
  season_id <- seq(0, n - 1) %% m
  season_e <- split(object$residuals, season_id)
  season_sd <- vapply(season_e, FUN = sd, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE, na.rm = TRUE)
  
  # Create forecast distributions
  fc_id <- (seq(0, h-1) + n %% m) %% m + 1
  mu <- coef[fc_id]
  sigma <- season_sd[fc_id]
  distributional::dist_normal(mu, sigma)
}

forecast(fit)

## -----------------------------------------------------------------------------
#' @importFrom fabletools stream
#' @export
stream.model_smean <- function(object, new_data, specials, ...){
  # Extract a vector of response data
  mv <- tsibble::measured_vars(new_data)
  y <- new_data[[mv]]
  
  # Compute the new seasonal averages
  m <- length(object$coef)
  season_id <- (seq(0, length(y) - 1) + object$n %% m) %% m
  season_y <- split(y, season_id)
  season_avg <- vapply(season_y, FUN = mean, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE)
  weight_new <- vapply(season_y, FUN = length, FUN.VALUE = integer(1L),
                       USE.NAMES = FALSE)
  
  # Update coefficients to include new estimates
  weight_orig <- rep(object$n %/% m, m) + c(rep(1, object$n %% m), rep(0, m - object$n %% m))
  new_coef <- (object$coef * weight_orig + season_avg * weight_new) / (weight_orig + weight_new)
  coef_change <- new_coef - object$coef
  
  # Update model
  new_fits <- new_coef[season_id+1]
  new_e <- y - new_fits
  object$coef <- new_coef
  object$fitted <- c(object$fitted + rep_len(coef_change, object$n), new_fits)
  object$residuals <- c(object$residuals - rep_len(coef_change, object$n), new_e)
  object$n <- object$n + length(y)
  object$sigma2 <- var(object$residuals, na.rm = TRUE)
  
  # Return updated model object
  object
}

us_acc_deaths <- as_tsibble(USAccDeaths)
fit_stream <- us_acc_deaths %>% 
  dplyr::slice(1:60) %>% 
  model(SMEAN(value))
report(fit_stream)

# Update the model with new data
us_acc_deaths_new <- us_acc_deaths %>% dplyr::slice(61:72)
fit_stream <- fit_stream %>% 
  stream(us_acc_deaths_new)
report(fit_stream)

# Check that it matches a model of the full data
us_acc_deaths %>% 
  model(SMEAN(value)) %>% 
  report()

## -----------------------------------------------------------------------------
#' @importFrom fabletools fitted
#' @export
fitted.model_smean <- function(object, ...){
  object$fitted
}

fitted(fit)

## -----------------------------------------------------------------------------
#' @importFrom fabletools residuals
#' @export
residuals.model_smean <- function(object, ...){
  object$residuals
}

residuals(fit)

## ----eval=FALSE---------------------------------------------------------------
# #' @importFrom fabletools components
# #' @export
# components.model_smean <- function(object, ...){
#   # Create a tsibble of the components
#   dcmp <- tibble::tibble(
#     !!object$y_name := fitted(object) + residuals(object),
#     season = fitted(object),
#     remainder = residuals(object)
#   )
# 
#   # Describe how the components combine into other columns
#   aliases <- tibble::lst(!!object$y_name := quote(season + remainder))
# 
#   # Define the behaviour of seasonal components
#   # This is used for automatic modelling of seasonal components in `decomposition_model()`
#   # It may also be used for plotting in the future.
#   seasonalities <- list(season = list(period = length(object$coef)))
# 
#   # Return a dable
#   as_dable(
#     dcmp,
#     resp = !!sym(object$y_name), method = model_sum(object),
#     seasons = seasonalities, aliases = aliases
#   )
# }
# 
# components(fit) # Need to store index somewhere. This workflow should improve.

#' Compute Impulse Response Function (IRF)
#'
#' This function calculates the impulse response function (IRF) of a time series model.
#' The IRF describes how a model's variables react to external shocks over time.
#'
#' If `new_data` contains the `.impulse` column, those values will be
#' treated as impulses for the calculated impulse responses.
#'
#' @param x A fitted model object, such as from a VAR or ARIMA model. This model is used to compute the impulse response.
#' @param ... Additional arguments to be passed to lower-level functions.
#' 
#' @details
#' The impulse response function provides insight into the dynamic behaviour of a system in 
#' response to external shocks. It traces the effect of a one-unit change in the impulse 
#' variable on the response variable over a specified number of periods.
#'
#' @export
IRF <- function(x, ...) {
  UseMethod("IRF")
}

#' @export
IRF.mdl_df <- function(x, ...){
  mdl_df_apply(x, IRF, ...)
}

#' @export
IRF.mdl_ts <- function(x, new_data = NULL, h = NULL, ...) {
  if(is.null(new_data)){
    new_data <- make_future_data(x$data, h)
  }
  
  # Compute specials with new_data
  x$model$stage <- "generate"
  x$model$add_data(new_data)
  specials <- tryCatch(parse_model_rhs(x$model),
                       error = function(e){
                         abort(sprintf(
                           "%s
Unable to compute required variables from provided `new_data`.
Does your model require extra variables to produce simulations?", e$message))
                       }, interrupt = function(e) {
                         stop("Terminated by user", call. = FALSE)
                       })
  x$model$remove_data()
  x$model$stage <- NULL
  
  IRF(x$fit, new_data, specials, ...)
}

#' Internal vctrs methods
#'
#' These methods are the extensions that allow mable objects to
#' work with vctrs.
#'
#' @keywords internal
#' @name mable-vctrs
NULL

#' @rdname mable-vctrs
#' @method vec_ptype2 mdl_df
#' @export
vec_ptype2.mdl_df <- function(x, y, ...) {
  UseMethod("vec_ptype2.mdl_df", y)
}

#' @export
vec_ptype2.mdl_df.mdl_df <- function(x, y, ...) {
  mable_ptype2(x, y, ...)
}

#' @export
vec_ptype2.data.frame.mdl_df <- function(x, y, ...) {
  mable_ptype2(y, x, ...)
}

#' @export
vec_ptype2.mdl_df.data.frame <- vec_ptype2.mdl_df.mdl_df

#' @export
vec_ptype2.tbl_df.mdl_df <- vec_ptype2.data.frame.mdl_df

#' @export
vec_ptype2.mdl_df.tbl_df <- vec_ptype2.mdl_df.mdl_df

mable_ptype2 <- function(x, y, ...) {
  key_x <- key_vars(x)
  mdl_x <- mable_vars(x)
  resp_x <- response_vars(x)
  if (is_mable(y)) {
    if (!identical(resp_x, response_vars(y))) {
      abort("Objects with different response variables cannot be combined.")
    }
    key_x <- union(key_x, key_vars(y))
    mdl_x <- union(mdl_x, mable_vars(y))
  }
  out <- df_ptype2(x, y, ...)
  build_mable_meta(out, key_data = group_data(group_by(out, !!!syms(key_x))),
                   model = mdl_x, response = resp_x)
}

#' @rdname mable-vctrs
#' @method vec_cast mdl_df
#' @export
vec_cast.mdl_df <- function(x, to, ...) {
  UseMethod("vec_cast.mdl_df")
}

#' @export
vec_cast.mdl_df.mdl_df <- function(x, to, ...) {
  is_identical <- identical(x, to)
  tbl <- tib_cast(x, to, ...)
  build_mable(tbl,
              key = !!key_vars(to), 
              key_data = if (is_identical) key_data(x) else NULL,
              model = mable_vars(to))
}

#' @export
vec_cast.mdl_df.tbl_df <- function(x, to, ...) {
  tbl <- tib_cast(x, to, ...)
  build_mable(tbl,
              key = !!key_vars(to), 
              key_data = NULL,
              model = mable_vars(to))
}

#' @export
vec_cast.mdl_df.data.frame <- vec_cast.mdl_df.tbl_df

#' @export
vec_cast.tbl_df.mdl_df <- function(x, to, ...) {
  tib_cast(x, to, ...)
}

#' @export
vec_cast.data.frame.mdl_df <- function(x, to, ...) {
  df_cast(x, to, ...)
}

#' Extract fitted values from models
#' 
#' Extracts the fitted values from each of the models in a mable. A tsibble will
#' be returned containing these fitted values. Fitted values will be 
#' automatically back-transformed if a transformation was specified.
#' 
#' @aliases  hfitted
#' 
#' @param object A mable or time series model.
#' @param ... Other arguments passed to the model method for `fitted()`
#' 
#' @importFrom stats fitted
#' @export
fitted.mdl_df <- function(object, ...){
  mbl_vars <- mable_vars(object)
  kv <- key_vars(object)
  object <- mutate(as_tibble(object), 
              dplyr::across(all_of(mbl_vars), function(x) lapply(x, fitted, ...)))
  object <- pivot_longer(object, all_of(mbl_vars), names_to = ".model", values_to = ".fitted")
  unnest_tsbl(object, ".fitted", parent_key = c(kv, ".model"))
}

#' @rdname fitted.mdl_df
#' 
#' @param h The number of steps ahead that these fitted values are computed from.
#' 
#' @export
fitted.mdl_ts <- function(object, h = 1, ...){
  bt <- map(object$transformation, invert_transformation)
  
  fits <- if(h==1) fitted(object$fit, ...) else hfitted(object, h = h, ...)
  if(h == 1){
    fits <- as.matrix(fits)
    # Backtransformation is required for fitted, but forecast() handles it already.
    fits <- map2(bt, split(fits, col(fits)), function(bt, fit) bt(fit))
  }
  
  nm <- if(length(fits) == 1) ".fitted" else map_chr(object$response, expr_name)
  
  out <- object$data[index_var(object$data)]
  out[nm] <- fits
  out
}

#' @export
hfitted <- function(object, ...) {
  UseMethod("hfitted")
}

#' @export
hfitted.mdl_ts <- function(object, h, ...) {
  fn <- tryCatch(utils::getS3method("hfitted", class(object[["fit"]])),
                 error = function(e) NULL)

  if(is.null(fn)) {
    # No direct hfitted method exists, so we will refit the model repeatedly to get the fitted values.
    dt <- object$data
    resp <- response_vars(object)
    n <- nrow(dt)
    fits <- rep(NA_real_, n)

    fn <- tryCatch(utils::getS3method("refit", mdl_cls <- class(object[["fit"]])), error = function(e) NULL)
    if (is.null(fn)) {
      cli::cli_warn(
        "No method for hfitted() or refit() for class {.cls {mdl_cls}}. Returning NA.",
        class = "fabletools_no_hfitted_method"
      )
    }
    
    # Undo transformations
    bt <- lapply(object$transformation, invert_transformation)
    mv <- match(measured_vars(dt), names(dt))
    dt[mv] <- mapply(calc, bt, dt[measured_vars(dt)], SIMPLIFY = FALSE)
    names(dt)[mv] <- resp
    
    for (i in seq_len(n-h)) {
      mdl <- tryCatch(refit(object, vec_slice(dt, seq_len(i))),
                      error = function(e) NULL)
      if(is.null(mdl)) next
      fits[i + h] <- mean(forecast(mdl, h = h, point_forecast = NULL)[[resp]][h])
    }
    fits <- list(fits)
  } else {
    # Direct hfitted method is 
    fits <- as.matrix(fn(object[["fit"]], h=h, ...))
    # Backtransform fits from model method
    bt <- map(object$transformation, invert_transformation)
    fits <- map2(bt, split(fits, col(fits)), function(bt, fit) bt(fit))
  }
  fits
}

# nocov start
.onLoad <- function(...) {
  register_s3_method("pillar", "type_sum", "mdl_ts")
  register_s3_method("pillar", "type_sum", "mdl_lst")
  register_s3_method("pillar", "type_sum", "fbl_ts")
  
  register_s3_method("pillar", "pillar_shaft", "agg_vec")
  
  register_s3_method("tibble", "tbl_sum", "dcmp_ts")
  register_s3_method("tibble", "tbl_sum", "mdl_df")
  register_s3_method("tibble", "tbl_sum", "fbl_ts")
  
  register_s3_method("ggplot2", "scale_type", "agg_vec")
  
  op <- options()
  op.fable <- list(
    fable.show_progress = TRUE
  )
  toset <- !(names(op.fable) %in% names(op))
  if (any(toset)) options(op.fable[toset])
  
  invisible()
}

.onAttach <- function(...) {
  # fabletools -> ggtime method migration
  ggtime_version <- as.character(suppressWarnings(
    utils::packageDescription("ggtime", fields = "Version")
  ))
  if (!is.na(ggtime_version)) {
    ggtime_ns <- getNamespace("ggtime")
    register_s3_method("ggplot2", "autoplot", "tbl_ts", utils::getS3method("autoplot", "tbl_ts", envir = ggtime_ns))
    register_s3_method("ggplot2", "autolayer", "tbl_ts", utils::getS3method("autolayer", "tbl_ts", envir = ggtime_ns))
    register_s3_method("ggplot2", "autoplot", "dcmp_ts", utils::getS3method("autoplot", "dcmp_ts", envir = ggtime_ns))
    if (package_version(ggtime_version) >= "0.2.0") {
      register_s3_method("ggplot2", "autoplot", "fbl_ts", utils::getS3method("autoplot", "fbl_ts", envir = ggtime_ns))
      register_s3_method("ggplot2", "autolayer", "fbl_ts", utils::getS3method("autolayer", "fbl_ts", envir = ggtime_ns))
      register_s3_method("ggplot2", "fortify", "fbl_ts", utils::getS3method("fortify", "fbl_ts", envir = ggtime_ns))
    } else {
      register_s3_method("ggplot2", "autoplot", "fbl_ts", autoplot.fbl_ts)
      register_s3_method("ggplot2", "autolayer", "fbl_ts", autolayer.fbl_ts)
      register_s3_method("ggplot2", "fortify", "fbl_ts", fortify.fbl_ts)
    }
  } else {
    register_s3_method("ggplot2", "autoplot", "fbl_ts", autoplot.fbl_ts)
    register_s3_method("ggplot2", "autoplot", "tbl_ts", autoplot.tbl_ts)
    register_s3_method("ggplot2", "autoplot", "dcmp_ts", autoplot.dcmp_ts)
    register_s3_method("ggplot2", "autolayer", "fbl_ts", autolayer.fbl_ts)
    register_s3_method("ggplot2", "autolayer", "tbl_ts", autolayer.fbl_ts)
    register_s3_method("ggplot2", "fortify", "fbl_ts", fortify.fbl_ts)
  }
}

register_s3_method <- function(pkg, generic, class, fun = NULL) {
  stopifnot(is.character(pkg), length(pkg) == 1)
  stopifnot(is.character(generic), length(generic) == 1)
  stopifnot(is.character(class), length(class) == 1)
  
  if (is.null(fun)) {
    fun <- get(paste0(generic, ".", class), envir = parent.frame())
  } else {
    stopifnot(is.function(fun))
  }
  
  if (pkg %in% loadedNamespaces()) {
    registerS3method(generic, class, fun, envir = asNamespace(pkg))
  }
  
  # Always register hook in case package is later unloaded & reloaded
  setHook(
    packageEvent(pkg, "onLoad"),
    function(...) {
      registerS3method(generic, class, fun, envir = asNamespace(pkg))
    }
  )
}
# nocov end

#' Extract residuals values from models
#' 
#' Extracts the residuals from each of the models in a mable. A tsibble will
#' be returned containing these residuals.
#' 
#' @param object A mable or time series model.
#' @param ... Other arguments passed to the model method for `residuals()`
#' 
#' @importFrom stats residuals
#' @export
residuals.mdl_df <- function(object, ...){
  mbl_vars <- mable_vars(object)
  kv <- key_vars(object)
  object <- mutate(as_tibble(object), 
              dplyr::across(all_of(mbl_vars), function(x) lapply(x, residuals, ...)))
  object <- pivot_longer(object, all_of(mbl_vars), names_to = ".model", values_to = ".resid")
  unnest_tsbl(object, ".resid", parent_key = c(kv, ".model"))
}

#' @param type The type of residuals to compute. If `type="response"`, residuals on the back-transformed data will be computed.
#' @rdname residuals.mdl_df
#' @export
residuals.mdl_ts <- function(object, type = "innovation", ...){
  if(type == "response"){
    .resid <- response(object)
    .fits <- fitted(object)
    .resid <- as.matrix(.resid[measured_vars(.resid)]) - as.matrix(.fits[measured_vars(.fits)])
  }
  else{
    .resid <- residuals(object$fit, type = type, ...)
    if(is.null(.resid)){
      if(type == "innovation") {
        .resid <- response(object)
        .resid <- map2(object$transformation, .resid[measured_vars(.resid)], calc)
        .fits <- fitted(object)
        .fits <- map2(object$transformation, .fits[measured_vars(.fits)], calc)
        .resid <- do.call(cbind, .resid) - do.call(cbind, as.matrix(.fits))
      } else {
        warn(sprintf(
'Residuals of type `%s` are not supported for %s models.
Defaulting to `type="response"`', type, model_sum(object)))
        return(residuals(object, type = "response", ...))
      }
    }
  }
  .resid <- as.matrix(.resid)
  
  .resid <- split(.resid, col(.resid))
  nm <- if(length(.resid) == 1) ".resid" else map_chr(object$response, expr_name)
  
  out <- object$data[index_var(object$data)]
  out[nm] <- .resid
  out
}

train_combination <- function(.data, specials, ..., cmbn_fn, cmbn_args){
  mdls <- dots_list(...)
  
  # Estimate model definitions
  mdl_def <- map_lgl(mdls, inherits, "mdl_defn")
  mdls[mdl_def] <- mdls[mdl_def] %>% 
    map(function(x) estimate(self$data, x))
  
  do.call(cmbn_fn, c(mdls, cmbn_args))[["fit"]]
}

#' Combination modelling
#' 
#' Combines multiple model definitions (passed via `...`) to produce a model
#' combination definition using some combination function (`cmbn_fn`). Currently
#' distributional forecasts are only supported for models producing normally
#' distributed forecasts.
#' 
#' A combination model can also be produced using mathematical operations.
#'
#' @param ... Model definitions used in the combination.
#' @param cmbn_fn A function used to produce the combination.
#' @param cmbn_args Additional arguments passed to `cmbn_fn`.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibble)
#' library(tsibbledata)
#' 
#' # cmbn1 and cmbn2 are equivalent and equally weighted.
#' aus_production %>%
#'   model(
#'     cmbn1 = combination_model(SNAIVE(Beer), TSLM(Beer ~ trend() + season())),
#'     cmbn2 = (SNAIVE(Beer) + TSLM(Beer ~ trend() + season()))/2
#'   )
#' 
#' # An inverse variance weighted ensemble.
#' aus_production %>%
#'   model(
#'     cmbn1 = combination_model(
#'       SNAIVE(Beer), TSLM(Beer ~ trend() + season()), 
#'       cmbn_args = list(weights = "inv_var")
#'     )
#'   )
#' @export
combination_model <- function(..., cmbn_fn = combination_ensemble,
                              cmbn_args = list()){
  mdls <- dots_list(...)
  if(!any(map_lgl(mdls, inherits, "mdl_defn"))){
    abort("`combination_model()` must contain at least one valid model definition.")
  }
  
  # Guess the response variable without transformations
  resp <- Reduce(intersect, lapply(Filter(function(x) inherits(x, "mdl_defn"), mdls), function(x) all.vars(model_lhs(x))))
  if(length(resp) == 0) abort("`combination_model()` must use component models with the same response variable.")
  
  cmbn_model <- new_model_class("cmbn_mdl", train = train_combination, 
                                specials = new_specials(xreg = function(...) NULL))
  fml <- mdls[[1]]$formula
  fml <- quo_set_expr(fml, sym(resp[[1]]))
  new_model_definition(cmbn_model, !!fml, ..., 
                       cmbn_fn = cmbn_fn, cmbn_args = cmbn_args)
}

#' Ensemble combination
#' 
#' @param ... Estimated models used in the ensemble.
#' @param weights The method used to weight each model in the ensemble.
#' 
#' @seealso [`combination_weighted()`]
#' 
#' @export
combination_ensemble <- function(..., weights = c("equal", "inv_var")){
  mdls <- dots_list(...)
  
  if(all(map_lgl(mdls, inherits, "mdl_defn"))){
    return(combination_model(..., cmbn_args = list(weights = weights)))
  }
  
  weights <- match.arg(weights)
  
  if(weights == "equal"){
    out <- reduce(mdls, `+`)/length(mdls)
  }
  else if(weights == "inv_var") {
    out <- if(is_model(mdls[[1]])) list(mdls) else transpose(mdls)
    out <- map(out, function(x){
      inv_var <- map_dbl(x, function(x) 1/var(residuals(x)$.resid, na.rm = TRUE))
      weights <- inv_var/sum(inv_var)
      reduce(map2(weights, x, `*`), `+`)
    })
    if(is_model(mdls[[1]])) out <- out[[1]]
  }
  
  if(is_model(out)){
    out$response <- mdls[[1]]$response
  }
  else{
    out <- structure(
      map(out, `[[<-`, "response", mdls[[1]][[1]]$response), 
      class = c("mdl_lst", "lst_mdl", "list"))
  }
  out
}

#' Weighted combination
#' 
#' @param ... Estimated models used in the ensemble.
#' @param weights The numeric weights applied to each model in `...`
#' 
#' @seealso [`combination_ensemble()`]
#' 
#' @export
combination_weighted <- function(..., weights = NULL){
  mdls <- dots_list(...)
  resp_var <- mdls[[1]]$response
  
  if(all(map_lgl(mdls, inherits, "mdl_defn"))){
    return(combination_model(..., cmbn_fn = combination_weighted,
                             cmbn_args = list(weights = weights)))
  }
  vctrs::vec_assert(weights, numeric(), length(mdls))
  
  # Standardise weights to sum 1
  weights <- weights/sum(weights)
  
  mdls <- map2(mdls, weights, `*`)
  out <- reduce(mdls, `+`)
  out$response <- resp_var
  out
}

new_model_combination <- function(x, combination){
  mdls <- map_lgl(x, is_model)
  
  mdls_response <- map(x, function(x) if(is_model(x)) x[["response"]] else x)
  comb_response <- map(transpose(mdls_response),
                       function(x) eval(expr(substitute(!!combination, x))))
  
  if(length(comb_response) > 1) abort("Combining multivariate models is not yet supported.")
  
  # Compute new response data
  resp <- map2(x, mdls, function(x, is_mdl) if(is_mdl) response(x)[[".response"]] else x)
  resp <- eval_tidy(combination, resp)
  
  if(any(!mdls)){
    # Simplify the response with a f(numeric, model)
    op <- deparse(combination[[1]])
    if(op == "*" || (op == "/" && mdls[1])){
      num <- x[[which(!mdls)]]
      num <- switch(op, `*` = 1/num, `/` = num)
      if(num%%1 == 0){
        cmp <- all.names(x[[which(mdls)]][["response"]][[1]])
        if(length(unique(cmp)) == 2 && sum(cmp == "+") + 1 == num){
          comb_response <- syms(cmp[cmp != "+"][1])
        }
      }
    }
  } else {
    # Simplify the response with f(model, model)
    
    # Assume both models come from the same root variable, find the data of the
    # mable's response variable.
    root_var <- traverse(
      x$e1, 
      .f = function(x, y) x[[1]][c("data", "response")],
      .g = function(mdl) Filter(function(x) inherits(x, "mdl_ts"), mdl$fit),
      base = function(mdl) !inherits(mdl$fit, "model_combination")
    )
    root_resp <- root_var[["response"]]
    root_resp_var <- rlang::as_label(root_resp[[1]])
    if(isTRUE(all.equal(resp, root_var$data[[root_resp_var]]))) {
      comb_response <- root_resp
    }
  }
  
  new_model(
    structure(x, combination = combination, class = c("model_combination")),
    model = x[[which(mdls)[1]]][["model"]],
    data = transmute(x[[which(mdls)[1]]][["data"]], !!expr_name(comb_response[[1]]) := resp),
    response = comb_response,
    transformation = list(new_transformation(identity, identity))
  )
}

#' @export
model_sum.model_combination <- function(x){
  "COMBINATION"
}

#' @export
Ops.mdl_defn <- function(e1, e2){
  e1_expr <- enexpr(e1)
  e2_expr <- enexpr(e2)
  ok <- switch(.Generic, `+` = , `-` = , `*` = , `/` = TRUE, FALSE)
  if (!ok) {
    warn(sprintf("`%s` not meaningful for model definitions", .Generic))
    return(rep.int(NA, max(length(e1), if (!missing(e2)) length(e2))))
  }
  if(.Generic == "/" && is_model(e2)){
    warn(sprintf("Cannot divide by a model definition"))
    return(rep.int(NA, max(length(e1), if (!missing(e2)) length(e2))))
  }
  
  combination_model(e1, e2, cmbn_fn = .Generic)
}

#' @export
Ops.mdl_ts <- function(e1, e2){
  e1_expr <- enexpr(e1)
  e2_expr <- enexpr(e2)
  ok <- switch(.Generic, `+` = , `-` = , `*` = , `/` = TRUE, FALSE)
  if (!ok) {
    warn(sprintf("`%s` not meaningful for models", .Generic))
    return(rep.int(NA, max(length(e1), if (!missing(e2)) length(e2))))
  }
  if(.Generic == "/" && is_model(e2)){
    warn(sprintf("Cannot divide by a model"))
    return(rep.int(NA, max(length(e1), if (!missing(e2)) length(e2))))
  }
  if(.Generic %in% c("-", "+") && missing(e2)){
    e2 <- e1
    .Generic <- "*"
    e1 <- 1
  }
  if(.Generic == "-"){
    .Generic <- "+"
    e2 <- -e2
  }
  else if(.Generic == "/"){
    .Generic <- "*"
    e2 <- 1/e2
  }
  
  # e_len <- c(length(e1), length(e2))
  # if(max(e_len) %% min(e_len) != 0){
  #   warn("longer object length is not a multiple of shorter object length")
  # }
  # 
  # if(e_len[[1]] != e_len[[2]]){
  #   if(which.min(e_len) == 1){
  #     is_mdl <- is_model(e1)
  #     cls <- class(e1)
  #     e1 <- rep_len(e1, e_len[[2]])
  #     if(is_mdl){
  #       e1 <- structure(e1, class = cls)
  #     }
  #   }
  #   else{
  #     is_mdl <- is_model(e2)
  #     cls <- class(e2)
  #     e2 <- rep_len(e2, e_len[[1]])
  #     if(is_mdl){
  #       e2 <- structure(e2, class = cls)
  #     }
  #   }
  # }
  
  if(is_model(e1) && is_model(e2)){
    if(.Generic == "*"){
      warn(sprintf("Multipling models is not supported"))
      return(rep.int(NA, length(e1)))
    }
  }
  
  new_model_combination(
    list(e1 = e1, e2 = e2),
    combination = call2(.Generic, sym("e1"), sym("e2"))
  )
}

#' @export
Ops.mdl_lst <- function(e1, e2){
  list_of_models(map2(e1, e2, .Generic))
}
#' @export
Ops.lst_mdl <- deprecate_lst_mdl(Ops.mdl_lst)

#' @importFrom stats var
#' @export
forecast.model_combination <- function(object, new_data, specials, ...){
  mdls <- map_lgl(object, is_model)
  expr <- attr(object, "combination")
  # Compute residual covariance to adjust the forecast variance
  # Assumes correlation across h is identical
  
  fbl <- object
  fbl[mdls] <- map(fbl[mdls], forecast, new_data = new_data, ...)
  fbl[mdls] <- map(fbl[mdls], function(x) x[[distribution_var(x)]])
  
  is_normal <- map_lgl(fbl[mdls], function(x) all(dist_types(x) == "dist_normal"))
  if(all(is_normal)){
    .dist <- eval_tidy(expr, fbl)
    
    # Adjust for covariance
    # var(x) + var(y) + 2*cov(x,y)
    if(all(mdls)) {
      fc_cov <- var(
        cbind(
          residuals(object[[1]], type = "response")[[".resid"]],
          residuals(object[[2]], type = "response")[[".resid"]]
        ),
        na.rm = TRUE
      )
      fc_sd <- fbl %>% 
        map(function(x) sqrt(distributional::variance(x))) %>% 
        transpose_dbl()
      fc_cov <- suppressWarnings(stats::cov2cor(fc_cov))
      fc_cov[!is.finite(fc_cov)] <- 0 # In case of perfect forecasts
      fc_cov <- map_dbl(fc_sd, function(sigma) (diag(sigma)%*%fc_cov%*%t(diag(sigma)))[1,2])
      .dist <- distributional::dist_normal(mean(.dist), sqrt(distributional::variance(.dist) + 2*fc_cov))
    }
  } else {
    .dist <- distributional::dist_degenerate(eval_tidy(expr, map(fbl, mean)))
  }
  
  .dist
}

#' @export
generate.model_combination <- function(x, new_data, specials, bootstrap = FALSE, ...){
  if(".innov" %in% names(new_data)){
    # Assume bootstrapped paths are requested (this needs future work)
    bootstrap <- TRUE
    new_data[[".innov"]] <- NULL
    # abort("Providing innovations for simulating combination models is not supported.")
  }
  
  mdls <- map_lgl(x, is_model)
  expr <- attr(x, "combination")
  x[mdls] <- map(x[mdls], generate, new_data, bootstrap = bootstrap, ...)
  out <- x[[which(mdls)[1]]]
  sims <- map(x, function(x) if(is_tsibble(x)) x[[".sim"]] else x)
  out[[".sim"]] <- eval_tidy(expr, sims)
  out
}

#' @export
fitted.model_combination <- function(object, ...){
  mdls <- map_lgl(object, is_model)
  expr <- attr(object, "combination")
  object[mdls] <- map(object[mdls], fitted, ...)
  fits <- map(object, function(x) if(is_tsibble(x)) x[[".fitted"]] else x)
  eval_tidy(expr, fits)
}

#' @export
residuals.model_combination <- function(object, type = "response", ...) {
  mdls <- map_lgl(object, is_model)
  expr <- attr(object, "combination")
  # Ignore type and always give response residuals here
  # if(type != "response") {
  #   stop("Only response residuals are supported for combination models.")
  # }
  object[mdls] <- map(object[mdls], residuals, type = "response", ...)
  res <- map(object, function(x) if(is_tsibble(x)) x[[".resid"]] else x)
  eval_tidy(expr, res)
}

context("test-combination")

test_that("Combination modelling", {
  skip_if_not_installed("fable")
  
  mbl_cmbn <- mbl %>% 
    transmute(combination = (ets + ets)/2)
  
  expect_equal(
    select(augment(mbl_cmbn), -.model, -.innov),
    select(augment(mbl), -.model, -.innov)
  )
  
  expect_equivalent(
    forecast(mbl_cmbn, h = 12)[,-1],
    fbl[,-1]
  )
  
  mbl_cmbn <- us_deaths_tr %>% 
    model(
      a = fable::SNAIVE(value),
      b = combination_model(fable::SNAIVE(value), fable::SNAIVE(value))
    )
  
  fbl_cmbn <- forecast(mbl_cmbn)
  
  expect_equivalent(
    fbl_cmbn[1:24, -1],
    fbl_cmbn[25:48, -1]
  )
  
  mbl_cmbn <- us_deaths_tr %>% 
    model(
      snaive = fable::SNAIVE(value),
      rw = fable::RW(value ~ drift()),
      cmbn = (fable::SNAIVE(value) + fable::RW(value ~ drift()))/2
    ) %>% 
    mutate(combination_ensemble(snaive, rw, weights = "inv_var"))
  fbl_snaive <- forecast(select(mbl_cmbn, 1))
  fbl_rw <- forecast(select(mbl_cmbn, 2))
  fbl_cmbn <- forecast(select(mbl_cmbn, 3))
  fbl_wt_cmbn <- forecast(select(mbl_cmbn, 4))
  expect_equal(
    mean(fbl_cmbn$value),
    mean((fbl_snaive$value + fbl_rw$value)/2)
  )
  expect_failure(
    expect_equal(
      mean(fbl_wt_cmbn$value),
      mean(fbl_cmbn$value)
    )
  )
})

#' Extend a fitted model with new data
#'
#' Extend the length of data used to fit a model and update the parameters to 
#' suit this new data.
#'
#' @param object An object (such as a model) which can be extended with additional data.
#' @param ... Additional arguments passed on to stream methods.
#'
#' @rdname stream
#' @export
stream <- function(object, ...){
  UseMethod("stream")
}

#' @param new_data A dataset of the same structure as was used to fit the model.
#' 
#' @rdname stream
#' @export
stream.mdl_df <- function(object, new_data, ...){
  mdls <- mable_vars(object)
  object <- bind_new_data(object, new_data)
  new_data <- object[["new_data"]]
  object[["new_data"]] <- NULL
  dplyr::mutate_at(object, vars(!!!mdls), stream, new_data, ...)
}

#' @export
stream.mdl_lst <- function(object, new_data, ...){
  add_class(map2(object, new_data, stream, ...), class(object))
}
#' @export
stream.lst_mdl <- deprecate_lst_mdl(stream.mdl_lst)

#' @export
stream.mdl_ts <- function(object, new_data, ...){
  # Compute specials with new_data
  object$model$add_data(new_data)
  specials <- parse_model_rhs(object$model)
  object$model$remove_data()
  
  resp <- map2(object$response, object$transformation, 
       function(y, t){
         eval_tidy(expr(t(!!y)), new_data)
       }
  )
  new_data <- new_data[index_var(new_data)]
  new_data[measured_vars(object$data)] <- resp
  
  object$fit <- stream(object[["fit"]], new_data, specials = specials, ...)
  object$data <- bind_rows(object$data, select(new_data, !!!syms(colnames(object$data))))
  object
}

#' Create a new mable
#' 
#' A mable (model table) data class (`mdl_df`) is a tibble-like data structure 
#' for applying multiple models to a dataset. Each row of the mable refers to a
#' different time series from the data (identified by the key columns). A mable
#' must contain at least one column of time series models (`mdl_ts`), where the
#' list column itself (`mdl_lst`) describes how these models are related.
#' 
#' @inheritParams tsibble::tsibble
#' 
#' @param key Structural variable(s) that identify each model.
#' @param model Identifiers for the columns containing model(s).
#'
#' @export
mable <- function(..., key = NULL, model = NULL){
  as_mable(tibble(...), key = !!enquo(key), model = !!enquo(model))
}

#' Is the object a mable
#' 
#' @param x An object.
#' 
#' @export
is_mable <- function(x){
  inherits(x, "mdl_df")
}

#' Coerce a dataset to a mable
#' 
#' @param x A dataset containing a list model column.
#' @param ... Additional arguments passed to other methods.
#' 
#' @rdname as_mable
#' @export
as_mable <- function(x, ...){
  UseMethod("as_mable")
}

#' @rdname as_mable
#' 
#' @inheritParams mable
#' 
#' @export
as_mable.data.frame <- function(x, key = NULL, model = NULL, ...){
  build_mable(x, key = !!enquo(key), model = model)
}

# TODO - allow empty mdl_df objects to be constructed with a given response variable
# which is used to check if the models use that response variable. The default
# response variable would then simply be that of the first model (violating the
# check if models have different response variables).
build_mable <- function (x, key = NULL, key_data = NULL, model = NULL) {
  model <- names(tidyselect::eval_select(all_of(model), data = x))
  
  if(length(resp_var <- unique(map(x[model], function(mdl) response_vars(mdl[[1]])))) > 1){
    abort("A mable can only contain models with the same response variable(s).")
  }
  if(length(resp_var) == 0) {
    abort("A mable must contain at least one model.")
  }
  
  if (!is_null(key_data)){
    assert_key_data(key_data)
    key <- utils::head(names(key_data), -1L)
  }
  else {
    key <- names(tidyselect::eval_select(enquo(key), data = x))
    key_data <- group_data(group_by(x, !!!syms(key)))
  }
  
  if(any(map_int(key_data[[length(key_data)]], length) > 1)){
    abort("The result is not a valid mable. The key variables must uniquely identify each row.")
  }
  
  build_mable_meta(x, key_data, model, response = resp_var[[1]])
}

build_mable_meta <- function(x, key_data, model, response){
  tibble::new_tibble(x, key = key_data, model = model, response = response,
                     nrow = NROW(x), class = "mdl_df", subclass = "mdl_df") 
}

#' @export
as_tibble.mdl_df <- function(x, ...){
  attr(x, "key") <- attr(x, "model") <- NULL
  class(x) <- c("tbl_df", "tbl", "data.frame")
  as_tibble(x, ...)
}

tbl_sum.mdl_df <- function(x){
  out <- c(`A mable` = paste(map_chr(dim(x), big_mark), collapse = " x "))
  
  if(!is_empty(key(x))){
    out <- c(out, c("Key" = sprintf("%s [%s]",
                                    paste0(key_vars(x), collapse = ", "),
                                    map_chr(n_keys(x), big_mark))))
  }
  
  out
}

restore_mable <- function(data, template){
  data <- as_tibble(data)
  data_cols <- names(data)
  
  # key_vars <- setdiff(key_vars(template), data_cols)
  # key_data <- select(key_data(template), key_vars)
  # if (vec_size(key_data) == 1) {
  #   template <- remove_key(template, setdiff(key_vars(template), key_vars))
  # }
  
  model_vars <- intersect(mable_vars(template), data_cols)
  # Variables to keep
  mbl_vars <- setdiff(key_vars(template), data_cols)
  res <- bind_cols(template[mbl_vars], data)
  
  build_mable(res, key = !!key_vars(template), model = model_vars)
}

#' @export
gather.mdl_df <- function(data, key = "key", value = "value", ..., na.rm = FALSE,
                          convert = FALSE, factor_key = FALSE){
  value <- enexpr(value)
  tbl <- gather(as_tibble(data), key = !!key, value = !!value, 
                ..., na.rm = na.rm, convert = convert, factor_key = factor_key)
  mdls <- names(which(map_lgl(tbl, inherits, "mdl_lst")))
  kv <- c(key_vars(data), key)
  build_mable(tbl, key = !!kv, model = mdls)
}

# Adapted from tsibble:::pivot_longer.tbl_ts
#' @importFrom tidyr pivot_longer
#' @export
pivot_longer.mdl_df <- function (data, ..., names_to = "name") {
  if (!has_length(names_to)) {
    abort("`pivot_longer(<mable>)` can't accept zero-length `names_to`.")
  }
  if (".value" %in% names_to) {
    abort("`pivot_longer(<mable>)` can't accept the special \".value\" in `names_to`.")
  }
  new_key <- c(key_vars(data), names_to)
  tbl <- tidyr::pivot_longer(as_tibble(data), ..., names_to = names_to)
  build_mable(tbl, key = !!new_key,
              model = which(vapply(tbl, inherits, logical(1L), "mdl_lst")))
}

#' @export
select.mdl_df <- function (.data, ...){
  res <- NextMethod()
  
  loc <- tidyselect::eval_select(expr(c(...)), .data)
  
  kv <- key_vars(.data)
  rm_kv <- intersect(kv, names(.data)[-loc])
  key_data <- vec_unique(select(key_data(.data), all_of(rm_kv)))
  
  # Drop/keep redundant/necessary key variables
  if(vec_size(key_data) == 1) {
    .data%@%"key" <- key_data(.data)[c(setdiff(kv, rm_kv), ".rows")]
  } else {
    res <- bind_cols(.data[rm_kv], res)
  }
  names(.data)[loc] <- names(loc)
  
  restore_mable(res, .data)
}
#' @export
transmute.mdl_df <- function (.data, ...){
  nm <- names(enquos(..., .named = TRUE))
  res <- mutate(.data, ...)
  select(res, all_of(nm))
}

#' @export
`$<-.mdl_df` <- function (x, name, value) {
  tbl <- NextMethod()
  mdls <- names(which(map_lgl(tbl, inherits, "mdl_lst")))
  as_mable(tbl, key = key_vars(x), model = mdls)
}

#' @export
`names<-.mdl_df` <- function(x, value) {
  nm <- colnames(x)
  key_pos <- match(key_vars(x), nm)
  kd <- key_data(x)
  colnames(kd) <- c(value[key_pos], ".rows")
  mdl_pos <- match(mable_vars(x), nm)
  res <- NextMethod()
  build_mable_meta(res, key_data = kd, model = value[mdl_pos], 
                   response = response_vars(x))
}

#' @export
`[.mdl_df` <- function (x, i, j, drop = FALSE){
  out <- as_tibble(NextMethod())
  cn <- names(out)
  new_kv <- intersect(old_kv <- key_vars(x), cn)
  mv <- intersect(mable_vars(x), cn)
  
  # If keys have changed, return a tibble if mable is invalid
  if(!identical(old_kv, new_kv)){
    return(out)
    key_data <- group_data(group_by(x, !!!syms(key)))
    if(any(lengths(key_data[[length(key_data)]]) > 1))
      return(out)
    else
      return(build_mable(out, key_data = key_data, model = mv))
  }
  
  # If all models are removed, return a tibble
  if(length(mv) == 0)
    return(out)
  
  build_mable(out, key = !!old_kv, model = mv)
}

#' @export
group_data.mdl_df <- function(.data){
  .data <- as_tibble(.data)
  NextMethod()
}

#' @export
key_data.mdl_df <- function(.data){
  .data%@%"key"
}

#' @export
key_vars.mdl_df <- function(x){
  keys <- key_data(x)
  names(keys)[-NCOL(keys)]
}

#' @export
key.mdl_df <- function(x){
  syms(key_vars(x))
}

context("setup-data.R")

us_deaths <- as_tsibble(USAccDeaths)
us_deaths_tr <- dplyr::filter(us_deaths, index < tsibble::yearmonth("1978 Jan"))
lung_deaths_long <- as_tsibble(cbind(mdeaths, fdeaths))
lung_deaths_long_tr <- dplyr::filter(lung_deaths_long, index < tsibble::yearmonth("1979 Jan"))
lung_deaths_wide <- as_tsibble(cbind(mdeaths, fdeaths), pivot_longer = FALSE)
lung_deaths_wide_tr <- dplyr::filter(lung_deaths_wide, index < tsibble::yearmonth("1979 Jan"))

if (requireNamespace("fable", quietly = TRUE)) {
mbl <- us_deaths_tr %>%
  model(ets = fable::ETS(value))
fbl <- mbl %>% forecast(h = 12)

mbl_multi <- lung_deaths_long_tr %>%
  model(ets = fable::ETS(value))
fbl_multi <- mbl_multi %>% forecast(h = 12)

mbl_complex <- lung_deaths_long_tr %>% 
  model(ets = fable::ETS(value), lm = fable::TSLM(value ~ trend() + season()))
fbl_complex <- mbl_complex %>% forecast(h = 12)

mbl_mv <- lung_deaths_wide_tr %>%
  model(var = fable::VAR(vars(mdeaths, fdeaths) ~ fourier(K = 4)))
fbl_mv <- mbl_mv %>% forecast(h = 12)
}

if (requireNamespace("feasts", quietly = TRUE)) {
dcmp <- us_deaths %>%
  model(feasts::STL(value)) %>% 
  components()

dcmp_multi <- lung_deaths_long %>%
  model(feasts::STL(value)) %>% 
  components()
}

#' @export
equation.mdl_df <- function(object, ...){
  if(NROW(object) > 1 || length(mable_vars(object)) > 1){
    abort("Model equations are only supported for individual models. To see the equation for a specific model, use `select()` and `filter()` to identify a single model.")
  }
  equation(object[[(mable_vars(object))[[1]]]][[1]])
}

#' @export
equation.mdl_ts <- function(object, ...){
  if(any(!map_lgl(object$transformation, compose(is.name, body)))){
    abort("Cannot display equations containing transformations.")
  }
  equation(object[["fit"]])
}

new_model <- function(fit = NULL, model, data, response, transformation){
  structure(list(fit = fit, model = model, data = data,
                 response = response, transformation = transformation),
            class = "mdl_ts")
}

#' @export
format.mdl_ts <-  function(x, ...){
  model_sum(x)
}

type_sum.mdl_ts <-  function(x){
  model_sum(x[["fit"]])
}

#' Is the object a model
#' 
#' @param x An object.
#' 
#' @export
is_model <- function(x){
  inherits(x, "mdl_ts")
}

#' Provide a succinct summary of a model
#' 
#' Similarly to pillar's type_sum and obj_sum, model_sum is used to provide brief model summaries.
#' 
#' @param x The model to summarise
#' 
#' @export
model_sum <- function(x){
  UseMethod("model_sum")
}

#' @export
model_sum.default <- function(x){
  tibble::type_sum(x)
}

#' @export
model_sum.mdl_ts <-  function(x){
  model_sum(x[["fit"]])
}

#' Internal vctrs methods
#'
#' These methods are the extensions that allow dable objects to
#' work with vctrs.
#'
#' @keywords internal
#' @name dable-vctrs
NULL

#' @rdname dable-vctrs
#' @method vec_ptype2 dcmp_ts
#' @export
vec_ptype2.dcmp_ts <- function(x, y, ...) {
  UseMethod("vec_ptype2.dcmp_ts", y)
}

#' @export
vec_ptype2.dcmp_ts.dcmp_ts <- function(x, y, ...) {
  dable_ptype2(x, y, ...)
}

#' @export
vec_ptype2.data.frame.dcmp_ts <- function(x, y, ...) {
  dable_ptype2(y, x, ...)
}

#' @export
vec_ptype2.dcmp_ts.data.frame <- vec_ptype2.dcmp_ts.dcmp_ts

#' @export
vec_ptype2.tbl_df.dcmp_ts <- vec_ptype2.data.frame.dcmp_ts

#' @export
vec_ptype2.dcmp_ts.tbl_df <- vec_ptype2.dcmp_ts.dcmp_ts

dable_ptype2 <- function(x, y, ...) {
  idx_x <- index_var(x)
  key_x <- key_vars(x)
  seas_x <- x%@%"seasons"
  alias_x <- x%@%"aliases"
  method_x <- x%@%"method"
  resp_x <- response_vars(x)
  if (is_dable(y)) {
    if (idx_x != index_var(y)) {
      abort("No common index variable for `x` and `y`.")
    }
    if (!identical(resp_x, response_vars(y))) {
      abort("Objects with different response variables cannot be combined.")
    }
    if(method_x != y%@%"method") method_x <- NULL
    seas_x <- vec_unique(c(seas_x, y%@%"seasons"))
    alias_x <- vec_unique(c(alias_x, y%@%"aliases"))
    key_x <- union(key_x, key_vars(y))
  }
  out <- df_ptype2(x, y, ...)
  tsbl <- build_tsibble_meta(
    out, key_data = tibble(!!!x[key_x], !!".rows" := list_of(.ptype = integer())),
    index = idx_x, index2 = idx_x, ordered = TRUE,
    interval = new_interval())
  build_dable(tsbl, response = resp_x, method = method_x, 
              seasons = seas_x, aliases = alias_x)
}

#' @rdname dable-vctrs
#' @method vec_cast dcmp_ts
#' @export
vec_cast.dcmp_ts <- function(x, to, ...) {
  UseMethod("vec_cast.dcmp_ts")
}

#' @export
vec_cast.dcmp_ts.dcmp_ts <- function(x, to, ...) {
  is_identical <- identical(x, to)
  tbl <- tib_cast(x, to, ...)
  tsbl <- build_tsibble(
    tbl, key = key_vars(to),
    key_data = if (is_identical) key_data(x) else NULL,
    index = index_var(to), index2 = index2_var(to),
    ordered = is_ordered(to),
    validate = FALSE, .drop = key_drop_default(to))
  build_dable(tsbl, response = response_vars(to), method = to%@%"method", 
              seasons = to%@%"seasons", aliases = to%@%"aliases")
}

#' @export
vec_cast.dcmp_ts.tbl_df <- function(x, to, ...) {
  tbl <- tib_cast(x, to, ...)
  tsbl <- build_tsibble(
    tbl, key = key_vars(to), index = index_var(to), index2 = index2_var(to),
    ordered = TRUE, validate = TRUE, .drop = key_drop_default(to))
  build_dable(tsbl, response = response_vars(to), method = to%@%"method", 
              seasons = to%@%"seasons", aliases = to%@%"aliases")
}

#' @export
vec_cast.dcmp_ts.data.frame <- vec_cast.dcmp_ts.tbl_df

#' @export
vec_cast.tbl_df.dcmp_ts <- function(x, to, ...) {
  tib_cast(x, to, ...)
}

#' @export
vec_cast.data.frame.dcmp_ts <- function(x, to, ...) {
  df_cast(x, to, ...)
}

#' @export
vec_cast.tbl_ts.dcmp_ts <- function(x, to, ...) {
  tbl <- tib_cast(x, to, ...)
  build_tsibble(
    tbl, key = key_vars(to), index = index_var(to), index2 = index2_var(to),
    ordered = TRUE, validate = TRUE, .drop = key_drop_default(to)
  )
}

#' Augment a mable
#' 
#' Uses a fitted model to augment the response variable with fitted values and
#' residuals. Response residuals (back-transformed) are stored in the `.resid`
#' column, while innovation residuals (transformed) are stored in the `.innov`
#' column.
#' 
#' @param x A mable.
#' @param ... Arguments for model methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' # Forecasting with an ETS(M,Ad,A) model to Australian beer production
#' aus_production %>%
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) %>% 
#'   augment()
#' 
#' @rdname augment
#' @export
augment.mdl_df <- function(x, ...){
  mbl_vars <- mable_vars(x)
  kv <- key_vars(x)
  x <- mutate(as_tibble(x), dplyr::across(all_of(mbl_vars), augment))
  x <- pivot_longer(x, all_of(mbl_vars), names_to = ".model", values_to = ".aug")
  unnest_tsbl(x, ".aug", parent_key = c(kv, ".model"))
}

#' @rdname augment
#' @export
augment.mdl_lst <- function(x, ...){
  lapply(x, augment, ...)
}
#' @export
augment.lst_mdl <- deprecate_lst_mdl(augment.mdl_lst)

#' @rdname augment
#' @param type Deprecated.
#' @export
augment.mdl_ts <- function(x, type = NULL, ...){
  if (!is.null(type)) { 
    lifecycle::deprecate_warn("0.2.1", "fabletools::augment(type = )", 
                              details = "The type argument is now deprecated for changes to broom v0.7.0.
Response residuals are now always found in `.resid` and innovation residuals are now found in `.innov`.")
  }
  tryCatch(augment(x[["fit"]], ...),
           error = function(e){
             idx <- index_var(x$data)
             resp <- x$response
             if(length(resp) > 1){
               response(x) %>% 
                 gather(".response", "value", !!!resp, factor_key = TRUE) %>% 
                 left_join(
                   gather(fitted(x, ...), ".response", ".fitted",
                          !!!resp, factor_key = TRUE),
                   by = c(".response", idx)
                 ) %>% 
                 left_join(
                   gather(residuals(x, type = "response", ...), ".response", ".resid",
                          !!!resp, factor_key = TRUE),
                   by = c(".response", idx)
                 ) %>% 
                 left_join(
                   gather(residuals(x, type = "innovation", ...), ".response", ".innov",
                          !!!resp, factor_key = TRUE),
                   by = c(".response", idx)
                 )
             } else {
               mutate(
                 set_names(response(x), c(idx, as_string(resp[[1]]))),
                 .fitted = fitted(x, ...)[[".fitted"]],
                 .resid = residuals(x, type = "response", ...)[[".resid"]],
                 .innov = residuals(x, type = "innovation", ...)[[".resid"]],
               )
             }
             
           })
}

#' Glance a mable
#' 
#' Uses the models within a mable to produce a one row summary of their fits.
#' This typically contains information about the residual variance,
#' information criterion, and other relevant summary statistics. Each model 
#' will be represented with a row of output.
#' 
#' @param x A mable.
#' @param ... Arguments for model methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' olympic_running %>%
#'   model(lm = TSLM(log(Time) ~ trend())) %>% 
#'   glance()
#' @rdname glance
#' @export
glance.mdl_df <- function(x, ...){
  mbl_vars <- mable_vars(x)
  x <- mutate(as_tibble(x), dplyr::across(all_of(mbl_vars), glance))
  x <- pivot_longer(x, all_of(mbl_vars), names_to = ".model", values_to = ".glanced")
  unnest(x, ".glanced")
}

#' @rdname glance
#' @export
glance.mdl_lst <- function(x, ...){
  lapply(x, glance, ...)
}
#' @export
glance.lst_mdl <- deprecate_lst_mdl(glance.mdl_lst)

#' @rdname glance
#' @export
glance.mdl_ts <- function(x, ...){
  glance(x$fit, ...)
}

#' Extract model coefficients from a mable
#' 
#' This function will obtain the coefficients (and associated statistics) for
#' each model in the mable.
#' 
#' @param x,object A mable.
#' @param ... Arguments for model methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' olympic_running %>%
#'   model(lm = TSLM(log(Time) ~ trend())) %>% 
#'   tidy()
#' 
#' @rdname tidy
#' @export
tidy.mdl_df <- function(x, ...){
  mbl_vars <- mable_vars(x)
  x <- mutate(as_tibble(x), dplyr::across(all_of(mbl_vars), tidy))
  x <- pivot_longer(x, all_of(mbl_vars), names_to = ".model", values_to = ".tidied")
  unnest(x, ".tidied")
}

#' @rdname tidy
#' @export
coef.mdl_df <- function(object, ...){
  tidy(object, ...)
}
#' @rdname tidy
#' @export
coef.mdl_lst <- function(object, ...){
  lapply(object, tidy, ...)
}
#' @export
coef.lst_mdl <- deprecate_lst_mdl(coef.mdl_lst)
#' @rdname tidy
#' @export
coef.mdl_ts <- function(object, ...){
  tidy(object, ...)
}

#' @rdname tidy
#' @export
tidy.mdl_lst <- function(x, ...){
  lapply(x, tidy, ...)
}
#' @export
tidy.lst_mdl <- deprecate_lst_mdl(tidy.mdl_lst)
#' @rdname tidy
#' @export
tidy.mdl_ts <- function(x, ...){
  tidy(x$fit, ...)
}

globalVariables("self")

train_decomposition <- function(.data, specials, ..., dcmp){
  # Extract raw original data
  dcmp <- components(model(self$data, dcmp))
  dcmp_method <- (dcmp%@%"aliases")[[response_vars(dcmp)]]
  structure <- dcmp%@%"seasons"
  aliases <- dcmp%@%"aliases"
  method <- dcmp%@%"method"
  
  xreg_vars <- setdiff(names(self$data), names(.data))
  dcmp[xreg_vars] <- self$data[xreg_vars]
  
  req_vars <- all.vars(dcmp_method)
  
  dcmp_ops <- traverse(dcmp_method,
    .f = function(x, y) c(y, x[[1]]),
    .g = function(x) x[-1],
    .h = function(x) if(is_call(x)) x[[1]] else NULL
  )
  
  if(any(map_chr(dcmp_ops, as_string) != "+")){
    abort("Only modelling of additive decompositions is supported.")
  }
  
  dcmp <- update_tsibble(dcmp, key = NULL)
  mdls <- dots_list(...) %>% 
    map(function(x) estimate(dcmp, x))
  
  alias_vars <- function(expr, aliases){
    vars <- all.vars(expr)
    alias_vars <- aliases[intersect(vars, names(aliases))] %>% 
      map(alias_vars, aliases = aliases) %>% 
      invoke(c, .)
    unname(c(setdiff(vars, names(aliases)), alias_vars))
  }
  
  mdl_vars <- map(mdls, `[[`, "response") %>% 
    squash() %>% 
    map(alias_vars, aliases = aliases) %>% 
    invoke(c, .)
                  
  miss_vars <- setdiff(req_vars, mdl_vars)
  
  if(!all(miss_vars %in% names(structure))) {
    abort(sprintf(
"Suitable defaults for these decomposition elements are not available: %s.
Please specify an appropriate model for these components",
      paste0(setdiff(miss_vars, names(structure)), collapse = ", "))
    )
  }
  
  mdls_default <- if(!is_empty(miss_vars)){
    check_installed("fable")
    lag <- NULL
    
    structure[miss_vars] %>% 
      imap(function(x, nm){
        estimate(dcmp, fable::SNAIVE(
          new_formula(lhs = sym(nm), rhs = expr(lag(!!x[["period"]])))
        ))
      })
  }
  else{
    list()
  }
  
  model <- reduce(c(mdls, mdls_default), `+`)
  
  all_match <- function(x, y) {
    non_missing <- stats::complete.cases(x, y)
    isTRUE(all.equal(x[non_missing], y[non_missing]))
  }
  if(!all_match(response(model)[[".response"]], .data[[measured_vars(.data)]])) {
    abort(
"The models specified do not combine to give the correct response.
Please check that you have specified the decomposition models appropriately.")
  }
  
  structure(model[["fit"]], dcmp_method = method,
            class = union("decomposition_model", class(model[["fit"]])))
}

#' Decomposition modelling
#' 
#' This function allows you to specify a decomposition combination model using 
#' any additive decomposition. It works by first decomposing the data using the
#' decomposition method provided to `dcmp_fn` with the given formula. Secondary
#' models are used to fit each of the components from the resulting 
#' decomposition. These models are specified after the decomposition formula.
#' All non-seasonal decomposition components must be specified, and any
#' unspecified seasonal components will be forecasted using seasonal naive. 
#' These component models will be combined according to the decomposition 
#' method, giving a combination model for the response of the decomposition.
#' 
#' @param dcmp A model definition which supports extracting decomposed [`components()`].
#' @param ... Model definitions used to model the components
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE) && requireNamespace("feasts", quietly = TRUE)
#' library(fable)
#' library(feasts)
#' library(tsibble)
#' library(dplyr)
#' library(ggtime)
#' 
#' vic_food <- tsibbledata::aus_retail %>% 
#'   filter(State == "Victoria", Industry == "Food retailing")
#'   
#' # Identify an appropriate decomposition
#' vic_food %>% 
#'   model(STL(log(Turnover) ~ season(window = Inf))) %>% 
#'   components() %>% 
#'   autoplot()
#'   
#' # Use an ETS model to seasonally adjusted data, and SNAIVE to season_year
#' # Any model can be used, and seasonal components will default to use SNAIVE.
#' my_dcmp_spec <- decomposition_model(
#'   STL(log(Turnover) ~ season(window = Inf)),
#'   ETS(season_adjust ~ season("N")), SNAIVE(season_year)
#' )
#' 
#' vic_food %>%
#'   model(my_dcmp_spec) %>% 
#'   forecast(h="5 years") %>% 
#'   autoplot(vic_food)
#' 
#' @seealso 
#' [*Forecasting: Principles and Practice* - Forecasting Decomposition](https://otexts.com/fpp3/forecasting-decomposition.html)
#' 
#' @export
decomposition_model <- function(dcmp, ...){
  dcmp_model <- new_model_class("dcmp_mdl", train = train_decomposition, 
                                specials = new_specials(xreg = function(...) NULL))
  new_model_definition(dcmp_model, !!dcmp$formula, ..., dcmp = dcmp)
}

#' @export
model_sum.decomposition_model <- function(x){
  paste(x%@%"dcmp_method", "decomposition model")
}

#' @export
report.model_combination <- function(object, ...){
  is_mdl_cmbn <- function(x) inherits(x, "model_combination")
  comb_expr <- traverse(
    object, .f = function(resp, comb) eval(expr(substitute(!!comb, resp))),
    .h = function(x) if(is_model(x)) x[["response"]][[1]] else if (is_mdl_cmbn(x)) x%@%"combination" else x,
    base = compose(`!`, is_mdl_cmbn))
  
  cmbn <- sprintf("Combination: %s", expr_text(comb_expr))
  cat(sprintf("%s\n\n%s\n\n", cmbn, strrep("=", nchar(cmbn))))
  
  traverse(object, .h = function(x) if(is_model(x)) {report(x);cat("\n")}, base = compose(`!`, is_mdl_cmbn))
  
  invisible(object)
}

context("test-decomposition-model")

test_that("Decomposition modelling", {
  skip_if_not_installed("fable")
  skip_if_not_installed("feasts")
  
  mdl_dcmp <- us_deaths %>%
    model(decomposition_model(feasts::STL(value), fable::NAIVE(season_adjust)))
  
  expect_output(
    report(mdl_dcmp),
    "Series: season_adjust \\nModel: NAIVE"
  )
  expect_output(
    report(mdl_dcmp),
    "Series: season_year \\nModel: SNAIVE"
  )
  
  fbl_dcmp <- forecast(mdl_dcmp)
  
  expect_equal(
    mean(fbl_dcmp$value),
    rep(dcmp$season_year[61:72], 2) + dcmp$season_adjust[72]
  )
})

#' Extract the response variable from a model
#' 
#' Returns a tsibble containing only the response variable used in the fitting
#' of a model.
#' 
#' @param object The object containing response data
#' @param ... Additional parameters passed on to other methods
#' 
#' @export
response <- function(object, ...){
  UseMethod("response")
}

#' @export
response.mdl_df <- function(object, ...){
  object <- tidyr::pivot_longer(object, all_of(mable_vars(object)),
                             names_to = ".model", values_to = ".fit")
  kv <- c(key_vars(object), ".model")
  object <- transmute(as_tibble(object),
                   !!!syms(kv),
                   !!sym(".model"),
                   response = map(!!sym(".fit"), response)
  )
  unnest_tsbl(object, "response", parent_key = kv)
}

#' @export
response.mdl_ts <- function(object, ...){
  # Extract response
  mv <- measured_vars(object$data)
  resp <- as.list(object$data)[mv]
  
  # Back transform response
  bt <- map(object$transformation, invert_transformation)
  resp <- map2(bt, resp, function(bt, fit) bt(fit))
  
  # Create object
  out <- object$data[index_var(object$data)]
  out[if(length(resp) == 1) ".response" else mv] <- resp
  out
}

#' Return response variables
#' 
#' `response_vars()` returns a character vector of the response variables in the
#' object.
#' 
#' @param x A dataset containing a response variable (such as a mable, fable, or dable).
#' @export
response_vars <- function(x){
  UseMethod("response_vars")
}

#' @export
response_vars.fbl_ts <- function(x){
  x%@%"response"
}
#' @export
response_vars.mdl_df <- function(x){
  x%@%"response"
}
#' @export
response_vars.mdl_ts <- function(x){
  vapply(x$response, rlang::as_label, FUN.VALUE = character(1L))
}
#' @export
response_vars.dcmp_ts <- function(x){
  x%@%"response"
}

#' Return distribution variable
#' 
#' `distribution_var()` returns a character vector of the distribution variable 
#' in the data.
#' 
#' @param x A dataset containing a distribution variable (such as a fable).
#' @export
distribution_var <- function(x){
  UseMethod("distribution_var")
}
#' @export
distribution_var.fbl_ts <- function(x){
  x%@%"dist"
}

#' Return model column variables
#' 
#' `mable_vars()` returns a character vector of the model variables in the
#' object.
#' 
#' @param x A dataset containing models (such as a mable).
#' @export
mable_vars <- function(x){
  UseMethod("mable_vars")
}
#' @export
mable_vars.mdl_df <- function(x){
  x%@%"model"
}

#' @export
dplyr_row_slice.fbl_ts <- function(data, i, ..., preserve = FALSE) {
  res <- NextMethod()
  build_fable(res, response = response_vars(data), distribution = distribution_var(data))
}

#' @export
dplyr_row_slice.grouped_fbl <- dplyr_row_slice.fbl_ts

#' @export
dplyr_col_modify.fbl_ts <- function(data, cols) {
  res <- NextMethod()
  build_fable(res, response = response_vars(data), distribution = distribution_var(data))
}

#' @export
dplyr_col_modify.grouped_fbl <- dplyr_col_modify.fbl_ts

#' @export
dplyr_reconstruct.fbl_ts <- function(data, template) {
  res <- NextMethod()
  dist <- distribution_var(template)
  if(dist %in% names(res)) {
    build_fable(res, response = response_vars(template), distribution = dist)
  } else {
    res
  }
}

#' @export
dplyr_reconstruct.grouped_fbl <- dplyr_reconstruct.fbl_ts

#' @export
summarise.fbl_ts <- function(.data, ..., .groups = NULL) { 
  dist_var <- distribution_var(.data)
  dist_ptype <- vec_ptype(.data[[dist_var]])
  resp_var <- response_vars(.data)
  .data <- summarise(as_tsibble(.data), ..., .groups = .groups)
  
  # If the distribution is lost, return a tsibble
  if(!(dist_var %in% names(.data))) {
    if(!vec_is(.data[[dist_var]], dist_ptype)){
      return(.data)
    }
  }
  
  build_fable(.data, response = resp_var, distribution = dist_var)
}

#' @export
summarise.grouped_fbl <- summarise.fbl_ts

#' @importFrom dplyr %>%
#' @export
dplyr::`%>%`

#' @export
tsibble::as_tsibble

#' @importFrom dplyr vars
#' @export
dplyr::vars

#' @importFrom ggplot2 autoplot
#' @export
ggplot2::autoplot

#' @importFrom ggplot2 autolayer
#' @export
ggplot2::autolayer

#' @importFrom generics accuracy
#' @export
generics::accuracy

#' @importFrom generics equation
#' @export
generics::equation

#' @importFrom generics interpolate
#' @export
generics::interpolate

#' @importFrom generics components
#' @export
generics::components

#' @importFrom generics augment
#' @export
generics::augment

#' @importFrom generics glance
#' @export
generics::glance

#' @importFrom generics tidy
#' @export
generics::tidy

#' @importFrom generics hypothesize
#' @export
generics::hypothesize

#' @importFrom generics generate
#' @export
generics::generate

#' @importFrom generics refit
#' @export
generics::refit

#' @importFrom generics forecast
#' @export
generics::forecast

#' @importFrom distributional hilo
#' @export
distributional::hilo

features_impl <- function(.tbl, .var, features, ...){
  dots <- dots_list(...)
  
  if(is_function(features) || is_formula(features)){
    features <- list(features)
  }
  features <- map(squash(features), as_function)
  
  dots$.period <- get_frequencies(dots$.period, .tbl, .auto = "smallest")
  
  # Compute response
  key_dt <- key_data(.tbl)
  idx <- index_var(.tbl)
  .tbl <- as_tibble(.tbl)
  if(NCOL(key_dt) > 1){
    .tbl <- dplyr::new_grouped_df(.tbl, key_dt)
  }
  .resp <- unclass(dplyr::transmute(.tbl, !!!.var))
  .resp <- .resp[seq_along(.var) + NCOL(key_dt) - 1]
  names(.resp) <- names(.var)
  
  # Compute features
  out <- map(.resp, function(x){
    res <- imap(features, function(fn, nm){
      fmls <- formals(fn)[-1]
      fn_safe <- safely(fn, tibble(.rows = 1))
      res <- transpose(map(key_dt[[".rows"]], function(i){
        # Add index to inputs
        dots$.index <- .tbl[[idx]][i]
        # Evaluate feature
        out <- do.call(fn_safe, c(list(x[i]), dots[intersect(names(fmls), names(dots))]))
        if(is.null(names(out[["result"]]))) 
          names(out[["result"]]) <- paste0("..?", seq_along(out[["result"]]))
        out
      }))
      err <- compact(res[["error"]])
      tbl <- vctrs::vec_rbind(!!!res[["result"]])
      
      names(tbl)[grepl("^\\.\\.?", names(tbl))] <- ""
      if(is.character(nm) && nzchar(nm)){
        names(tbl) <- sprintf("%s%s%s", nm, ifelse(nzchar(names(tbl)), "_", ""), names(tbl))
      }
      list(error = err, result = tbl)
    })
    res <- transpose(res)
    res[["result"]] <- invoke(bind_cols, res[["result"]])
    res
  })
  out <- transpose(out)
  
  # Report errors
  err <- flatten(unname(out$error))
  imap(err, function(err, nm){
    err <- compact(err)
    if((tot_err <- length(err)) > 0){
      err_msg <- table(map_chr(err, function(x) x[["message"]]))
      warn(
        sprintf("%i error%s encountered for feature %s\n%s\n",
                tot_err,
                if(tot_err > 1) sprintf("s (%i unique)", length(err_msg)) else "", 
                nm,
                paste0("[", err_msg, "] ", names(err_msg), collapse = "\n")
        )
      )
    }
  })
  
  out <- out[["result"]]
  
  if(!is.null(names(out))){
    out <- imap(out, function(tbl, nm){
      set_names(tbl, sprintf("%s_%s", nm, colnames(tbl)))
    })
  }
  
  bind_cols(
    key_dt[-NCOL(key_dt)],
    !!!out,
    .name_repair = "minimal"
  )
}

#' Extract features from a dataset
#'
#' Create scalar valued summary features for a dataset from feature functions.
#' 
#' Lists of available features can be found in the following pages:
#' - [Features by package][features_by_pkg]
#' - [Features by tag][features_by_tag]
#' 
#' @param .tbl A dataset
#' @param .var An expression that produces a vector from which the features are computed.
#' @param .vars A tidyselect compatible selection of the column(s) to compute features on.
#' @param features A list of functions (or lambda expressions) for the features to compute. [`feature_set()`] is a useful helper for building sets of features.
#' @param .predicate A predicate function (or lambda expression) to be applied to the columns or a logical vector. The variables for which .predicate is or returns TRUE are selected.
#' @param ... Additional arguments to be passed to each feature. These arguments will only be passed to features which use it in their formal arguments ([`base::formals()`]), and not via their `...`. While passing `na.rm = TRUE` to [`stats::var()`] will work, it will not for [`base::mean()`] as its formals are `x` and `...`. To more precisely pass inputs to each function, you should use lambdas in the list of features (`~ mean(., na.rm = TRUE)`).
#'
#' @seealso [`feature_set()`]
#'
#' @examples 
#' # Provide a set of functions as a named list to features.
#' library(tsibble)
#' tourism %>% 
#'   features(Trips, features = list(mean = mean, sd = sd))
#'
#' # Search and use useful features with `feature_set()`. 
#' 
#' @examplesIf requireNamespace("feasts", quietly = TRUE)
#' library(feasts)
#' @examples
#' tourism %>% 
#'   features(Trips, features = feature_set(tags = "autocorrelation"))
#' 
#' # Best practice is to use anonymous functions for additional arguments
#' tourism %>% 
#'   features(Trips, list(~ quantile(., probs=seq(0,1,by=0.2))))
#'
#'
#' @export
features <- function(.tbl, .var, features, ...){
  UseMethod("features")
}

#' @export
features.tbl_ts <- function(.tbl, .var = NULL, features = list(), ...){
  .var <- enquo(.var)
  if(quo_is_null(.var)){
    inform(sprintf(
      "Feature variable not specified, automatically selected `.var = %s`",
      measured_vars(.tbl)[1]
    ))
    .var <- as_quosure(sym(measured_vars(.tbl)[[1]]), env = empty_env())
  }
  else if(possibly(compose(is_quosures, eval_tidy), FALSE)(.var)){
    abort("`features()` only supports a single variable. To compute features across multiple variables consider scoped variants like `features_at()`")
  }
  
  features_impl(.tbl, list(.var), features, ...)
}

#' @rdname features
#' @export
features_at <- function(.tbl, .vars, features, ...){
  UseMethod("features_at")
}

#' @export
features_at.tbl_ts <- function(.tbl, .vars = NULL, features = list(), ...){
  .vars <- syms(tidyselect::vars_select(names(.tbl), !!!.vars))
  features_impl(.tbl, syms(.vars), features = features, ...)
}

#' @rdname features
#' @export
features_all <- function(.tbl, features, ...){
  UseMethod("features_all")
}

#' @export
features_all.tbl_ts <- function(.tbl, features = list(), ...){
  .vars <- measured_vars(.tbl)
  features_impl(.tbl, set_names(syms(.vars), .vars), features = features, ...)
}

#' @rdname features
#' @export
features_if <- function(.tbl, .predicate, features, ...){
  UseMethod("features_if")
}

#' @export
features_if.tbl_ts <- function(.tbl, .predicate, features = list(), ...){
  .vars <- measured_vars(.tbl)
  .vars <- .vars[map_lgl(.tbl[.vars], rlang::as_function(.predicate))]
  features_impl(.tbl, set_names(syms(.vars), .vars), features = features, ...)
}

# Lookup table for features
feature_table <- function() {
  table <- new.env(parent = emptyenv())
  list(
    add = function(fn, fn_name, tags) {
      pkg <- environmentName(environment(fn))
      table[[pkg]] <- as.list(table[[pkg]])
      table[[pkg]][[fn_name]] <- list(fn = fn, tags = tags)
    },
    get = function(pkg) {
      if(is.null(pkg)){
        as.list(table)
      }
      else{
        as.list(table)[pkg]
      }
    },
    list = function() {
      map(table, function(x) map(x, `[[`, "tags"))
    }
  )
}

feature_table <- feature_table()

#' Register a feature function
#' 
#' Allows users to find and use features from your package using [`feature_set()`].
#' If the features are being registered from within a package, this feature
#' registration should happen at load time using `[.onLoad()]`.
#' 
#' @param fn The feature function
#' @param tags Identifying tags
#' 
#' @examples 
#' 
#' \dontrun{
#' tukey_five <- function(x){
#'   setNames(fivenum(x), c("min", "hinge_lwr", "med", "hinge_upr", "max"))
#' }
#' 
#' register_feature(tukey_five, tags = c("boxplot", "simple"))
#' 
#' }
#' 
#' @export
register_feature <- function(fn, tags){
  nm <- enexpr(fn)
  nm <- if(is_call(nm)) call_name(fn) else as_string(nm)
  feature_table$add(fn, nm, tags)
}

#' Create a feature set from tags
#' 
#' Construct a feature set from features available in currently loaded packages.
#' Lists of available features can be found in the following pages:
#' - [Features by package][features_by_pkg]
#' - [Features by tag][features_by_tag]
#' 
#' @param pkgs The package(s) from which to search for features. If `NULL`, 
#' all registered features from currently loaded packages will be searched.
#' @param tags Tags used to identify similar groups of features. If `NULL`,
#' all tags will be included.
#' 
#' @section Registering features:
#' Features can be registered for use with the `feature_set()` function using
#' [`register_feature()`]. This function allows you to register a feature along
#' with the tags associated with it. If the features are being registered from
#' within a package, this feature registration should happen at load time using
#' `[.onLoad()]`.
#' 
#' @export
feature_set <- function(pkgs = NULL, tags = NULL){
  f_set <- flatten(unname(feature_table$get(pkgs)))
  if(!is.null(tags)){
    f_set <- f_set[map_lgl(f_set, function(x) any(x[["tags"]] %in% tags))]
  }
  unname(map(f_set, `[[`, "fn"))
}

rd_features_pkg <- function(){
  features <- map(feature_table$list(), names)
  
  if (length(features) == 0) {
    return("No features found in currently loaded packages.")
  }
  
  feature_links <- paste0(
    map2_chr(features, names(features), function(fns, pkg) {
      sprintf(
        "\\subsection{%s}{\n\\itemize{\n%s\n}\n}", pkg, 
        paste0(
          map_chr(fns, function(fn){
            sprintf("\\item \\code{\\link[%s]{%s}}", pkg, fn)
          }),
          collapse = "\n"
        )
      )
    }),
    collapse = "\n"
  )
  
  sprintf(
"See the following help topics for more details about currently available features:\n%s",
feature_links
  )
}

rd_features_tag <- function(){
  features <- imap(feature_table$list(), function(fns, pkg){
    fns <- set_names(
      unlist(fns, use.names = FALSE), 
      rep(names(fns), map_dbl(fns, length))
    )
    set_names(fns, sprintf("\\item \\code{\\link[%s]{%s}}", pkg, names(fns)))
  })
  
  if (length(features) == 0) {
    return("No features found in currently loaded packages.")
  }
  
  features <- invoke(c, unname(features))
  features <- split(names(features), features)
  
  feature_links <- paste0(
    map2_chr(features, names(features), function(fns, tag) {
      sprintf(
        "\\subsection{%s}{\n\\itemize{\n%s\n}\n}",
        tag, paste0(fns, collapse = "\n")
      )
    }),
    collapse = "\n"
  )
  
  sprintf(
    "See the following help topics for more details about each feature:\n%s",
    feature_links
  )
}

#' Features by package
#' 
#' This documentation lists all available in currently loaded packages. This is
#' a useful reference for making a [`feature_set()`] from particular package(s).
#' 
#' \Sexpr[stage=render,results=rd]{fabletools:::rd_features_pkg()}
#' 
#' @seealso [features_by_tag]
#' 
#' @keywords internal
#' @name features_by_pkg
NULL

#' Features by tag
#' 
#' This documentation lists all available in currently loaded packages. This is
#' a useful reference for making a [`feature_set()`] from particular tag(s).
#' 
#' \Sexpr[stage=render,results=rd]{fabletools:::rd_features_tag()}
#' 
#' @seealso [features_by_pkg]
#' 
#' @keywords internal
#' @name features_by_tag
NULL

context("test-features")

www_usage <- as_tsibble(WWWusage)
lung_deaths_long <- as_tsibble(cbind(mdeaths, fdeaths))
lung_deaths_wide <- as_tsibble(cbind(mdeaths, fdeaths), pivot_longer = FALSE)

first <- function(x) c(first = x[1])
last <- function(x) c(last = x[length(x)])

test_that("features()", {
  expect_error(
    features(lung_deaths_wide, vars(mdeaths, fdeaths), first),
    "only supports a single variable"
  )
  
  expect_message(
    features(lung_deaths_wide, features = list(first, last)),
    "Feature variable not specified, automatically selected \\`.var = mdeaths\\`"
  ) %>%
    colnames() %>%
    expect_equal(c("first", "last"))
  
  features(lung_deaths_wide, mdeaths, features = list(a = mean, b = min, max)) %>% 
    colnames() %>% 
    expect_equal(c("a", "b", "...3"))
})

test_that("Scoped variants of features()", {
  ft_at <- features_at(lung_deaths_wide, vars(mdeaths:fdeaths), list(first, last))
  expect_equal(
    substr(colnames(ft_at), 1, 7),
    c(rep("mdeaths", 2), rep("fdeaths", 2))
  )
  ft_if <- features_if(lung_deaths_wide, is.numeric, list(first, last))
  expect_identical(
    ft_at, ft_if
  )
  ft_all <- features_all(lung_deaths_wide, list(first, last))
  expect_identical(
    ft_if, ft_all
  )
})

names_no_null <- function(x){
  names(x) %||% rep_along(x, "")
}

# Small function to combine named lists
merge_named_list <- function(...){
  flat <- flatten(list(...))
  nm <- names_no_null(flat)
  map(split(flat, nm), function(x) flatten(unname(x)))
}

add_class <- function(x, new_class){
  `class<-`(x, union(new_class, class(x)))
}

rm_class <- function(x, class){
  `class<-`(x, setdiff(class(x), class))
}

custom_error <- function(.f, error){
  force(error)
  function(...){
    res <- capture_error(.f(...))
    if(!is.null(res$error)){
      abort(error)
    }
    res$result
  }
}

#' @importFrom tibble new_tibble
make_future_data <- function(.data, h = NULL){
  n <- get_frequencies(h, .data, .auto = "smallest")
  if(length(n) > 1){
    warn("More than one forecast horizon specified, using the smallest.")
    n <- min(n)
  }
  if(is.null(h)) n <- n*2
  
  tsibble::new_data(.data, round(n))
  # Re-implemented here using a simpler/faster method
  # n <- round(n)
  # 
  # idx <- index_var(.data)
  # itvl <- interval(.data)
  # tunit <- default_time_units(itvl)
  # 
  # idx_max <- max(.data[[idx]])
  # if(is.factor(idx_max)){
  #   abort("Cannot automatically create `new_data` from a factor/ordered time index. Please provide `new_data` directly.")
  # }
  # 
  # .data <- list2(!!idx := seq(idx_max, by = tunit, length.out = n+1)[-1])
  # build_tsibble_meta(
  #   new_tibble(.data, nrow = n),
  #   key_data = new_tibble(list(.rows = list(seq_len(n))), nrow = 1),
  #   index = idx, index2 = idx, ordered = TRUE, interval = itvl
  # )
}

bind_new_data <- function(object, new_data){
  # Handle multiple scenarios provided to new_data.
  if(inherits(new_data, "list")) {
    scenario_nm <- attr(new_data, "names_to") %||% ".scenario"
    new_data <- vec_rbind(
      !!!map(new_data, compose(as_tibble, bind_new_data), object = object),
      .names_to = scenario_nm
    )
    return(
      build_mable(new_data, 
                  key = c(scenario_nm, key_vars(object)), 
                  model = mable_vars(object))
    )
  }
  if(!is.data.frame(new_data)){
    abort(sprintf("`new_data` requires a data frame. Perhaps you intended to specify the forecast horizon? If so, use `h = %s`.", deparse(new_data)))
  }
  if(!identical(key_vars(object), key_vars(new_data))){
    abort("Provided data contains a different key structure to the models.")
  }
  
  new_data <- nest_keys(new_data, "new_data")
  
  if(length(key_vars(object)) > 0){
    object <- left_join(object, new_data, by = key_vars(object))
    # Use empty tsibbles for non-matches
    no_new_data <- map_lgl(object[["new_data"]], is_null)
    if(any(no_new_data)){
      object[["new_data"]][no_new_data] <- rep(list(new_data[["new_data"]][[1]][0,]), sum(no_new_data))
    }
  }
  else{
    object[["new_data"]] <- new_data[["new_data"]]
  }
  object
}

# From R6
assign_func_envs <- function(objs, target_env) {
  if (!is.null(target_env)){
    objs <- lapply(objs, function(x) {
      if (is.function(x)) environment(x) <- target_env
      x
    })
  }
  objs
}

# tibble:::big_mark
big_mark <- function (x, ...) {
  mark <- if (identical(getOption("OutDec"), ",")) 
    "."
  else ","
  ret <- formatC(x, big.mark = mark, ...)
  ret[is.na(x)] <- "??"
  ret
}

calc <- function(f, ...){
  f(...)
}

# Unnest nested tibble/tsibble
# 
# Similar to tsibble::unnest_tsibble, but less general and faster validation
# 
# @param .data A dataset containing a listed column of tsibbles
# @param tbl_col/tsbl_col The column containing the tibble/tsibble to be unnested
# @param parent_key The keys from data to be joined with keys from the nested tsibble
# @param interval If NULL, the interval will be taken from the first tsibble, otherwise defaults to [[tsibble::build_tsibble()]] functionality.
unnest_tbl <- function(.data, tbl_col, .sep = NULL){
  row_indices <- rep.int(seq_len(NROW(.data)), map_int(.data[[tbl_col[[1]]]], NROW))
  
  nested_cols <- map(tbl_col, function(x){
    lst_col <- .data[[x]]
    if(is.data.frame(lst_col[[1]])){
      lst_col <- map(lst_col, as_tibble)
      vctrs::vec_rbind(!!!lst_col)
    }
    else{
      unlist(lst_col)
    }
  })
  
  if(!is.null(.sep)){
    nested_cols <- map2(
      nested_cols, tbl_col,
      function(x, nm) set_names(x, paste(nm, colnames(x), sep = .sep))
    )
  }
  
  is_df <- map_lgl(nested_cols, is.data.frame)
  vctrs::vec_cbind(
    .data[row_indices, setdiff(names(.data), tbl_col), drop = FALSE], # Parent cols
    !!!set_names(nested_cols[!is_df], tbl_col[!is_df]), # Nested cols
    !!!nested_cols[is_df] # Nested df
  )
}

unnest_tsbl <- function(.data, tsbl_col, parent_key = NULL, interval = NULL){
  tsbl <- .data[[tsbl_col]][[1L]]
  if (!is_tsibble(tsbl)) {
    abort("Unnested column is not a tsibble object.")
  }
  idx <- index(tsbl)
  idx_chr <- as_string(idx)
  key <- c(parent_key, key_vars(tsbl))
  
  .data <- unnest_tbl(.data, tsbl_col)
  
  build_tsibble(.data, key = !!key, index = !!idx,
                index2 = !!index2(tsbl), ordered = is_ordered(tsbl),
                interval = interval%||%interval(tsbl))
}

nest_grps <- function(.data, nm = "data"){
  out <- dplyr::group_data(.data)
  grps <- dplyr::group_vars(.data)
  row_indices <- out[[NCOL(out)]]
  out[[NCOL(out)]] <- NULL
  col_nest <- -match(grps, colnames(.data))
  if(is_empty(col_nest)){
    col_nest <- rlang::missing_arg()
  }
  out[[nm]] <- map(row_indices, function(x, i, j){
    out <- x[i,j]
  }, x = .data, j = col_nest)
  out
}

nest_keys <- function(.data, nm = "data"){
  out <- unclass(key_data(.data))
  key <- key_vars(.data)
  row_indices <- out[[length(out)]]
  out[[length(out)]] <- NULL
  col_nest <- -match(key, colnames(.data))
  if(is_empty(col_nest)){
    col_nest <- NULL
  }
  idx <- index_var(.data)
  idx2 <- index2_var(.data)
  ordered <- is_ordered(.data)
  regular <- is_regular(.data)
  out[[nm]] <- map(row_indices, function(x, i, j){
    out <- if(is.null(j)) x[i,] else x[i,j]
    build_tsibble_meta(
      out, 
      key_data = as_tibble(list(.rows = list(seq_along(i)))),
      index = idx, index2 = idx2, ordered = ordered, 
      interval = if(length(i) > 1 && regular) interval_pull(out[[idx]]) else interval(.data)
    )
  }, x = as_tibble(.data), j = col_nest)
  as_tibble(out)
}

bind_row_attrb <- function(x){
  attrb <- transpose(map(x, function(dt) map(dt, attributes)))
  simple_attrb <- map_lgl(attrb, function(x) length(unique(x)) == 1)
  
  x <- vctrs::vec_rbind(!!!x)
  
  for (col in which(simple_attrb)){
    attributes(x[[col]]) <- attrb[[col]][[1]]
  }
  x
}

is.formula <- function(x) {
  inherits(x, "formula")
}

# tsibble:::assert_key_data
assert_key_data <- function (x) {
  nc <- NCOL(x)
  if (is_false(is.data.frame(x) && nc > 0 && is.list(x[[nc]]) && 
               names(x)[[nc]] == ".rows")) {
    abort("The `key` attribute must be a data frame with its last column called `.rows`.")
  }
}

flatten_with_names <- function (x, sep = "_") {
  if(is_empty(x)) return(x)
  if(!is.list(x[[1]])) return(as.list(x))
  x <- map2(x, names(x), function(x, nm){
    if(!is.null(names(x))) set_names(x, paste(nm, names(x), sep = sep)) else x
  })
  flatten(unname(map(x, flatten_with_names, sep = sep)))
}

mapply_maybe_parallel <- function (.f, ..., MoreArgs = list(), SIMPLIFY = FALSE) {
  p <- progressr::progressor(length(..1))
  .fp <- function(...) {
    p()
    .f(...)
  }
  if(is_attached("package:future")){
    check_installed("future.apply")
    future.apply::future_mapply(
      FUN = .fp,
      ...,
      MoreArgs = MoreArgs,
      SIMPLIFY = SIMPLIFY,
      future.globals = FALSE,
      future.seed = TRUE
    )
  }
  else{
    mapply(
      FUN = .fp,
      ...,
      MoreArgs = MoreArgs,
      SIMPLIFY = SIMPLIFY
    )
  }
}

mable_apply <- function (.data, .f, ..., names_to = ".model") {
  mv <- mable_vars(.data)
  kv <- key_vars(.data)
  
  # Compute .f for all models in mable
  result <- lapply(
    as_tibble(.data)[mv],
    mapply_maybe_parallel,
    .f = .f,
    .data, 
    MoreArgs = dots_list(...)
  )
  num_rows <- lapply(result, vapply, nrow, integer(1L))
  # Assume same tsibble structure for all outputs
  first_result <- result[[1]][[1]]
  result <- vec_rbind(!!!lapply(result, function(x) vec_rbind(!!!x)))
  
  # Compute .model label
  model <- rep.int(mv, vapply(num_rows, sum, integer(1L)))
  
  # Repeat key structure as needed
  .data <- .data[rep.int(seq_along(num_rows[[1]]), num_rows[[1]]), kv]
  
  # Combine into single
  .data <- bind_cols(.data, !!names_to := model, result)
  
  if (is_tsibble(first_result)) {
    .data <- build_tsibble(
      .data, key = c(kv, names_to, key_vars(first_result)), 
      index = index_var(first_result), index2 = !!index2(first_result), 
      ordered = is_ordered(first_result),
      interval = interval(first_result))
  }
  
  return(.data)
}

dist_types <- function(dist) {
  map_chr(vec_data(dist), function(x) class(x)[1])
}

mdl_df_apply <- function(x, f, ...) {
  mbl_vars <- mable_vars(x)
  kv <- key_vars(x)
  x <- mutate(as_tibble(x), 
              dplyr::across(all_of(mbl_vars), function(x) lapply(x, f, ...)))
  x <- pivot_longer(x, all_of(mbl_vars), names_to = ".model", values_to = "__results__")
  unnest_tsbl(x, "__results__", parent_key = c(kv, ".model"))
}

#' Estimate models
#' 
#' Trains specified model definition(s) to a dataset. This function will 
#' estimate the a set of model definitions (passed via `...`) to each series
#' within `.data` (as identified by the key structure). The result will be a
#' mable (a model table), which neatly stores the estimated models in a tabular
#' structure. Rows of the data identify different series within the data, and
#' each model column contains all models from that model definition. Each cell
#' in the mable identifies a single model.
#' 
#' @param .data A data structure suitable for the models (such as a `tsibble`)
#' @param ... Definitions for the models to be used. All models must share the
#' same response variable.
#'
#' @rdname model
#' @export
model <- function(.data, ...){
  UseMethod("model")
}

#' @rdname model
#' 
#' @param .safely If a model encounters an error, rather than aborting the process a [NULL model][null_model()] will be returned instead. This allows for an error to occur when computing many models, without losing the results of the successful models.
#' 
#' @section Parallel:
#' 
#' It is possible to estimate models in parallel using the
#' [future](https://cran.r-project.org/package=future) package. By specifying a
#' [`future::plan()`] before estimating the models, they will be computed 
#' according to that plan.
#' 
#' @section Progress:
#' 
#' Progress on model estimation can be obtained by wrapping the code with 
#' `progressr::with_progress()`. Further customisation on how progress is 
#' reported can be controlled using the `progressr` package.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' # Training an ETS(M,Ad,A) model to Australian beer production
#' aus_production %>%
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A")))
#' 
#' # Training a seasonal naive and ETS(A,A,A) model to the monthly 
#' # "Food retailing" turnover for selected Australian states.
#' library(dplyr)
#' progressr::with_progress(
#' aus_retail %>% 
#'   filter(
#'     Industry == "Food retailing",
#'     State %in% c("Victoria", "New South Wales", "Queensland")
#'   ) %>% 
#'   model(
#'     snaive = SNAIVE(Turnover),
#'     ets = ETS(log(Turnover) ~ error("A") + trend("A") + season("A")),
#'   )
#' )
#' 
#' @export
model.tbl_ts <- function(.data, ..., .safely = TRUE){
  nm <- map(enexprs(...), expr_text)
  models <- dots_list(...)
  
  if(length(models) == 0){
    abort("At least one model must be specified.")
  }
  if(!all(is_mdl <- map_lgl(models, inherits, "mdl_defn"))){
    abort(sprintf("Model definition(s) incorrectly created: %s
Check that specified model(s) are model definitions.", nm[which(!is_mdl)[1]]))
  }
  
  num_key <- n_keys(.data)
  num_mdl <- length(models)
  num_est <- num_mdl * num_key
  p <- progressr::progressor(num_est)
  
  kv <- key_vars(.data)
  .data <- nest_keys(.data, "lst_data")
  
  if(.safely){
    estimate <- function(dt, mdl){
      out <- safely(fabletools::estimate)(dt, mdl)
      if(is.null(out$result)){
        f <- quo(!!mdl$formula)
        f <- set_env(f, mdl$env)
        out$result <- fabletools::estimate(dt, null_model(!!f))
      }
      out
    }
  }
  
  estimate_progress <- function(dt, mdl){
    out <- estimate(dt, mdl)
    p()
    out
  }
  
  if(is_attached("package:future")){
    check_installed("future.apply")
    eval_models <- function(models, lst_data){
      out <- future.apply::future_mapply(
        rep(lst_data, length(models)),
        rep(models, each = length(lst_data)),
        FUN = estimate_progress,
        SIMPLIFY = FALSE,
        future.globals = FALSE,
        future.seed = TRUE
      )
      unname(split(out, rep(seq_len(num_mdl), each = num_key)))
    }
  }
  else{
    eval_models <- function(models, lst_data){
      map(models, function(model){
        map(lst_data, estimate_progress, model)
      })
    }
  }
  
  fits <- eval_models(models, .data[["lst_data"]])
  names(fits) <- ifelse(nchar(names(models)), names(models), nm)
  
  # Report errors if estimated safely
  if(.safely){
    fits <- imap(fits, function(x, nm){
      err <- map_lgl(x, function(x) !is.null(x[["error"]]))
      if((tot_err <- sum(err)) > 0){
        err_msg <- table(map_chr(x[err], function(x) x[["error"]][["message"]]))
        warn(
          sprintf("%i error%s encountered for %s\n%s\n",
                  tot_err,
                  if(tot_err > 1) sprintf("s (%i unique)", length(err_msg)) else "", 
                  nm,
                  paste0("[", err_msg, "] ", names(err_msg), collapse = "\n")
          )
        )
      }
      map(x, function(x) x[["result"]])
    })
  }
  
  fits <- map(fits, list_of_models)
  
  .data %>% 
    transmute(
      !!!syms(kv),
      !!!fits
    ) %>% 
    as_mable(key = !!kv, model = names(fits))
}

#' Extract the left hand side of a model
#' 
#' @param model A formula
#' 
#' @export
model_lhs <- function(model){
  f <- model$formula
  if(is_quosure(f)){
    f <- get_expr(f)
  }
  if(is.formula(f)){
    return(f_lhs(f))
  }
  if(is.call(f)) {
    if(call_name(f) == "~")
      return(f[[2]])
  }
  f
}

#' Extract the right hand side of a model
#' 
#' @param model A formula
#' 
#' @export
model_rhs <- function(model){
  if(is.formula(model$formula)){
    f_rhs(model$formula)
  }
  else{
    expr(NULL)
  }
}

#' @export
dplyr_row_slice.mdl_df <- function(data, i, ..., preserve = FALSE) {
  res <- dplyr_row_slice(as_tibble(data), i, ..., preserve = preserve)
  build_mable_meta(
    res,
    key_data = dplyr::group_data(dplyr::group_by(res, !!!syms(key_vars(data)))),
    model = mable_vars(data),
    response = response_vars(data)
  )
}

#' @export
dplyr_col_modify.mdl_df <- function(data, cols) {
  res <- dplyr_col_modify(as_tibble(data), cols)
  is_mdl <- map_lgl(cols, inherits, c("lst_mdl", "mdl_lst"))
  # val_key <- any(key_vars(data) %in% cols)
  # if (val_key) {
  #   key_vars <- setdiff(names(res), measured_vars(data))
  #   data <- remove_key(data, key_vars)
  # }
  build_mable(res, 
              key = !!key_vars(data), 
              model = union(mable_vars(data), names(which(is_mdl))))
}

#' @export
dplyr_reconstruct.mdl_df <- function(data, template) {
  res <- NextMethod()
  mbl_vars <- names(which(vapply(data, inherits, logical(1L), "mdl_lst")))
  kv <- key_vars(template)
  if(all(kv %in% names(res))) {
    build_mable(data, key = !!kv, model = mbl_vars)
  } else {
    as_tibble(res)
  }
}

#' Box Cox Transformation
#'
#' `box_cox()` returns a transformation of the input variable using a Box-Cox
#' transformation. `inv_box_cox()` reverses the transformation.
#'
#' The Box-Cox transformation is given by \deqn{f_\lambda(x) =\frac{x^\lambda -
#' 1}{\lambda}}{f(x;lambda)=(x^lambda - 1)/lambda} if \eqn{\lambda\ne0}{lambda
#' is not equal to 0}. For \eqn{\lambda=0}{lambda=0},
#' \deqn{f_0(x)=\log(x)}{f(x;0)=log(x)}.
#'
#' @param x a numeric vector.
#' @param lambda a numeric value for the transformation parameter.
#' @return a transformed numeric vector of the same length as x.
#' @author Rob J Hyndman & Mitchell O'Hara-Wild
#' 
#' @references Box, G. E. P. and Cox, D. R. (1964) An analysis of
#' transformations. \emph{JRSS B} \bold{26} 211--246.
#' 
#' @examples
#' library(tsibble)
#' library(dplyr)
#' airmiles %>% 
#'   as_tsibble() %>% 
#'   mutate(box_cox = box_cox(value, lambda = 0.3))
#'
#' @export
box_cox <- function(x, lambda) {
  lambda <- vec_recycle(lambda, vec_size(x))
  x[lambda < 0 & x < 0] <- NA
  lambda_0 <- lambda == 0
  x[lambda_0] <- log(x[lambda_0])
  x[!lambda_0] <- (sign(x[!lambda_0]) * abs(x[!lambda_0]) ^ lambda[!lambda_0] - 1) / lambda[!lambda_0]
  x
}

#' @rdname box_cox
#' @export
inv_box_cox <- function(x, lambda) {
  lambda <- vec_recycle(lambda, vec_size(x))
  x[lambda < 0 & (x > -1 / lambda)] <- NA
  lambda_0 <- lambda == 0
  x[lambda_0] <- exp(x[lambda_0])
  z <- x[!lambda_0] * lambda[!lambda_0] + 1
  x[!lambda_0] <- sign(z) * abs(z) ^ (1 / lambda[!lambda_0])
  x
}

context("test-transformations")

simple_data <- tsibble(idx = 1:10, y = abs(rnorm(10)), x = 1:10, index = idx)
test_transformation <- function(..., dt = simple_data){
  mdl <- estimate(dt, no_specials(...))
  trans <- mdl$transformation[[1]]
  resp <- mdl$response[[1]]
  expect_equal(
    dt[[expr_name(resp)]],
    fabletools:::invert_transformation(trans)(trans(dt[[expr_name(resp)]]))
  )
}

test_that("single transformations", {
  test_transformation(y)
  test_transformation(y + 10)
  test_transformation(10 + y)
  test_transformation(+y)
  test_transformation(y - 10)
  test_transformation(10 - y)
  test_transformation(-y)
  test_transformation(3*y)
  test_transformation(y*3)
  test_transformation(3/y)
  test_transformation(y/3)
  test_transformation(log(y))
  test_transformation(logb(y, 10))
  test_transformation(log10(y))
  test_transformation(log2(y))
  test_transformation(log1p(y))
  test_transformation(expm1(y))
  test_transformation(exp(y))
  test_transformation(box_cox(y, 0.4))
  test_transformation(inv_box_cox(y, 0.4))
  test_transformation(sqrt(y))
  test_transformation(y^2)
  test_transformation(2^y)
  test_transformation((y))
})


test_that("transformation chains", {
  test_transformation(y + 10 - 10)
  test_transformation(10 + y * 10)
  test_transformation(+y - y)
  test_transformation(y^2 + 3)
  test_transformation(log(sqrt(y)))
  test_transformation(log(y + 1))
  test_transformation(box_cox(y^2,0.3))
  test_transformation(box_cox(y,0.3) + 1)
  
  # Something too complex
  expect_error(
    test_transformation(box_cox(y,0.3)^2),
    "Could not identify a valid back-transformation"
  )
  
  # Something rediculous
  test_transformation(log(sqrt(sqrt(sqrt(sqrt(sqrt(y)))+3))))
})

#' Report information about an object
#' 
#' Displays the object in a suitable format for reporting.
#' 
#' @param object The object to report
#' @param ... Additional options for the reporting function
#' 
#' @export
report <- function(object, ...){
  UseMethod("report")
}

#' @export
report.mdl_df <- function(object, ...){
  if(NROW(object) > 1 || length(mable_vars(object)) > 1){
    warning("Model reporting is only supported for individual models, so a glance will be shown. To see the report for a specific model, use `select()` and `filter()` to identify a single model.")
    return(glance(object))
  }
  else{
    report(object[[mable_vars(object)[[1]]]][[1]], ...)
  }
  invisible(object)
}

#' @export
report.mdl_ts <- function(object, ...){
  cat(paste("Series:", paste0(map(object$response, expr_name), collapse = ", "), "\n"))
  cat(paste("Model:", model_sum(object), "\n"))
  if(!is_symbol(body(object$transformation[[1]])) && length(object$response) == 1){
    cat(paste("Transformation:", expr_name(body(object$transformation[[1]])), "\n"))
  }
  tryCatch(
    report(object[["fit"]], ...),
    error = function(e){
      cat("\nA model specific report is not available for this model class.")
    }
  )
}

parse_specials <- function(call = NULL, specials = NULL){
  if(!is.null(call)){ # No model specified
    call <- enexpr(call)
    
    # Don't parse xreg_specials - leave them to the xregs
    nm <- setdiff(names(specials), attr(specials, "xreg_specials"))
    
    parsed <- traverse_call(!!call,
                            .f = function(.x, ...) {
                              merge_named_list(.x[[1]], .x[[2]])},
                            .g = function(.x){ 
                              map(as.list(get_expr(.x))[-1], expr)
                            },
                            .h = function(x){ # Base types
                              x <- get_expr(x)
                              if(!is_call(x) || !(call_name(x) %in% nm)){
                                list(list(x))
                              }
                              else{# Current call is a special function
                                set_names(list(list(x)), call_name(x))
                              }
                            },
                            base = function(.x){
                              .x <- get_expr(.x)
                              !is_call(.x) || call_name(.x) != "+"
                            }
    )
  } else {
    parsed <- list()
  }
  
  bare_xreg <- names_no_null(parsed) == ""
  if(any(bare_xreg)){
    parsed$xreg[[length(parsed$xreg) + 1]] <- expr((!!sym("xreg"))(!!!parsed[[which(bare_xreg)]]))
    parsed[[which(bare_xreg)]] <- NULL
  }
  
  # Add required_specials
  missing_specials <- setdiff(attr(specials, "required_specials"), names(parsed))
  parsed[missing_specials] <- map(missing_specials, function(x) list(call(x)))
  
  parsed
}

#' Validate the user provided model
#' 
#' Appropriately format the user's model for evaluation. Typically ran as one of the first steps
#' in a model function.
#' @param model A quosure for the user's model specification
#' @param data A dataset used for automatic response selection
#' 
#' @keywords internal
validate_formula <- function(model, data = NULL){
  # Clean inputs
  if(!is_quosure(model$formula)){
    model$formula <- new_quosure(model$formula)
  }
  
  if(quo_is_missing(model$formula)){
    model$formula <- guess_response(data)
  }
  else{
    if(possibly(compose(is.formula, eval_tidy), FALSE)(model$formula)){
      f_env <- get_env(model$formula)
      model$formula <- eval_tidy(model$formula)
      environment(model$formula) <- f_env
      
      # Add response if missing
      if(is.null(model_lhs(model))){
        model$formula <- new_formula(
          lhs = guess_response(data),
          rhs = model_rhs(model),
          env = get_env(model$formula)
        )
      }
    }
    else{
      model$formula <- get_expr(model$formula)
    }
  }
  
  invisible(model$formula)
}

#' Parse the model specification for specials
#' 
#' Using a list of defined special functions, the user's formula specification and data
#' is parsed to extract important modelling components.
#' 
#' @param model A model definition
#' 
#' @keywords internal
parse_model <- function(model){
  # Parse model
  list2(
    model = model,
    !!!parse_model_lhs(model),
    specials = parse_model_rhs(model)
  )
}

#' Parse the RHS of the model formula for specials
#' 
#' @inheritParams parse_model
#' @keywords internal
parse_model_rhs <- function(model){
  # if(length(model$specials) == 0){
  #   return(list(specials = NULL))
  # }
  rhs <- model_rhs(model)
  specials <- parse_specials(rhs, specials = model$specials)
  map(specials, function(.x){
      map(.x, eval_tidy, data = model$data, env = model$specials)
  })
}

#' Parse the RHS of the model formula for transformations
#' 
#' @inheritParams parse_model
#' @keywords internal
parse_model_lhs <- function(model){
  model_lhs <- model_lhs(model)
  if(is_call(model_lhs) && call_name(model_lhs) == "vars"){
    model_lhs[[1]] <- sym("exprs")
    model_lhs <- eval(model_lhs)
  }
  else{
    model_lhs <- list(model_lhs)
  }
  is_resp <- function(x) is_call(x) && x[[1]] == sym("resp")
  
  # Traverse call removing all resp() usage
  # This is used to evaluate the response from the input data
  response_exprs <- lapply(
    model_lhs, traverse,
    .f = function(x, y) {
      if(is_resp(y)) x[[1]] else call2(y[[1]], !!!x)
    },
    .g = function(x) x[-1],
    # .h = function(x) if(is_resp(x)) x[[length(x)]] else x,
    base = function(x) is_syntactic_literal(x) || is_symbol(x)
  )
  
  has_resp <- function(x) traverse(
    x,
    .f = function(x,y) x[[1]]||y, .g = function(x) x[-1], .h = is_resp,
    base = function(x) is_syntactic_literal(x) || is_symbol(x)
  )
  
  # Traverse call AST to parse out order of transformations
  #
  # If the response is set via resp(), remove all usage of resp() from the traversal
  # If the response is not set via resp(), identify the response by the maximum length object until encountering ties
  # 
  # Returns a list of increasing depth of transformations
  traversed_lhs <- lapply(model_lhs, 
    function(x) {
      len1vals <- list()
      path <- traverse(
        x,
        .f = function(x, y) {
          # Special handling for if the response was found by a length tie
          if((length(x[[1]]$response == 1L) && (as_label(x[[1]]$response) == as_label(y[[1]])))) {
            return(x[[1]])
          }
          
          # Rebuild the expression
          args <- lapply(x, function(y) {
            y[[length(y)]]
          })
          y <- as.call(c(y[[1]][[1]], args))
          
          # Search for response
          path <- compact(lapply(x, `[[`, "response"))
          if(length(path) > 0) return(list(response = c(path[[1]], y)))
          
          # Otherwise keep the path that isn't length 1
          path <- x[[which(map_lgl(x, function(x) is.name(x[[1]]) && !(as_label(x[[1]]) %in% names(len1vals))))]]
          c(path, y)
        },
        .g = function(x) {
          # traverse only call arguments
          args <- x[-1]
          
          # search for resp() to avoid unneeded evaluation
          resp_loc <- which(map_lgl(args, has_resp))
          if(length(resp_loc) > 1) {
            abort("The `resp()` function can only be used once per response variable. For multivariate modelling, use `vars()`.")
          }
          non_resp <- if(length(resp_loc) == 0) seq_along(args) else -resp_loc
          
          res <- map(args[non_resp], function(y) eval(y, envir = model$data, enclos = model$specials))
          len <- map_dbl(res, length)
          
          if(length(unique(len[len!=1])) > 1){
            abort(
              sprintf(
                "Response variable transformation has incompatible lengths, all arguments must be the length of the data %i or 1.",
                max(len)
              )
            )
          }
          
          # store length 1 arguments for transformation environment
          len1check <- function(len, arg) {
            (len == 1) && (is.name(arg) || (is.call(arg) && !(as_label(arg[[1]]) %in% "length")))
          }
          
          if(any(is_singular <- map2_lgl(len, args[non_resp], len1check))) {
            nm <- map_chr(args[non_resp][is_singular], as_label)
            args[non_resp][is_singular] <- syms(nm)
            len1vals[nm] <<- res[is_singular]
          }
          
          # handle unspecified response with equal length args
          if((length(resp_loc) == 0) && (sum(len == max(len)) > 1)) {
            return(list(call("resp", x)))
          }
          
          args
        },
        .h = function(x) {
          if(is_resp(x)){
            if(length(x[-1]) > 2) abort("The response variable accepts only one input. For multivariate modelling, use `vars()`.")
            list(response = sym(as_label(x[[2]])))
          } else list(x)
        },
        base = function(x) {
          is_syntactic_literal(x) || is_symbol(x) || is_resp(x)
        }
      )
      if("response" %in% names(path)) path <- path$response
      if(!is.list(path)) path <- list(path)
      list(path = path, len1vals = len1vals)
    }
  )
  
  # Obtain parsed out response variable
  responses <- map(traversed_lhs, function(x) x$path[[1]])
  responses <- map_chr(responses, as_label)
  
  # Obtain transformation expression applied to response variable
  transform_exprs <- lapply(traversed_lhs, function(x) x$path[[length(x$path)]])
  
  # Invert transformation applied to response variable
  inverse_exprs <- lapply(traversed_lhs, function(x){
    x <- rev(x$path)
    result <- x[[length(x)]]
    for (i in seq_len(length(x) - 1)){
      result <- undo_transformation(x[[i]], x[[i + 1]], result)
    }
    result
  })
  
  # Create evaluation environment for transformation functions
  # Includes cached values of single length arguments
  transform_args <- lapply(traversed_lhs, `[[`, "len1vals")
  envs <- lapply(transform_args, new_environment,  parent = model$env)
  
  # Produce transformation class functions for bt() usage
  make_transforms <- function(exprs, responses, envs){
    .mapply(
      function(x, response, env){
        new_function(args = set_names(list(missing_arg()), response), x, env = env)
      },
      dots = list(x = exprs, response = responses, env = envs),
      MoreArgs = list()
    )
  }
  
  transformations <- map2(
    make_transforms(transform_exprs, responses, envs),
    make_transforms(inverse_exprs, responses, envs),
    new_transformation
  )
  
  # Test combinations of transformations to see if they are valid
  comb_trans <- map_lgl(transformations, function(x) 
    length(all.names(body(x))) > length(all.vars(body(x))) + 1)
  map2(transformations[comb_trans], responses[comb_trans], function(trans, resp){
    dt <- model$data
    environment(trans) <- new_environment(dt, get_env(trans))
    inv <- invert_transformation(trans)
    environment(inv) <- new_environment(dt, get_env(inv))
    y <- eval_tidy(rlang::parse_expr(resp), data = dt, env = model$env)
    valid <- all.equal(y, inv(trans(y)))
    if(!isTRUE(valid)){
      abort(
"Could not identify a valid back-transformation for this transformation.
Please specify a valid form of your transformation using `new_transformation()`.")
    }
  })
  
  list(
    expressions = response_exprs,
    response = syms(responses),
    transformation = transformations
  )
}

#' Refit a mable to a new dataset
#' 
#' Applies a fitted model to a new dataset. For most methods this can be done
#' with or without re-estimation of the parameters.
#' 
#' @param object A mable.
#' @param new_data A tsibble dataset used to refit the model.
#' @param ... Additional optional arguments for refit methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' 
#' fit <- as_tsibble(mdeaths) %>% 
#'   model(ETS(value ~ error("M") + trend("A") + season("A")))
#' fit %>% report()
#'
#' fit %>% 
#'   refit(as_tsibble(fdeaths)) %>% 
#'   report(reinitialise = TRUE)
#' 
#' @rdname refit
#' @export
refit.mdl_df <- function(object, new_data, ...){
  mdls <- mable_vars(object)
  object <- bind_new_data(object, new_data)
  new_data <- object[["new_data"]]
  object[["new_data"]] <- NULL
  dplyr::mutate_at(object, vars(!!!mdls), refit, new_data, ...)
}

#' @export
refit.mdl_lst <- function(object, new_data, ...){
  attrb <- attributes(object)
  object <- map2(object, new_data, refit, ...)
  attributes(object) <- attrb
  object
}
#' @export
refit.lst_mdl <- deprecate_lst_mdl(refit.mdl_lst)

#' @rdname refit
#' @export
refit.mdl_ts <- function(object, new_data, ...){
  # Compute specials with new_data
  object$model$stage <- "refit"
  object$model$add_data(new_data)
  specials <- parse_model_rhs(object$model)
  object$model$remove_data()
  object$model$stage <- NULL
  
  resp <- map2(seq_along(object$response), object$response, function(i, resp){
    expr(object$transformation[[!!i]](!!resp))
  }) %>% 
    set_names(map_chr(object$response, as_string))
  
  new_data <- transmute(new_data, !!!resp)
  object$fit <- refit(object[["fit"]], new_data, specials = specials, ...)
  object$data <- new_data
  object
}

#' Interpolate missing values
#' 
#' Uses a fitted model to interpolate missing values from a dataset.
#' 
#' @param object A mable containing a single model column.
#' @param new_data A dataset with the same structure as the data used to fit the model.
#' @param ... Other arguments passed to interpolate methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' # The fastest running times for the olympics are missing for years during 
#' # world wars as the olympics were not held.
#' olympic_running
#' 
#' olympic_running %>% 
#'   model(TSLM(Time ~ trend())) %>% 
#'   interpolate(olympic_running)
#' 
#' @rdname interpolate
#' @export
interpolate.mdl_df <- function(object, new_data, ...){
  if(length(mable_vars(object)) > 1){
abort("Interpolation can only be done using one model. 
Please use select() to choose the model to interpolate with.")
  }
  
  object <- bind_new_data(object, new_data)
  kv <- key_vars(object)
  object <- transmute(as_tibble(object),
      !!!syms(kv),
      interpolated = map2(!!sym(mable_vars(object)), new_data, interpolate, ...)
    )
  unnest_tsbl(object, "interpolated", parent_key = kv)
}

#' @rdname interpolate
#' @export
interpolate.mdl_ts <- function(object, new_data, ...){
  # Compute specials with new_data
  object$model$stage <- "interpolate"
  object$model$add_data(new_data)
  specials <- tryCatch(parse_model_rhs(object$model),
                       error = function(e){
                         abort(sprintf(
                           "%s
Unable to compute required variables from provided `new_data`.
Does your interpolation data include all variables required by the model?", e$message))
                       }, interrupt = function(e) {
                         stop("Terminated by user", call. = FALSE)
                       })
  
  object$model$remove_data()
  object$model$stage <- NULL
  
  resp <- map2(seq_along(object$response), object$response, function(i, resp){
    expr(object$transformation[[!!i]](!!resp))
  }) %>% 
    set_names(map_chr(object$response, as_string))
  
  new_data <- transmute(new_data, !!!resp)
  new_data <- interpolate(object[["fit"]], new_data = new_data, specials = specials, ...)
  new_data[names(resp)] <- map2(new_data[names(resp)], object$transformation,
                                function(x, f) invert_transformation(f)(x))
  new_data
}

model_definition <- R6::R6Class(NULL,
  public = list(
    model = "Unknown model",
    specials = list(),
    formula = NULL,
    extra = NULL,
    origin = NULL,
    env = global_env(),
    check = function(.data){
    },
    prepare = function(...){
    },
    initialize = function(formula, ..., .env){
      self$formula <- enquo(formula)
      
      # self$env <- .env
      # Create specials environment with user's scoping
      specials_env <- new_environment(parent = .env)
      # Set `self` `super`, and `specials` in eval env for special functions
      fn_env <- new_environment(as.list(self$.__enclos_env__), self$env)
      env_bind(specials_env, !!!assign_func_envs(self$specials, fn_env))
      self$specials <- structure(
        specials_env,
        required_specials = self$specials%@%"required_specials",
        xreg_specials = self$specials%@%"xreg_specials"
      )
      
      # Define custom lag() xreg special for short term memory
      xreg_env <- get_env(self$specials$xreg)
      xreg_env$lag <- self$recall_lag
      
      self$prepare(formula, ...)
      
      self$extra <- list2(...)
    },
    recall_lag = function(x, n = 1L, ...){
      start <- NULL
      if(self$stage %in% c("generate", "forecast")){
        x_expr <- enexpr(x)
        start <- eval_tidy(x_expr, self$recent_data)
      }
      else if(self$stage == "estimate" && NROW(self$recent_data) < n){
        self$recent_data <- self$data[NROW(self$data) - n + seq_len(n),]
      }
      dplyr::lag(c(start, x), n = n, ...)[seq_along(x) + length(start)]
    },
    recent_data = NULL, # Used in short term memory for lagged operators
    stage = NULL, # Identifies the current operation
    data = NULL,
    add_data = function(.data){
      self$check(.data)
      # Add data origin if not yet known (fitting model)
      if(is.null(self$origin)) {
        self$origin <- .data[[index_var(.data)]][[1]]
      }
      self$data <- .data
    },
    remove_data = function(){
      self$data <- NULL
    },
    train = function(...){
      abort("This model has not defined a training method.")
    },
    print = function(...){
      cat(sprintf("<%s model definition>\n", self$model), sep = "")
    }
  ),
  lock_objects = FALSE
)

#' Create a new class of models
#' 
#' Suitable for extension packages to create new models for fable.
#' 
#' This function produces a new R6 model definition. An understanding of R6 is
#' not required, however could be useful to provide more sophisticated model
#' interfaces. All functions have access to `self`, allowing the functions for 
#' training the model and evaluating specials to access the model class itself.
#' This can be useful to obtain elements set in the %TODO
#' 
#' @param model The name of the model
#' @param train A function that trains the model to a dataset. `.data` is a tsibble
#' containing the data's index and response variables only. `formula` is the 
#' user's provided formula. `specials` is the evaluated specials used in the formula.
#' @param specials Special functions produced using [new_specials()]
#' @param check A function that is used to check the data for suitability with 
#' the model. This can be used to check for missing values (both implicit and 
#' explicit), regularity of observations, ordered time index, and univariate
#' responses.
#' @param prepare This allows you to modify the model class according to user
#' inputs. `...` is the arguments passed to `new_model_definition`, allowing
#' you to perform different checks or training procedures according to different
#' user inputs.
#' @param ... Further arguments to [R6::R6Class()]. This can be useful to set up
#' additional elements used in the other functions. For example, to use 
#' `common_xregs`, an `origin` element in the model is used to store
#' the origin for `trend()` and `fourier()` specials. To use these specials, you
#' must add an `origin` element to the object (say with `origin = NULL`).
#' @param .env The environment from which functions should inherit from.
#' @param .inherit A model class to inherit from.
#' 
#' @rdname new-model-class
#' 
#' @export
new_model_class <- function(model = "Unknown model", 
                            train = function(.data, formula, specials, ...) abort("This model has not defined a training method."),
                            specials = new_specials(),
                            check = function(.data){},
                            prepare = function(...){},
                            ...,
                            .env = caller_env(),
                            .inherit = model_definition){
  R6::R6Class(NULL, inherit = .inherit,
    public = list(
      model = model,
      train = train,
      specials = specials%||%new_specials(),
      check = check,
      prepare = prepare,
      env = .env,
      ...
    ),
    parent_env = env_bury(.env, .inherit = .inherit)
  )
}

#' @rdname new-model-class
#' @param formula The user's model formula.
#' @param .class A model class (typically created with [new_model_class()]).
#' @export
new_model_definition <- function(.class, formula, ..., .env = caller_env(n = 2)){
  add_class(.class$new({{formula}}, ..., .env = .env), "mdl_defn")
}

context("test-mable.R")

test_that("Mable classes", {
  skip_if_not_installed("fable")
  expect_s3_class(mbl, "mdl_df")
  expect_s3_class(mbl[[attr(mbl,"model")[[1]]]], "lst_mdl")
})

test_that("Mable print output", {
  skip_if_not_installed("fable")
  expect_output(print(mbl), "A mable:")
})

test_that("Mable fitted values", {
  skip_if_not_installed("fable")
  fits <- fitted(mbl)
  expect_true(is_tsibble(fits))
  expect_true(all(colnames(fits) %in% c(".model", "index", ".fitted")))
  expect_equal(fits[["index"]], us_deaths_tr[["index"]])
  expect_equal(
    fits[[".fitted"]],
    fitted(mbl[[attr(mbl,"model")[[1]]]][[1]])[[".fitted"]]
  )
  
  fits <- fitted(mbl_multi)
  expect_true(is_tsibble(fits))
  expect_equal(key_vars(fits), c("key", ".model"))
  expect_true(all(colnames(fits) %in% c("key", ".model", "index", ".fitted")))
  expect_equal(unique(fits[["key"]]), mbl_multi[["key"]])
  expect_equal(fits[["index"]], lung_deaths_long_tr[["index"]])
  expect_equal(fits[[".fitted"]],
               as.numeric(c(
                 fitted(mbl_multi[[attr(mbl,"model")[[1]]]][[1]])[[".fitted"]],
                 fitted(mbl_multi[[attr(mbl,"model")[[1]]]][[2]])[[".fitted"]]
               ))
  )
})

test_that("Mable residuals", {
  skip_if_not_installed("fable")
  resids <- residuals(mbl)
  expect_true(is_tsibble(resids))
  expect_true(all(colnames(resids) %in% c(".model", "index", ".resid")))
  expect_equal(resids[["index"]], us_deaths_tr[["index"]])
  expect_equal(resids[[".resid"]], as.numeric(residuals(mbl[[attr(mbl,"model")[[1]]]][[1]])[[".resid"]]))
  
  resids <- residuals(mbl_multi)
  expect_true(is_tsibble(resids))
  expect_equal(key_vars(resids), c("key", ".model"))
  expect_true(all(colnames(resids) %in% c("key", ".model", "index", ".resid")))
  expect_equal(unique(resids[["key"]]), mbl_multi[["key"]])
  expect_equal(resids[["index"]], lung_deaths_long_tr[["index"]])
  expect_equal(resids[[".resid"]], 
               as.numeric(c(
                 residuals(mbl_multi[[attr(mbl,"model")[[1]]]][[1]])[[".resid"]],
                 residuals(mbl_multi[[attr(mbl,"model")[[1]]]][[2]])[[".resid"]]
               ))
  )
})

test_that("mable dplyr verbs", {
  skip_if_not_installed("fable")
  library(dplyr)
  expect_output(mbl_complex %>% select(key, ets) %>% print, "mable: 2 x 2") %>% 
    colnames %>% 
    expect_identical(c("key", "ets"))
  
  expect_output(mbl_complex %>% select(key, ets) %>% print, "mable: 2 x 2") %>% 
    colnames %>% 
    expect_identical(c("key", "ets"))
  
  # Test for negative tidyselect with keyed data (#120)
  mbl_complex %>% 
    select(-lm) %>%
    colnames() %>% 
    expect_identical(c("key", "ets"))
  
  # expect_error(select(mbl_complex, -key),
  #              "not a valid mable")
  
  expect_output(mbl_complex %>% filter(key == "mdeaths") %>% print, "mable") %>% 
    .[["key"]] %>% 
    expect_identical("mdeaths")
})

#' Create a dable object
#'
#' A dable (decomposition table) data class (`dcmp_ts`) which is a tsibble-like
#' data structure for representing decompositions. This data class is useful for
#' representing decompositions, as its print method describes how its columns
#' can be combined to produce the original data, and has a more appropriate
#' `autoplot()` method for displaying decompositions provided by ggtime.
#' Beyond this, a dable (`dcmp_ts`) behaves very similarly to a tsibble (`tbl_ts`).
#'
#' @param ... Arguments passed to [tsibble::tsibble()].
#' @param response The name of the response variable column.
#' @param method The name of the decomposition method.
#' @param seasons A named list describing the structure of seasonal components
#' (such as `period`, and `base`).
#' @param aliases A named list of calls describing common aliases computed from
#' components.
#'
#' @export
dable <- function(..., response, method = NULL, seasons = list(), aliases = list()){
  build_dable(tsibble(...), method = method, response = !!enquo(response), 
              seasons = seasons, aliases = aliases)
}

#' Is the object a dable
#' 
#' @param x An object.
#' 
#' @export
is_dable <- function(x){
  inherits(x, "dcmp_ts")
}

#' Coerce to a dable object
#' 
#' @inheritParams as_fable
#' @param x Object to be coerced to a dable (`dcmp_ts`)
#' 
#' @rdname as-dable
#' @export
as_dable <- function(x, ...){
  UseMethod("as_dable")
}

#' @rdname as-dable
#' @export
as_dable.tbl_df <- function(x, response, method = NULL, seasons = list(), aliases = list(), ...){
  build_dable(x, method = method, response = !!enquo(response), 
              seasons = seasons, aliases = aliases)
}

#' @rdname as-dable
#' 
#' @inheritParams dable
#' 
#' @export
as_dable.tbl_ts <- function(x, response, method = NULL, seasons = list(), aliases = list(), ...){
  build_dable(x, method = method, response = !!enquo(response), 
              seasons = seasons, aliases = aliases)
}

build_dable <- function (x, response, method = NULL, seasons = list(), aliases = list()) {
  response <- names(x)[tidyselect::eval_select(enquo(response), x)]
  new_tsibble(x, method = method, response = response, 
              seasons = seasons, aliases = aliases, class = "dcmp_ts")
}

#' @export
as_tsibble.dcmp_ts <- function(x, ...){
  new_tsibble(x)
}

#' @export
`[.dcmp_ts` <- function (x, i, j, drop = FALSE){
  out <- NextMethod()
  # Drop dable if tsibble is dropped
  
  cn <- colnames(out)
  not_dable <- !(response_vars(x) %in% cn) || !is_tsibble(out)
  
  if(not_dable)
    return(out)
  else
    as_dable(out, response = response_vars(x), method = x%@%"method",
             seasons = x%@%"seasons", aliases = x%@%"aliases")
}

tbl_sum.dcmp_ts <- function(x){
  response <- response_vars(x)
  method <- expr_text((x%@%"aliases")[[response]])
  out <- NextMethod()
  names(out)[1] <- "A dable"
  if(!is.null(method)) {
    out[[length(out) + 1]] <- set_names(paste(response, method, sep = " = "),
                                        paste(x%@%"method", "Decomposition"))
  }
  out
}

#' @export
rbind.dcmp_ts <- function(...){
  deprecate_warn("0.2.0", "rbind.fbl_ts()", "bind_rows()")
  dots <- dots_list(...)
  
  attrs <- combine_dcmp_attr(dots)
  
  as_dable(invoke("rbind", map(dots, as_tsibble)), 
           method = attrs[["method"]], response = !!attrs[["response"]], 
           seasons = attrs[["seasons"]], aliases = attrs[["aliases"]])
}

combine_dcmp_attr <- function(lst_dcmp){
  resp <- map(lst_dcmp, response_vars)
  method <- map(lst_dcmp, function(x) x%@%"method")
  strc <- map(lst_dcmp, function(x) x%@%"seasons")
  aliases <- map(lst_dcmp, function(x) x%@%"aliases")
  if(length(resp <- unique(resp)) > 1){
    abort("Decomposition response variables must be the same for all models.")
  }
  
  strc <- unlist(unique(strc), recursive = FALSE)
  strc <- strc[!duplicated(names(strc))]
  
  aliases <- unlist(unique(aliases), recursive = FALSE)
  aliases <- split(aliases, names(aliases)) %>% 
    map(function(x){
      vars <- map(x, all.vars)
      x[[which.max(map_dbl(vars, length))]]
    })
  
  list(response = resp[[1]], method = paste0(unique(method), collapse = " & "),
       seasons = strc, aliases = aliases)
}

#' Create a fable object
#'
#' A fable (forecast table) data class (`fbl_ts`) which is a tsibble-like data
#' structure for representing forecasts. In extension to the key and index from 
#' the tsibble (`tbl_ts`) class, a fable (`fbl_ts`) must also contain a single
#' distribution column that uses values from the distributional package.
#'
#' @param ... Arguments passed to [tsibble::tsibble()].
#' @param response The character vector of response variable(s).
#' @param distribution The name of the distribution column (can be provided
#' using a bare expression).
#'
#' @export
fable <- function(..., response, distribution){
  # Create tsibble from ... inputs
  x <- tsibble(...)
  
  # If the response (from user input) needs converting
  response <- eval_tidy(enquo(response))
  distribution <- names(x)[tidyselect::eval_select(enquo(distribution), x)]
  
  build_fable(x, response, distribution)
}

#' Is the object a fable
#' 
#' @param x An object.
#' 
#' @export
is_fable <- function(x){
  inherits(x, "fbl_ts")
}

#' Coerce to a fable object
#' 
#' @inheritParams fable
#' @param x Object to be coerced to a fable (`fbl_ts`)
#' @param ... Additional arguments passed to methods
#' 
#' @rdname as-fable
#' @export
as_fable <- function(x, ...){
  UseMethod("as_fable")
}

#' @rdname as-fable
#' @export
as_fable.tbl_ts <- function(x, response, distribution, ...){
  # If the response (from user input) needs converting
  response <- eval_tidy(enquo(response))
  distribution <- names(x)[tidyselect::eval_select(enquo(distribution), x)]
  
  build_fable(x, response = response, distribution = distribution)
}

#' @rdname as-fable
#' @export
as_fable.grouped_ts <- as_fable.tbl_ts

#' @rdname as-fable
#' @export
as_fable.tbl_df <- function(x, response, distribution, ...){
  as_fable(
    as_tsibble(x, ...), 
    response = !!enquo(response),
    distribution = !!enquo(distribution)
  )
}

#' @rdname as-fable
#' @export
as_fable.fbl_ts <- function(x, response, distribution, ...){
  if(missing(response)){
    response <- response_vars(x)
  }
  else{
    response <- eval_tidy(enquo(response))
  }
  if(missing(distribution)){
    distribution <- distribution_var(x)
  }
  else{
    distribution <- names(x)[tidyselect::eval_select(enquo(distribution), x)]
  }
  build_fable(update_tsibble(x, ...),
              response = response, distribution = distribution)
}

#' @rdname as-fable
#' @export
as_fable.grouped_df <- as_fable.tbl_df

#' @inheritParams forecast.mdl_df
#' @rdname as-fable
#' @export
as_fable.forecast <- function(x, ..., point_forecast = list(.mean = mean)){
  if(is.null(x$upper)){
    # Without intervals, the best guess is the point forecast
    dist <- distributional::dist_degenerate(x$mean)
  } else {
    if(!is.null(x$lambda)){
      x$upper <- box_cox(x$upper, x$lambda)
      x$lower <- box_cox(x$lower, x$lambda)
    }
    warn("Assuming intervals are computed from a normal distribution.")
    level <- colnames(x$upper)[1]
    level <- as.numeric(gsub("^[^0-9]+|%", "", level))/100
    mid <- (x$upper[,1] - x$lower[,1])/2
    mu <- x$lower[,1] + mid
    sigma <- mid/(stats::qnorm((1+level)/2))
    dist <- distributional::dist_normal(mu = as.numeric(mu), sigma = as.numeric(sigma))
    if(!is.null(x$lambda)){
      dist <- distributional::dist_transformed(
        dist, 
        transform = rlang::new_function(exprs(x = ), expr(inv_box_cox(x, !!x$lambda)), env = rlang::pkg_env("fabletools")), 
        inverse = rlang::new_function(exprs(x = ), expr(inv_box_cox(x, !!x$lambda)), env = rlang::pkg_env("fabletools"))
      )
    }
  }
  out <- as_tsibble(x$mean)
  dimnames(dist) <- "value"
  out[["value"]] <- dist
  
  point_fc <- compute_point_forecasts(dist, point_forecast)
  out[names(point_fc)] <- point_fc
  
  build_fable(
    out,
    response = "value",
    distribution = "value"
  )
}

build_fable <- function (x, response, distribution) {
  if(is_grouped_ts(x)){
    fbl <- structure(x, class = c("grouped_fbl", "grouped_ts", "grouped_df", 
                                  "fbl_ts", "tbl_ts", "tbl_df", "tbl", "data.frame"),
                     response = response, dist = distribution,
                     model_cn = ".model")
  } else {
    fbl <- tsibble::new_tsibble(
      x, response = response, dist = distribution, model_cn = ".model",
      class = "fbl_ts")
  }
  if(is.null(dimnames(fbl[[distribution]]))) {
    warn("The dimnames of the fable's distribution are missing and have been set to match the response variables.")
    dimnames(fbl[[distribution]]) <- response
  }
  if(!identical(response, dimnames(fbl[[distribution]]))) {
    dimnames(fbl[[distribution]]) <- response
  }
  validate_fable(fbl)
  fbl
}

#' @export
as_tsibble.fbl_ts <- function(x, ...){
  new_tsibble(x)
}

#' @export
as_tsibble.grouped_fbl <- function(x, ...){
  structure(x, class=setdiff(class(x), c("grouped_fbl", "fbl_ts")),
            response = NULL, dist = NULL, model_cn = NULL)
}

#' @export
as_tibble.fbl_ts <- function(x, ...) {
  new_tibble(vec_data(x), nrow = nrow(x))
}

#' @export
as_tibble.grouped_fbl <- function(x, ...) {
  dplyr::new_grouped_df(as_tibble(vec_data(x)), groups = group_data(x))
}

validate_fable <- function(fbl){
  stopifnot(inherits(fbl, "fbl_ts"))
  chr_dist <- distribution_var(fbl)
  if (!(chr_dist %in% names(fbl))){
    abort(sprintf("Could not find distribution variable `%s` in the fable. A fable must contain a distribution, if you want to remove it convert to a tsibble with `as_tsibble()`.",
                  chr_dist))
  }
  vec_assert(fbl[[chr_dist]], distributional::new_dist(dimnames = response_vars(fbl)))
}

tbl_sum.fbl_ts <- function(x){
  out <- NextMethod()
  names(out)[1] <- "A fable"
  out
}

#' @importFrom distributional hilo
#' @export
hilo.fbl_ts <- function(x, level = c(80, 95), ...){
  hilo_exprs <- map(level,function(.x) expr(hilo(!!sym(distribution_var(x)), !!.x)))
  names(hilo_exprs) <- paste0(level, "%")
  as_tsibble(x) %>%
    mutate(!!!hilo_exprs)
}

restore_fable <- function(data, template){
  data_cols <- names(data)
  
  # Variables to keep
  fbl_vars <- setdiff(distribution_var(template), data_cols)
  res <- bind_cols(data, template[fbl_vars])
  
  build_fable(data, response = response_vars(template), distribution = distribution_var(template))
}

#' @export
select.fbl_ts <- function (.data, ...){
  res <- select(as_tsibble(.data), ...)
  restore_fable(res, .data)
}

#' @export
select.grouped_fbl <- select.fbl_ts

#' @export
transmute.fbl_ts <- function (.data, ...) {
  res <- transmute(as_tsibble(.data), ...)
  restore_fable(res, .data)
}

#' @export
transmute.grouped_fbl <- transmute.fbl_ts

#' @export
group_by.fbl_ts <- function(.data, ...) {
  build_fable(NextMethod(), response_vars(.data), distribution_var(.data))
}

#' @export
group_by.grouped_fbl <- group_by.fbl_ts

#' @export
ungroup.fbl_ts <- function(x, ...) {
  build_fable(NextMethod(), response_vars(x), distribution_var(x))
}

#' @export
ungroup.grouped_fbl <- function(x, ...) {
  build_fable(NextMethod(), response_vars(x), distribution_var(x))
}

#' @export
fill_gaps.fbl_ts <- function(.data, ..., .full = FALSE) {
  vec_restore(NextMethod(.data), .data)
}

#' @export
rbind.fbl_ts <- function(...){
  deprecate_warn("0.2.0", "rbind.fbl_ts()", "bind_rows()")
  fbls <- dots_list(...)
  response <- map(fbls, response_vars)
  dist <- map(fbls, distribution_var)
  if(length(response <- unique(response)) > 1){
    abort("Cannot combine fables with different response variables.")
  }
  if(length(dist <- unique(dist)) > 1){
    abort("Cannot combine fables with different distribution names.")
  }
  out <- suppressWarnings(invoke(bind_rows, map(fbls, as_tsibble)))
  build_fable(out, response[[1]], dist[[1]])
}

#' @export
`[.fbl_ts` <- function (x, i, j, drop = FALSE){
  out <- NextMethod()
  # Drop fable if tsibble is dropped
  
  cn <- colnames(out)
  not_fable <- !(distribution_var(x) %in% cn) || !is_tsibble(out)
  
  if(not_fable)
    return(out)
  else
    build_fable(out, response_vars(x), distribution_var(x))
}

type_sum.fbl_ts <- function(x){
  "fable"
}

#' Expand a dataset to include other levels of aggregation
#' 
#' Uses the structural specification given in `.spec` to aggregate a time
#' series. A grouped structure is specified using `grp1 * grp2`, and a nested 
#' structure is specified via `parent / child`. Aggregating the key structure is
#' commonly used with forecast reconciliation to produce coherent forecasts over
#' some hierarchy.
#' 
#' This function is experimental, and is subject to change in the future.
#' 
#' The way in which the measured variables are aggregated is specified in a
#' similar way to how `[dplyr::summarise()]` is used.
#' 
#' @param .data A tsibble.
#' @param .spec The specification of aggregation structure.
#' @inheritParams dplyr::summarise
#' 
#' @seealso 
#' [`reconcile()`], [`is_aggregated()`]
#' 
#' @examples 
#' library(tsibble)
#' tourism %>% 
#'   aggregate_key(Purpose * (State / Region), Trips = sum(Trips))
#' 
#' @export
aggregate_key <- function(.data, .spec, ...){
  UseMethod("aggregate_key")
}

#' @export
aggregate_key.tbl_ts <- function(.data, .spec = NULL, ...){#, dev = FALSE){
  .spec <- enexpr(.spec)
  if(is.null(.spec)){
    kv <- syms(key_vars(.data))
    message(
      sprintf("Key structural specification not found, defaulting to `.spec = %s`",
              paste(kv, collapse = "*"))
    )
    .spec <- reduce(kv, call2, .fn = "*")
  }

  key_comb <- parse_agg_spec(.spec)
  
  idx <- index2_var(.data)
  intvl <- interval(.data)
  kd <- key_data(.data)
  cn <- colnames(.data)
  has_varied_index <- any(has_gaps(.data, .full = TRUE)[[".gaps"]]) && !is_ordered(.data)
  .data <- as_tibble(.data)
  
  kv <- unique(unlist(key_comb, recursive = FALSE))
  # if(dev){
  #   .data <- map(unname(key_comb), function(x){
  #     agg <- summarise_alt(.data, ..., 
  #                          .grps = group_by_alt(.data, !!sym(idx), !!!syms(x)))
  #     agg[x] <- map(agg[x], agg_vec)
  #     agg_keys <- setdiff(kv, x)
  #     agg[agg_keys] <- rep(list(agg_vec(NA_character_, aggregated = TRUE)), length(agg_keys))
  #     agg
  #   })
  #   .data <- vctrs::vec_rbind(!!!.data)
  # } else {
  agg_dt <- map(unname(key_comb), function(x){
    gd <- group_data(group_by(.data, !!sym(idx), !!!set_names(map(x, function(.) expr(agg_vec(!!sym(.)))), x)))
    agg_keys <- setdiff(kv, x)
    agg_cols <- rep(list(agg_vec(NA, aggregated = TRUE)), length(agg_keys))
    gd[agg_keys] <- agg_cols
    gd[c(idx, kv, ".rows")]
  })
  agg_dt <- vctrs::vec_rbind(!!!agg_dt)
  .data <- dplyr::new_grouped_df(.data, groups = agg_dt)
  .data <- summarise(.data, ...)
  # }
  
  # Re-order columns into index, keys, values order
  .data <- .data[c(idx, kv, setdiff(colnames(.data), c(idx,kv)))]
  
  key_dt <- group_data(group_by(.data, !!!syms(kv)))
  .data <- ungroup(.data)
  
  # Return tsibble
  build_tsibble_meta(.data, key_data = key_dt, index = idx, 
                     index2 = as_string(idx), ordered = TRUE,
                     interval = intvl)
}

parse_agg_spec <- function(expr){
  # Key combinations
  tm <- stats::terms(new_formula(lhs = NULL, rhs = expr), env = empty_env())
  key_comb <- attr(tm, "factors")
  key_vars <- sub("^`(.*)`$", "\\1", rownames(key_comb))
  key_comb <- map(split(key_comb, col(key_comb)), function(x) key_vars[x!=0])
  if(attr(tm, "intercept")){
    key_comb <- c(list(chr()), key_comb)
  }
  unname(key_comb)
}


# #' @rdname aggregate_key
# #' 
# #' @param .times Temporal aggregations to include. The default (NULL) will
# #' automatically identify appropriate temporal aggregations. This can be specified
# #' as a list of [`lubridate::period()`] elements, or a character vector describing the
# #' temporal aggregations.
# #' 
# #' @examples
# #' library(tsibble)
# #' pedestrian %>% 
# #'   aggregate_index()
# aggregate_index <- function(.data, .times, ...){
#   UseMethod("aggregate_index")
# }
# 
# #' @export
# aggregate_index.tbl_ts <- function(.data, .times = NULL, ...){
#   warn("Temporal aggregation is highly experimental. The interface will be refined in the near future.")
#   
#   require_package("lubridate")
#   idx <- index(.data)
#   kv <- key_vars(.data)
#   
#   # Parse times as lubridate::period
#   if(is.null(.times)){
#     interval <- with(interval(.data), lubridate::years(year) + 
#            lubridate::period(3*quarter + month, units = "month") + lubridate::weeks(week) +
#            lubridate::days(day) + lubridate::hours(hour) + lubridate::minutes(minute) + 
#            lubridate::seconds(second) + lubridate::milliseconds(millisecond) + 
#            lubridate::microseconds(microsecond) + lubridate::nanoseconds(nanosecond))
#     periods <- common_periods(.data)
#     .times <- c(set_names(names(periods), names(periods)), list2(!!format(interval(.data)) := interval))
#   }
#   .times <- set_names(map(.times, lubridate::as.period), names(.times) %||% .times)
#   
#   secs <- map_dbl(.times, lubridate::period_to_seconds)
#   .times <- .times[order(secs, decreasing = TRUE)]
#   
#   # Temporal aggregations
#   .data <- as_tibble(.data)
#   agg_dt <- vctrs::vec_rbind(
#     !!!map(seq_along(.times), function(tm){
#       group_data(
#         group_by(.data,
#                  !!!set_names(names(.times), names(.times))[seq_len(tm-1) + 1],
#                  !!as_string(idx) := lubridate::floor_date(!!idx, .times[[tm]]),
#                  !!!syms(kv))
#       )
#     })
#   )
#   kv <- setdiff(colnames(agg_dt), c(as_string(idx), ".rows"))
#   agg_dt <- agg_dt[c(as_string(idx), kv, ".rows")]
#   
#   .data <- dplyr::new_grouped_df(.data, groups = agg_dt)
#   
#   # Compute aggregates and repair index attributes
#   idx_attr <- attributes(.data[[as_string(idx)]])
#   .data <- ungroup(summarise(.data, ...))
#   attributes(.data[[as_string(idx)]]) <- idx_attr
#   
#   # Return tsibble
#   as_tsibble(.data, key = kv, index = !!idx) %>% 
#     mutate(!!!set_names(map(kv, function(x) expr(agg_vec(!!sym(x)))), kv))
# }

#' Create an aggregation vector
#' 
#' \lifecycle{maturing}
#' 
#' An aggregation vector extends usual vectors by adding `<aggregated>` values.
#' These vectors are typically produced via the [`aggregate_key()`] function,
#' however it can be useful to create them manually to produce more complicated
#' hierarchies (such as unbalanced hierarchies).
#' 
#' @param x The vector of values.
#' @param aggregated A logical vector to identify which values are `<aggregated>`.
#' 
#' @examples
#' agg_vec(
#'   x = c(NA, "A", "B"),
#'   aggregated = c(TRUE, FALSE, FALSE)
#' )
#' 
#' @export
agg_vec <- function(x = character(), aggregated = logical(vec_size(x))){
  is_agg <- is_aggregated(x)
  if (inherits(x, "agg_vec")) x <- field(x, "x")
  x[is_agg] <- NA
  vec_assert(aggregated, ptype = logical())
  vctrs::new_rcrd(list(x = x, agg = is_agg | aggregated), class = "agg_vec")
}

#' @export
format.agg_vec <- function(x, ..., agg_chr = "<aggregated>"){
  n <- vec_size(x)
  x <- vec_data(x)
  is_agg <- x[["agg"]]
  out <- character(length = n)
  out[is_agg] <- agg_chr
  out[!is_agg] <- format(x[["x"]][!is_agg], ...)
  out 
}

pillar_shaft.agg_vec <- function(x, ...) {
  if(requireNamespace("crayon")){
    agg_chr <- crayon::style("<aggregated>", crayon::make_style("#999999", grey = TRUE))
  }
  else{
    agg_chr <- "<aggregated>"
  }
  
  out <- format(x, agg_chr = agg_chr)
  
  pillar::new_pillar_shaft_simple(out, align = "left", min_width = 10)
}

#' Internal vctrs methods
#'
#' These methods are the extensions that allow aggregation vectors to work with
#' vctrs.
#'
#' @keywords internal
#' @name aggregation-vctrs
NULL

#' @rdname aggregation-vctrs
#' @importFrom vctrs vec_ptype2
#' @method vec_ptype2 agg_vec
#' @export
vec_ptype2.agg_vec <- function(x, y, ...) UseMethod("vec_ptype2.agg_vec", y)
#' @rdname aggregation-vctrs
#' @export
vec_ptype2.agg_vec.agg_vec <- function(x, y, ...) {
  x <- vec_data(x)[["x"]]
  y <- vec_data(y)[["x"]]
  ptype <- if(!is_logical(x) && !is_logical(y)) {
    vec_ptype2(x, y)
  } else if (is_logical(x)) {
    y
  } else {
    x
  }
  agg_vec(ptype)
}
#' @rdname aggregation-vctrs
#' @export
vec_ptype2.agg_vec.default <- function(x, y, ...) agg_vec()
#' @rdname aggregation-vctrs
#' @export
vec_ptype2.agg_vec.character <- function(x, y, ...) agg_vec()
#' @rdname aggregation-vctrs
#' @export
vec_ptype2.character.agg_vec <- function(x, y, ...) agg_vec()

#' @rdname aggregation-vctrs
#' @export
vec_ptype_abbr.agg_vec <- function(x, ...) {
  paste0(vctrs::vec_ptype_abbr(vec_data(x)[["x"]], ...), "*")
}

#' @rdname aggregation-vctrs
#' @method vec_cast agg_vec
#' @export
vec_cast.agg_vec <- function(x, to, ...) UseMethod("vec_cast.agg_vec")
#' @rdname aggregation-vctrs
#' @export
vec_cast.agg_vec.agg_vec <- function(x, to, ...) {
  x <- vec_proxy(x)
  if(all(x$agg)) x$x <- vec_rep(vec_cast(NA, vec_proxy(to)$x), length(x$x))
  vec_restore(x, to)
}
#' @rdname aggregation-vctrs
#' @export
vec_cast.agg_vec.default <- function(x, to, ...) agg_vec(x)
#' @export
vec_cast.agg_vec.character <- function(x, to, ...) agg_vec(x)
#' @rdname aggregation-vctrs
#' @export
vec_cast.character.agg_vec <- function(x, to, ...) trimws(format(x))

#' @rdname aggregation-vctrs
#' @export
vec_proxy_compare.agg_vec <- function(x, ...) {
  vec_proxy(x)[c(2,1)]
}

#' @export
`==.agg_vec` <- function(e1, e2){
  e1_agg <- inherits(e1, "agg_vec")
  e2_agg <- inherits(e2, "agg_vec")
  
  if(!e1_agg || !e2_agg){
    x <- list(e1,e2)[[which(!c(e1_agg, e2_agg))]]
    is_agg <- x == "<aggregated>"
    if(any(is_agg)){
      warn("<aggregated> character values have been converted to aggregated values.
Hint: If you're trying to compare aggregated values, use `is_aggregated()`.")
    }
    x <- agg_vec(ifelse(is_agg, NA, x), aggregated = is_agg)
    if(!e1_agg) e1 <- x else e2 <- x
  }
  
  x <- vec_recycle_common(e1, e2)
  e1 <- vec_proxy(x[[1]])
  e2 <- vec_proxy(x[[2]])
  out <- logical(vec_size(e1))
  (e1$agg & e2$agg) | vec_equal(e1$x, e2$x, na_equal = TRUE)
}

#' @export
`!=.agg_vec` <- function(e1, e2) {
  !(e1 == e2)
}

#' @export
is.na.agg_vec <- function(x) {
  is.na(field(x, "x")) & !field(x, "agg")
}

#' @importFrom dplyr recode
#' @export
recode.agg_vec <- function(.x, ...) {
  field(.x, "x") <- recode(field(.x, "x"), ...)
  .x
}

#' Is the element an aggregation of smaller data
#' 
#' @param x An object.
#' 
#' @seealso [`aggregate_key`]
#' 
#' @export
is_aggregated <- function(x){
  if(!inherits(x, "agg_vec")){
    logical(vec_size(x))
  } else {
    vec_proxy(x)[["agg"]]
  }
}

scale_type.agg_vec <- function(x) {
  abort("Cannot add an aggregated vector to a plot, use format() to plot with your aggregations.")
}
# # Space efficient group identifiers, storing group positions as vectors instead of lists
# group_by_alt <- function(.data, ...){
#   .data <- transmute(as_tibble(.data), ...)
#   grps <- vctrs::vec_group_id(.data)
#   list(pos = order(grps, method = "radix"), len = tabulate(grps), var = names(.data))
# }
# 
# # Simplified summarise(), which uses the group identifiers from group_by_alt()
# summarise_alt <- function(.data, ..., .grps = group_by_alt(.data)){
#   .data <- as.data.frame(.data)
#   dots <- enquos(..., .named = TRUE)
#   dots_names <- names(dots)
#   
#   grp_pos <- .grps[["pos"]]
#   grp_len <- .grps[["len"]]
#   
#   n_grps <- length(grp_len)
#   grp_start <- 1 + cumsum(c(0, grp_len[-n_grps]))
#   out <- vec_slice(.data[.grps[["var"]]], grp_pos[grp_start])
#   
#   # Promise based group aware data mask (adapted from dplyr:::DataMask)
#   bindings <- env(empty_env())
#   resolved <- rep_len(TRUE, ncol(.data))
#   promise_fn <- function(index) {
#     resolved[[index]] <<- TRUE
#     vec_slice(.subset2(.data, index), rows)
#   }
#   promise_env <- get_env(promise_fn)
#   nm <- names2(.data)
#   promises <- map(seq_len(ncol(.data)), function(.x) expr(delayedAssign(!!nm[[.x]], promise_fn(!!.x), env, bindings)))
#   env <- current_env()
#   # env_bind_lazy(bindings, !!!promises)
#   mask <- new_data_mask(bindings)
#   mask$.data <- as_data_pronoun(mask)
#   
#   # Compute dots over groups
#   for(i in seq_along(dots)){
#     res <- NULL
#     dot <- dots[[i]]
#     for(grp in seq_len(n_grps)){
#       # Set groups and reset resolved promises
#       size <- grp_len[grp]
#       idx <- grp_start[grp]
#       promise_env$rows <- .subset(grp_pos, idx:(idx+size-1))
#       for(j in which(resolved)){
#         eval(promises[[j]])
#         resolved[j] <- FALSE
#       }
#       
#       # Compute and store calculation
#       val <- eval_tidy(dot, mask)
#       if(is.null(res)){
#         res <- vec_init(val, n = n_grps)
#       }
#       res[grp] <- val
#     }
#     out[names(dots)[[i]]] <- res
#   }
#   out
# }

# Adapted from cut.Date
date_breaks <- function(x, breaks, start_monday = TRUE, offset = TRUE){
  # Currently only dates are supported
  x <- as.Date(x)
  by2 <- strsplit(breaks, " ", fixed = TRUE)[[1L]]
  if (length(by2) > 2L || length(by2) < 1L) 
    stop("invalid specification of 'breaks'")
  valid <- pmatch(by2[length(by2)], c("days", "weeks", 
                                      "months", "years", "quarters"))
  if (is.na(valid)) 
    stop("invalid specification of 'breaks'")
  start <- as.POSIXlt(min(x, na.rm = TRUE))
  if (valid == 1L) 
    incr <- 1L
  if (valid == 2L) {
    start$mday <- start$mday - start$wday
    if (start_monday) 
      start$mday <- start$mday + ifelse(start$wday > 
                                          0L, 1L, -6L)
    start$isdst <- -1L
    incr <- 7L
  }
  if (valid == 3L) {
    start$mday <- 1L
    start$isdst <- -1L
    end <- as.POSIXlt(max(x, na.rm = TRUE))
    step <- if (length(by2) == 2L) 
      as.integer(by2[1L])
    else 1L
    end <- as.POSIXlt(end + (31 * step * 86400))
    end$mday <- 1L
    end$isdst <- -1L
    breaks <- as.Date(seq(start, end, breaks))
  }
  else if (valid == 4L) {
    start$mon <- 0L
    start$mday <- 1L
    start$isdst <- -1L
    end <- as.POSIXlt(max(x, na.rm = TRUE))
    step <- if (length(by2) == 2L) 
      as.integer(by2[1L])
    else 1L
    end <- as.POSIXlt(end + (366 * step * 86400))
    end$mon <- 0L
    end$mday <- 1L
    end$isdst <- -1L
    breaks <- as.Date(seq(start, end, breaks))
  }
  else if (valid == 5L) {
    qtr <- rep(c(0L, 3L, 6L, 9L), each = 3L)
    start$mon <- qtr[start$mon + 1L]
    start$mday <- 1L
    start$isdst <- -1L
    maxx <- max(x, na.rm = TRUE)
    end <- as.POSIXlt(maxx)
    step <- if (length(by2) == 2L) 
      as.integer(by2[1L])
    else 1L
    end <- as.POSIXlt(end + (93 * step * 86400))
    end$mon <- qtr[end$mon + 1L]
    end$mday <- 1L
    end$isdst <- -1L
    breaks <- as.Date(seq(start, end, paste(step * 3L, 
                                            "months")))
    lb <- length(breaks)
    if (maxx < breaks[lb - 1]) 
      breaks <- breaks[-lb]
  }
  else {
    start <- as.Date(start)
    if (length(by2) == 2L) 
      incr <- incr * as.integer(by2[1L])
    maxx <- max(x, na.rm = TRUE)
    breaks <- seq(start, maxx + incr, breaks)
    breaks <- breaks[seq_len(1L + max(which(breaks <= 
                                              maxx)))]
  }
  if(offset == "end" || (is.logical(offset) && offset)) {
    breaks <- breaks + (x[length(x)] - breaks[length(breaks)])
  } else if (offset == "start") {
    breaks <- breaks + (x[1] - breaks[1])
  }
  breaks
}

bin_date <- function(time, breaks, offset){
  if(is.character(breaks) && length(breaks) == 1){
    breaks <- date_breaks(time, breaks, offset = offset)
  }
  bincode <- .bincode(unclass(as.Date(time)), unclass(breaks), right = FALSE)
  list(
    bin = bincode,
    breaks = breaks,
    complete_size = diff(breaks)
  )
}

#' Expand a dataset to include temporal aggregates
#' 
#' \lifecycle{experimental}
#' 
#' This feature is very experimental. It currently allows for temporal 
#' aggregation of daily data as a proof of concept.
#' 
#' @inheritParams aggregate_key
#' @param .window Temporal aggregations to include. The default (NULL) will
#' automatically identify appropriate temporal aggregations. This can be 
#' specified in several ways (see details).
#' @param .offset Offset the temporal aggregation windows to align with the start
#' or end of the data. If FALSE, no offset will be applied (giving common 
#' breakpoints for temporal bins.)
#' @param .bin_size Temporary. Define the number of observations in each temporal bucket
#' 
#' @details 
#' The aggregation `.window` can be specified in several ways:
#' * A character string, containing one of "day", "week", "month", "quarter" or 
#' "year". This can optionally be preceded by a (positive or negative) integer
#'  and a space, or followed by "s".
#' * A number, taken to be in days.
#' * A [`difftime`] object.
#' 
#' @examples
#' library(tsibble)
#' pedestrian %>%
#'   # Currently only supports daily data
#'   index_by(Date) %>% 
#'   dplyr::summarise(Count = sum(Count)) %>% 
#'   # Compute weekly aggregates
#'   fabletools:::aggregate_index("1 week", Count = sum(Count))
aggregate_index <- function(.data, .window, ..., .offset = "end", .bin_size = NULL){
  idx <- index_var(.data)
  # Compute temporal bins and bin sizes
  new_index <- bin_date(.data[[idx]], .window, .offset)
  if(!is.null(.bin_size)) new_index$complete_size <- vec_recycle(.bin_size, length(new_index$complete_size))
  as_tibble(.data) %>% 
    # Compute groups of temporal bins
    group_by(
      !!idx := !!{new_index$breaks[new_index$bin]},
      !!!key(.data)
    ) %>% 
    # Keep only complete windows, currently assumes daily base interval
    filter(dplyr::n() == (!!new_index$complete_size)[match((!!sym(idx))[1], !!new_index$breaks)]) %>%
    # Compute aggregates
    summarise(..., .groups = "drop") %>% 
    # Rebuild tsibble
    as_tsibble(key = key_vars(.data), index = !!index(.data))
}

list_of_models <- function(x = list()){
  vctrs::new_vctr(x, class = c("mdl_lst", "lst_mdl"))
}

type_sum.mdl_lst <- function(x){
  "model"
}

#' @export
format.mdl_lst <- function(x, ...){
  map_chr(x, function(x) paste0("<", model_sum(x), ">"))
}
#' @export
format.lst_mdl <- deprecate_lst_mdl(format.mdl_lst)

#' @export
vec_cast.character.mdl_lst <- function(x, to, ...) format(x)
#' @export
vec_cast.character.lst_mdl <- deprecate_lst_mdl(vec_cast.character.mdl_lst)

#' @export
vec_ptype2.mdl_lst.mdl_lst <- function(x, y, ...){
  list_of_models()
}

#' @export
vec_cast.mdl_lst.mdl_lst <- function(x, to, ...){
  if(!identical(class(x), class(to))){
    abort("Cannot combine model lists with different reconciliation strategies.")
  }
  x
}

#' Structural (summing) matrix for coherent time series
#'
#' Constructs the \eqn{n \times n_b} structural matrix \eqn{\mathbf{S}} for a
#' set of linearly related time series, where \eqn{n} is the total number
#' of series and \eqn{n_b} is the number of bottom-level series.
#'
#' Let \eqn{\boldsymbol{b}_t} be the \eqn{n_b}-vector of bottom-level series at
#' time \eqn{t}, and \eqn{\boldsymbol{a}_t = \mathbf{A}\boldsymbol{b}_t} be the
#' \eqn{n_a = n - n_b} aggregated series, where \eqn{\mathbf{A}} is the
#' \eqn{n_a \times n_b} aggregation matrix. The full \eqn{n}-vector of series is
#' \eqn{\boldsymbol{y}_t = [\boldsymbol{a}_t', \boldsymbol{b}_t']'}, and the
#' structural matrix satisfies
#'
#' \deqn{\boldsymbol{y}_t = \mathbf{S}\boldsymbol{b}_t, \quad
#'   \mathbf{S} = \begin{bmatrix} \mathbf{A} \\ \mathbf{I}_{n_b} \end{bmatrix}.}
#'
#' Any linear reconciliation method can be written as
#' \eqn{\tilde{\boldsymbol{y}}_h = \mathbf{S}\mathbf{G}\hat{\boldsymbol{y}}_h},
#' where \eqn{\hat{\boldsymbol{y}}_h} are the \eqn{h}-step base forecasts.
#' Optimal reconciled forecasts use
#' \eqn{\mathbf{G} = (\mathbf{S}'\mathbf{W}^{-1}\mathbf{S})^{-1}\mathbf{S}'\mathbf{W}^{-1}},
#' with \eqn{\mathbf{W}} a positive definite weight matrix (see
#' [`min_trace()`]).
#'
#' @param data A data object which contains linearly related coherent structures.
#' @param sparse If `TRUE`, a sparse matrix (class `"dgCMatrix"` from the
#'   \pkg{Matrix} package) is returned. Defaults to `FALSE`.
#'
#' @return An \eqn{n \times n_b} matrix (dense or sparse) encoding the
#'   structural relationships among all series.
#'
#' @seealso [coherent_cmat()] for the corresponding zero-constraint matrix
#'   \eqn{\mathbf{C}} and [aggregate_key()] for computing cross-sectional
#'   aggregations with tsibble data sets.
#'
#' @references
#' Hyndman, R. J., & Athanasopoulos, G. (2022). *Notation for forecast
#' reconciliation*. <https://robjhyndman.com/hyndsight/reconciliation-notation.html>
#'
#' @examples
#' tsibble::tourism %>% 
#'   aggregate_key(Purpose, Trips = sum(Trips)) %>%
#'   coherent_smat()
#' 
#' @export
coherent_smat <- function(data, sparse = FALSE) {
  UseMethod("coherent_smat")
}

#' @export
coherent_smat.tbl <- function(data, sparse = FALSE) {
  if (names(data)[[ncol(data)]] != ".rows") {
    tryCatch(
      data <- key_data(data),
      error = function(e) {
        cli::cli_abort(
          "The data must be a `key_data` object, or an object with a suitable `key_data()` method.",
          call = NULL
        )
      }
    )
  }

  spos <- build_key_data_smat(data)

  if (sparse) {
    row_btm <- spos$leaf
    row_agg <- seq_len(nrow(data))[-row_btm]
    Matrix::sparseMatrix(
      i = rep(seq_along(spos$agg), lengths(spos$agg)),
      j = unlist(spos$agg),
      x = rep(1, sum(lengths(spos$agg)))
    )
  } else {
    out <- matrix(0L, nrow = nrow(data), ncol = length(spos$leaf))
    out[nrow(data)*(unlist(spos$agg)-1) + rep(seq_along(spos$agg), lengths(spos$agg))] <- 1L
    out
  }
}

#' @export
coherent_smat.default <- function(data, sparse = FALSE) {
  coherent_smat(key_data(data), sparse = sparse)
}

#' Zero-constraint matrix for coherent time series
#'
#' Constructs the \eqn{n_a \times n} zero-constraint matrix \eqn{\mathbf{C}}
#' for a set of linearly related time series, where \eqn{n_a = n - n_b}
#' is the number of aggregated series.
#'
#' Given the aggregation matrix \eqn{\mathbf{A}} (see [`coherent_smat()`]), the
#' constraint matrix is defined as
#'
#' \deqn{\mathbf{C} = [\mathbf{I}_{n_a} \;\; {-\mathbf{A}}],}
#'
#' so that \eqn{\mathbf{C}\boldsymbol{y}_t = \boldsymbol{0}_{n_a}} for all
#' coherent vectors \eqn{\boldsymbol{y}_t}. This zero-constrained representation
#' yields the mapping matrix
#'
#' \deqn{\mathbf{M} = \mathbf{I}_n -
#'   \mathbf{W}\mathbf{C}'(\mathbf{C}\mathbf{W}\mathbf{C}')^{-1}\mathbf{C},}
#'
#' which requires inverting an \eqn{n_a \times n_a} matrix rather than the
#' \eqn{n_b \times n_b} matrix in the structural form, and is therefore more
#' efficient when \eqn{n_a < n_b}.
#'
#' @inheritParams coherent_smat
#'
#' @return An \eqn{n_a \times n} matrix (dense or sparse) whose rows encode
#'   each aggregation constraint as \eqn{\mathbf{C}\boldsymbol{y}_t =
#'   \boldsymbol{0}_{n_a}}.
#'
#' @seealso [coherent_smat()] for the corresponding structural matrix
#'   \eqn{\mathbf{S}} and [aggregate_key()] for computing cross-sectional
#'   aggregations with tsibble data sets.
#' 
#' @references
#' Hyndman, R. J., & Athanasopoulos, G. (2022). *Notation for forecast
#' reconciliation*. <https://robjhyndman.com/hyndsight/reconciliation-notation.html>
#'
#' Di Fonzo, T., & Girolimetto, D. (2021). Cross-temporal forecast
#' reconciliation: Optimal combination method and heuristic alternatives.
#' *International Journal of Forecasting*. \doi{10.1016/j.ijforecast.2021.08.004}
#'
#' @export
coherent_cmat <- function(data, sparse = FALSE) {
  UseMethod("coherent_cmat")
}

#' @export
coherent_cmat.default <- function(data, sparse = FALSE) {
  S <- coherent_smat(data, sparse = sparse)
  row_btm <- apply(S, 1, sum) == 1L
  row_agg <- which(!row_btm)
  row_btm <- which(row_btm)
  if (sparse) diag <- Matrix::Diagonal

  U <- cbind(
    diag(diff(rev(dim(S)))),
    -S[row_agg,,drop = FALSE]
  )
  U[, order(c(row_agg, row_btm)), drop = FALSE]
}

guess_response <- function(.data){
  all_vars <- custom_error(
    measured_vars,
    "This model function does not support automatic selection of response variables. Please specify this in the model formula."
  )(.data)
  
  if(length(all_vars)!=1){
    abort("Could not automatically determine the response variable, please provide the response variable in the model specification")
  }
  
  out <- sym(all_vars[[1]])
  inform(sprintf(
    "Model not specified, defaulting to automatic modelling of the `%s` variable. Override this using the model formula.",
    expr_name(out)
  ))
  out
}

# Lookup table for function inverses
inverse_table <- function() {
  table <- new.env(parent = emptyenv())
  list(
    add = function(ns, fn, inv) {
      table[[ns]] <- as.list(table[[ns]])
      table[[ns]][[fn]] <- inv
    },
    
    get = function(ns, fn) {
      ns_name <- environmentName(ns)
      if(nchar(ns_name) == 0){
        ns_name <- "base"
      }
      ret <- table[[ns_name]][[fn]]
      if (is.null(ret)) {
        t_fn <- get(fn, envir = ns)
        if(inherits(t_fn, "transformation")){
          ret <- function(operation, target, result){
            args <- call_args(operation)
            target_pos <- match(list(target), args)
            call2(expr(invert_transformation(!!t_fn)), !!!replace(args, target_pos, list(result)))
          }
        }
        else{
          abort(sprintf("No supported inverse for the `%s` transformation.", fn))
        }
      }
      ret
    })
}

undo_transformation <- function(operation, target, result){
  fn <- call_name(operation)
  env <- get_env(operation, caller_env())
  ns <- eval_tidy(expr(environment(get(!!fn))), env = env)
  inverse_table$get(ns, fn)(operation, get_expr(target), result)
}

traverse_transformation <- function(transformation){
  transformation <- enquo(transformation)
  
  # Evaluate transformation stack via call traversal along longest argument (hopefully, the data)
  traverse_call(!!transformation,
                .f = function(.x, .y) append(.y, .x[[1]]),
                .g = function(.x) {
                  .x %>%
                  get_expr %>%
                  as.list %>% 
                  map(new_quosure, env = get_env(.x)) %>%
                  .[which.max(map(., function(.x) length(eval_tidy(.x))))]
                },
                .h = list)
}

#' Create a new modelling transformation
#' 
#' Produces a new transformation for fable modelling functions which will be used to transform, back-transform, and adjust forecasts.
#' 
#' For more details about transformations, read the vignette:
#' `vignette("transformations", package = "fable")`
#' 
#' @param transformation A function which transforms the data
#' @param inverse A function which is the inverse of a transformation
#' 
#' @examples
#' 
#' scaled_logit <- function(x, lower=0, upper=1){
#'   log((x-lower)/(upper-x))
#' }
#' inv_scaled_logit <- function(x, lower=0, upper=1){
#'   (upper-lower)*exp(x)/(1+exp(x)) + lower
#' }
#' my_scaled_logit <- new_transformation(scaled_logit, inv_scaled_logit)
#' 
#' t_vals <- my_scaled_logit(1:10, 0, 100)
#' t_vals
#' 
#' @export
new_transformation <- function(transformation, inverse){
  structure(transformation, class = "transformation", inverse = inverse)
}

#' Bias adjust back-transformation functions
#' 
#' To produce forecast means (instead of forecast medians) it is necessary to adjust the back-transformation function relative to the forecast variance.
#' 
#' More details about bias adjustment can be found in the transformations vignette: read the vignette:
#' `vignette("transformations", package = "fable")`
#' 
#' @param bt The back-transformation function
#' @param sd The forecast standard deviation
#' 
#' @examples 
#' 
#' adj_fn <- bias_adjust(function(x) exp(x), 1:10)
#' y <- rnorm(10)
#' exp(y)
#' adj_fn(y)
#' 
#' @keywords internal
#' @export
bias_adjust <- function(bt, sd){
  fvar <- sd^2
  if(any(is.na(fvar))){
    warn("Could not bias adjust the point forecasts as the forecast standard deviation is unknown. Perhaps your series is too short or insufficient bootstrap samples are used.")
    return(bt)
  }
  if(is.name(body(bt))){
    return(bt)
  }
  function(x){
    h <- .Machine$double.eps^(1/4)
    f0 <- bt(x)
    
    # Set up hessian function
    .x_var <- sym(names(formals(bt)))
    bt_hessian <- bt
    
    # Try for symbolic derivatives
    body(bt_hessian) <- possibly(stats::deriv, NULL)(
      body(bt), names(formals(bt))[1], hessian = TRUE)
    
    # Fall back to numerical derivatives
    if(is.null(body(bt_hessian))){
      body(bt_hessian) <- expr({
        h <- abs(1e-4 * !!.x_var) + 1e-4 * (abs(!!.x_var) < sqrt(.Machine$double.eps/7e-07))
        structure(list(), hessian = (bt(!!.x_var + h) - 2 * f0 + bt(!!.x_var - h))/h^2)
      })
      environment(bt_hessian) <- env_bury(get_env(bt), bt = bt, f0 = f0, h = h)
    }
    
    hessian <- as.numeric(attr(bt_hessian(x), "hessian"))

    if(any(!is.na(f0) & !is.finite(hessian))){
      warn("Could not bias adjust the point forecasts as the back-transformation's hessian is not well behaved. Consider using a different transformation.")
      return(f0)
    }
    f0 + fvar/2*hessian
  }
}

#' @export
print.transformation <- function(x, ...){
  cat("Transformation: ", expr_text(body(x)), "\n",
      "Backtransformation: ", expr_text(body(x%@%"inverse")), sep="")
}

#' @rdname new_transformation
#' @param x A transformation (such as one created with `new_transformation`).
#' @param ... Further arguments passed to other methods.
#' @export
invert_transformation <- function(x, ...){
  UseMethod("invert_transformation")
}

#' @export
invert_transformation.transformation <- function(x, ...){
  new_transformation(attr(x, "inverse"), `attributes<-`(x, NULL))
}

inverse_table <- inverse_table()

map(c("log", "logb"),
    function(.x) inverse_table$add("base", .x, 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    if(length(args) == 1){
                      expr(exp(!!result))
                    }
                    else{
                      expr((!!args[[2]]) ^ !!result)
                    }
                  }
    )
)

inverse_table$add("base", "log10",
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(10^!!result)
                  }
)

inverse_table$add("base", "log2",
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(2^!!result)
                  }
)

inverse_table$add("base", "log1p",
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(expm1(!!result))
                  }
)

inverse_table$add("base", "expm1",
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(log1p(!!result))
                  }
)

inverse_table$add("base", "exp", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(log(!!!replace(args, target_pos, list(result))))
                  }
)

inverse_table$add("fabletools", "box_cox", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(inv_box_cox(!!!replace(args, target_pos, list(result))))
                  }
)

inverse_table$add("fabletools", "inv_box_cox", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(box_cox(!!!replace(args, target_pos, list(result))))
                  }
)

inverse_table$add("base", "sqrt", 
                  function(operation, target, result){
                    expr((!!result) ^ 2)
                  }
)

inverse_table$add("base", "^", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    if(target_pos == 1){
                      if(args[[2]] == 0){
                        abort(sprintf("Cannot invert %s.", expr_text(operation)))
                      }
                      expr((!!result) ^ (1 / !!args[[2]]))
                    }
                    else{
                      expr(log(!!result) / log(!!args[[1]]))
                    }
                  }
)

inverse_table$add("base", "+", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    if(length(args) == 1){
                      result
                    }
                    else{
                      expr(!!result - !!args[[-target_pos]])
                    }
                  }
)

inverse_table$add("base", "-", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    if(length(args) == 1){
                      expr(-!!result)
                    }
                    else{
                      if(target_pos == 1){
                        expr(!!result + !!args[[2]])
                      }
                      else{
                        expr(!!args[[1]] - !!result)
                      }
                    }
                  }
)

inverse_table$add("base", "/", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    if(target_pos == 1){
                      expr(!!args[[2]] * !!result)
                    }
                    else{
                      expr(!!args[[1]] / !!result)
                    }
                  }
)

inverse_table$add("base", "*", 
                  function(operation, target, result){
                    args <- call_args(operation)
                    target_pos <- match(list(target), args)
                    expr(!!result / !!args[[-target_pos]])
                  }
)

inverse_table$add("base", "(", 
                  function(operation, target, result){
                    call2("(", !!!exprs(!!result))
                  }
)

#' Recursively traverse an object
#' 
#' @param x The object to traverse
#' @param .f A function for combining the recursed components
#' @param .g A function applied to the object before recursion
#' @param .h A function applied to the base case
#' @param base The base case for the recursion
#' 
#' @keywords internal
traverse <- function(x, .f = list, .g = identity, .h = identity, base = function(.x) is_syntactic_literal(.x) || is_symbol(.x)){
  # base case
  if(base(x))
    return(.h(x))
  # recursive case
  .f(lapply(.g(x), traverse, .f=.f, .g=.g, .h=.h, base=base), .h(x))
}

traverse_list <- function(x,
                          .f = function(.x, .y) as.list(.x),
                          .g = identity, 
                          .h = identity, 
                          base = function(.x) !is.list(.x)){
  traverse(x, .f=.f, .g=.g, .h=.h, base=base)
}

traverse_call <- function(x,
                          .f = function(.x, .y) map(.x, quo_get_expr) %>% as.call %>% new_quosure(env = get_env(.x[[1]])), # (quo_listcall_to_call)
                          .g = function(.x) .x %>% get_expr %>% as.list %>% map(new_quosure, env = get_env(.x)), # (quo_call_to_listcall)
                          .h = identity,
                          base = function(.x) !quo_is_call(.x)){
  x <- enquo(x)
  traverse(x, .f=.f, .g=.g, .h=.h, base=base)
}

big_mark <- function (x, ...){
  mark <- if (identical(getOption("OutDec"), ",")) 
    "."
  else ","
  formatC(x, big.mark = mark, ...)
}

dim_tbl <- function (x){
  dim_x <- dim(x)
  format_dim <- map_chr(dim_x, big_mark)
  paste(format_dim, collapse = " x ")
}

train_null_mdl <- function(.data, ...){
  structure(list(n = NROW(.data), vars = measured_vars(.data)), class = "null_mdl")
}

#' NULL model
#' 
#' Create a NULL model definition. This model produces NA forecasts and does not
#' require any estimation of the data. It is generally used as a placeholder for
#' models which have encountered an error (see `.safely` in [`model()`]).
#' 
#' @param formula Model specification (response variable)
#' @param ... Unused
#' 
#' @keywords internal
#' @export
null_model <- function(formula, ...){
  null_model <- new_model_class("null_mdl", train = train_null_mdl, 
                                specials = new_specials(xreg = function(...) NULL))
  new_model_definition(null_model, formula = !!enquo(formula), ...)
}

#' @rdname null_model
#' @param x The object to be tested.
#' @export
is_null_model <- function(x){
  if(is_model(x)) return(is_null_model(x[["fit"]]))
  if(inherits(x, "mdl_lst")) return(map_lgl(x, is_null_model))
  is.null(x) || inherits(x, "null_mdl")
}

#' @export
forecast.null_mdl <- function(object, new_data, ...){
  h <- NROW(new_data)
  vec_cast(rep(NA_real_, h), distributional::new_dist())
}

#' @export
forecast.NULL <- forecast.null_mdl

#' @export
generate.null_mdl <- function(x, new_data, ...){
  mutate(new_data, .sim = NA_real_)
}
#' @export
generate.NULL <- generate.null_mdl

#' @export
stream.null_mdl <- function(object, new_data, ...){
  object$n <- object$n + NROW(new_data)
  object
}
#' @export
stream.NULL <- function(object, new_data, ...) {
  NULL
}

#' @export
refit.null_mdl <- function(object, new_data, ...){
  object$n <- NROW(new_data)
  object
}
#' @export
refit.NULL <- function(object, new_data, ...) {
  NULL
}

#' @export
residuals.null_mdl <- function(object, ...){
  matrix(NA_real_, nrow = object$n, ncol = length(object$vars), 
         dimnames = list(NULL, object$vars))
}
#' @export
residuals.NULL <- function(object, new_data, ...) {
  NA_real_
}

#' @export
fitted.null_mdl <- function(object, ...){
  matrix(NA_real_, nrow = object$n, ncol = length(object$vars), 
         dimnames = list(NULL, object$vars))
}
#' @export
fitted.NULL <- function(object, new_data, ...) {
  NA_real_
}

#' @export
glance.null_mdl <- function(x, ...){
  tibble()
}

#' @export
tidy.null_mdl <- function(x, ...){
  tibble(term = character(), estimate = numeric())
}

#' @export
report.null_mdl <- function(object, ...){
  cat("NULL model")
}
#' @export
report.NULL <- report.null_mdl

#' @export
model_sum.null_mdl <- function(x){
  "NULL model"
}
#' @export
report.NULL <- report.null_mdl

#' Identify outliers
#' 
#' Return a table of outlying observations using a fitted model.
#' 
#' @param object An object which can identify outliers.
#' @param ... Arguments for further methods.
#' 
#' @rdname outliers
#' @export
outliers <- function(object, ...){
  UseMethod("outliers")
}

#' @rdname outliers
#' @export
outliers.mdl_df <- function(object, ...){
  mbl_vars <- mable_vars(object)
  kv <- key_vars(object)
  object <- mutate(as_tibble(object), 
                   dplyr::across(all_of(mbl_vars), function(x) lapply(x, outliers, ...)))
  object <- pivot_longer(object, all_of(mbl_vars), names_to = ".model", values_to = ".outliers")
  unnest_tsbl(object, ".outliers", parent_key = c(kv, ".model"))
}

#' @rdname outliers
#' @export
outliers.mdl_ts <- function(object, ...){
  object$data[outliers(object$fit, ...),]
}

#' @export
dplyr_row_slice.dcmp_ts <- function(data, i, ..., preserve = FALSE) {
  res <- NextMethod()
  build_dable(res, response = response_vars(data), method = data%@%"method", 
              seasons = data%@%"seasons", aliases = data%@%"aliases")
}

#' @export
dplyr_col_modify.dcmp_ts <- function(data, cols) {
  res <- NextMethod()
  build_dable(res, response = response_vars(data), method = data%@%"method", 
              seasons = data%@%"seasons", aliases = data%@%"aliases")
}

#' @export
dplyr_reconstruct.dcmp_ts <- function(data, template) {
  res <- NextMethod()
  build_dable(res, response = response_vars(template), method = template%@%"method", 
              seasons = template%@%"seasons", aliases = template%@%"aliases")
}

#' Extract components from a fitted model
#' 
#' Allows you to extract elements of interest from the model which can be
#' useful in understanding how they contribute towards the overall fitted values.
#' 
#' A dable will be returned, which will allow you to easily plot the components
#' and see the way in which components are combined to give forecasts.
#' 
#' The components can also be visualised using the [`autoplot()`] method provided
#' by the ggtime package.
#' 
#' @param object A mable.
#' @param ... Other arguments passed to methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' # Forecasting with an ETS(M,Ad,A) model to Australian beer production
#' aus_production %>%
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) %>% 
#'   components()
#' 
#' @rdname components
#' @export
components.mdl_df <- function(object, ...){
  object <- tidyr::pivot_longer(object, all_of(mable_vars(object)),
                             names_to = ".model", values_to = ".fit")
  kv <- key_vars(object)
  object <- transmute(as_tibble(object),
                      !!!syms(kv), !!sym(".model"),
                      cmp = map(!!sym(".fit"), components))
  attrs <- combine_dcmp_attr(object[["cmp"]])
  object <- unnest_tsbl(object, "cmp", parent_key = kv)
  as_dable(object, method = attrs[["method"]], resp = !!attrs[["response"]],
           seasons = attrs[["seasons"]], aliases = attrs[["aliases"]])
}

#' @rdname components
#' @export
components.mdl_ts <- function(object, ...){
  components(object$fit, ...)
}

#' Plot time series from a tsibble
#' 
#' Produces a time series plot of one or more variables from a tsibble. If the
#' tsibble contains a multiple keys, separate time series will be identified by
#' colour.
#' 
#' @param object A tsibble.
#' @param .vars A bare expression containing data you wish to plot. Multiple variables can be plotted using [`ggplot2::vars()`].
#' @param ... Further arguments passed to [`ggplot2::geom_line()`], which can be used to specify fixed aesthetics such as `colour = "red"` or `linewidth = 3`. 
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' library(tsibble)
#' 
#' tsibbledata::gafa_stock %>%
#'  autoplot(vars(Close, log(Close)))
#' 
#' @importFrom ggplot2 ggplot aes geom_line guides guide_legend xlab
autoplot.tbl_ts <- function(object, .vars = NULL, ...){
  lifecycle::deprecate_soft(
    when = "0.6.0",
    what = "autoplot.tbl_ts()",
    with = "ggtime::autoplot.tbl_ts()",
    details = "Graphics functions have been moved to the {ggtime} package. Please use `library(ggtime)` instead."
  )
  quo_vars <- enquo(.vars)
  
  kv <- key_vars(object)
  nk <- n_keys(object)
  
  if(quo_is_null(quo_vars)){
    mv <- measured_vars(object)
    pos <- which(vapply(object[mv], is.numeric, logical(1L)))
    if(is_empty(pos)) {
      abort("Could not automatically identify an appropriate plot variable, please specify the variable to plot.")
    }
    inform(sprintf(
      "Plot variable not specified, automatically selected `.vars = %s`",
      mv[pos[1]]
    ))
    y <- sym(mv[pos[1]])
    .vars <- as_quosures(list(y), env = empty_env())
  }
  else if(possibly(compose(is_quosures, eval_tidy), FALSE)(.vars)){
    .vars <- eval_tidy(.vars)
    object <- gather(
      mutate(object, !!!.vars),
      ".response", "value", !!!map(.vars, quo_name), factor_key = TRUE
    )
    y <- sym("value")
  }
  else{
    y <- quo_vars
    .vars <- list(y)
  }
  
  aes_spec <- list(x = index(object), y = y)
  
  if(nk > 1){
    object <- dplyr::mutate_if(object, ~inherits(., "agg_vec"), compose(trimws, format))
    aes_spec["colour"] <- list(expr(interaction(!!!syms(kv), sep = "/")))
  }
  
  p <- ggplot(object, eval_tidy(expr(aes(!!!aes_spec)))) + 
    geom_line(...) +
    xlab(paste0(index_var(object), " [", format(interval(object)), "]"))
  
  if(nk > 1){
    p <- p + 
      guides(colour = guide_legend(paste0(kv, collapse = "/")))
  }
  
  if(length(.vars) > 1){
    p <- p + facet_wrap(vars(!!sym(".response")), scales = "free_y", 
                        ncol = length(.vars)) + ggplot2::ylab(NULL)
  }
  
  p
}

#' @rdname autoplot.tbl_ts
#' @importFrom ggplot2 ggplot aes geom_line guides guide_legend xlab
autolayer.tbl_ts <- function(object, .vars = NULL, ...){
  lifecycle::deprecate_soft(
    when = "0.6.0",
    what = "autolayer.tbl_ts()",
    with = "ggtime::autolayer.tbl_ts()",
    details = "Graphics functions have been moved to the {ggtime} package. Please use `library(ggtime)` instead."
  )
  quo_vars <- enquo(.vars)
  kv <- key_vars(object)
  nk <- n_keys(object)
  
  if(quo_is_null(quo_vars)){
    mv <- measured_vars(object)
    pos <- which(vapply(object[mv], is.numeric, logical(1L)))
    if(is_empty(pos)) {
      abort("Could not automatically identify an appropriate plot variable, please specify the variable to plot.")
    }
    inform(sprintf(
      "Plot variable not specified, automatically selected `.vars = %s`",
      mv[pos[1]]
    ))
    y <- sym(mv[pos[1]])
    .vars <- as_quosures(list(y), env = empty_env())
  }
  else if(possibly(compose(is_quosures, eval_tidy), FALSE)(.vars)){
    .vars <- eval_tidy(.vars)
    object <- gather(
      mutate(object, !!!.vars),
      ".response", "value", !!!map(.vars, quo_name), factor_key = TRUE
    )
    y <- sym("value")
  }
  else{
    y <- quo_vars
    .vars <- list(y)
  }
  
  aes_spec <- list(x = index(object), y = y)
  
  if(nk > 1){
    aes_spec["colour"] <- list(expr(interaction(!!!syms(kv), sep = "/")))
  }
  if(n_keys(object) > 1){
    aes_spec["group"] <- list(expr(interaction(!!!syms(key_vars(object)), sep = "/")))
  }
  
  geom_line(eval_tidy(expr(aes(!!!aes_spec))), data = object, ..., inherit.aes = FALSE)
}

#' @importFrom ggplot2 fortify
fortify.fbl_ts <- function(model, data = NULL, level = c(80, 95), ...){
  return(as_tibble(model))
}

#' Plot a set of forecasts
#' 
#' Produces a forecast plot from a fable. As the original data is not included
#' in the fable object, it will need to be specified via the `data` argument.
#' The `data` argument can be used to specify a shorter period of data, which is
#' useful to focus on the more recent observations.
#' 
#' @param object A fable.
#' @param data A tsibble with the same key structure as the fable.
#' @param level The confidence level(s) for the plotted intervals.
#' @param show_gap Setting this to `FALSE` will connect the most recent value in `data` with the forecasts.
#' @param ... Further arguments passed used to specify fixed aesthetics for the forecasts such as `colour = "red"` or `linewidth = 3`.
#' @param point_forecast The point forecast measure to be displayed in the plot.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' fc <- aus_production %>%
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) %>% 
#'   forecast(h = "3 years") 
#' 
#' fc %>% 
#'   autoplot(aus_production)
#' 
#' @importFrom ggplot2 facet_wrap
autoplot.fbl_ts <- function(object, data = NULL, level = c(80, 95), show_gap = TRUE, ...){
  lifecycle::deprecate_soft(
    when = "0.6.0",
    what = "autoplot.fbl_ts()",
    with = "ggtime::autoplot.fbl_ts()",
    details = "Graphics functions have been moved to the {ggtime} package. Please use `library(ggtime)` instead."
  )

  fc_resp <- response_vars(object)
  fc_key <- setdiff(key_vars(object), ".model")
  common_models <- duplicated(key_data(object)[[".model"]] %||% rep(TRUE, NROW(key_data(object))))
  
  aes_y <- if(length(fc_resp) > 1){
    sym("value")
  }
  else{
    sym(fc_resp)
  }

  # Structure input data to match fable
  if (!is.null(data)){
    data <- as_tsibble(data)
    if(!identical(fc_key, key_vars(data))){
      abort("Provided data contains a different key structure to the forecasts.")
    }
    
    if(!is_empty(key(data))){
      data <- semi_join(data, object, by = key_vars(data))
    }
    
    if(length(fc_resp) > 1){
      data <- gather(data, ".response", "value", !!!syms(fc_resp), factor_key = TRUE)
    }
    
    data <- data %>% 
      dplyr::mutate_if(~inherits(., "agg_vec"), compose(trimws, format))
  }

  # Compute facets, if any
  facet_vars <- if(length(fc_resp) > 1) ".response" else NULL
  if(any(common_models)) facet_vars <- c(facet_vars, fc_key)
  
  # Change colours to be more appropriate for later facets
  fc_layer <- build_fbl_layer(object, data = data, level = level, 
                              show_gap = show_gap, facet_vars = facet_vars, ...)
  
  # Add forecasts on base plot
  p <- ggplot(data, aes(x = !!index(object))) + 
    fc_layer
    
  # Add historical data
  if(!is.null(data)){
    hist_aes <- aes(y = !!aes_y)
    if(length(key_vars(data)) > 0) {
      hist_aes[["group"]] <- expr(interaction(!!!syms(key_vars(data))))
    }
    p <- p + geom_line(hist_aes)
  }
  
  # Add facets
  if(!is_empty(facet_vars)){
    p <- p + facet_wrap(vars(!!!syms(facet_vars)), ncol = length(fc_resp),
                        scales = "free_y")
  }
  if(length(fc_resp) > 1) p <- p + ggplot2::ylab(NULL)
  p
}

#' @rdname autoplot.fbl_ts
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' aus_production %>% 
#'   autoplot(Beer) + 
#'   autolayer(fc)
autolayer.fbl_ts <- function(object, data = NULL, level = c(80, 95), 
                             point_forecast = list(mean = mean), show_gap = TRUE, ...){
  lifecycle::deprecate_soft(
    when = "0.6.0",
    what = "autolayer.fbl_ts()",
    with = "ggtime::autolayer.fbl_ts()",
    details = "Graphics functions have been moved to the {ggtime} package. Please use `library(ggtime)` instead."
  )
  build_fbl_layer(object = object, data = data, level = level, 
                  point_forecast = point_forecast, show_gap = show_gap, ...)

}

build_fbl_layer <- function(object, data = NULL, level = c(80, 95), 
                            colour = NULL, color = NULL, fill = NULL,
                            point_forecast = list(mean = mean), show_gap = TRUE, 
                            linetype = 1,
                            ..., facet_vars = NULL){
  mdl_key <- object%@%"model_cn"
  fc_key <- setdiff(key_vars(object), mdl_key)
  key_data <- key_data(object)
  key_vars <- key_vars(object)
  resp_var <- response_vars(object)
  dist_var <- distribution_var(object)
  idx <- index(object)
  common_models <- duplicated(key_data[[mdl_key]] %||% rep(TRUE, NROW(key_data(object))))
  colour <- colour %||% color %||% fill %||% "#446ffc"
  without <- function(x, el) x[setdiff(names(x), el)]
  
  if(isFALSE(level)){
    warn("Plot argument `level` should be a numeric vector of levels to display. Setting `level = NULL` will remove the intervals from the plot.")
    level <- NULL
  }
  
  if(!show_gap && is.null(data)){
    warn("Could not connect forecasts to last observation as `data` was not provided. Setting `show_gap = FALSE`.")
  }
  if(!show_gap){
    gap <- key_data
    gap[ncol(gap)] <- NULL
    last_obs <- filter(group_by_key(data), !!idx == max(!!idx))
    if (length(key_vars(last_obs)) == 0) {
      gap[names(last_obs)] <- last_obs
    }
    else {
      gap <- left_join(gap, last_obs, by = key_vars(last_obs))
    }
    if (length(resp_var) > 1) abort("`show_gap = FALSE` is not yet supported for multivariate forecasts.")
    gap[[dist_var]] <- distributional::dist_degenerate(gap[[resp_var]])
    dimnames(gap[[dist_var]]) <- resp_var
    gap <- as_fable(gap, index = !!idx, key = key_vars(object),
                    response = resp_var,
                    distribution = dist_var)
    object <- bind_rows(gap, object)
  }
  
  if(length(resp_var) > 1){
    resp <- sym("value")
    grp <- syms(".response")
  }
  else{
    resp <- sym(resp_var)
    grp <- NULL
  }
  
  mapping <- aes(
    x = !!idx,
  )
  
  useful_keys <- fc_key[map_lgl(key_data[fc_key], function(x) sum(!duplicated(x)) > 1)]
  col <- c(
    if(any(common_models)) useful_keys else NULL,
    if(sum(!common_models) > 1) ".model" else NULL
  )
  col <- setdiff(col, facet_vars)
  if(!is_empty(col)){
    col_nm <- paste0(col, collapse = "/")
    col <- if(length(col)==1) sym(col) else expr(interaction(!!!syms(col), sep = "/"))
  } else {
    col <- NULL
  }
  
  grp <- c(grp, syms(key_vars(object)))
  if(length(grp) > 0){
    mapping$group <- expr(interaction(!!!map(grp, function(x) expr(format(!!x))), sep = "/"))
  }
  
  # Single time point forecasts
  kd <- key_data(object)
  single_row <- lapply(split(kd$.rows, lengths(kd$.rows) == 1), unlist)
  
  out <- list()
  object <- object %>% 
    dplyr::mutate_if(~inherits(., "agg_vec"), compose(trimws, format))
  
  # Adapted from ggdist:::draw_key_lineribbon
  draw_key_ribbon <- function (self, data, params, size) {
    data$alpha <- data$alpha %||% NA
    if (is.null(data[["fill"]]) && (!is.null(data[["fill_ramp"]]) || !all(is.na(data[["alpha"]])))) {
      data$fill = "gray65"
    }
    if (is.null(data[["colour"]]) && (!is.null(data[["colour_ramp"]]) || !all(is.na(data[["alpha"]])))) {
      data$colour = "black"
    }
    # Apply ramped fill
    if (!is.null(data[["fill_ramp"]])) {
      if (utils::packageVersion("ggdist") > "3.3.1") {
        data$fill <- get("ramp_colours", asNamespace("ggdist"), mode = "function")(data$fill, data$fill_ramp)
      } else {
        data$fill <- mapply(function(color, amount){
          (scales::seq_gradient_pal(attr(amount, "from") %||% "white", color))(amount %||% NA)
        }, data$fill, data$fill_ramp)
      }
    }
    # Apply ramped colour
    if (!is.null(data[["colour_ramp"]])) {
      if (utils::packageVersion("ggdist") > "3.3.1") {
        data$colour <- get("ramp_colours", asNamespace("ggdist"), mode = "function")(data$colour, data$colour_ramp)
      } else {
        data$colour <- mapply(function(color, amount){
          (scales::seq_gradient_pal(attr(amount, "from") %||% "white", color))(amount %||% NA)
        }, data$colour, data$colour_ramp)
      }
    }
    ggplot2::draw_key_rect(data, params, size)
  }
  
  # Add forecast interval ribbons to plot
  if(!is.null(level)){
    intvl_mapping <- mapping
    # intvl_mapping$dist <- sym(dist_var)
    intvl_mapping$ymin <- sym(".lower")
    intvl_mapping$ymax <- sym(".upper")
    intvl_mapping$fill_ramp <- intvl_mapping$colour_ramp <- sym(".width")
    intvl_mapping$fill <- intvl_mapping$colour <- col
    
    qi_marginal <- function(x, .width = 0.95, na.rm = FALSE) {
      if (!na.rm && anyNA(x)) {
        return(matrix(c(NA_real_, NA_real_), ncol = 2))
      }
      
      if (utils::packageVersion("distributional") > "0.4.0") {
        do.call(rbind, lapply(quantile(x, (1 + c(-1, 1) * .width)/2, type = "marginal", na.rm = na.rm), t))
      } else if (utils::packageVersion("distributional") >= "0.6.0") {
        do.call(rbind, lapply(quantile(x, (1 + c(-1, 1) * .width)/2, kind = "marginal", na.rm = na.rm), t))
      } else {
        do.call(rbind, lapply(quantile(x, (1 + c(-1, 1) * .width)/2, na.rm = na.rm), t))
      }
    }
    
    dist_qi_frame <- function(data, level) {
      data <- ggdist::point_interval(as_tibble(data), !!sym(dist_var), .interval = qi_marginal, .width = level/100)
      names(data)[match(".index", names(data))] <- ".response"
      data
    }
    
    if(!is.null(col)){
      if(length(single_row[["FALSE"]]) > 0) {
        out[[length(out) + 1L]] <- ggdist::geom_lineribbon(without(intvl_mapping, "colour_ramp"), data = dist_qi_frame(object[single_row[["FALSE"]],], level), ..., inherit.aes = FALSE, key_glyph = draw_key_ribbon)
      }
      if(length(single_row[["TRUE"]]) > 0) {
        out[[length(out) + 1L]] <- ggdist::geom_interval(intvl_mapping, data = dist_qi_frame(object[single_row[["TRUE"]],], level), orientation = "vertical", ..., inherit.aes = FALSE, key_glyph = draw_key_ribbon)
      }
      out[[length(out) + 1L]] <- ggplot2::labs(fill = col_nm)
    } else {
      if(length(single_row[["FALSE"]]) > 0) {
        out[[length(out) + 1L]] <- ggdist::geom_lineribbon(without(intvl_mapping, "colour_ramp"), data = dist_qi_frame(object[single_row[["FALSE"]],], level), fill = colour, ..., inherit.aes = FALSE, key_glyph = draw_key_ribbon)
      }
      if(length(single_row[["TRUE"]]) > 0) {
        out[[length(out) + 1L]] <- ggdist::geom_interval(intvl_mapping, data = dist_qi_frame(object[single_row[["TRUE"]],], level), colour = colour, orientation = "vertical", ..., inherit.aes = FALSE, key_glyph = draw_key_ribbon)
      }
    }
    
    # Add scale for confidence level ramp
    if(length(level) > 6) {
      level_breaks <- ggplot2::waiver()
      level_guide <- ggdist::guide_rampbar()
    } else {
      level_breaks <- level/100
      level_guide <- "legend"
    }
    if(length(single_row[["FALSE"]]) > 0) {
      out[[length(out) + 1L]] <- ggdist::scale_fill_ramp_continuous(name = "level", from = "white", breaks = level_breaks, limits = function(l) range(l), range = c(0.7, 0.3), labels = function(x) scales::percent(as.numeric(x)), guide = level_guide)
    }
    if(length(single_row[["TRUE"]]) > 0) {
      out[[length(out) + 1L]] <- ggdist::scale_colour_ramp_continuous(name = "level", from = "white", breaks = level_breaks, limits = function(l) range(l), range = c(0.7, 0.3), labels = function(x) scales::percent(as.numeric(x)), guide = level_guide)
    }
  }
  
  # Calculate point forecasts
  object <- as_tibble(object)
  object[names(point_forecast)] <- map(point_forecast, calc, object[[dist_var]])
  
  unpack_data <- function(x) {
    x <- tidyr::pivot_longer(x[-match(dist_var, names(x))], names(point_forecast), names_to = "Point forecast", values_to = dist_var)
    if(length(resp_var) > 1){
      x[[dist_var]] <- as_tibble(x[[dist_var]])
      x <- x[setdiff(names(x), resp_var)] %>% 
        tidyr::unpack(!!dist_var) %>% 
        tidyr::pivot_longer(names(x[[dist_var]]), names_to = ".response", values_to = dist_var)
    }
    x
  }
  
  # Add point forecasts to plot
  mapping$y <- sym(dist_var)
  if(length(point_forecast) > 1){
    mapping$linetype <- mapping$shape <- sym("Point forecast")
    grp <- c(grp, mapping$linetype)
    mapping$group <- expr(interaction(!!!map(grp, function(x) expr(format(!!x))), sep = "/"))
  }
  object <- as_tibble(object)
  if(!is.null(col)){
    mapping$colour <- col
    
    if(length(single_row[["FALSE"]]) > 0) {
      out[[length(out) + 1L]] <- geom_line(mapping = without(mapping, "shape"), data = unpack_data(object[single_row[["FALSE"]],]), ..., inherit.aes = FALSE)#, key_glyph = ggplot2::draw_key_timeseries)
    }
    if(length(single_row[["TRUE"]]) > 0) {
      out[[length(out) + 1L]] <- ggplot2::geom_point(mapping = without(mapping, "linetype"), data = unpack_data(object[single_row[["TRUE"]],]), size = 3, ..., inherit.aes = FALSE)
    }
    out[[length(out) + 1L]] <- ggplot2::labs(colour = col_nm)
  } else {
    if(length(single_row[["FALSE"]]) > 0) {
      out[[length(out) + 1L]] <- geom_line(mapping = without(mapping, "shape"), data = unpack_data(object[single_row[["FALSE"]],]), color = colour, ..., inherit.aes = FALSE)#, key_glyph = ggplot2::draw_key_timeseries)
    }
    if(length(single_row[["TRUE"]]) > 0) {
      out[[length(out) + 1L]] <- ggplot2::geom_point(mapping = without(mapping, "linetype"), data = unpack_data(object[single_row[["TRUE"]],]), color = colour, size = 3, ..., inherit.aes = FALSE)
    }
  }
  
  out
}

#' Decomposition plots
#' 
#' Produces a faceted plot of the components used to build the response 
#' variable of the dable. Useful for visualising how the components contribute
#' in a decomposition or model.
#' 
#' @param object A dable.
#' @param .vars The column of the dable used to plot. By default, this will be the response variable of the decomposition.
#' @param scale_bars If `TRUE`, each facet will include a scale bar which represents the same units across each facet.
#' @param level If the decomposition contains distributions, which levels should be used to display intervals?
#' @inheritParams autoplot.tbl_ts
#' 
#' @examplesIf requireNamespace("feasts", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)
#' library(feasts)
#' library(tsibbledata)
#' aus_production %>% 
#'   model(STL(Beer)) %>%
#'   components() %>%  
#'   autoplot()
#' 
#' @importFrom ggplot2 ggplot geom_line geom_rect facet_grid vars ylab labs after_stat
autoplot.dcmp_ts <- function(object, .vars = NULL, scale_bars = TRUE, 
                             level = c(80, 95), ...){
  lifecycle::deprecate_soft(
    when = "0.6.0",
    what = "autoplot.dcmp_ts()",
    with = "ggtime::autoplot.dcmp_ts()",
    details = "Graphics functions have been moved to the {ggtime} package. Please use `library(ggtime)` instead."
  )
  method <- object%@%"method"
  idx <- index(object)
  keys <- key(object)
  n_keys <- n_keys(object)
  
  .vars <- enquo(.vars)
  if(quo_is_null(.vars)){
    .vars <- sym(response_vars(object))
  }
  dcmp_str <- dcmp <- (object%@%"aliases")[[expr_name(get_expr(.vars))]]
  if(!is.null(dcmp_str)){
    dcmp_str <- expr_text(dcmp_str)
  }
  object <- as_tsibble(object) %>% 
    transmute(!!.vars, !!!syms(all.vars(dcmp))) %>% 
    pivot_longer(measured_vars(.), values_to = ".val",
                 names_to = ".var", names_transform = list(.var = ~ factor(., levels = unique(.))))
  
  if(has_dist <- inherits(object[[".val"]], "distribution")) { 
    interval_data <- as_tibble(object)
    interval_data[paste0(level, "%")] <- lapply(level, hilo, x = interval_data[[".val"]])
    interval_data <- tidyr::pivot_longer(
      interval_data, paste0(level, "%"), names_to = NULL, values_to = "hilo"
    )
    intvl_aes <- aes(x = !!idx, dist = !!sym(".val"), fill_ramp = after_stat(level))
    line_aes <- aes(x = !!idx, y = mean(!!sym(".val")))
    if(n_keys > 1){
      line_aes$colour <- intvl_aes$fill <- intvl_aes$group <- expr(interaction(!!!keys, sep = "/"))
    }
    dcmp_geom <- list(
      if(n_keys > 1) {
        ggdist::stat_ribbon(intvl_aes, .width = level/100, ...)
      } else {
        ggdist::stat_ribbon(intvl_aes, fill = "gray65", .width = level/100, ...)
      },
      ggdist::scale_fill_ramp_discrete(from = "white", range = c(0.3, 0.7), labels = function(x) scales::percent(as.numeric(x))),
      geom_line(line_aes, ...)
    )
  } else {
    line_aes <- aes(x = !!idx, y = !!sym(".val"))
    if(n_keys > 1){
      line_aes$colour <- expr(interaction(!!!keys, sep = "/"))
    }
    dcmp_geom <- geom_line(line_aes, ...)
  }

  p <- object %>% 
    ggplot() + 
    dcmp_geom + 
    facet_grid(vars(!!sym(".var")), scales = "free_y") + 
    ylab(NULL) + 
    labs(
      title = paste(method%||%"A", "decomposition"), 
      subtitle = paste(c(expr_text(get_expr(.vars)), dcmp_str), collapse = " = ")
    )
  
  # Rangebars
  if (scale_bars) {
    xranges <- range(object[[expr_name(idx)]])
    barwidth <- pmax(1, round((1 / 64) * diff(as.double(xranges))))
    
    # Avoid issues with visible bindings
    ymin <- ymax <- center <- diff <- NULL
    
    min_fn <- if(has_dist) function(x, ...) min(quantile(x, (100-max(level))/200), ...) else min
    max_fn <- if(has_dist) function(x, ...) max(quantile(x, (100 + max(level))/200), ...) else max
    
    range_data <- as_tibble(object) %>%
      group_by(!!sym(".var")) %>% 
      summarise(ymin = min_fn(!!sym(".val"), na.rm = TRUE), ymax = max_fn(!!sym(".val"), na.rm = TRUE)) %>% 
      mutate(
        center = (ymin + ymax) / 2,
        diff = min(ymax - ymin),
        xmin = xranges[1] - barwidth * 2, xmax = xranges[1] - barwidth,
        ymin = center - diff/2, ymax = center + diff/2
      )
    
    p <- p + geom_rect(data = range_data,
      aes(ymin = !!sym("ymin"), ymax = !!sym("ymax"),
          xmin = !!sym("xmin"), xmax = !!sym("xmax")),
      fill = "gray75", colour = "black", linewidth = 1 / 3
    )
  }
  
  if (n_keys > 1) {
    colour_title <- paste0(map_chr(keys, expr_name), collapse = "/")
    p <- p + labs(colour = colour_title)
    if(has_dist) {
      p <- p + labs(fill = colour_title)
    }
  }
  
  p
}

#' Create evaluation environment for specials
#' 
#' Allows extension packages to make use of the formula parsing of specials.
#' 
#' @param ... A named set of functions which used to parse formula inputs
#' @param .required_specials The names of specials which must be provided (and if not, are included with no inputs).
#' @param .xreg_specials The names of specials which will be only used as inputs to other specials (most commonly `xreg`).
#' 
#' @export
new_specials <- function(..., .required_specials = NULL, .xreg_specials = NULL){
  specials <- squash(list2(...))
  if(is.null(specials$xreg)){
    specials$xreg <- function(...) abort(sprintf("Exogenous regressors are not supported for %s.", self$model))
  }
  structure(specials,
            required_specials = .required_specials,
            xreg_specials = .xreg_specials,
            class="fable_specials")
}

#' Helper special for producing a model matrix of exogenous regressors
#' 
#' @param ... Arguments for `fable_xreg_matrix` (see Details)
#' 
#' @details 
#' 
#' Currently the `fable_xreg_matrix` helper supports a single argument named
#' `default_intercept`. If this argument is TRUE (passed via `...` above), then
#' the intercept will be returned in the matrix if not specified (much like the
#' behaviour of `lm()`). If FALSE, then the intercept will only be included if
#' explicitly requested via `1` in the formula.
#' 
#' @importFrom stats model.frame model.matrix terms
#' @export
special_xreg <- function(...) {
  new_function(
    args = pairlist2(...=),
    body = call2(call2(":::", sym("fabletools"), sym("fable_xreg_matrix")),
                 sym("..."), ..., .data = parse_expr("self$data")),
    env = base_env()
  )
}

fable_xreg_matrix <- function(..., .data, default_intercept = TRUE) {
  dots <- enexprs(...)
  # Remove default intercept if needed.
  if(!default_intercept) {
    constants <- map_lgl(dots, inherits, "numeric")
    constant_specified <- any(map_lgl(dots[constants], `%in%`, c(-1, 0, 1)))
    # If the constant isn't specified, remove it.
    if(!constant_specified) dots <- c(dots, list(0))
  }
  # Remove index and keys from .
  is_dot <- vapply(dots, function(x) expr_text(x) == ".", logical(1L))
  if(any(is_dot)) {
    new_dot <- reduce(syms(c(".", index_var(.data), key_vars(.data))), call2, .fn = "-")
    dots <- c(new_dot, dots[!is_dot])
  }
  
  # Combine `...` into a model formula, then evaluate terms() to substitute `.`
  model_formula <- new_formula(
    lhs = NULL,
    rhs = reduce(dots, function(.x, .y) call2("+", .x, .y))
  )
  model_formula <- terms(model_formula, data = .data)
  
  # Produce appropriate evaluation environment with specials
  env <- map(enquos(...), get_env)
  env[map_lgl(env, compose(is_empty, env_parents))] <- NULL
  env <- if (!is_empty(env)) get_env(env[[1]]) else base_env()
  
  # Produce xreg matrix
  xreg <- model.frame(model_formula, data = env, na.action = stats::na.pass)
  mm <- model.matrix(terms(xreg), xreg)
  if (NROW(mm) == 0 && identical(colnames(mm), "(Intercept)")) {
    return(matrix(data = 1, nrow = NROW(.data), dimnames = list(NULL, "(Intercept)")))
  }
  mm
}

#' Run a hypothesis test from a mable
#' 
#' This function will return the results of a hypothesis test for each model in 
#' the mable.
#' 
#' @param x A mable.
#' @param ... Arguments for model methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibbledata)
#' 
#' olympic_running %>%
#'   model(lm = TSLM(log(Time) ~ trend())) %>% 
#'   hypothesize()
#' 
#' @rdname hypothesize.mdl_df
#' @importFrom generics hypothesize
#' @export
hypothesize.mdl_df <- function(x, ...){
  mbl_vars <- mable_vars(x)
  x <- mutate(as_tibble(x), 
              dplyr::across(all_of(mbl_vars), function(x) lapply(x, hypothesize, ...)))
  x <- pivot_longer(x, all_of(mbl_vars), names_to = ".model", values_to = ".hypothesis")
  unnest(x, ".hypothesis")
}

#' @param tests a list of test functions to perform on the model
#' @rdname hypothesize.mdl_df
#' @export
hypothesize.mdl_ts <- function(x, tests = list(), ...){
  if(is_function(tests)){
    tests <- list(tests)
  }
  vctrs::vec_rbind(
    !!!map(tests, calc, x$fit, ...),
    .names_to = ".test"
  )
}

#' Forecast reconciliation 
#' 
#' This function allows you to specify the method used to reconcile forecasts
#' in accordance with its key structure.
#' 
#' @param .data A mable.
#' @param ... Reconciliation methods applied to model columns within `.data`.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' lung_deaths_agg <- as_tsibble(cbind(mdeaths, fdeaths)) %>%
#'   aggregate_key(key, value = sum(value))
#' 
#' lung_deaths_agg %>%
#'   model(lm = TSLM(value ~ trend() + season())) %>%
#'   reconcile(lm = min_trace(lm)) %>% 
#'   forecast()
#' 
#' @export
reconcile <- function(.data, ...){
  UseMethod("reconcile")
}

#' @rdname reconcile
#' @export
reconcile.mdl_df <- function(.data, ...){
  mutate(.data, ...)
}

#' Minimum trace forecast reconciliation
#' 
#' Reconciles a hierarchy using the minimum trace combination method. The 
#' response variable of the hierarchy must be aggregated using sums. The 
#' forecasted time points must match for all series in the hierarchy (caution:
#' this is not yet tested for beyond the series length).
#' 
#' @param models A column of models in a mable.
#' @param method The reconciliation method to use.
#' @param sparse If TRUE, the reconciliation will be computed using sparse 
#' matrix algebra? By default, sparse matrices will be used if the MatrixM 
#' package is installed.
#' 
#' @seealso 
#' [`reconcile()`], [`aggregate_key()`]
#' 
#' @references 
#' Wickramasuriya, S. L., Athanasopoulos, G., & Hyndman, R. J. (2019). Optimal forecast reconciliation for hierarchical and grouped time series through trace minimization. Journal of the American Statistical Association, 1-45. https://doi.org/10.1080/01621459.2018.1448825 
#' 
#' @export
min_trace <- function(models, method = c("wls_var", "ols", "wls_struct", "mint_cov", "mint_shrink"),
                 sparse = NULL){
  if(is.null(sparse)){
    sparse <- requireNamespace("Matrix", quietly = TRUE)
  }
  structure(models, class = c("lst_mint_mdl", "mdl_lst", "list"),
            method = match.arg(method), sparse = sparse)
}

#' @export
forecast.lst_mint_mdl <- function(object, key_data, 
                                  new_data = NULL, h = NULL,
                                  point_forecast = list(.mean = mean), ...){
  method <- object%@%"method"
  sparse <- object%@%"sparse"
  if(sparse){
    check_installed("Matrix")
    as.matrix <- Matrix::as.matrix
    t <- Matrix::t
    diag <- function(x) if(is.vector(x)) Matrix::Diagonal(x = x) else Matrix::diag(x)
    solve <- Matrix::solve
    cov2cor <- Matrix::cov2cor
  } else {
    cov2cor <- stats::cov2cor
  }
  
  point_method <- point_forecast
  point_forecast <- list()
  # Get forecasts
  fc <- NextMethod()
  if(length(unique(map(fc, interval))) > 1){
    abort("Reconciliation of temporal hierarchies is not yet supported.")
  }
  
  # Compute weights (sample covariance)
  res <- map(object, function(x, ...) residuals(x, ...), type = "response")
  if(length(unique(map_dbl(res, nrow))) > 1){
    # Join residuals by index #199
    res <- unname(as.matrix(reduce(res, full_join, by = index_var(res[[1]]))[,-1]))
  } else {
    res <- matrix(invoke(c, map(res, `[[`, 2)), ncol = length(object))
  }
  
  # Construct S matrix - ??GA: have moved this here as I need it for Structural scaling
  agg_data <- build_key_data_smat(key_data)
  
  n <- nrow(res)
  covm <- crossprod(stats::na.omit(res)) / n
  if(method == "ols"){
    # OLS
    W <- diag(rep(1L, nrow(covm)))
  } else if(method == "wls_var"){
    # WLS variance scaling
    W <- diag(diag(covm))
  } else if (method == "wls_struct"){
    # WLS structural scaling
    W <- diag(vapply(agg_data$agg,length,integer(1L)))
  } else if (method == "mint_cov"){
    # min_trace covariance
    W <- covm
  } else if (method == "mint_shrink"){
    # min_trace shrink
    tar <- diag(apply(res, 2, compose(crossprod, stats::na.omit))/n)
    corm <- cov2cor(covm)
    xs <- scale(res, center = FALSE, scale = sqrt(diag(covm)))
    xs <- xs[stats::complete.cases(xs),]
    v <- (1/(n * (n - 1))) * (crossprod(xs^2) - 1/n * (crossprod(xs))^2)
    diag(v) <- 0
    corapn <- cov2cor(tar)
    d <- (corm - corapn)^2
    lambda <- sum(v)/sum(d)
    lambda <- max(min(lambda, 1), 0)
    W <- lambda * tar + (1 - lambda) * covm
  } else {
    abort("Unknown reconciliation method")
  }
  
  # Check positive definiteness of weights
  eigenvalues <- eigen(W, only.values = TRUE)[["values"]]
  if (any(eigenvalues < 1e-8)) {
    abort("min_trace needs covariance matrix to be positive definite.", call. = FALSE)
  }
  
  # Reconciliation matrices
  if(sparse){ 
    row_btm <- agg_data$leaf
    row_agg <- seq_len(nrow(key_data))[-row_btm]
    S <- Matrix::sparseMatrix(
      i = rep(seq_along(agg_data$agg), lengths(agg_data$agg)),
      j = vec_c(!!!agg_data$agg),
      x = rep(1, sum(lengths(agg_data$agg))))
    J <- Matrix::sparseMatrix(i = S[row_btm,,drop = FALSE]@i+1, j = row_btm, x = 1L, 
                              dims = rev(dim(S)))
    if (length(row_agg) == 0) {
      # Simple case of no constraints, to avoid unnecessary matrix algebra
      P <- J
    } else {
      U <- cbind(
        Matrix::Diagonal(diff(dim(J))),
        -S[row_agg,,drop = FALSE]
      )
      U <- U[, order(c(row_agg, row_btm)), drop = FALSE]
      Ut <- t(U)
      WUt <- W %*% Ut
      P <- J - J %*% WUt %*% solve(U %*% WUt, U)
      # P <- J - J%*%W%*%t(U)%*%solve(U%*%W%*%t(U))%*%U
    }
  }
  else {
    S <- matrix(0L, nrow = length(agg_data$agg), ncol = max(vec_c(!!!agg_data$agg)))
    S[length(agg_data$agg)*(vec_c(!!!agg_data$agg)-1) + rep(seq_along(agg_data$agg), lengths(agg_data$agg))] <- 1L
    R <- t(S)%*%solve(W)
    P <- solve(R%*%S)%*%R
  }
  
  reconcile_fbl_list(fc, S, P, W, point_forecast = point_method)
}

#' Bottom up forecast reconciliation
#' 
#' \lifecycle{experimental}
#' 
#' Reconciles a hierarchy using the bottom up reconciliation method. The 
#' response variable of the hierarchy must be aggregated using sums. The 
#' forecasted time points must match for all series in the hierarchy.
#' 
#' @param models A column of models in a mable.
#' 
#' @seealso 
#' [`reconcile()`], [`aggregate_key()`]
#' @export
bottom_up <- function(models){
  structure(models, class = c("lst_btmup_mdl", "mdl_lst", "list"))
}

#' @export
forecast.lst_btmup_mdl <- function(object, key_data, 
                                   point_forecast = list(.mean = mean),
                                   new_data = NULL, ...){
  # Keep only bottom layer
  agg_data <- build_key_data_smat(key_data)
  
  S <- matrix(0L, nrow = length(agg_data$agg), ncol = max(vec_c(!!!agg_data$agg)))
  S[length(agg_data$agg)*(vec_c(!!!agg_data$agg)-1) + rep(seq_along(agg_data$agg), lengths(agg_data$agg))] <- 1L
  
  btm <- agg_data$leaf
  # object <- object[btm]
  # if(!is.null(new_data)){
  #   new_data <- new_data[btm]
  # }
  
  point_method <- point_forecast
  point_forecast <- list()
  
  # Get base forecasts
  # fc <- vector("list", nrow(S))
  # fc[btm] <- NextMethod()
  fc <- NextMethod()
  
  # Add dummy forecasts to unused levels
  # fc[seq_along(fc)[-btm]] <- fc[btm[1]]
  
  P <- matrix(0L, nrow = ncol(S), ncol = nrow(S))
  P[(btm-1L)*nrow(P) + seq_len(nrow(P))] <- 1L
  
  reconcile_fbl_list(fc, S, P, W = diag(nrow(S)),
                     point_forecast = point_method)
}


#' Top down forecast reconciliation
#' 
#' \lifecycle{experimental}
#' 
#' Reconciles a hierarchy using the top down reconciliation method. The 
#' response variable of the hierarchy must be aggregated using sums. The 
#' forecasted time points must match for all series in the hierarchy.
#' 
#' @param models A column of models in a mable.
#' @param method The reconciliation method to use.
#' 
#' @seealso 
#' [`reconcile()`], [`aggregate_key()`]
#' 
#' @export
top_down <- function(models, method = c("forecast_proportions", "average_proportions", "proportion_averages")){
  structure(models, class = c("lst_topdwn_mdl", "mdl_lst", "list"),
            method = match.arg(method))
}

#' @export
forecast.lst_topdwn_mdl <- function(object, key_data, 
                                    point_forecast = list(.mean = mean), ...){
  method <- object%@%"method"
  point_method <- point_forecast
  point_forecast <- list()
  
  agg_data <- build_key_data_smat(key_data)
  S <- matrix(0L, nrow = length(agg_data$agg), ncol = max(vec_c(!!!agg_data$agg)))
  S[length(agg_data$agg)*(vec_c(!!!agg_data$agg)-1) + rep(seq_along(agg_data$agg), lengths(agg_data$agg))] <- 1L
  # Identify top and bottom level
  top <- which.max(rowSums(S))
  btm <- agg_data$leaf
  
  kv <- names(key_data)[-ncol(key_data)]
  agg_shadow <- as_tibble(map(key_data[kv], is_aggregated))
  agg_struct <- vctrs::vec_unique(agg_shadow)
  agg_depth <- nrow(agg_struct)
  if(length(kv) != (agg_depth - 1)) {
    abort("Top down reconciliation requires strictly hierarchical structures.")
  }
  agg_order <- kv[order(vapply(agg_struct, sum, integer(1L)))]
  
  if(method == "forecast_proportions") {
    fc <- NextMethod()
    fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])
    fc_mean <- lapply(fc_dist, mean)
    fc_mean <- do.call(cbind, fc_mean)
    fc_prop <- matrix(1, nrow = nrow(fc_mean), ncol = ncol(fc_mean))
    
    # Ensure key structure matches order of fc object
    key_data <- key_data[order(vec_c(!!!key_data$.rows)),]
    for(i in seq_len(agg_depth - 1)) {
      agg_layer <- key_data[agg_order[seq_len(i)]]
      agg_nodes <- vec_group_loc(is_aggregated(agg_layer[[length(agg_layer)]]))
      # Drop nodes which aren't in this layer
      if((i+1) < agg_depth) {
        agg_keep <- which(is_aggregated(key_data[[agg_order[[i+1]]]]))
        agg_nodes$loc <- lapply(agg_nodes$loc, intersect, agg_keep)
      }
      # Identify index position of the layer's nodes and their parents
      agg_child_loc <- agg_nodes$loc[[which(!agg_nodes$key)]]
      agg_parent_loc <- agg_nodes$loc[[which(agg_nodes$key)]]
      agg_parent_key <- agg_layer[,-length(agg_layer)]
      agg_parent <- vec_match(agg_parent_key[agg_child_loc,,drop=FALSE], agg_parent_key[agg_parent_loc,,drop=FALSE])
      # Compute forecast proportions for this layer
      fc_prop[,agg_child_loc] <- fc_prop[,agg_parent_loc,drop=FALSE][,agg_parent,drop=FALSE] * fc_mean[,agg_child_loc,drop=FALSE] / t(rowsum(t(fc_mean[,agg_child_loc,drop=FALSE]), agg_parent))[,agg_parent]
    }
    # Code adapted from reconcile_fbl_list to handle changing weights over horizon
    # This will need to be refactored later so that reconcile_fbl_list is broken up into more sub-problems
    # As the weight matrix is an identity, this code and computation is much simpler.
    is_normal <- all(map_lgl(fc_dist, function(x) all(dist_types(x) == "dist_normal")))
    # Point forecast means can be computed in one step
    fc_mean <- split(fc_mean[,top]*fc_prop,col(fc_prop))
    if(is_normal) {
      fc_var <- map(fc_dist, distributional::variance)
      fc_var <- fc_prop * fc_var[[top]] * fc_prop
      fc_var <- split(fc_var, col(fc_var))
      fc_dist <- map2(fc_mean, map(fc_var, sqrt), distributional::dist_normal)
    } else {
      fc_dist <- lapply(fc_mean, distributional::dist_degenerate)
    }
    # Update fables
    fc <- map2(fc, fc_dist, function(fc, dist){
      dimnames(dist) <- dimnames(fc[[distribution_var(fc)]])
      fc[[distribution_var(fc)]] <- dist
      point_fc <- compute_point_forecasts(dist, point_method)
      fc[names(point_fc)] <- point_fc
      fc
    })
    return(fc)
    
  } else {
    # Compute dis-aggregation matrix
    history <- lapply(object, function(x) response(x)[[".response"]])
    top_y <- history[[top]]
    btm_y <- history[btm]
    if (method == "average_proportions") { 
      prop <- map_dbl(btm_y, function(y) mean(y/top_y))
    } else if (method == "proportion_averages") {
      prop <- map_dbl(btm_y, mean) / mean(top_y)
    } else {
      abort("Unkown `top_down()` reconciliation `method`.")
    }
    
    # Keep only top layer
    object <- object[top]
    
    # Get base forecasts
    fc <- vector("list", nrow(S))
    fc[top] <- NextMethod()
    
    # Add dummy forecasts to unused levels
    fc[seq_along(fc)[-top]] <- fc[top]
  }
  
  P <- matrix(0L, nrow = ncol(S), ncol = nrow(S))
  P[,top] <- prop
  
  reconcile_fbl_list(fc, S, P, W = diag(nrow(S)),
                     point_forecast = point_method)
}


#' Middle out forecast reconciliation
#' 
#' \lifecycle{experimental}
#' 
#' Reconciles a hierarchy using the middle out reconciliation method. The 
#' response variable of the hierarchy must be aggregated using sums. The 
#' forecasted time points must match for all series in the hierarchy.
#' 
#' @param models A column of models in a mable.
#' @param split The middle level of the hierarchy from which the bottom-up and
#' top-down approaches are used above and below respectively.
#' 
#' @seealso 
#' [`reconcile()`], [`aggregate_key()`]
#' [*Forecasting: Principles and Practice* - Middle-out approach](https://otexts.com/fpp3/single-level.html#middle-out-approach)
#' 
#' @export
middle_out <- function(models, split = 1){
  structure(models, class = c("lst_midout_mdl", "mdl_lst", "list"),
            split = split)
}

#' @export
forecast.lst_midout_mdl <- function(object, key_data, 
                                    point_forecast = list(.mean = mean),
                                    new_data = NULL, ...){
  split <- object%@%"split"
  point_method <- point_forecast
  point_forecast <- list()
  
  agg_data <- build_key_data_smat(key_data)
  S <- matrix(0L, nrow = length(agg_data$agg), ncol = max(vec_c(!!!agg_data$agg)))
  S[length(agg_data$agg)*(vec_c(!!!agg_data$agg)-1) + rep(seq_along(agg_data$agg), lengths(agg_data$agg))] <- 1L
  
  # Identify top and bottom level
  top <- which.max(rowSums(S))
  btm <- agg_data$leaf
  
  kv <- names(key_data)[-ncol(key_data)]
  agg_shadow <- as_tibble(map(key_data[kv], is_aggregated))
  agg_struct <- vctrs::vec_unique(agg_shadow)
  agg_depth <- nrow(agg_struct)
  if(length(kv) != (agg_depth - 1)) {
    abort("Middle out reconciliation requires strictly hierarchical structures.")
  }
  agg_order <- kv[order(vapply(agg_struct, sum, integer(1L)))]
  if(is.character(split)) {
    split <- match(split, agg_order)
  }
  nodes_above <- which(agg_shadow[[split]])
  object <- object[-nodes_above]
  if(!is.null(new_data)){
    new_data <- new_data[-nodes_above]
  }
  fc <- NextMethod()
  
  fc_dist <- lapply(fc, function(x) x[[distribution_var(x)]])
  h <- vec_size(fc_dist[[1]])
  fc_mean <- matrix(0, h, nrow(key_data))
  fc_mean[,-nodes_above] <- do.call(cbind, lapply(fc_dist, mean))
  
  fc_prop <- key_data[[split]]
  fc_prop <- matrix(1, nrow = nrow(fc_mean), ncol = vec_unique_count(fc_prop[!is_aggregated(fc_prop)]))
  
  # Ensure key structure matches order of fc object
  mid_root_nodes <- NULL
  key_data <- key_data[order(vec_c(!!!key_data$.rows)),]
  for(i in seq(split+1, length.out = agg_depth - 1 - split)) {
    agg_layer <- key_data[agg_order[seq_len(i)]][-nodes_above,]
    agg_nodes <- vec_group_loc(is_aggregated(agg_layer[[length(agg_layer)]]))
    # Drop nodes which aren't in this layer
    if((i+1) < agg_depth) {
      agg_keep <- which(is_aggregated(key_data[[agg_order[[i+1]]]]))
      agg_nodes$loc <- lapply(agg_nodes$loc, intersect, agg_keep)
    }
    # Identify index position of the layer's nodes and their parents
    agg_child_loc <- agg_nodes$loc[[which(!agg_nodes$key)]]
    agg_parent_loc <- agg_nodes$loc[[which(agg_nodes$key)]]
    agg_parent_key <- agg_layer[,-length(agg_layer)]
    agg_parent <- vec_match(agg_parent_key[agg_child_loc,,drop=FALSE], agg_parent_key[agg_parent_loc,,drop=FALSE])
    
    # Produce matching replications of middle layer node positions
    if(is.null(mid_root_nodes)) mid_root_nodes <- agg_parent_loc
    mid_root_nodes <- mid_root_nodes[agg_parent]
    # Compute forecast proportions for this layer
    fc_prop <- fc_prop[,agg_parent,drop=FALSE] * fc_mean[,agg_child_loc,drop=FALSE] / t(rowsum(t(fc_mean[,agg_child_loc,drop=FALSE]), agg_parent))[,agg_parent]
  }
  
  fc_mean <- (fc_prop*fc_mean[,mid_root_nodes])%*%t(S)
  # Code adapted from reconcile_fbl_list to handle changing weights over horizon
  # This will need to be refactored later so that reconcile_fbl_list is broken up into more sub-problems
  # As the weight matrix is an identity, this code and computation is much simpler.
  is_normal <- all(map_lgl(fc_dist, function(x) all(dist_types(x) == "dist_normal")))
  # Point forecast means can be computed in one step
  fc_mean <- split(fc_mean,col(fc_mean))
  if(is_normal) {
    fc_var <- vector("list", nrow(key_data))
    fc_var[-nodes_above] <- map(fc_dist, distributional::variance)
    fc_var[nodes_above] <- rep_len(list(double(h)), length(nodes_above))
    
    P <- matrix(0L, nrow = ncol(S), ncol = nrow(S))
    
    # (S%*%P)%*%t(fc_mean)
    fc_var <- map(seq_len(h), function(i) {
      # Add top down structure
      P[seq_along(mid_root_nodes) + (mid_root_nodes-1)*nrow(P)] <- fc_prop[i,]
      SP <- S%*%P
      diag(SP%*%diag(map_dbl(fc_var, `[[`, i))%*%t(SP))
    })
    fc_dist <- map2(fc_mean, transpose_dbl(map(fc_var, sqrt)), distributional::dist_normal)
    
  } else {
    fc_dist <- lapply(fc_mean, distributional::dist_degenerate)
  }
  
  # Update fables
  map2(rep(fc[1], nrow(key_data)), fc_dist, function(fc, dist){
    dimnames(dist) <- dimnames(fc[[distribution_var(fc)]])
    fc[[distribution_var(fc)]] <- dist
    point_fc <- compute_point_forecasts(dist, point_method)
    fc[names(point_fc)] <- point_fc
    fc
  })
}

reconcile_fbl_list <- function(fc, S, P, W, point_forecast, SP = NULL) {
  if(length(unique(map(fc, interval))) > 1){
    abort("Reconciliation of temporal hierarchies is not yet supported.")
  }
  if(!inherits(S, "matrix")) {
    # Use sparse functions
    check_installed("Matrix")
    as.matrix <- Matrix::as.matrix
    t <- Matrix::t
    diag <- function(x) if(is.vector(x)) Matrix::Diagonal(x = x) else Matrix::diag(x)
    cov2cor <- Matrix::cov2cor
  } else {
    cov2cor <- stats::cov2cor
  }
  if(is.null(SP)) {
    SP <- S%*%P
  }
  
  fc_dist <- map(fc, function(x) x[[distribution_var(x)]])
  dist_type <- lapply(fc_dist, function(x) unique(dist_types(x)))
  dist_type <- unique(unlist(dist_type))
  is_normal <- all(map_lgl(fc_dist, function(x) all(dist_types(x) == "dist_normal")))
  
  fc_mean <- as.matrix(invoke(cbind, map(fc_dist, mean)))
  fc_var <- transpose_dbl(map(fc_dist, distributional::variance))
  
  # Apply to forecasts
  fc_mean <- as.matrix(SP%*%t(fc_mean))
  fc_mean <- split(fc_mean, row(fc_mean))
  if(identical(dist_type, "dist_normal")){
    R1 <- cov2cor(W)
    W_h <- map(fc_var, function(var) diag(sqrt(var))%*%R1%*%t(diag(sqrt(var))))
    fc_var <- map(W_h, function(W) diag(SP%*%W%*%t(SP)))
    fc_dist <- map2(fc_mean, transpose_dbl(map(fc_var, sqrt)), distributional::dist_normal)
  } else if (identical(dist_type, "dist_sample")) {
    sample_size <- unique(unlist(lapply(fc_dist, function(x) unique(lengths(distributional::parameters(x)$x)))))
    if(length(sample_size) != 1L) stop("Cannot reconcile sample paths with different replication sizes.")
    sample_horizon <- unique(lengths(fc_dist))
    if(length(sample_horizon) != 1L) stop("Cannot reconcile sample paths with different forecast horizon lengths.")
    # Extract sample paths
    samples <- lapply(fc_dist, function(x) distributional::parameters(x)$x)
    # Convert to array [samples,horizon,nodes]
    samples <- array(unlist(samples, use.names = FALSE), dim = c(sample_size, sample_horizon, length(fc_dist)))
    # Reconcile
    samples <- apply(samples, 1, function(x) as.matrix(SP%*%t(x)), simplify = FALSE)
    # Convert to array [nodes, horizon, samples]
    samples <- array(unlist(samples), dim = c(length(fc_dist), sample_horizon, sample_size))
    # Convert to distributions
    fc_dist <- apply(
      samples, 1L, simplify = FALSE,
      function(x) unname(distributional::dist_sample(split(x, row(x))))
    )
  } else {
    fc_dist <- map(fc_mean, distributional::dist_degenerate)
  }
  
  # Update fables
  map2(fc, fc_dist, function(fc, dist){
    dimnames(dist) <- dimnames(fc[[distribution_var(fc)]])
    fc[[distribution_var(fc)]] <- dist
    point_fc <- compute_point_forecasts(dist, point_forecast)
    fc[names(point_fc)] <- point_fc
    fc
  })
}

build_smat_rows <- function(key_data){
  lifecycle::deprecate_warn("0.2.1", "fabletools::build_smat_rows()", "fabletools::build_key_data_smat()")
  row_col <- sym(colnames(key_data)[length(key_data)])
  
  smat <- key_data %>%
    unnest(!!row_col) %>% 
    dplyr::arrange(!!row_col) %>% 
    select(!!expr(-!!row_col))
  
  agg_struc <- group_data(dplyr::group_by_all(as_tibble(map(smat, is_aggregated))))
  
  # key_unique <- map(smat, function(x){
  #   x <- unique(x)
  #   x[!is_aggregated(x)]
  # })
  
  agg_struc$.smat <- map(agg_struc$.rows, function(n) diag(1, nrow = length(n), ncol = length(n)))
  agg_struc <- map(seq_len(nrow(agg_struc)), function(i) agg_struc[i,])
  
  out <- reduce(agg_struc, function(x, y){
    # For now, assume x is aggregated into y somehow
    n_key <- ncol(x)-2
    nm_key <- names(x)[seq_len(n_key)]
    agg_vars <- map2_lgl(x[seq_len(n_key)], y[seq_len(n_key)], `<`)
    
    if(!any(agg_vars)) abort("Something unexpected happened, please report this bug at https://github.com/tidyverts/fabletools/issues/ with a description of what you're trying to do.")
    
    # Match rows between summation matrices
    not_agg <- names(Filter(`!`, y[seq_len(n_key)]))
    cols <- group_data(group_by(smat[x$.rows[[1]][seq_len(ncol(x$.smat[[1]]))],], !!!syms(not_agg)))$.rows
    cols_pos <- unlist(cols)
    cols <- rep(seq_along(cols), map_dbl(cols, length))
    cols[cols_pos] <- cols
    
    x$.rows[[1]] <- c(x$.rows[[1]], y$.rows[[1]])
    x$.smat <- list(rbind(
      x$.smat[[1]],
      y$.smat[[1]][, cols, drop = FALSE]
    ))
    x
  })
  
  smat <- out$.smat[[1]]
  smat[out$.rows[[1]],] <- smat
  
  return(smat)
}

build_key_data_smat <- function(x){
  if (any(lengths(x[[ncol(x)]]) > 1L)) {
    # Find the order based on the first entry position of each key
    x[[ncol(x)]] <- as.list(rank(vapply(x[[ncol(x)]], min, integer(1L))))
  }

  kv <- names(x)[-ncol(x)]
  agg_shadow <- as_tibble(map(x[kv], is_aggregated))
  grp <- as_tibble(vctrs::vec_group_loc(agg_shadow))
  num_agg <- rowSums(grp$key)
  # Initialise comparison leafs with known/guaranteed leafs
  x_leaf <- x[unlist(grp$loc[which(num_agg == min(num_agg))]),]
  
  # Sort by disaggregation to identify aggregated leafs in order
  grp <- grp[order(num_agg),]
  
  grp$match <- lapply(unname(split(grp, seq_len(nrow(grp)))), function(level){
    disagg_col <- which(!vec_c(!!!level$key))
    agg_idx <- level[["loc"]][[1]]
    pos <- vec_match(x_leaf[disagg_col], x[agg_idx, disagg_col])
    pos <- vec_group_loc(pos)
    pos <- pos[!is.na(pos$key),]
    # Add non-matches as leaf nodes
    agg_leaf <- setdiff(seq_along(agg_idx), pos$key)
    if(!is_empty(agg_leaf)){
      pos <- vec_rbind(
        pos,
        structure(list(key = agg_leaf, loc = as.list(seq_along(agg_leaf) + nrow(x_leaf))), 
                  class = "data.frame", row.names = agg_leaf)
      )
      x_leaf <<- vec_rbind(
        x_leaf, 
        x[agg_idx[agg_leaf],]
      )
    }
    pos$loc[order(pos$key)]
  })
  if(any(lengths(grp$loc) != lengths(grp$match))) {
    abort("An error has occurred when constructing the summation matrix.\nPlease report this bug here: https://github.com/tidyverts/fabletools/issues")
  }
  idx_leaf <- unlist(x_leaf$.rows)
  x$.rows[unlist(x$.rows)[unlist(grp$loc)]] <- unlist(grp$match, recursive = FALSE)
  return(list(agg = x$.rows, leaf = idx_leaf))
}

context("test-reconciliation")

test_that("reconciliation", {
  lung_deaths_agg <- lung_deaths_long %>% 
    aggregate_key(key, value = sum(value))
  expect_equal(n_keys(lung_deaths_agg), 3)
  expect_equal(
    lung_deaths_agg$value[1:72], 
    lung_deaths_long$value[1:72] + lung_deaths_long$value[72 + (1:72)]
  )
  expect_output(
    print(lung_deaths_agg$key),
    "<aggregated>"
  )
  expect_output(
    print(lung_deaths_agg),
    "<aggregated>"
  )
  
  skip_if_not_installed("fable")
  
  fit_agg <- lung_deaths_agg %>% 
    model(snaive = fable::SNAIVE(value))
  
  fc_agg <- fit_agg %>% forecast()
  fc_agg_reconciled <- fit_agg %>% reconcile(snaive = min_trace(snaive)) %>% forecast()
  
  expect_equal(
    mean(fc_agg$value),
    mean(fc_agg_reconciled$value)
  )
  expect_failure(
    expect_equal(
      fc_agg$value,
      fc_agg_reconciled$value
    ) 
  )
  
  fit_agg <- lung_deaths_agg %>% 
    model(ses = fable::ETS(value ~ error("A") + trend("A") + season("A")))
  fc_agg <- fit_agg %>% forecast()
  fc_agg_reconciled <- fit_agg %>% reconcile(ses = min_trace(ses)) %>% forecast()
  expect_equal(
    mean(fc_agg_reconciled$value[48 + (1:24)]),
    mean(fc_agg_reconciled$value[(1:24)] + fc_agg_reconciled$value[24 + (1:24)]),
  )
  expect_failure(
    expect_equal(
      fc_agg$value,
      fc_agg_reconciled$value
    )
  )
  
  fc_agg_reconciled <- fit_agg %>% reconcile(ses = min_trace(ses, method = "wls_var")) %>% forecast()
  expect_equal(
    mean(fc_agg_reconciled$value[48 + (1:24)]),
    mean(fc_agg_reconciled$value[(1:24)] + fc_agg_reconciled$value[24 + (1:24)])
  )
  expect_failure(
    expect_equal(
      fc_agg$value,
      fc_agg_reconciled$value
    )
  )
  
  fc_agg_reconciled <- fit_agg %>% reconcile(ses = min_trace(ses, method = "ols")) %>% forecast()
  expect_equal(
    mean(fc_agg_reconciled$value[48 + (1:24)]),
    mean(fc_agg_reconciled$value[(1:24)] + fc_agg_reconciled$value[24 + (1:24)])
  )
  expect_failure(
    expect_equal(
      fc_agg$value,
      fc_agg_reconciled$value
    )
  )
  
  fc_agg_reconciled <- fit_agg %>% reconcile(ses = min_trace(ses, method = "mint_cov")) %>% forecast()
  expect_equal(
    mean(fc_agg_reconciled$value[48 + (1:24)]),
    mean(fc_agg_reconciled$value[(1:24)] + fc_agg_reconciled$value[24 + (1:24)])
  )
  expect_failure(
    expect_equal(
      fc_agg$value,
      fc_agg_reconciled$value
    )
  )
})

context("setup-models.R")

test_specials <- new_specials(
  rnorm = function(m,s){
    stats::rnorm(NROW(self$data), m, s)
  },
  log5 = function(x){
    logb(x, base = 5)
  },
  oops = function(){
    stop("Not allowed")
  },
  xreg = function(...){
    deparse(match.call())
  }
)

test_train <- function(specials, ...) specials

no_specials <- function(formula, ...){
  no_specials_model <- new_model_class(model = "test model", train = test_train, specials = NULL)
  new_model_definition(no_specials_model, !!enquo(formula), ...)
}

specials <- function(formula, ...){
  specials_model <- new_model_class(model = "test model", train = test_train, specials = test_specials)
  new_model_definition(specials_model, !!enquo(formula), ...)
}

context("test-validate_model.R")

test_that("validate_model", {
  # Test expression capturing
  expect_identical(validate_formula(no_specials(user_model)), as.name("user_model"))
  
  # Test formula evaluating
  user_model <- y~x
  expect_identical(validate_formula(no_specials(user_model)), user_model)
  
  # Test bare formula
  expect_identical(validate_formula(no_specials(y~x)), y~x)
  
  tsbl1 <- tsibble::tsibble(
    date = seq(as.Date("2017-01-01"), as.Date("2017-01-10"), by = 1),
    value = rnorm(10), index = date
  )
  
  # Test automatic response selection
  expect_message(res <- validate_formula(no_specials(), data = tsbl1),
                 "Model not specified, defaulting to automatic modelling of the `value` variable.")
  expect_identical(res, as.name("value"))
  
  # Test LHS automatic response selection
  expect_message(res <- validate_formula(no_specials(~x), data = tsbl1), "Model not specified, defaulting to automatic modelling of the `value` variable.")
  expect_identical(res, value ~ x)
  
  # Test failed response selection
  tsbl1[["value2"]] <- rnorm(10) 
  expect_error(validate_formula(no_specials(), data = tsbl1), "Could not automatically determine the response variable")
  
  # Test failed LHS response variable selection
  expect_error(validate_formula(no_specials(~ x), data = tsbl1), "Could not automatically determine the response variable")
  
  # Test not supported automatic variable selection
  expect_error(validate_formula(no_specials(), data = NULL), "This model function does not support automatic selection")
})

context("test-generate")

test_that("generate", {
  skip_if_not_installed("fable")
  
  gen <- mbl %>% generate()
  expect_equal(NROW(gen), 24)
  expect_equal(gen$index, yearmonth("1978 Jan") + 0:23)
  
  gen_multi <- mbl_multi %>% generate()
  expect_equal(NROW(gen_multi), 48)
  expect_equal(gen_multi$index, yearmonth("1979 Jan") + rep(0:23, 2))
  expect_equal(unique(gen_multi$key), c("fdeaths", "mdeaths"))
  
  gen_complex <- mbl_complex %>% generate(times = 3)
  expect_equal(NROW(gen_complex), 24*2*2*3)
  expect_equal(gen_complex$index, yearmonth("1979 Jan") + rep(0:23, 2*2*3))
  expect_equal(unique(gen_complex$key), c("fdeaths", "mdeaths"))
  expect_equal(unique(gen_complex$.model), c("ets", "lm"))
})

test_that("generate seed setting", {
  skip_if_not_installed("fable")
  
  seed <- rnorm(1)
  expect_equal(
    mbl %>% generate(seed = seed),
    mbl %>% generate(seed = seed) 
  )
  
  expect_failure(
    expect_equal(
      mbl %>% generate(),
      mbl %>% generate()
    )
  )
})

context("test-parser.R")

test_that("Model parsing variety", {
  # Parse with no rhs and no specials
  parse1 <- model(us_deaths, no_specials(value))
  expect_equal(parse1[[1]][[1]]$fit, list())
  
  # Parse with no specials
  expect_warning(model(us_deaths, no_specials(value ~ rhs)), 
                 "Exogenous regressors are not supported")
  
  # Parse xreg
  parse_xreg <- model(us_deaths, specials(value ~ xreg(value, 3) + log(value)))
  expect_identical(parse_xreg[[1]][[1]]$fit$xreg[[1]], "xreg(value, 3)")
  
  parse_xreg <- model(us_deaths, specials(value ~ value + log(value)))
  expect_identical(parse_xreg[[1]][[1]]$fit$xreg[[1]], "xreg(value, log(value))")
  
  # Parse special
  parse_log5 <- model(us_deaths, specials(value ~ log5(value)))
  expect_identical(parse_log5[[1]][[1]]$fit$log5[[1]], logb(us_deaths$value, 5))
  
  # Parse specials using .vals
  parse_rnorm <- model(us_deaths, specials(value ~ rnorm(0,1)))
  expect_length(parse_rnorm[[1]][[1]]$fit$rnorm[[1]], NROW(us_deaths))
  
  # Parse multiple specials
  parse_multi <- model(us_deaths, specials(value ~ value + log(value) + rnorm(0,1) + log5(value)))
  expect_length(parse_multi[[1]][[1]]$fit, 3)
  expect_identical(parse_xreg[[1]][[1]]$fit$xreg[[1]], parse_multi[[1]][[1]]$fit$xreg[[1]])
  expect_identical(parse_log5[[1]][[1]]$fit$log5[[1]], parse_multi[[1]][[1]]$fit$log5[[1]])
  expect_identical(length(parse_rnorm[[1]][[1]]$fit$rnorm[[1]]), length(parse_multi[[1]][[1]]$fit$rnorm[[1]]))
  
  # Special causing error
  expect_warning(model(us_deaths, specials(value ~ oops())), "Not allowed")
  
  # Parse lhs transformation with no rhs
  parse_log1 <- model(us_deaths, specials(log(value)))
  mdl1_trans <- parse_log1[[1]][[1]]$transformation[[1]]
  log_trans <- new_transformation(
    function(value) log(value),
    function(value) exp(value)
  )
  expect_identical(capture.output(mdl1_trans), capture.output(log_trans))
  expect_equal(response_vars(parse_log1), "value")
  
  # Parse lhs transformation with rhs
  parse_log2 <- model(us_deaths, specials(log(value) ~ 1))
  mdl2_trans <- parse_log2[[1]][[1]]$transformation[[1]]
  expect_identical(capture.output(mdl1_trans), capture.output(mdl2_trans))
  expect_identical(response_vars(parse_log1), response_vars(parse_log2))
  
  # Parse lhs transformation with specials
  parse_log3 <- model(us_deaths, specials(log(value) ~ value + log(value) + rnorm(0,1) + log5(value)))
  mdl3_trans <- parse_log3[[1]][[1]]$transformation[[1]]
  expect_identical(capture.output(mdl1_trans), capture.output(mdl3_trans))
  expect_identical(response_vars(parse_log1), response_vars(parse_log3))
})


test_that("Model parsing scope", {
  # Test scoping without provided formula
  mdl <- eval({
    model(us_deaths, no_specials())
  }, envir = new_environment(list(no_specials = no_specials)))
  expect_equal(response_vars(mdl), "value")
  
  mdl <- eval({
    model(us_deaths, no_specials(value))
  }, envir = new_environment(list(no_specials = no_specials)))
  expect_equal(response_vars(mdl), "value")
  
  expect_error(
    eval({
      model(us_deaths, no_specials(nothing))
    }, envir = new_environment(list(no_specials = no_specials))),
    "nothing"
  )
  
  # Response variable from env
  mdl <- eval({
    something <- 1:72
    model(us_deaths, no_specials(something))
  }, envir = new_environment(list(no_specials = no_specials)))
  
  expect_equal(response_vars(mdl), "something")
  
  # Transformation from scalar
  mdl <- eval({
    scale <- pi
    model(us_deaths, no_specials(value/scale))
  }, envir = new_environment(list(no_specials = no_specials)))
  
  expect_equal(response_vars(mdl), "value")
  
  # Transformation from scalar in function env
  mdl <- eval({
    {function() {
      scale <- pi
      model(us_deaths, no_specials(value/scale))
    }} ()
  }, envir = new_environment(list(no_specials = no_specials)))
  
  expect_equal(response_vars(mdl), "value")
  
  # Specials missing values
  expect_warning(
    eval({
      model(us_deaths, specials(value ~ log5(mytrend)))
    }, envir = new_environment(list(specials = specials))),
    "mytrend"
  )
  
  # Specials with data from scope
  mdl <- eval({
    mytrend <- 1:72
    model(us_deaths, specials(value ~ log5(mytrend)))
  }, envir = new_environment(list(specials = specials)))
  
  expect_equal(mdl[[1]][[1]]$fit[[1]][[1]], log(1:72, 5))
})


test_that("Model response identification", {
  dt <- tsibble(
    idx = Sys.Date() - 1:10, GDP = rnorm(10), CPI = rnorm(10),
    index = idx
  )
  
  # Untransformed response
  mdl <- model(dt, no_specials(GDP))
  expect_equal(response_vars(mdl), "GDP")
  mdl <- model(dt, no_specials(resp(GDP)))
  expect_equal(response_vars(mdl), "GDP")
  
  # Scalar transformed response
  mdl <- model(dt, no_specials(GDP/pi))
  expect_equal(response_vars(mdl), "GDP")
  mdl <- model(dt, no_specials(resp(GDP)/pi))
  expect_equal(response_vars(mdl), "GDP")
  
  # Transformation with a tie
  mdl <- model(dt, no_specials(GDP/CPI))
  expect_equal(response_vars(mdl), "GDP/CPI")
  mdl <- model(dt, no_specials(resp(GDP)/CPI))
  mdl_trans <- mdl[[1]][[1]]$transformation[[1]]
  cpi_trans <- new_transformation(
    function(GDP) GDP/CPI,
    function(GDP) CPI * GDP
  )
  expect_identical(response_vars(mdl), "GDP")
  expect_identical(capture.output(mdl_trans), capture.output(cpi_trans))
  mdl <- model(dt, no_specials(GDP/resp(CPI)))
  mdl_trans <- mdl[[1]][[1]]$transformation[[1]]
  gdp_trans <- new_transformation(
    function(CPI) GDP/CPI,
    function(CPI) GDP/CPI
  )
  expect_equal(response_vars(mdl), "CPI")
  expect_identical(capture.output(mdl_trans), capture.output(gdp_trans))
})

#' Generate responses from a mable
#' 
#' Use a model's fitted distribution to simulate additional data with similar
#' behaviour to the response. This is a tidy implementation of 
#' [stats::simulate()].
#' 
#' Innovations are sampled by the model's assumed error distribution. 
#' If `bootstrap` is `TRUE`, innovations will be sampled from the model's 
#' residuals. If `new_data` contains the `.innov` column, those values will be
#' treated as innovations for the simulated paths.
#' 
#' @param x A mable.
#' @param new_data The data to be generated (time index and exogenous regressors)
#' @param h The simulation horizon (can be used instead of `new_data` for regular
#' time series with no exogenous regressors).
#' @param times The number of replications.
#' @param seed The seed for the random generation from distributions.
#' @param ... Additional arguments for individual simulation methods.
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(dplyr)
#' UKLungDeaths <- as_tsibble(cbind(mdeaths, fdeaths), pivot_longer = FALSE)
#' UKLungDeaths %>% 
#'   model(lm = TSLM(mdeaths ~ fourier("year", K = 4) + fdeaths)) %>% 
#'   generate(UKLungDeaths, times = 5)
#' @export
generate.mdl_df <- function(x, new_data = NULL, h = NULL, times = 1, seed = NULL, ...){
  mdls <- mable_vars(x)
  if(!is.null(new_data)){
    x <- bind_new_data(x, new_data)
  }
  kv <- c(key_vars(x), ".model")
  x <- tidyr::pivot_longer(as_tibble(x), all_of(mdls),
                           names_to = ".model", values_to = ".sim")
  
  # Evaluate simulations
  x[[".sim"]] <- map2(x[[".sim"]], 
                 x[["new_data"]] %||% rep(list(NULL), length.out = NROW(x)),
                 generate, h = h, times = times, seed = seed, ...)
  x[["new_data"]] <- NULL
  unnest_tsbl(x, ".sim", parent_key = kv)
}

#' @export
generate.mdl_lst <- function(x, new_data = NULL, h = NULL, times = 1, seed = NULL, ...) {
  map2(x, 
       new_data %||% rep(list(NULL), length.out = NROW(x)),
       generate, h = h, times = times, seed = seed, ...)
}
#' @export
generate.lst_mdl <- deprecate_lst_mdl(generate.mdl_lst)

#' @rdname generate.mdl_df
#' 
#' @param bootstrap If TRUE, then forecast distributions are computed using simulation with resampled errors.
#' @param bootstrap_block_size The bootstrap block size specifies the number of contiguous residuals to be taken in each bootstrap sample. 
#' 
#' @export
generate.mdl_ts <- function(x, new_data = NULL, h = NULL, times = 1, seed = NULL, 
                            bootstrap = FALSE, bootstrap_block_size = 1, ...){
  if (!exists(".Random.seed", envir = .GlobalEnv, inherits = FALSE)) 
    stats::runif(1)
  if (is.null(seed))
    RNGstate <- get(".Random.seed", envir = .GlobalEnv)
  else {
    R.seed <- get(".Random.seed", envir = .GlobalEnv)
    set.seed(seed)
    RNGstate <- structure(seed, kind = as.list(RNGkind()))
    on.exit(assign(".Random.seed", R.seed, envir = .GlobalEnv))
  }
  
  if(is.null(new_data)){
    new_data <- make_future_data(x$data, h)
  }
  
  if(is.null(new_data[[".rep"]])){
    kv <- c(".rep", key_vars(new_data))
    idx <- index_var(new_data)
    intvl <- tsibble::interval(new_data)
    new_data <- vctrs::vec_rbind(
      !!!set_names(rep(list(as_tibble(new_data)), times), seq_len(times)),
      .names_to = ".rep"
    )
    new_data <- build_tsibble(new_data, index = !!idx, key = !!kv, interval = intvl)
  }
  
  if(bootstrap) {
    res <- residuals(x$fit, type = "innovation")
    f_mean <- if(length(x$response) == 1) mean else colMeans
    res <- stats::na.omit(res) - f_mean(res, na.rm = TRUE)
    i <- if(bootstrap_block_size == 1) {
      sample.int(NROW(res), nrow(new_data), replace = TRUE)
    } else {
      if(any(has_gaps(x$data)$.gaps)) abort("Residuals must be regularly spaced without gaps to use a block bootstrap method.")
      kr <- tsibble::key_rows(new_data)
      ki <- lapply(lengths(kr), function(n) block_bootstrap(NROW(res), bootstrap_block_size, size = n))
      vec_c(!!!ki)
    }
    new_data$.innov <- if(length(x$response) == 1) res[i] else res[i,]
  }
  
  # Compute specials with new_data
  x$model$stage <- "generate"
  x$model$add_data(new_data)
  specials <- tryCatch(parse_model_rhs(x$model),
                       error = function(e){
                         abort(sprintf(
                           "%s
Unable to compute required variables from provided `new_data`.
Does your model require extra variables to produce simulations?", e$message))
                       }, interrupt = function(e) {
                         stop("Terminated by user", call. = FALSE)
                       })
  x$model$remove_data()
  x$model$stage <- NULL
  
  .sim <- generate(x[["fit"]], new_data = new_data, specials = specials, ...)
  .sim_cols <- setdiff(names(.sim), names(new_data))
  
  # TODO: Breaking change, .sim should be response variable name
  # For now, only do this with multivariate models but this change will be made for v1.0.0
  resp_vars <- vapply(x$response, expr_name, character(1L), USE.NAMES = FALSE)
  if (length(resp_vars) > 1) {
    .sim_cols <- resp_vars
    .sim[resp_vars] <- split(.sim$.sim, col(.sim$.sim))
    .sim$.sim <- NULL
  }
  
  # Back-transform forecast distributions
  bt <- map(x$transformation, function(x){
    bt <- invert_transformation(x)
    env <- new_environment(new_data, get_env(bt))
    set_env(bt, env)
  })
  
  .sim[.sim_cols] <- .mapply(function(f, x) f(x), list(bt, .sim[.sim_cols]), NULL)
  .sim
}

block_bootstrap <- function (n, window_size, size = n) {
  n_blocks <- size%/%window_size + 2
  bx <- numeric(n_blocks * window_size)
  for (i in seq_len(n_blocks)) {
    block_pos <- sample(seq_len(n - window_size + 1), 1)
    bx[((i - 1) * window_size + 1):(i * window_size)] <- block_pos:(block_pos + window_size - 1)
  }
  start_from <- sample.int(window_size, 1)
  bx[seq(start_from, length.out = size)]
}

#' @docType package
#' @keywords package
"_PACKAGE"

globalVariables(".")

## usethis namespace: start
#' @import rlang
#' @import vctrs
#' @import tsibble
#' @importFrom dplyr mutate transmute summarise filter select rename group_by ungroup groups group_data 
#' @importFrom dplyr full_join anti_join left_join semi_join
#' @importFrom dplyr dplyr_row_slice dplyr_col_modify dplyr_reconstruct
#' @importFrom dplyr bind_rows bind_cols
#' @importFrom tidyr nest unnest gather spread
#' @importFrom tidyselect all_of
#' @importFrom lifecycle deprecate_warn deprecated
## usethis namespace: end
NULL

#' @rdname point_accuracy_measures
#' @export
ME <- function(.resid, na.rm = TRUE, ...){
  mean(.resid, na.rm = na.rm)
}

#' @rdname point_accuracy_measures
#' @export
MSE <- function(.resid, na.rm = TRUE, ...){
  mean(.resid ^ 2, na.rm = na.rm)
}

#' @rdname point_accuracy_measures
#' @export
RMSE <- function(.resid, na.rm = TRUE, ...){
  sqrt(MSE(.resid, na.rm = na.rm))
}

#' @rdname point_accuracy_measures
#' @export
MAE <- function(.resid, na.rm = TRUE, ...){
  mean(abs(.resid), na.rm = na.rm)
}

#' @rdname point_accuracy_measures
#' @export
MPE <- function(.resid, .actual, na.rm = TRUE, ...){
  mean(.resid / .actual * 100, na.rm = na.rm)
}
#' @rdname point_accuracy_measures
#' @export
MAPE <- function(.resid, .actual, na.rm = TRUE, ...){
  mean(abs(.resid / .actual * 100), na.rm = na.rm)
}

#' @rdname point_accuracy_measures
#' @export
MASE <- function(.resid, .train, demean = FALSE, na.rm = TRUE, .period, d = .period == 1, D = .period > 1, ...){
  if (D > 0) { # seasonal differencing
    .train <- diff(.train, lag = .period, differences = D)
  }
  if (d > 0) {
    .train <- diff(.train, differences = d)
  }
  if(demean){
    .train <- .train - mean(.train, na.rm = na.rm)
  }
  scale <- mean(abs(.train), na.rm = na.rm)
  mean(abs(.resid / scale), na.rm = na.rm)
}

#' @rdname point_accuracy_measures
#' @export
RMSSE <- function(.resid, .train, demean = FALSE, na.rm = TRUE, .period, d = .period == 1, D = .period > 1, ...){
  if (D > 0) { # seasonal differencing
    .train <- diff(.train, lag = .period, differences = D)
  }
  if (d > 0) {
    .train <- diff(.train, differences = d)
  }
  if(demean){
    .train <- .train - mean(.train, na.rm = na.rm)
  }
  scale <- mean(.train^2, na.rm = na.rm)
  sqrt(mean(.resid^2 / scale, na.rm = na.rm))
}

#' @rdname point_accuracy_measures
#' @export
ACF1 <- function(.resid, na.action = stats::na.pass, demean = TRUE, ...){
  if(length(.resid) > 1) {
    stats::acf(.resid, plot = FALSE, lag.max = 2, na.action = na.action, 
             demean = demean)$acf[2, 1, 1]
  } else {
    return(NA_real_)
  }
}

#' Mean Arctangent Absolute Percentage Error
#' 
#' @inheritParams point_accuracy_measures
#' 
#' @references 
#' Kim, Sungil and Heeyoung Kim (2016) "A new metric of absolute percentage error
#' for intermittent demand forecasts". \emph{International Journal of Forecasting},
#' \bold{32}(3), 669-679.
#' 
#' @export
MAAPE <- function(.resid, .actual, na.rm = TRUE, ...){
  mean(atan(abs(.resid / .actual * 100)), na.rm = na.rm)
}

#' Point estimate accuracy measures
#' 
#' @param .resid A vector of residuals from either the training (model accuracy)
#'  or test (forecast accuracy) data.
#' @param .actual A vector of responses matching the fitted values 
#' (for forecast accuracy, `new_data` must be provided).
#' @param .train A vector of responses used to train the model
#' (for forecast accuracy, the `orig_data` must be provided).
#' @param .period The seasonal period of the data (defaulting to 'smallest' seasonal period).
# #' @param .fc A vector containing the one-step-ahead fitted (forecasted) values 
#' from a model, or forecasted values from the forecast.
# #' @param .expr_resp An expression for the response variable.
#' @param na.rm Remove the missing values before calculating the accuracy measure
#' @param ... Additional arguments for each measure.
#' @param demean Should the response be demeaned (MASE)
#' @param d Should the response model include a first difference?
#' @param D Should the response model include a seasonal difference?
#' @param na.action Function to handle missing values.
#'
#' @export
point_accuracy_measures <- list(ME = ME, RMSE = RMSE, MAE = MAE,
                       MPE = MPE, MAPE = MAPE, MASE = MASE, RMSSE = RMSSE,
                       ACF1 = ACF1)

#' @rdname directional_accuracy_measures
#' 
#' @param reward,penalty The weights given to correct and incorrect predicted
#'   directions.
#' 
#' @export
MDA <- function(.resid, .actual, na.rm = TRUE, reward = 1, penalty = 0, ...){
  actual_change <- diff(.actual)
  actual_direction <- sign(actual_change)
  predicted_change <- actual_change - .resid[-1]
  predicted_direction <- sign(predicted_change)
  directional_error <- actual_direction == predicted_direction
  
  (reward-penalty) * mean(directional_error, na.rm = na.rm) + penalty
}

#' @rdname directional_accuracy_measures
#' @export
MDV <- function(.resid, .actual, na.rm = TRUE, ...){
  actual_change <- diff(.actual)
  actual_direction <- sign(actual_change)
  predicted_change <- actual_change - .resid[-1]
  predicted_direction <- sign(predicted_change)
  directional_accuracy <- ifelse(actual_direction == predicted_direction, 1, -1)
  mean(abs(actual_change) * directional_accuracy, na.rm = na.rm)
}

#' @rdname directional_accuracy_measures
#' @export
MDPV <- function(.resid, .actual, na.rm = TRUE, ...){
  actual_change <- diff(.actual)
  actual_direction <- sign(actual_change)
  predicted_change <- actual_change - .resid[-1]
  predicted_direction <- sign(predicted_change)
  directional_accuracy <- ifelse(actual_direction == predicted_direction, 1, -1)
  
  mean(abs(actual_change / .actual[-1]) * directional_accuracy, na.rm = na.rm) * 100
}

#' Directional accuracy measures
#' 
#' A collection of accuracy measures based on the accuracy of the prediction's 
#' direction (say, increasing or decreasing).
#' 
#' `MDA()`: Mean Directional Accuracy
#' `MDV()`: Mean Directional Value
#' `MDPV()`: Mean Directional Percentage Value
#' 
#' @inheritParams point_accuracy_measures
#' 
#' @references 
#' Blaskowitz and H. Herwartz (2011) "On economic evaluation of directional forecasts". \emph{International Journal of Forecasting},
#' \bold{27}(4), 1058-1065.
#' 
#' @export
directional_accuracy_measures <- list(MDA = MDA, MDV = MDV, MDPV = MDPV)

#' @rdname interval_accuracy_measures
#' @export
winkler_score <- function(.dist, .actual, level = 95, na.rm = TRUE, ...){
  interval <- hilo(.dist, level)
  if(!inherits(interval, "hilo")) abort("Winkler scores are not supported for multivariate distributions.")
  alpha <- 1-level/100
  lt <- interval$lower
  ut <- interval$upper
  score <- ifelse(
    .actual < lt, 
      (ut - lt) + (2/alpha)*(lt-.actual),
  ifelse(
    .actual > ut,
      (ut - lt) + (2/alpha)*(.actual-ut),
  # else     
    ut-lt)
  )
  mean(score, na.rm = na.rm)
}

#' @rdname interval_accuracy_measures
#' @importFrom stats quantile
#' @export
pinball_loss <- function(.dist, .actual, level = 95, na.rm = TRUE, ...){
  q <- stats::quantile(.dist, level/100)
  loss <- ifelse(.actual>=q, level/100 * (.actual-q), (1-level/100) * (q-.actual))
  2*mean(loss, na.rm = na.rm)
}

#' @rdname interval_accuracy_measures
#' @export
scaled_pinball_loss <- function(.dist, .actual, .train, level = 95, na.rm = TRUE,
                                demean = FALSE, .period, d = .period == 1, 
                                D = .period > 1, ...){
  if (D > 0) { # seasonal differencing
    .train <- diff(.train, lag = .period, differences = D)
  }
  if (d > 0) {
    .train <- diff(.train, differences = d)
  }
  if(demean){
    .train <- .train - mean(.train, na.rm = na.rm)
  }
  scale <- mean(abs(.train), na.rm = na.rm)
  
  q <- stats::quantile(.dist, level/100)
  loss <- ifelse(.actual>=q, level/100 * (.actual-q), (1-level/100) * (q-.actual))
  mean(loss/scale, na.rm = na.rm)
}

#' Interval estimate accuracy measures
#' 
#' @inheritParams point_accuracy_measures
#' @param .dist The distribution of fitted values from the model, or forecasted values from the forecast.
#' @param level The level of the forecast interval.
#' 
#' @export
interval_accuracy_measures <- list(winkler = winkler_score, pinball = pinball_loss, scaled_pinball = scaled_pinball_loss)

#' @rdname distribution_accuracy_measures
#' @export
percentile_score <- function(.dist, .actual, na.rm = TRUE, ...){
  quantile_score(.dist, .actual, probs = seq(0.01, 0.99, 0.01), na.rm = TRUE, ...)
}

#' @rdname distribution_accuracy_measures
#' @param probs A vector of probabilities at which the metric is evaluated.
#' @export
quantile_score <- function(.dist, .actual, probs = c(0.05,0.25,0.5,0.75,0.95),
                           na.rm = TRUE, ...){
  percentiles <- map(probs, quantile, x = .dist)
  if(!is.numeric(percentiles[[1]])) abort("Quantile scores are not supported for multivariate distributions.")
  2*mean(map2_dbl(percentiles, probs, function(percentile, prob){
    L <- ifelse(.actual < percentile, (1-prob), prob)*abs(percentile-.actual)
    mean(L, na.rm = na.rm)
  }), na.rm = na.rm)
}

#' @rdname distribution_accuracy_measures
#' @export
CRPS <- function(.dist, .actual, n_quantiles = 1000, na.rm = TRUE, ...){
  dist_type <- dist_types(.dist)
  is_normal <- dist_type == "dist_normal"
  is_sample <- dist_type == "dist_sample"
  z <- rep(NA_real_, length(.dist))
  
  if(any(is_normal)){
    mean <- mean(.dist[is_normal])
    sd <- sqrt(distributional::variance(.dist[is_normal]))
    zn <- (.actual[is_normal]-mean)/sd
    zn <- sd*(zn*(2*stats::pnorm(zn)-1)+2*stats::dnorm(zn)-1/sqrt(pi))
    z[is_normal] <- zn
  }
  if(any(is_sample)){
    z[is_sample] <- map2_dbl(
      .dist[is_sample], .actual[is_sample], 
      function(d, y){
        x <- sort(d$x)
        m <- length(x)
        (2/m) * mean((x-y)*(m*(y<x) - seq_len(m) + 0.5))
      }
    )
  }
  is_default <- is.na(z)
  if(any(is_default)){
    probs <- seq(0, 1, length.out = n_quantiles + 2)[seq_len(n_quantiles) + 1]
    percentiles <- map(probs, quantile, x = .dist[is_default])
    if(!is.numeric(percentiles[[1]])) abort("Percentile scores are not supported for multivariate distributions.")
    za <- transpose_dbl(map2(percentiles, probs, function(percentile, prob){
      ifelse(.actual[is_default] < percentile, (1-prob), prob)*abs(percentile-.actual[is_default])
    }))
    
    z[is_default] <- vapply(za, mean, numeric(1L), na.rm = na.rm)*2
  }
  
  mean(z, na.rm = na.rm)
}

#' Distribution accuracy measures
#' 
#' These accuracy measures can be used to evaluate how accurately a forecast 
#' distribution predicts a given actual value.
#' 
#' @section Quantile/percentile score (pinball loss):
#' 
#' A quantile (or percentile) score evaluates how accurately a set of quantiles
#' (or percentiles) from the distribution match the given actual value. This 
#' score uses a pinball loss function, and can be calculated via the average of
#' the score function given below:
#' 
#' The score function \eqn{s_p(q_p,y)} is given by \eqn{(1-p)(q_p-y)} if 
#' \eqn{y < q_p}, and \eqn{p(y-q_p)} if \eqn{y \ge q_p}. Where \eqn{p} is the 
#' quantile probability, \eqn{q_p = F^{-1}(p)} is the quantile with probability 
#' \eqn{p}, and \eqn{y} is the actual value.
#' 
#' The resulting accuracy measure will average this score over all predicted
#' points at all desired quantiles (defined via the `probs` argument).
#' 
#' The percentile score is uses the same method with `probs` set to all 
#' percentiles `probs = seq(0.01, 0.99, 0.01)`.
#' 
#' @section Continuous ranked probability score (CRPS):
#' 
#' The continuous ranked probability score (CRPS) is the continuous analogue of
#' the pinball loss quantile score defined above. Its value is twice the 
#' integral of the quantile score over all possible quantiles:
#' 
#' \deqn{
#'   CRPS(F,y) = 2 \int_0^1 s_p(q_p,y) dp
#' }{
#'   CRPS(F,y) = 2 integral_0^1 s_p(q_p,y) dp
#' }
#' 
#' It can be computed directly from the distribution via:
#' 
#' \deqn{
#'   CRPS(F,y) = \int_{-\infty}^\infty (F(x) - 1{y\leq x})^2 dx
#' }{
#'   CRPS(F,y) = integral_{-\infty}^\infty (F(x) - 1{y\leq x})^2 dx
#' }
#' 
#' For some forecast distribution \eqn{F} and actual value \eqn{y}.
#' 
#' Calculating the CRPS accuracy measure is computationally difficult for many
#' distributions, however it can be computed quickly and exactly for Normal and
#' emperical (sample) distributions. For other distributions the CRPS is 
#' approximated using the quantile score of many quantiles (using the number of 
#' quantiles specified in the `n_quantiles` argument).
#' 
#' @inheritParams interval_accuracy_measures
#' @param n_quantiles The number of quantiles to use in approximating CRPS when an exact solution is not available.
#' 
#' @export
distribution_accuracy_measures <- list(percentile = percentile_score, CRPS = CRPS)


#' Forecast skill score measure
#' 
#' This function converts other error metrics such as `MSE` into a skill score.
#' The reference or benchmark forecasting method is the Naive method for
#' non-seasonal data, and the seasonal naive method for seasonal data.
#' When used within \code{\link{accuracy.fbl_ts}}, it is important that the data
#' contains both the training and test data, as the training data is used to
#' compute the benchmark forecasts.
#'
#' @param measure The accuracy measure to use in computing the skill score.
#' 
#' @examples 
#' 
#' skill_score(MSE)
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE)
#' library(fable)
#' library(tsibble)
#' 
#' lung_deaths <- as_tsibble(cbind(mdeaths, fdeaths))
#' lung_deaths %>% 
#'   dplyr::filter(index < yearmonth("1979 Jan")) %>%
#'   model(
#'     ets = ETS(value ~ error("M") + trend("A") + season("A")),
#'     lm = TSLM(value ~ trend() + season())
#'   ) %>%
#'   forecast(h = "1 year") %>%
#'   accuracy(lung_deaths, measures = list(skill = skill_score(MSE)))
#' 
#' @export
skill_score <- function(measure) {
  function(...) {
    # Compute accuracy measure for forecasts
    score <- measure(...)
    
    # Compute arguments of benchmark method using .train
    bench <- list(...)
    lag <- bench$.period
    n <- length(bench$.train)
    y <- bench$.train
    
    ## Compute point forecast from benchmark
    bench$.fc <- rep_len(
      y[c(rep(NA, max(0, lag - n)), seq_len(min(n, lag)) + n - min(n, lag))],
      length(bench$.fc)
    )
    bench$.resid <- bench$.actual - bench$.fc
    
    # Compute forecast distribution from benchmark
    e <- y - c(rep(NA, min(lag, n)), y[seq_len(length(y) - lag)])
    mse <- mean(e^2, na.rm = TRUE)
    h <- length(bench$.actual)
    fullperiods <- (h - 1) / lag + 1
    steps <- rep(seq_len(fullperiods), rep(lag, fullperiods))[seq_len(h)]
    bench$.dist <- distributional::dist_normal(bench$.fc, sqrt(mse * steps))
    ref_score <- do.call(measure, bench)
    
    1 - score / ref_score
  }
}

#' Evaluate accuracy of a forecast or model
#' 
#' Summarise the performance of the model using accuracy measures. Accuracy
#' measures can be computed directly from models as the one-step-ahead fitted
#' residuals are available. When evaluating accuracy on forecasts, you will
#' need to provide a complete dataset that includes the future data and data
#' used to train the model.
#' 
#' @param object A model or forecast object
#' @param measures A list of accuracy measure functions to compute (such as [`point_accuracy_measures`], [`interval_accuracy_measures`], or [`distribution_accuracy_measures`])
#' @param ... Additional arguments to be passed to measures that use it.
#' 
#' @seealso 
#' [Evaluating forecast accuracy](https://otexts.com/fpp3/accuracy.html)
#' 
#' @examplesIf requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)
#' library(fable)
#' library(tsibble)
#' library(tsibbledata)
#' library(dplyr)
#' 
#' fit <- aus_production %>%
#'   filter(Quarter < yearquarter("2006 Q1")) %>% 
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A")))
#' 
#' # In-sample training accuracy does not require extra data provided.
#' accuracy(fit)
#' 
#' # Out-of-sample forecast accuracy requires the future values to compare with.
#' # All available future data will be used, and a warning will be given if some
#' # data for the forecast window is unavailable.
#' fc <- fit %>% 
#'   forecast(h = "5 years")
#' fc %>% 
#'   accuracy(aus_production)
#'   
#' # It is also possible to compute interval and distributional measures of
#' # accuracy for models and forecasts which give forecast distributions.
#' fc %>% 
#'   accuracy(
#'     aus_production,
#'     measures = list(interval_accuracy_measures, distribution_accuracy_measures)
#'   )
#' 
#' @rdname accuracy
#' @export
accuracy.mdl_df <- function(object, measures = point_accuracy_measures, ...){
  if(is_tsibble(measures)){
    abort("The `measures` argument must contain a list of accuracy measures.
Hint: A tsibble of future values is only required when computing accuracy of a fable. To compute forecast accuracy, you'll need to compute the forecasts first.")
  }
  as_tibble(object) %>% 
    tidyr::pivot_longer(mable_vars(object), names_to = ".model", values_to = "fit") %>% 
    mutate(fit = map(!!sym("fit"), accuracy, measures = measures, ...)) %>% 
    unnest_tbl("fit")
}

#' @rdname accuracy
#' @export
accuracy.mdl_ts <- function(object, measures = point_accuracy_measures, ...){
  dots <- dots_list(...)
  resp <- if(length(object$response) > 1) sym("value") else object$response[[1]]
  
  aug <- as_tibble(augment(object))
  
  # Compute inputs for each response variable
  if(length(object$response) > 1){
    aug <- group_by(aug, !!sym(".response"))
  }
  
  aug <- aug %>% 
    rename(
      ".actual" := !!resp,
    ) %>% 
    summarise(
      .resid = list(!!sym(".actual") - !!sym(".fitted")),
      .train = list(!!sym(".actual")),
      .actual = list(!!sym(".actual")), 
      .fc = list(!!sym(".fitted")),
    )
  
  # Re-group after summarise
  if(length(object$response) > 1){
    aug <- group_by(aug, !!sym(".response"))
  }
  
  # Add user inputs
  aug <- mutate(aug, ...)
  
  if(is.null(aug[[".period"]])){
    aug <- mutate(aug, .period = get_frequencies(NULL, object[["data"]], .auto = "smallest"))
  }
  aug <- aug %>% 
    nest_grps(nm = ".accuracy")
  
  measures <- map(aug[[".accuracy"]], function(measures, inputs){
    map(measures, safely(do.call, otherwise = NA_real_), flatten(transpose(inputs)))
  }, measures = squash(measures))
  
  measures <- imap(measures, function(x, nm){
    err <- map_lgl(x, function(x) !is.null(x[["error"]]))
    if((tot_err <- sum(err)) > 0){
      err_msg <- table(map_chr(x[err], function(x) x[["error"]][["message"]]))
      warn(
        sprintf("%i error%s encountered\n%s\n",
                tot_err,
                if(tot_err > 1) sprintf("s (%i unique)", length(err_msg)) else "", 
                paste0("[", err_msg, "] ", names(err_msg), collapse = "\n")
        )
      )
    }
    as_tibble(map(x, function(x) x[["result"]]))
  })
  
  aug %>% 
    mutate(
      .type = "Training",
      .accuracy = measures
    ) %>% 
    unnest_tbl(".accuracy")
}

#' @param data A dataset containing the complete model dataset (both training and test data). The training portion of the data will be used in the computation of some accuracy measures, and the test data is used to compute the forecast errors.
#' @param by Variables over which the accuracy is computed (useful for computing across forecast horizons in cross-validation). If `by` is `NULL`, groups will be chosen automatically from the key structure.
#' 
#' @rdname accuracy
#' @export
accuracy.fbl_ts <- function(object, data, measures = point_accuracy_measures, ..., 
                            by = NULL){
  resp <- response_vars(object)
  dist <- sym(distribution_var(object))
  
  if(is.null(by)){
    by <- intersect(c(".model", key_vars(data)), colnames(object))
    if(length(resp) > 1) by <- c(by, ".response")
  }
  
  grp <- c(syms(by), groups(object))
  by <- union(index_var(object), by)
  
  
  if(!(".model" %in% by) & ".model" %in% names(object)){
    warn('Accuracy measures should be computed separately for each model, have you forgotten to add ".model" to your `by` argument?')
  }
  
  if(NROW(missing_test <- suppressWarnings(anti_join(object, data, by = intersect(colnames(data), by)))) > 0){
    n_miss <- length(unique(missing_test[[index_var(missing_test)]]))
    warn(sprintf(
      "The future dataset is incomplete, incomplete out-of-sample data will be treated as missing. 
%i %s %s", 
      n_miss, 
      ifelse(n_miss==1, "observation is missing at", "observations are missing between"),
      paste(unique(range(missing_test[[index_var(missing_test)]])), collapse = " and ")
    ))
  }
  
  # Compute .fc, .dist, .actual and .resid
  object <- as_tsibble(object)
  
  if(mv <- length(resp) > 1){
    aug <- object <- tidyr::pivot_longer(
      transmute(object, as_tibble(mean(!!dist)), .dist = !!dist, !!!syms(intersect(by, colnames(object)))),
      !!resp, names_to = ".response", values_to = ".fc")
    aug_dt <- data <- tidyr::pivot_longer(
      transmute(data, !!index(data), !!!syms(resp)),
      !!resp, names_to = ".response", values_to = ".actual")
    resp <- ".actual"
  } else {
    aug <- transmute(object, .fc = mean(!!dist), .dist = !!dist, !!!syms(by))
    aug_dt <- transmute(data, !!index(data), .actual = !!sym(resp))
  }
  aug <- left_join(as_tibble(aug), as_tibble(aug_dt),
      by = c(key_vars(aug_dt), index_var(aug_dt)),
      suffix = c("", ".y")
    )
  aug <- summarise(group_by(aug, !!!grp),
                   .resid = list(!!sym(".actual") - !!sym(".fc")),
                   .fc = list(!!sym(".fc")), .dist = list(!!sym(".dist")), 
                   .actual = list(!!sym(".actual")))
  
  # Extract training data (min key index, max grp index)
  extract_train <- function(idx, ...){
    cnds <- dots_list(...)
    cnds <- map2(syms(names(cnds)), cnds, function(x, y){
      if(is.na(y)) expr(is.na(!!x)) else expr(!!x == !!y)
    })
    eval_tidy(sym(resp), data = filter(data, !!index(data) < idx, !!!cnds))
  }
  mutual_keys <- intersect(key(data), key(object))
  mutual_keys <- set_names(mutual_keys, map_chr(mutual_keys, as_string))
  .train <- as_tibble(object) %>% 
    group_by(!!!union(key(object), grp)) %>% 
    filter(dplyr::row_number() == which.min(!!index(object))) %>% 
    group_by(!!!grp) %>% 
    filter(dplyr::row_number() == which.max(!!index(object))) %>% 
    transmute(.train = pmap(list2(idx = !!index(object), !!!mutual_keys), extract_train))
  aug <- left_join(aug, .train, by = map_chr(grp, as_string))

  # Add user inputs
  aug <- mutate(aug, ...)
  
  if(is.null(aug[[".period"]])){
    aug <- mutate(aug, .period = get_frequencies(NULL, object, .auto = "smallest"))
  }
  
  aug <- aug %>% 
    group_by(!!!grp) %>% 
    nest_grps(nm = ".accuracy")
  
  measures <- map(aug[[".accuracy"]], function(measures, inputs){
    map(measures, safely(do.call, otherwise = NA_real_), flatten(transpose(inputs)))
  }, measures = squash(measures))
  
  measures <- imap(measures, function(x, nm){
    err <- map_lgl(x, function(x) !is.null(x[["error"]]))
    if((tot_err <- sum(err)) > 0){
      err_msg <- table(map_chr(x[err], function(x) x[["error"]][["message"]]))
      warn(
        sprintf("%i error%s encountered\n%s\n",
                tot_err,
                if(tot_err > 1) sprintf("s (%i unique)", length(err_msg)) else "", 
                paste0("[", err_msg, "] ", names(err_msg), collapse = "\n")
        )
      )
    }
    as_tibble(map(x, function(x) x[["result"]]))
  })
  
  aug %>% 
    mutate(
      .type = "Test",
      .accuracy = measures
    ) %>% 
    unnest_tbl(".accuracy")
}

context("test-accuracy")

test_that("accuracy() hints", {
  skip_if_not_installed("fable")
  expect_error(
    accuracy(mbl, us_deaths),
    "To compute forecast accuracy, you'll need to compute the forecasts first.",
    fixed = TRUE
  )
})

test_that("In-sample accuracy", {
  skip_if_not_installed("fable")

  expect_warning(
    accuracy(mbl, measures = interval_accuracy_measures),
    'argument ".dist" is missing'
  ) %>%
    {is.na(.[["winkler"]])} %>%
    expect_true()

  acc <- accuracy(mbl)
  expect_equal(acc$.type, "Training")
  expect_equal(dim(acc), c(1,10))
  expect_true(!any(map_lgl(acc, compose(any, is.na))))
  expect_equal(
    as.list(acc),
    as_tibble(augment(mbl)) %>%
      group_by(.model) %>%
      summarise(.type = "Training", ME = mean(.resid), RMSE = sqrt(mean(.resid^2)),
                MAE = mean(abs(.resid)), MPE = mean(.resid/value*100),
                MAPE = mean(abs(.resid/value)*100),
                MASE = MASE(.resid, value, .period = 12),
                RMSSE = RMSSE(.resid, value, .period = 12),
                ACF1 = ACF1(.resid)) %>%
      as.list()
  )

  acc_multi <- accuracy(mbl_multi)
  expect_equal(acc_multi$key, c("fdeaths", "mdeaths"))
  expect_equal(dim(acc_multi), c(2,11))
  expect_true(!any(map_lgl(acc_multi, compose(any, is.na))))

  acc_complex <- accuracy(mbl_complex)
  expect_equal(acc_complex$key, rep(c("fdeaths", "mdeaths"), each = 2))
  expect_equal(acc_complex$.model, rep(c("ets", "lm"), 2))
  expect_equal(dim(acc_complex), c(4,11))
  expect_equal(acc_complex[c(1,3), -2], acc_multi[,-2])
  expect_true(!any(map_lgl(acc_complex, compose(any, is.na))))

  acc_mv <- accuracy(mbl_mv)
  expect_equal(
    acc_mv$.response,
    factor(c("mdeaths", "fdeaths"), levels = unique(c("mdeaths", "fdeaths")))
  )
  expect_true(!any(map_lgl(acc_mv, compose(any, is.na))))
})


test_that("Out-of-sample accuracy", {
  skip_if_not_installed("fable")

  expect_warning(
    accuracy(fbl, utils::head(us_deaths)),
    "12 observations are missing between 1978 Jan and 1978 Dec"
  )

  acc <- accuracy(fbl, us_deaths)
  expect_equal(acc$.type, "Test")
  expect_equal(dim(acc), c(1,10))
  expect_true(!any(map_lgl(acc, compose(any, is.na))))
  expect_equal(
    as.list(acc),
    as_tibble(fbl) %>%
      mutate(
        actual = semi_join(us_deaths, fbl, by = "index")$value,
        .resid = actual - mean(value)
      ) %>%
      group_by(.model) %>%
      summarise(.type = "Test", ME = mean(.resid), RMSE = sqrt(mean(.resid^2)),
                MAE = mean(abs(.resid)), MPE = mean(.resid/actual*100),
                MAPE = mean(abs(.resid/actual)*100),
                MASE = MASE(.resid, us_deaths_tr$value, .period = 12), 
                RMSSE = RMSSE(.resid, us_deaths_tr$value, .period = 12), 
                ACF1 = ACF1(.resid)) %>%
      as.list()
  )

  acc <- accuracy(fbl, us_deaths, measures = list(interval_accuracy_measures, distribution_accuracy_measures))
  expect_equal(acc$.type, "Test")
  expect_equal(colnames(acc), c(".model", ".type", "winkler", "pinball", "scaled_pinball", "percentile", "CRPS"))
  expect_true(!any(map_lgl(acc, compose(any, is.na))))

  acc_multi <- accuracy(fbl_multi, lung_deaths_long)
  expect_equal(acc_multi$key, c("fdeaths", "mdeaths"))
  expect_equal(dim(acc_multi), c(2,11))
  expect_true(!any(map_lgl(acc_multi, compose(any, is.na))))

  acc_complex <- accuracy(fbl_complex, lung_deaths_long)

  expect_equal(acc_complex$key, rep(c("fdeaths", "mdeaths"), 2))
  expect_equal(acc_complex$.model, rep(c("ets", "lm"), each = 2))
  expect_equal(dim(acc_complex), c(4,11))
  expect_equal(acc_complex[1:2, -(1:2)], acc_multi[,-(1:2)])
  expect_true(!any(map_lgl(acc_complex, compose(any, is.na))))

  acc_mv <- accuracy(fbl_mv, lung_deaths_wide)
  expect_equal(
    acc_mv$.response,
    c("fdeaths", "mdeaths")
  )
  expect_true(!any(map_lgl(acc_mv, compose(any, is.na))))
})

fbl_trend <- function(x, knots = NULL, origin = NULL) {
  idx_num <- as.double(x[[index_var(x)]])
  knots_num <- if (is.null(knots)) {
    NULL
  } else {
    as.double(knots)
  }
  index_interval <- default_time_units(interval(x))
  idx_num <- idx_num / index_interval
  knots_num <- knots_num / index_interval
  if (!is.null(origin)) {
    # trend should count from 1
    origin <- as.double(origin) / index_interval - 1
    idx_num <- idx_num - origin
    knots_num <- knots_num - origin
  }
  
  knots_exprs <- map(knots_num, function(.x) pmax(0, idx_num - .x))
  knots_exprs <- set_names(
    knots_exprs, 
    map_chr(knots_num, function(.x) paste0("trend_", format(.x)))
  )
  tibble(
    trend = idx_num,
    !!!knots_exprs
  )
}

fbl_season <- function(x, period) {
  idx_num <- as.double(x[[index_var(x)]])
  index_interval <- default_time_units(interval(x))
  idx_num <- idx_num / index_interval
  period <- get_frequencies(period, x, .auto = "smallest")
  season_exprs <- map(period, function(.x) expr(factor(floor((idx_num %% (!!.x)) + 1), levels = seq_len(!!.x))))
  season_exprs <- set_names(season_exprs, names(period) %||% paste0("season_", period))
  tibble(!!!season_exprs)
}

fbl_fourier <- function(x, period, K, origin = NULL) {
  idx_num <- as.double(x[[index_var(x)]])
  index_interval <- default_time_units(interval(x))
  idx_num <- idx_num / index_interval
  if (!is.null(origin)) {
    origin <- as.double(origin) / index_interval
  }
  period <- get_frequencies(period, x, .auto = "smallest")
  
  if (length(period) != length(K)) {
    abort("Number of periods does not match number of orders")
  }
  if (any(2 * K > period)) {
    abort("K must be not be greater than period/2")
  }
  
  fourier_exprs <- map2(
    as.numeric(period), K,
    function(period, K) {
      set_names(seq_len(K) / period, paste0(seq_len(K), "_", round(period)))
    }
  ) %>%
    invoke(c, .) %>%
    .[!duplicated(.)] %>%
    map2(., names(.), function(p, name) {
      out <- exprs(C = cospi(2 * !!p * idx_num))
      if (abs(2 * p - round(2 * p)) > .Machine$double.eps) {
        out <- c(out, exprs(S = sinpi(2 * !!p * idx_num)))
      }
      names(out) <- paste0(names(out), name)
      out
    }) %>%
    set_names(NULL) %>%
    unlist(recursive = FALSE)
  
  tibble(!!!fourier_exprs)
}

#' Common exogenous regressors
#'
#' These special functions provide interfaces to more complicated functions within
#' the model formulae interface.
#'
#' @aliases trend season fourier
#' 
#' @section Specials:
#'
#' \subsection{trend}{
#' The `trend` special includes common linear trend regressors in the model. It also supports piecewise linear trend via the `knots` argument.
#' \preformatted{
#' trend(knots = NULL, origin = NULL)
#' }
#'
#' \tabular{ll}{
#'   `knots`    \tab A vector of times (same class as the data's time index) identifying the position of knots for a piecewise linear trend.\cr
#'   `origin`   \tab An optional time value to act as the starting time for the trend.
#' }
#' }
#'
#' \subsection{season}{
#' The `season` special includes seasonal dummy variables in the model.
#' \preformatted{
#' season(period = NULL)
#' }
#'
#' \tabular{ll}{
#'   `period`   \tab The periodic nature of the seasonality. This can be either a number indicating the number of observations in each seasonal period, or text to indicate the duration of the seasonal window (for example, annual seasonality would be "1 year").
#' }
#' }
#'
#' \subsection{fourier}{
#' The `fourier` special includes seasonal fourier terms in the model. The maximum order of the fourier terms must be specified using `K`.
#' \preformatted{
#' fourier(period = NULL, K, origin = NULL)
#' }
#'
#' \tabular{ll}{
#'   `period`   \tab The periodic nature of the seasonality. This can be either a number indicating the number of observations in each seasonal period, or text to indicate the duration of the seasonal window (for example, annual seasonality would be "1 year"). \cr
#'   `K`        \tab The maximum order of the fourier terms.\cr
#'   `origin`   \tab An optional time value to act as the starting time for the fourier series.
#' }
#' }
#'
#' @format NULL
#' 
#' @export
common_xregs <- list(
  trend = function(knots = NULL, origin = NULL) {
    if (is.null(origin)) {
      if (is.null(self$origin)) {
        self$origin <- self$data[[index_var(self$data)]][[1]]
      }
      origin <- self$origin
    }
    as.matrix(fabletools:::fbl_trend(self$data, knots, origin))
  },
  season = function(period = NULL) {
    out <- as_model_matrix(fabletools:::fbl_season(self$data, period))
    stats::model.matrix(~., data = out)[, -1, drop = FALSE]
  },
  fourier = function(period = NULL, K, origin = NULL) {
    if (is.null(origin)) {
      if (is.null(self$origin)) {
        self$origin <- self$data[[index_var(self$data)]][[1]]
      }
      origin <- self$origin
    }
    as.matrix(fabletools:::fbl_fourier(self$data, period, K, origin))
  }
)

context("test-interpolate.R")

test_that("Test interpolation", {
  skip_if_not_installed("fable")
  NA_pos <- c(0, sample(seq_len(NROW(us_deaths)), 10), NROW(us_deaths))
  us_deaths$value[NA_pos] <- NA
  mbl_miss <- us_deaths %>%
    model(fable::TSLM(value ~ trend() + season()))
  interpolated <- interpolate(mbl_miss, us_deaths)
  expect_true(all(!is.na(interpolated$value)))
})

#' Unpack a hilo column
#'
#' @description 
#' `r lifecycle::badge('superseded')`
#' 
#' This function is superceded. It is recommended that you use the functionality
#' from the [distributional](https://pkg.mitchelloharawild.com/distributional/)
#' package to extract elements from a `<hilo>` object. For example, you can access
#' the lower bound with `<hilo>$lower`.
#' 
#' Allows a hilo column to be unpacked into its component columns: "lower", 
#' "upper", and "level".
#' 
#' @inheritParams tidyr::pack
#' @param cols Name of hilo columns to unpack.
#' 
#' @seealso [`tidyr::unpack()`]
#' 
#' @keywords internal
#' @export
unpack_hilo <- function(data, cols, names_sep = "_", names_repair = "check_unique"){
  orig <- data
  cols <- tidyselect::eval_select(enexpr(cols), data)
  if(any(bad_col <- !map_lgl(data[cols], inherits, "hilo"))){
    abort(sprintf(
      "Not all unpacking columns are hilo objects (%s). All unpacking columns of unpack_hilo() must be hilo vectors.",
      paste(names(bad_col)[bad_col], collapse = ", ")
    ))
  }
  data[cols] <- map(data[cols], function(x) vctrs::vec_proxy(x)[c("lower", "upper")])
  data <- tidyr::unpack(data, names(cols), names_sep = names_sep, names_repair = names_repair)
  vctrs::vec_restore(data, orig)
}

#' Produce forecasts
#' 
#' The forecast function allows you to produce future predictions of a time series
#' from fitted models. If the response variable has been transformed in the
#' model formula, the transformation will be automatically back-transformed
#' (and bias adjusted if `bias_adjust` is `TRUE`). More details about 
#' transformations in the fable framework can be found in
#' `vignette("transformations", package = "fable")`.
#' 
#' The forecasts returned contain both point forecasts and their distribution.
#' A specific forecast interval can be extracted from the distribution using the
#' [`hilo()`] function, and multiple intervals can be obtained using [`report()`].
#' These intervals are stored in a single column using the `hilo` class, to
#' extract the numerical upper and lower bounds you can use [`unpack_hilo()`].
#' 
#' @param object The time series model used to produce the forecasts
#' @param new_data A `tsibble` containing future information used to forecast.
#' @param h The forecast horison (can be used instead of `new_data` for regular
#' time series with no exogenous regressors).
#' @param simulate Should forecasts be based on simulated future paths instead
#' of analytical results.
#' @param bootstrap Should innovations from simulated forecasts be bootstrapped
#' from the model's fitted residuals. This allows the forecast distribution to
#' have a different underlying shape which could better represent the nature
#' of your data.
#' @param times The number of future paths for simulations if `simulate = TRUE`.
#' @param point_forecast The point forecast measure(s) which should be returned 
#' in the resulting fable. Specified as a named list of functions which accept
#' a distribution and return a vector. To compute forecast medians, you can use
#' `list(.median = median)`.
#' @param bias_adjust Deprecated. Please use `point_forecast` to specify the 
#' desired point forecast method.
#' @param ... Additional arguments for forecast model methods.
#' 
#' @return
#' A fable containing the following columns:
#' - `.model`: The name of the model used to obtain the forecast. Taken from
#'   the column names of models in the provided mable.
#' - The forecast distribution. The name of this column will be the same as the 
#'   dependent variable in the model(s). If multiple dependent variables exist,
#'   it will be named `.distribution`.
#' - Point forecasts computed from the distribution using the functions in the
#'   `point_forecast` argument.
#' - All columns in `new_data`, excluding those whose names conflict with the
#'   above.
#'   
#' @examplesIf requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE) && requireNamespace("urca", quietly = TRUE)
#' library(fable)
#' library(tsibble)
#' library(tsibbledata)
#' library(dplyr)
#' library(tidyr)
#' 
#' # Forecasting with an ETS(M,Ad,A) model to Australian beer production
#' beer_fc <- aus_production %>%
#'   model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) %>% 
#'   forecast(h = "3 years")
#' 
#' # Compute 80% and 95% forecast intervals
#' beer_fc %>% 
#'   hilo(level = c(80, 95))
#' 
#' beer_fc %>% 
#'   autoplot(aus_production)
#' 
#' # Forecasting with a seasonal naive and linear model to the monthly 
#' # "Food retailing" turnover for each Australian state/territory.
#' library(dplyr)
#' aus_retail %>% 
#'   filter(Industry == "Food retailing") %>% 
#'   model(
#'     snaive = SNAIVE(Turnover),
#'     ets = TSLM(log(Turnover) ~ trend() + season()),
#'   ) %>% 
#'   forecast(h = "2 years 6 months") %>% 
#'   autoplot(filter(aus_retail, Month >= yearmonth("2000 Jan")), level = 90)
#'   
#' # Forecast GDP with a dynamic regression model on log(GDP) using population and
#' # an automatically chosen ARIMA error structure. Assume that population is fixed
#' # in the future.
#' aus_economy <- global_economy %>% 
#'   filter(Country == "Australia")
#' fit <- aus_economy %>% 
#'   model(lm = ARIMA(log(GDP) ~ Population))
#' 
#' future_aus <- new_data(aus_economy, n = 10) %>% 
#'   mutate(Population = last(aus_economy$Population))
#' 
#' fit %>% 
#'   forecast(new_data = future_aus) %>% 
#'   autoplot(aus_economy)
#' 
#' @rdname forecast
#' @export
forecast.mdl_df <- function(object, new_data = NULL, h = NULL, 
                            point_forecast = list(.mean = mean), ...){
  mdls <- mable_vars(object)
  if(!is.null(h) && !is.null(new_data)){
    warn("Input forecast horizon `h` will be ignored as `new_data` has been provided.")
    h <- NULL
  }
  if(!is.null(new_data)){
    object <- bind_new_data(object, new_data)
  }
  kv <- c(key_vars(object), ".model")
  
  # Evaluate forecasts
  object <- dplyr::mutate_at(as_tibble(object), vars(!!!mdls),
                             forecast, new_data = object[["new_data"]],
                             h = h, point_forecast = point_forecast, ...,
                             key_data = key_data(object))
  
  object <- tidyr::pivot_longer(object, !!mdls, names_to = ".model", values_to = ".fc") 
  
  # Combine and re-construct fable
  fbl_attr <- attributes(object$.fc[[1]])
  out <- suppressWarnings(
    unnest_tsbl(as_tibble(object)[c(kv, ".fc")], ".fc", parent_key = kv)
  )
  build_fable(out, response = fbl_attr$response, distribution = fbl_attr$dist)
}

#' @export
forecast.mdl_lst <- function(object, new_data = NULL, key_data, ...){
  mapply_maybe_parallel(
    .f = forecast,
    object, 
    new_data %||% rep(list(NULL), length.out = length(object)),
    MoreArgs = dots_list(...)
  )
}

#' @export
forecast.lst_mdl <- deprecate_lst_mdl(forecast.mdl_lst)

#' @rdname forecast
#' @export
forecast.mdl_ts <- function(object, new_data = NULL, h = NULL, bias_adjust = NULL,
                            simulate = FALSE, bootstrap = FALSE, times = 5000,
                            point_forecast = list(.mean = mean), ...){
  if(!is.null(h) && !is.null(new_data)){
    warn("Input forecast horizon `h` will be ignored as `new_data` has been provided.")
    h <- NULL
  }
  if(!is.null(bias_adjust)){
    deprecate_warn("0.2.0", "forecast(bias_adjust = )", "forecast(point_forecast = )")
    point_forecast <- if(bias_adjust) list(.mean = mean) else list(.median = stats::median)
  }
  if(is.null(new_data)){
    new_data <- make_future_data(object$data, h)
  }
  
  # Useful variables
  idx <- index_var(new_data)
  mv <- measured_vars(new_data)
  resp_vars <- vapply(object$response, expr_name, character(1L), USE.NAMES = FALSE)
  dist_col <- if(length(resp_vars) > 1) ".distribution" else resp_vars
  
  # If there's nothing to forecast, return an empty fable.
  if(NROW(new_data) == 0){
    new_data[[dist_col]] <- distributional::new_dist(dimnames = resp_vars)
    fbl <- build_fable(new_data, response = resp_vars, distribution =  dist_col)
    return(fbl)
  }
  # Compute forecasts
  if(simulate || bootstrap) {
    fc <- generate(object, new_data, bootstrap = bootstrap, times = times, ...)
    fc_idx <- fc[[index_var(fc)]]
    fc <- if (length(resp_vars) > 1) {
      do.call(cbind, fc[resp_vars])
    } else {
      fc[[".sim"]]
    }
    
    fc <- distributional::dist_sample(vctrs::vec_split(fc, fc_idx)$val)
  } else {
    # Compute specials with new_data
    object$model$stage <- "forecast"
    object$model$add_data(new_data)
    specials <- tryCatch(parse_model_rhs(object$model),
                         error = function(e){
                           abort(sprintf(
  "%s
  Unable to compute required variables from provided `new_data`.
  Does your model require extra variables to produce forecasts?", e$message))
                         }, interrupt = function(e) {
                           stop("Terminated by user", call. = FALSE)
                         })
    object$model$remove_data()
    object$model$stage <- NULL
    fc <- forecast(object$fit, new_data, specials = specials, times = times, ...)
    
    # Back-transform forecast distributions
    bt <- map(object$transformation, function(x){
      trans <- x%@%"inverse"
      inv_trans <- `attributes<-`(x, NULL)
      req_vars <- setdiff(all.vars(body(trans)), names(formals(trans)))
      if(any(req_vars %in% names(new_data))) {
        trans <- lapply(
          vec_chop(new_data[req_vars]),
          function(transform_data) {
            set_env(trans, new_environment(transform_data, get_env(trans)))
          }
        )
        attr(trans, "inverse") <- lapply(
          vec_chop(new_data[req_vars]),
          function(transform_data) {
            set_env(inv_trans, new_environment(transform_data, get_env(inv_trans)))
          }
        )
        trans
      } else {
        structure(list(trans), inverse = list(inv_trans))
      }
    })
    
    is_transformed <- vapply(bt, function(x) !is_symbol(body(x[[1]])), logical(1L))
    if(length(bt) > 1) {
      if(any(is_transformed)){
        abort("Transformations of multivariate forecasts distributions are not supported, use simulate = TRUE or bootstrap = TRUE.")
      }
    }
    if(any(is_transformed)) {
      if (identical(unique(dist_types(fc)), "dist_sample")) {
        fc <- distributional::dist_sample(
          .mapply(exec, list(bt[[1]], distributional::parameters(fc)$x), MoreArgs = NULL)
        )
      } else {
        bt <- bt[[1]]
        fc <- distributional::dist_transformed(fc, `attributes<-`(bt, NULL), bt%@%"inverse")
      }
    }
  }
  
  dimnames(fc) <- resp_vars
  
  new_data[[dist_col]] <- fc
  point_fc <- compute_point_forecasts(fc, point_forecast)
  new_data[names(point_fc)] <- point_fc
  
  cn <- c(dist_col, names(point_fc))
  
  fbl <- build_tsibble_meta(
    as_tibble(new_data)[unique(c(idx, cn, mv))],
    key_data(new_data),
    index = idx, index2 = idx, ordered = is_ordered(new_data),
    interval = interval(new_data)
  )
  
  build_fable(fbl, response = resp_vars, distribution = dist_col)
}

#' Construct a new set of forecasts
#' 
#' @description 
#' `r lifecycle::badge('deprecated')`
#' 
#' This function is deprecated. `forecast()` methods for a model should return
#' a vector of distributions using the distributional package.
#' 
#' Backtransformations are automatically handled, and so no transformations should be specified here.
#' 
#' @param point The transformed point forecasts
#' @param sd The standard deviation of the transformed forecasts
#' @param dist The forecast distribution (typically produced using `new_fcdist`)
#' 
#' @keywords internal
#' @export
construct_fc <- function(point, sd, dist){
  lifecycle::deprecate_stop("0.3.0", what = "fabletools::construct_fc()",
                            details = "The forecast function should now return a vector of distributions from the 'distributional' package.")
}

compute_point_forecasts <- function(distribution, measures){
  map(measures, calc, distribution)
}

#' @export
forecast.fbl_ts <- function(object, ...){
  abort("Did you try to forecast a fable? Forecasts can only be computed from model objects (such as a mable).")
}

#' A set of future scenarios for forecasting
#' 
#' @param ... Input data for each scenario
#' @param names_to The column name used to identify each scenario
#' 
#' @export
scenarios <- function(..., names_to = ".scenario"){
  structure(list2(
    ...
  ), names_to = names_to)
}

#' Estimate a model
#' 
#' @param .data A data structure suitable for the models (such as a `tsibble`).
#' @param ... Further arguments passed to methods.
#' 
#' @rdname estimate
#' 
#' @export
estimate <- function(.data, ...){
  UseMethod("estimate")
}

#' @param .model Definition for the model to be used.
#' 
#' @rdname estimate
#' @export
estimate.tbl_ts <- function(.data, .model, ...){
  if(!inherits(.model, "mdl_defn")){
    abort("Model definition incorrectly created. Check that specified model(s) are model definitions.")
  }
  .model$stage <- "estimate"
  .model$add_data(.data)
  validate_formula(.model, .data)
  parsed <- parse_model(.model)
  
  # Compute response data (as attributes shouldn't change, using this approach should be much faster)
  .dt_attr <- attributes(.data)
  resp <- map(parsed$expressions, eval_tidy, data = .data, env = .model$specials)
  .data <- unclass(.data)[index_var(.data)]
  .data[map_chr(parsed$expressions, expr_name)] <- resp
  attributes(.data) <- c(attributes(.data), .dt_attr[setdiff(names(.dt_attr), names(attributes(.data)))])
  
  fit <- eval_tidy(
    expr(.model$train(.data = .data, specials = parsed$specials, !!!.model$extra))
  )
  .model$remove_data()
  .model$stage <- NULL
  new_model(fit, .model, .data, parsed$response, parsed$transformation)
}

deprecate_lst_mdl <- function(f) {
  fsym <- ensym(f)
  fmls <- formals(args(f))
  args <- names(fmls)
  args <- `names<-`(syms(args), args)
  # Remove name of `...` if it exists
  arg_dots <- match("...", names(args), 0L)
  names(args)[arg_dots] <- ""

  rlang::new_function(
    args = fmls,
    body = rlang::expr({
      lifecycle::deprecate_soft(
          "0.7.0",
          I("The `lst_mdl` class"),
          I("the `mdl_lst` class for lists of models")
        )
        (!!fsym)(!!!args)
      }
    ),
    env = rlang::caller_env()
  )
}

# nocov start

# This file serves as a reference for compatibility functions for
# purrr. They are not drop-in replacements but allow a similar style
# of programming. This is useful in cases where purrr is too heavy a
# package to depend on.

# Slightly adapted from the version found in rlang

map <- function(.x, .f, ...) {
  lapply(.x, .f, ...)
}
map_mold <- function(...) {
  out <- vapply(..., USE.NAMES = FALSE)
  names(out) <- names(..1)
  out
}
map_lgl <- function(.x, .f, ...) {
  map_mold(.x, .f, logical(1), ...)
}
map_int <- function(.x, .f, ...) {
  map_mold(.x, .f, integer(1), ...)
}
map_dbl <- function(.x, .f, ...) {
  map_mold(.x, .f, double(1), ...)
}
map_chr <- function(.x, .f, ...) {
  map_mold(.x, .f, character(1), ...)
}
map_cpl <- function(.x, .f, ...) {
  map_mold(.x, .f, complex(1), ...)
}

pluck <- function(.x, .f) {
  map(.x, `[[`, .f)
}
pluck_lgl <- function(.x, .f) {
  map_lgl(.x, `[[`, .f)
}
pluck_int <- function(.x, .f) {
  map_int(.x, `[[`, .f)
}
pluck_dbl <- function(.x, .f) {
  map_dbl(.x, `[[`, .f)
}
pluck_chr <- function(.x, .f) {
  map_chr(.x, `[[`, .f)
}
pluck_cpl <- function(.x, .f) {
  map_cpl(.x, `[[`, .f)
}

map2 <- function(.x, .y, .f, ...) {
  mapply(.f, .x, .y, MoreArgs = list(...), SIMPLIFY = FALSE)
}
map2_lgl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "logical")
}
map2_int <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "integer")
}
map2_dbl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "double")
}
map2_chr <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "character")
}
map2_cpl <- function(.x, .y, .f, ...) {
  as.vector(map2(.x, .y, .f, ...), "complex")
}

args_recycle <- function(args) {
  vctrs::vec_recycle_common(!!!args)
}
pmap <- function(.l, .f, ...) {
  args <- args_recycle(.l)
  do.call("mapply", c(
    FUN = list(quote(.f)),
    args, MoreArgs = quote(list(...)),
    SIMPLIFY = FALSE, USE.NAMES = FALSE
  ))
}

probe <- function(.x, .p, ...) {
  if (is_logical(.p)) {
    stopifnot(length(.p) == length(.x))
    .p
  } else {
    map_lgl(.x, .p, ...)
  }
}

keep <- function(.x, .f, ...) {
  .x[probe(.x, .f, ...)]
}
discard <- function(.x, .p, ...) {
  sel <- probe(.x, .p, ...)
  .x[is.na(sel) | !sel]
}
map_if <- function(.x, .p, .f, ...) {
  matches <- probe(.x, .p)
  .x[matches] <- map(.x[matches], .f, ...)
  .x
}

compact <- function(.x) {
  Filter(length, .x)
}

transpose <- function(.l) {
  inner_names <- names(.l[[1]])

  result <- map(seq_along(.l[[1]]), function(i) {
    map(.l, .subset2, i)
  })
  
  set_names(result, inner_names)
}

transpose_dbl <- function(.l) {
  inner_names <- names(.l[[1]])
  
  result <- map(seq_along(.l[[1]]), function(i) {
    map_dbl(.l, .subset2, i)
  })
  
  set_names(result, inner_names)
}

every <- function(.x, .p, ...) {
  for (i in seq_along(.x)) {
    if (!rlang::is_true(.p(.x[[i]], ...))) return(FALSE)
  }
  TRUE
}
some <- function(.x, .p, ...) {
  for (i in seq_along(.x)) {
    if (rlang::is_true(.p(.x[[i]], ...))) return(TRUE)
  }
  FALSE
}
negate <- function(.p) {
  function(...) !.p(...)
}

reduce <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init)
}
reduce_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE)
}
accumulate <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(x, y, ...)
  Reduce(f, .x, init = .init, accumulate = TRUE)
}
accumulate_right <- function(.x, .f, ..., .init) {
  f <- function(x, y) .f(y, x, ...)
  Reduce(f, .x, init = .init, right = TRUE, accumulate = TRUE)
}

invoke <- function(.f, .x, ..., .env = NULL){
  .env <- .env %||% parent.frame()
  args <- c(as.list(.x), list(...))
  do.call(.f, args, envir = .env)
}
imap <- function(.x, .f, ...){
  map2(.x, names(.x) %||% seq_along(.x), .f, ...)
}

capture_error <- function (code, otherwise = NULL, quiet = TRUE) 
{
  tryCatch(list(result = code, error = NULL), error = function(e) {
    if (!quiet) 
      message("Error: ", e$message)
    list(result = otherwise, error = e)
  }, interrupt = function(e) {
    stop("Terminated by user", call. = FALSE)
  })
}
safely <- function (.f, otherwise = NULL, quiet = TRUE) 
{
  function(...) capture_error(.f(...), otherwise, quiet)
}
possibly <- function (.f, otherwise, quiet = TRUE) 
{
  force(otherwise)
  function(...) capture_error(.f(...), otherwise, quiet)$result
}
compose <- function (...) {
  fs <- lapply(list(...), match.fun)
  n <- length(fs)
  last <- fs[[n]]
  rest <- fs[-n]
  function(...) {
    out <- last(...)
    for (f in rev(rest)) {
      out <- f(out)
    }
    out
  }
}

flatten <- function(x) {
  unlist(x, recursive = FALSE)
}

squash <- function(x) {
  unlist(x, recursive = TRUE)
}

# nocov end

#' Extract frequencies for common seasonal periods
#'
#' @param x An object containing temporal data (such as a `tsibble`, `interval`, `datetime` and others.)
#'
#' @return A named vector of frequencies appropriate for the provided data.
#'
#' @references <https://robjhyndman.com/hyndsight/seasonal-periods/>
#'
#' @rdname freq_tools
#'
#' @examples
#' common_periods(tsibble::pedestrian)
#'
#' @export
common_periods <- function(x){
  UseMethod("common_periods")
}

#' @rdname freq_tools
#' @export
common_periods.default <- function(x){
  common_periods(interval_pull(x))
}

#' @rdname freq_tools
#' @export
common_periods.tbl_ts <- function(x){
  common_periods(tsibble::interval(x))
}

#' @rdname freq_tools
#' @export
common_periods.interval <- function(x){
  if(inherits(x, "vctrs_vctr")){
    x <- vctrs::vec_data(x)
  }
  freq_sec <- c(year = 31557600, week = 604800, day = 86400, hour = 3600, minute = 60, second = 1,
                millisecond = 1e-3, microsecond = 1e-6, nanosecond = 1e-9)
  nm <- names(x)[x!=0]
  if(is_empty(x)) return(NULL)
  switch(paste(nm, collapse = ""),
         "unit" = c("none" = 1),
         "year" = c("year" = 1),
         "quarter" = c("year" = 4/x[["quarter"]]),
         "month" = c("year" = 12/x[["month"]]),
         "week" = c("year" = 52/x[["week"]]),
         "day" = c("year" = 365.25, "week" = 7)/x[["day"]],
         with(list(secs = freq_sec/sum(as.numeric(x)*freq_sec[nm])), secs[secs>1])
  )
}

#' @rdname freq_tools
#' @param period Specification of the time-series period
#' @param ... Other arguments to be passed on to methods
#' @export
get_frequencies <- function(period, ...){
  UseMethod("get_frequencies")
}

#' @rdname freq_tools
#' @export
get_frequencies.numeric <- function(period, ...){
  period
}

#' @rdname freq_tools
#' @param data A tsibble
#' @param .auto The method used to automatically select the appropriate seasonal
#' periods
#' @export
get_frequencies.NULL <- function(period, data, ...,
                                 .auto = c("smallest", "largest", "all")){
  .auto <- match.arg(.auto)
  frequencies <- Filter(function(x) x >= 1, common_periods(data))
  if(is_empty(frequencies)) frequencies <- 1
  if(.auto == "smallest") {
    return(frequencies[which.min(frequencies)])
  }
  else if(.auto == "largest"){
    return(frequencies[which.max(frequencies)])
  }
  else {
    return(frequencies)
  }
}

#' @rdname freq_tools
#' @export
get_frequencies.character <- function(period, data, ...){
  check_installed("lubridate")
  m <- lubridate::as.period(period)
  if(is.na(m)) abort(paste("Unknown period:", period))
  get_frequencies(m, data, ...)
}

#' @rdname freq_tools
#' @export
get_frequencies.Period <- function(period, data, ...){
  check_installed("lubridate")
  
  interval <- tsibble::interval(data)
  
  interval <- with(interval, lubridate::years(year) + 
    lubridate::period(3*quarter + month, units = "month") + lubridate::weeks(week) +
    lubridate::days(day) + lubridate::hours(hour) + lubridate::minutes(minute) + 
    lubridate::seconds(second) + lubridate::milliseconds(millisecond) + 
    lubridate::microseconds(microsecond) + lubridate::nanoseconds(nanosecond))
  
  suppressMessages(period / interval)
}

context("test-fable")

test_that("fable dplyr verbs", {
  skip_if_not_installed("fable")
  
  fbl_complex %>% filter(key == "mdeaths") %>% 
    expect_s3_class("fbl_ts") %>% 
    NROW %>% 
    expect_equal(24)
  
  # tsibble now automatically selects keys
  # expect_error(
  #   fbl_complex %>% select(index, .model, value, .distribution),
  #   "not a valid tsibble"
  # )
  
  fbl_complex %>%
    filter(key == "mdeaths") %>%
    select(index, .model, value, .mean) %>% 
    n_keys() %>% 
    expect_equal(2)
  
  expect_equal(
    colnames(hilo(fbl_complex, level = c(50, 80, 95))),
    c("key", ".model", "index", "value", ".mean", "50%", "80%", "95%")
  )
  
  expect_equivalent(
    as.list(fbl_multi),
    as.list(bind_rows(fbl_multi[1:12,], fbl_multi[13:24,]))
  )
})

context("test-multivariate.R")

test_that("multiple univariate", {
  skip_if_not_installed("fable")
  
  expect_equal(sort(mbl_multi$key), c("fdeaths", "mdeaths"))
  expect_s3_class(mbl_multi[[attr(mbl_multi,"model")[[1]]]], "lst_mdl")
  
  expect_equal(sort(unique(fbl_multi$key)), c("fdeaths", "mdeaths"))
  expect_s3_class(fbl_multi[[attr(fbl_multi,"dist")]], "distribution")
})

context("test-hilo.R")

test_that("Extracting intervals from a distribution", {
  skip_if_not_installed("fable")
  
  fc_sum <- fbl %>% 
    mutate(`80%` = hilo(value, 80))
  
  expect_s3_class(fc_sum$`80%`, "hilo")
})

context("test-broom")

test_that("augment", {
  skip_if_not_installed("fable")
  
  aug <- augment(mbl)
  expect_equal(aug$index, us_deaths_tr$index)
  expect_equal(aug$.fitted, fitted(mbl)$.fitted)
  expect_equal(aug$.resid, residuals(mbl, type ="response")$.resid)
  expect_equal(aug$.innov, residuals(mbl)$.resid)
  
  aug <- augment(mbl_multi)
  expect_equal(aug$index, lung_deaths_long_tr$index)
  expect_equal(aug$.fitted, fitted(mbl_multi)$.fitted)
  expect_equal(aug$.resid, residuals(mbl_multi, type = "response")$.resid)
  expect_equal(aug$.innov, residuals(mbl_multi)$.resid)
  
  aug <- augment(mbl_complex)
  expect_equal(aug$index, rep(lung_deaths_long_tr$index, 2))
  expect_equal(aug$.fitted, fitted(mbl_complex)$.fitted)
  expect_equal(aug$.resid, residuals(mbl_complex, type = "response")$.resid)
  expect_equal(aug$.innov, residuals(mbl_complex)$.resid)

  aug <- augment(mbl_mv)
  expect_equal(aug$index, rep(lung_deaths_wide_tr$index, 2))
  expect_equal(aug$.fitted, c(fitted(mbl_mv)$mdeaths, fitted(mbl_mv)$fdeaths))
  expect_equal(aug$.resid, c(residuals(mbl_mv)$mdeaths, residuals(mbl_mv)$fdeaths))
})

test_that("glance", {
  skip_if_not_installed("fable")
  
  gl <- glance(mbl)
  expect_equal(NROW(gl), 1)
  gl_multi <- glance(mbl_multi)
  expect_equal(NROW(gl_multi), 2)
  expect_equal(gl_multi$key, c("fdeaths", "mdeaths"))
  gl_complex <- glance(mbl_complex)
  expect_equal(NROW(gl_complex), 4)
  expect_equal(gl_complex$key, rep(c("fdeaths", "mdeaths"), each = 2))
  expect_equal(gl_multi[-2], gl_complex[c(1,3), names(gl_multi)][-2])
  
  gl_mv <- glance(mbl_mv)
  expect_equal(NROW(gl_mv), 1)
})

test_that("tidy", {
  skip_if_not_installed("fable")
  
  td <- tidy(mbl)
  expect_equal(unique(td$.model), "ets")
  td_multi <- tidy(mbl_multi)
  expect_equal(unique(td_multi$.model), "ets")
  expect_equal(unique(td_multi$key), c("fdeaths", "mdeaths"))
  td_complex <- tidy(mbl_complex)
  expect_equal(unique(td_complex$.model), c("ets", "lm"))
  expect_equal(unique(td_complex$key), c("fdeaths", "mdeaths"))
  
  td_mv <- tidy(mbl_mv)
  expect_equal(unique(td_mv$.model), "var")
  expect_equal(unique(td_mv$.response), c("mdeaths", "fdeaths"))
})

---
title: "Extending fabletools: Models"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Extending fabletools: Models}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(fabletools)
```

Writing a time series model using fabletools provides your model with many additional features without extra effort. Features that aren't model specific are handled by fabletools, allowing you to spend more time writing methods for your model. Some functionality handled by fabletools includes:

* Seamless integration with data manipulation and visualisation tools from the tidyverse.
* Consistent formula interface with other tidy time series models.
* Transformations with automatic back-transformation of response variables.
* Batch modelling of many time series and models with parallel support.
* Accuracy evaluation tools (in-sample, out-of-sample, and cross validation).
* Visualisation functions for decompositions, models, and forecasts.
* Support for ensemble and combination modelling (such as decomposition forecasting).
* Hierarchical, grouped, and temporal reconciliation of forecasts.

The fabletools package promotes consistent interfaces and output structures that allow various time series models to work well together. This vignette will guide you through creating a fabletools model, and provide a glimpse into the steps used to convert user data to modelling inputs, and modelling outputs to user data.

As an example, we'll create a fabletools model that uses seasonal averages: `SMEAN()`. It can be thought of as a seasonal version of `fable::MEAN()`, which instead of averaging the entire series, it averages values from each season.

## The model interface

Much like cross-sectional models (such as `lm()`), tidy time-series models use a formula based interface. Of course not all arguments need to be specified from within the formula (much like `na.action` in `lm()`). The model formula is a familiar and user friendly interface for specifying key model concepts (like `pdq()` in `ARIMA()`), and data-varying inputs (such as holidays and exogenous regressors). Model specific formula functions (like `pdq()`) are known as specials (much like `specials` from `stats::terms.formula()`).

Before writing code, it is a good idea to think about what interface best suits your model. This is often model specific, however you may find it useful to look at existing interfaces to see how yours could be written consistently. A good example of this is with seasonality: fourier terms are specified with the `fourier(period, K)` special, and seasonal dummy variables use `season(period)`.

A potential interface for the `SMEAN()` model could be:

```r
SMEAN(y ~ season(period))
```

## Minimum implementation of a model

At minimum, a model consists of a model function (something that returns a model definition), a set of specials, and a training function.

### The model function

Model functions typically consist of two function calls. A model class (defining the training method, the specials, and data checks) with `new_model_class()`, and `new_model_definition` to return the model definition:

```{r}
#' Seasonal mean models
#' 
#' Add the rest of your documentation here.
#' Typically this includes a "Specials" section
#' 
#' @export
SMEAN <- function(formula, ...) {
  # Create a model class which combines the training method, specials, and data checks
  model_smean <- new_model_class("smean",
    # The training method (more on this later)
    train = train_smean,
    # The formula specials (the next section)
    specials = specials_smean,
    # Any checks of the unprocessed data, like gaps, ordered, regular, etc.
    check = function(.data) { 
      if (!tsibble::is_regular(.data)) stop("Data must be regular") 
    }
  )
  
  # Return a model definition which stores the user's model specification
  new_model_definition(model_smean, {{formula}}, ...)
}
```

Anything passed to `...` of `new_model_definition()` will be passed onward to the model training function. Note that the formula needs to be embraced with `{{formula}}` in order to allow for non-formula inputs like `SMEAN(y)`.

### The specials

The specials for a model are created using `new_specials()`. The functions specified here will be used to compute specials each time the model is provided with new data (model training, forecasting, refitting, etc.). The results of these functions will be passed to the subsequent method via the `specials` argument (more on this later).

To enable automatic model specification (with `SMEAN(y)`), the `.required_specials` argument will ensure that the special is called at least once. By setting `season()` as a required special, `SMEAN(y)` will be parsed as `SMEAN(y ~ season())`. As no arguments will be provided for omitted required specials, make sure they have good defaults. The `fabletools::get_frequencies()` function is a useful helper for handling seasonal periods, as automatically chooses appropriate seasonalities when `period = NULL`, and is able to handle inputs like `period = "week"`.

Anything not handled by defined specials will be treated as exogenous regressors and passed to the `xreg()` special. That is to say `SMEAN(y ~ season("year") + x)` will be parsed as `SMEAN(y ~ season("year") + xreg(x))`. The `xreg()` special should be defined by all models, even if your model doesn't support it.

```{r}
specials_smean <- new_specials(
  season = function(period = NULL) {
    # Your input handling code here.
    get_frequencies(period, self$data, .auto = "smallest")
  },
  xreg = function(...) {
    # This model doesn't support exogenous regressors, time to error.
    stop("Exogenous regressors aren't supported by `SMEAN()`")
  },
  # This model requires `season()`
  # Adding this allows `SMEAN(y)` to automatically include the `season()` special
  .required_specials = "season"
)
```

The specials are the only thing needed for the formula to work, as the fabletools handles the transformations and response variables specified in the formula's left side.

### The training function

This function is used to apply the model definition created by the model function (`SMEAN()`) to users data when they use `model(data, SMEAN())`.

The `.data` argument is a single series tsibble (no keys), representing the parsed left side of the formula. The index of `.data` is the time of the measurement, and the measured variables are the transformed response variable(s).

The `specials` argument is a list of results from parsing the specials used in the right side of the formula. The result from the `season()` special in `SMEAN(y ~ season("year"))` would be accessible from `specials$season[[1]]`. As specials can be used more than once, the nth usage of special `xyz()` can be accessed with `specials$xyz[[n]]`.

As mentioned earlier, `...` will contain additional parameters passed `...` of `new_model_definition()`.

The function should return an S3 object that contains everything you need for your future methods (such as forecasting, getting fitted values, refitting, etc.).

```{r}
train_smean <- function(.data, specials, ...){
  # Extract a vector of response data
  mv <- tsibble::measured_vars(.data)
  if(length(mv) > 1) stop("SMEAN() is a univariate model.")
  y <- .data[[mv]]
  
  # Pull out inputs from the specials
  if(length(specials$season) > 1) stop("The `season()` special of `SMEAN()` should only be used once.")
  m <- specials$season[[1]]
  
  # Compute the seasonal averages
  season_id <- seq(0, length(y) - 1) %% m
  season_y <- split(y, season_id)
  season_avg <- vapply(season_y, FUN = mean, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE)
  
  # Compute fitted values and residuals
  fit <- season_avg[season_id+1]
  e <- y - fit
  
  # Create S3 model object
  # It should be small, but contain everything needed for methods below
  structure(
    list(
      coef = season_avg,
      n = length(y),
      y_name = mv,
      fitted = fit,
      residuals = e,
      sigma2 = var(e, na.rm = TRUE)
    ),
    class = "model_smean"
  )
}
```

Great, that's the bare minimum for a model complete with interface and training method. Let's try it out.

```{r first-mable}
fit <- tsibbledata::aus_production %>%
  model(SMEAN(Beer))
fit
```

It doesn't look like much, but it has used the above specials and training method to compute the seasonal average and store it in the object. However we can't see any details about the model yet. To make the model useful, we'll need to define some methods.

## Methods for models

```{r, echo = FALSE}
tibble::tribble(
  ~ Method, ~ Value, ~ Description,
  "`model_sum()`", "`character(1L)`", "A short summary of the model to display in the mable",
  "`report()`", "console output", "A detailed summary of the model, similar to `summary()`",
  "`equation()`", "character(1L)", "The mathematical equation for the fitted model",
  "`forecast()`", "distribution", "Produce forecasts from the model",
  "`stream()`", "updated model", "Extend the fit of the model with additional data",
  "`generate()`", "tsibble", "Generate potential reponse values at certain times from the model",
  "`interpolate()`", "tsibble", "Interpolate missing values using the model",
  "`refit()`", "refitted model", "Apply the model to a new dataset",
  "`tidy()`", "tibble of coefficients", "Extract coefficients from the model",
  "`glance()`", "tibble of statistics", "Extract summary statistics from the model",
  "`augment()`", "tibble of data", "Augment a dataset with information from the model",
  "`components()`", "dable of components", "Extract decomposed elements from the model",
  "`fitted()`", "numeric", "Extract fitted values from the model",
  "`residuals()`", "numeric", "Extract residuals from the model"
) %>% 
  knitr::kable()
```


```{r}
#' @importFrom fabletools model_sum
#' @export
model_sum.model_smean <- function(x){
  sprintf("SMEAN[%i]", length(x$coef))
}

fit
```

```{r}
#' @importFrom fabletools report
#' @export
report.model_smean <- function(x){
  m <- length(x$coef)
  
  cat("\n")
  cat(paste("Seasonal period:", m))
  cat("\n\n")
  cat("Seasonal averages:\n")
  
  print.default(
    setNames(x$coef, paste0("s", seq_len(m))),
    print.gap = 2
  )
  cat(paste("\nsigma^2:", round(x$sigma2, 4), "\n"))
}

report(fit)
```

```{r}
#' @importFrom fabletools tidy
#' @export
tidy.model_smean <- function(x){
  tibble::tibble(
    term = paste0("season_", seq_along(x$coef)), 
    estimate = x$coef
  )
}

tidy(fit)
```

```{r}
#' @importFrom fabletools glance
#' @export
glance.model_smean <- function(x){
  tibble::tibble(
    sigma2 = x$sigma2
  )
}

glance(fit)
```

```{r}
#' @importFrom fabletools forecast
#' @export
forecast.model_smean <- function(object, new_data, ...){
  # Extract required parameters
  h <- NROW(new_data)
  n <- object$n
  m <- length(object$coef)
  coef <- object$coef
  
  # Compute forecast variance
  season_id <- seq(0, n - 1) %% m
  season_e <- split(object$residuals, season_id)
  season_sd <- vapply(season_e, FUN = sd, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE, na.rm = TRUE)
  
  # Create forecast distributions
  fc_id <- (seq(0, h-1) + n %% m) %% m + 1
  mu <- coef[fc_id]
  sigma <- season_sd[fc_id]
  distributional::dist_normal(mu, sigma)
}

forecast(fit)
```

```{r}
#' @importFrom fabletools stream
#' @export
stream.model_smean <- function(object, new_data, specials, ...){
  # Extract a vector of response data
  mv <- tsibble::measured_vars(new_data)
  y <- new_data[[mv]]
  
  # Compute the new seasonal averages
  m <- length(object$coef)
  season_id <- (seq(0, length(y) - 1) + object$n %% m) %% m
  season_y <- split(y, season_id)
  season_avg <- vapply(season_y, FUN = mean, FUN.VALUE = numeric(1L), 
                       USE.NAMES = FALSE)
  weight_new <- vapply(season_y, FUN = length, FUN.VALUE = integer(1L),
                       USE.NAMES = FALSE)
  
  # Update coefficients to include new estimates
  weight_orig <- rep(object$n %/% m, m) + c(rep(1, object$n %% m), rep(0, m - object$n %% m))
  new_coef <- (object$coef * weight_orig + season_avg * weight_new) / (weight_orig + weight_new)
  coef_change <- new_coef - object$coef
  
  # Update model
  new_fits <- new_coef[season_id+1]
  new_e <- y - new_fits
  object$coef <- new_coef
  object$fitted <- c(object$fitted + rep_len(coef_change, object$n), new_fits)
  object$residuals <- c(object$residuals - rep_len(coef_change, object$n), new_e)
  object$n <- object$n + length(y)
  object$sigma2 <- var(object$residuals, na.rm = TRUE)
  
  # Return updated model object
  object
}

us_acc_deaths <- as_tsibble(USAccDeaths)
fit_stream <- us_acc_deaths %>% 
  dplyr::slice(1:60) %>% 
  model(SMEAN(value))
report(fit_stream)

# Update the model with new data
us_acc_deaths_new <- us_acc_deaths %>% dplyr::slice(61:72)
fit_stream <- fit_stream %>% 
  stream(us_acc_deaths_new)
report(fit_stream)

# Check that it matches a model of the full data
us_acc_deaths %>% 
  model(SMEAN(value)) %>% 
  report()
```

```{r}
#' @importFrom fabletools fitted
#' @export
fitted.model_smean <- function(object, ...){
  object$fitted
}

fitted(fit)
```

```{r}
#' @importFrom fabletools residuals
#' @export
residuals.model_smean <- function(object, ...){
  object$residuals
}

residuals(fit)
```

```{r, eval=FALSE}
#' @importFrom fabletools components
#' @export
components.model_smean <- function(object, ...){
  # Create a tsibble of the components
  dcmp <- tibble::tibble(
    !!object$y_name := fitted(object) + residuals(object),
    season = fitted(object),
    remainder = residuals(object)
  )
  
  # Describe how the components combine into other columns
  aliases <- tibble::lst(!!object$y_name := quote(season + remainder))
  
  # Define the behaviour of seasonal components
  # This is used for automatic modelling of seasonal components in `decomposition_model()`
  # It may also be used for plotting in the future.
  seasonalities <- list(season = list(period = length(object$coef)))
  
  # Return a dable
  as_dable(
    dcmp,
    resp = !!sym(object$y_name), method = model_sum(object),
    seasons = seasonalities, aliases = aliases
  )
}

components(fit) # Need to store index somewhere. This workflow should improve.
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_decomposition.R
\name{decomposition_model}
\alias{decomposition_model}
\title{Decomposition modelling}
\usage{
decomposition_model(dcmp, ...)
}
\arguments{
\item{dcmp}{A model definition which supports extracting decomposed \code{\link[=components]{components()}}.}

\item{...}{Model definitions used to model the components}
}
\description{
This function allows you to specify a decomposition combination model using
any additive decomposition. It works by first decomposing the data using the
decomposition method provided to \code{dcmp_fn} with the given formula. Secondary
models are used to fit each of the components from the resulting
decomposition. These models are specified after the decomposition formula.
All non-seasonal decomposition components must be specified, and any
unspecified seasonal components will be forecasted using seasonal naive.
These component models will be combined according to the decomposition
method, giving a combination model for the response of the decomposition.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE) && requireNamespace("feasts", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(feasts)
library(tsibble)
library(dplyr)
library(ggtime)

vic_food <- tsibbledata::aus_retail \%>\% 
  filter(State == "Victoria", Industry == "Food retailing")
  
# Identify an appropriate decomposition
vic_food \%>\% 
  model(STL(log(Turnover) ~ season(window = Inf))) \%>\% 
  components() \%>\% 
  autoplot()
  
# Use an ETS model to seasonally adjusted data, and SNAIVE to season_year
# Any model can be used, and seasonal components will default to use SNAIVE.
my_dcmp_spec <- decomposition_model(
  STL(log(Turnover) ~ season(window = Inf)),
  ETS(season_adjust ~ season("N")), SNAIVE(season_year)
)

vic_food \%>\%
  model(my_dcmp_spec) \%>\% 
  forecast(h="5 years") \%>\% 
  autoplot(vic_food)
\dontshow{\}) # examplesIf}
}
\seealso{
\href{https://otexts.com/fpp3/forecasting-decomposition.html}{\emph{Forecasting: Principles and Practice} - Forecasting Decomposition}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{validate_formula}
\alias{validate_formula}
\title{Validate the user provided model}
\usage{
validate_formula(model, data = NULL)
}
\arguments{
\item{model}{A quosure for the user's model specification}

\item{data}{A dataset used for automatic response selection}
}
\description{
Appropriately format the user's model for evaluation. Typically ran as one of the first steps
in a model function.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{skill_score}
\alias{skill_score}
\title{Forecast skill score measure}
\usage{
skill_score(measure)
}
\arguments{
\item{measure}{The accuracy measure to use in computing the skill score.}
}
\description{
This function converts other error metrics such as \code{MSE} into a skill score.
The reference or benchmark forecasting method is the Naive method for
non-seasonal data, and the seasonal naive method for seasonal data.
When used within \code{\link{accuracy.fbl_ts}}, it is important that the data
contains both the training and test data, as the training data is used to
compute the benchmark forecasts.
}
\examples{

skill_score(MSE)

\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibble)

lung_deaths <- as_tsibble(cbind(mdeaths, fdeaths))
lung_deaths \%>\% 
  dplyr::filter(index < yearmonth("1979 Jan")) \%>\%
  model(
    ets = ETS(value ~ error("M") + trend("A") + season("A")),
    lm = TSLM(value ~ trend() + season())
  ) \%>\%
  forecast(h = "1 year") \%>\%
  accuracy(lung_deaths, measures = list(skill = skill_score(MSE)))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fable.R
\name{as_fable}
\alias{as_fable}
\alias{as_fable.tbl_ts}
\alias{as_fable.grouped_ts}
\alias{as_fable.tbl_df}
\alias{as_fable.fbl_ts}
\alias{as_fable.grouped_df}
\alias{as_fable.forecast}
\title{Coerce to a fable object}
\usage{
as_fable(x, ...)

\method{as_fable}{tbl_ts}(x, response, distribution, ...)

\method{as_fable}{grouped_ts}(x, response, distribution, ...)

\method{as_fable}{tbl_df}(x, response, distribution, ...)

\method{as_fable}{fbl_ts}(x, response, distribution, ...)

\method{as_fable}{grouped_df}(x, response, distribution, ...)

\method{as_fable}{forecast}(x, ..., point_forecast = list(.mean = mean))
}
\arguments{
\item{x}{Object to be coerced to a fable (\code{fbl_ts})}

\item{...}{Additional arguments passed to methods}

\item{response}{The character vector of response variable(s).}

\item{distribution}{The name of the distribution column (can be provided
using a bare expression).}

\item{point_forecast}{The point forecast measure(s) which should be returned
in the resulting fable. Specified as a named list of functions which accept
a distribution and return a vector. To compute forecast medians, you can use
\code{list(.median = median)}.}
}
\description{
Coerce to a fable object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_null.R
\name{null_model}
\alias{null_model}
\alias{is_null_model}
\title{NULL model}
\usage{
null_model(formula, ...)

is_null_model(x)
}
\arguments{
\item{formula}{Model specification (response variable)}

\item{...}{Unused}

\item{x}{The object to be tested.}
}
\description{
Create a NULL model definition. This model produces NA forecasts and does not
require any estimation of the data. It is generally used as a placeholder for
models which have encountered an error (see \code{.safely} in \code{\link[=model]{model()}}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fable.R
\name{fable}
\alias{fable}
\title{Create a fable object}
\usage{
fable(..., response, distribution)
}
\arguments{
\item{...}{Arguments passed to \code{\link[tsibble:tsibble]{tsibble::tsibble()}}.}

\item{response}{The character vector of response variable(s).}

\item{distribution}{The name of the distribution column (can be provided
using a bare expression).}
}
\description{
A fable (forecast table) data class (\code{fbl_ts}) which is a tsibble-like data
structure for representing forecasts. In extension to the key and index from
the tsibble (\code{tbl_ts}) class, a fable (\code{fbl_ts}) must also contain a single
distribution column that uses values from the distributional package.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecast.R
\name{scenarios}
\alias{scenarios}
\title{A set of future scenarios for forecasting}
\usage{
scenarios(..., names_to = ".scenario")
}
\arguments{
\item{...}{Input data for each scenario}

\item{names_to}{The column name used to identify each scenario}
}
\description{
A set of future scenarios for forecasting
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stream.R
\name{stream}
\alias{stream}
\alias{stream.mdl_df}
\title{Extend a fitted model with new data}
\usage{
stream(object, ...)

\method{stream}{mdl_df}(object, new_data, ...)
}
\arguments{
\item{object}{An object (such as a model) which can be extended with additional data.}

\item{...}{Additional arguments passed on to stream methods.}

\item{new_data}{A dataset of the same structure as was used to fit the model.}
}
\description{
Extend the length of data used to fit a model and update the parameters to
suit this new data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dable.R
\name{is_dable}
\alias{is_dable}
\title{Is the object a dable}
\usage{
is_dable(x)
}
\arguments{
\item{x}{An object.}
}
\description{
Is the object a dable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{register_feature}
\alias{register_feature}
\title{Register a feature function}
\usage{
register_feature(fn, tags)
}
\arguments{
\item{fn}{The feature function}

\item{tags}{Identifying tags}
}
\description{
Allows users to find and use features from your package using \code{\link[=feature_set]{feature_set()}}.
If the features are being registered from within a package, this feature
registration should happen at load time using \verb{[.onLoad()]}.
}
\examples{

\dontrun{
tukey_five <- function(x){
  setNames(fivenum(x), c("min", "hinge_lwr", "med", "hinge_upr", "max"))
}

register_feature(tukey_five, tags = c("boxplot", "simple"))

}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mable.R
\name{mable}
\alias{mable}
\title{Create a new mable}
\usage{
mable(..., key = NULL, model = NULL)
}
\arguments{
\item{...}{A set of name-value pairs.}

\item{key}{Structural variable(s) that identify each model.}

\item{model}{Identifiers for the columns containing model(s).}
}
\description{
A mable (model table) data class (\code{mdl_df}) is a tibble-like data structure
for applying multiple models to a dataset. Each row of the mable refers to a
different time series from the data (identified by the key columns). A mable
must contain at least one column of time series models (\code{mdl_ts}), where the
list column itself (\code{mdl_lst}) describes how these models are related.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{aggregate_key}
\alias{aggregate_key}
\title{Expand a dataset to include other levels of aggregation}
\usage{
aggregate_key(.data, .spec, ...)
}
\arguments{
\item{.data}{A tsibble.}

\item{.spec}{The specification of aggregation structure.}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Name-value pairs of
summary functions. The name will be the name of the variable in the result.

The value can be:
\itemize{
\item A vector of length 1, e.g. \code{min(x)}, \code{n()}, or \code{sum(is.na(y))}.
\item A data frame with 1 row, to add multiple columns from a single expression.
}}
}
\description{
Uses the structural specification given in \code{.spec} to aggregate a time
series. A grouped structure is specified using \code{grp1 * grp2}, and a nested
structure is specified via \code{parent / child}. Aggregating the key structure is
commonly used with forecast reconciliation to produce coherent forecasts over
some hierarchy.
}
\details{
This function is experimental, and is subject to change in the future.

The way in which the measured variables are aggregated is specified in a
similar way to how \verb{[dplyr::summarise()]} is used.
}
\examples{
library(tsibble)
tourism \%>\% 
  aggregate_key(Purpose * (State / Region), Trips = sum(Trips))

}
\seealso{
\code{\link[=reconcile]{reconcile()}}, \code{\link[=is_aggregated]{is_aggregated()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mable.R
\name{is_mable}
\alias{is_mable}
\title{Is the object a mable}
\usage{
is_mable(x)
}
\arguments{
\item{x}{An object.}
}
\description{
Is the object a mable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{response}
\alias{response}
\title{Extract the response variable from a model}
\usage{
response(object, ...)
}
\arguments{
\item{object}{The object containing response data}

\item{...}{Additional parameters passed on to other methods}
}
\description{
Returns a tsibble containing only the response variable used in the fitting
of a model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecast.R
\name{construct_fc}
\alias{construct_fc}
\title{Construct a new set of forecasts}
\usage{
construct_fc(point, sd, dist)
}
\arguments{
\item{point}{The transformed point forecasts}

\item{sd}{The standard deviation of the transformed forecasts}

\item{dist}{The forecast distribution (typically produced using \code{new_fcdist})}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#deprecated}{\figure{lifecycle-deprecated.svg}{options: alt='[Deprecated]'}}}{\strong{[Deprecated]}}

This function is deprecated. \code{forecast()} methods for a model should return
a vector of distributions using the distributional package.

Backtransformations are automatically handled, and so no transformations should be specified here.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/irf.R
\name{IRF}
\alias{IRF}
\title{Compute Impulse Response Function (IRF)}
\usage{
IRF(x, ...)
}
\arguments{
\item{x}{A fitted model object, such as from a VAR or ARIMA model. This model is used to compute the impulse response.}

\item{...}{Additional arguments to be passed to lower-level functions.}
}
\description{
This function calculates the impulse response function (IRF) of a time series model.
The IRF describes how a model's variables react to external shocks over time.
}
\details{
If \code{new_data} contains the \code{.impulse} column, those values will be
treated as impulses for the calculated impulse responses.

The impulse response function provides insight into the dynamic behaviour of a system in
response to external shocks. It traces the effect of a one-unit change in the impulse
variable on the response variable over a specified number of periods.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dable.R
\name{as_dable}
\alias{as_dable}
\alias{as_dable.tbl_df}
\alias{as_dable.tbl_ts}
\title{Coerce to a dable object}
\usage{
as_dable(x, ...)

\method{as_dable}{tbl_df}(x, response, method = NULL, seasons = list(), aliases = list(), ...)

\method{as_dable}{tbl_ts}(x, response, method = NULL, seasons = list(), aliases = list(), ...)
}
\arguments{
\item{x}{Object to be coerced to a dable (\code{dcmp_ts})}

\item{...}{Additional arguments passed to methods}

\item{response}{The character vector of response variable(s).}

\item{method}{The name of the decomposition method.}

\item{seasons}{A named list describing the structure of seasonal components
(such as \code{period}, and \code{base}).}

\item{aliases}{A named list of calls describing common aliases computed from
components.}
}
\description{
Coerce to a dable object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fable.R
\name{is_fable}
\alias{is_fable}
\title{Is the object a fable}
\usage{
is_fable(x)
}
\arguments{
\item{x}{An object.}
}
\description{
Is the object a fable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent.R
\name{coherent_smat}
\alias{coherent_smat}
\title{Structural (summing) matrix for coherent time series}
\usage{
coherent_smat(data, sparse = FALSE)
}
\arguments{
\item{data}{A data object which contains linearly related coherent structures.}

\item{sparse}{If \code{TRUE}, a sparse matrix (class \code{"dgCMatrix"} from the
\pkg{Matrix} package) is returned. Defaults to \code{FALSE}.}
}
\value{
An \eqn{n \times n_b} matrix (dense or sparse) encoding the
structural relationships among all series.
}
\description{
Constructs the \eqn{n \times n_b} structural matrix \eqn{\mathbf{S}} for a
set of linearly related time series, where \eqn{n} is the total number
of series and \eqn{n_b} is the number of bottom-level series.
}
\details{
Let \eqn{\boldsymbol{b}_t} be the \eqn{n_b}-vector of bottom-level series at
time \eqn{t}, and \eqn{\boldsymbol{a}_t = \mathbf{A}\boldsymbol{b}_t} be the
\eqn{n_a = n - n_b} aggregated series, where \eqn{\mathbf{A}} is the
\eqn{n_a \times n_b} aggregation matrix. The full \eqn{n}-vector of series is
\eqn{\boldsymbol{y}_t = [\boldsymbol{a}_t', \boldsymbol{b}_t']'}, and the
structural matrix satisfies

\deqn{\boldsymbol{y}_t = \mathbf{S}\boldsymbol{b}_t, \quad
  \mathbf{S} = \begin{bmatrix} \mathbf{A} \\ \mathbf{I}_{n_b} \end{bmatrix}.}

Any linear reconciliation method can be written as
\eqn{\tilde{\boldsymbol{y}}_h = \mathbf{S}\mathbf{G}\hat{\boldsymbol{y}}_h},
where \eqn{\hat{\boldsymbol{y}}_h} are the \eqn{h}-step base forecasts.
Optimal reconciled forecasts use
\eqn{\mathbf{G} = (\mathbf{S}'\mathbf{W}^{-1}\mathbf{S})^{-1}\mathbf{S}'\mathbf{W}^{-1}},
with \eqn{\mathbf{W}} a positive definite weight matrix (see
\code{\link[=min_trace]{min_trace()}}).
}
\examples{
tsibble::tourism \%>\% 
  aggregate_key(Purpose, Trips = sum(Trips)) \%>\%
  coherent_smat()

}
\references{
Hyndman, R. J., & Athanasopoulos, G. (2022). \emph{Notation for forecast
reconciliation}. \url{https://robjhyndman.com/hyndsight/reconciliation-notation.html}
}
\seealso{
\code{\link[=coherent_cmat]{coherent_cmat()}} for the corresponding zero-constraint matrix
\eqn{\mathbf{C}} and \code{\link[=aggregate_key]{aggregate_key()}} for computing cross-sectional
aggregations with tsibble data sets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{is_aggregated}
\alias{is_aggregated}
\title{Is the element an aggregation of smaller data}
\usage{
is_aggregated(x)
}
\arguments{
\item{x}{An object.}
}
\description{
Is the element an aggregation of smaller data
}
\seealso{
\code{\link{aggregate_key}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{percentile_score}
\alias{percentile_score}
\alias{quantile_score}
\alias{CRPS}
\alias{distribution_accuracy_measures}
\title{Distribution accuracy measures}
\usage{
percentile_score(.dist, .actual, na.rm = TRUE, ...)

quantile_score(
  .dist,
  .actual,
  probs = c(0.05, 0.25, 0.5, 0.75, 0.95),
  na.rm = TRUE,
  ...
)

CRPS(.dist, .actual, n_quantiles = 1000, na.rm = TRUE, ...)

distribution_accuracy_measures
}
\arguments{
\item{.dist}{The distribution of fitted values from the model, or forecasted values from the forecast.}

\item{.actual}{A vector of responses matching the fitted values
(for forecast accuracy, \code{new_data} must be provided).}

\item{na.rm}{Remove the missing values before calculating the accuracy measure}

\item{...}{Additional arguments for each measure.}

\item{probs}{A vector of probabilities at which the metric is evaluated.}

\item{n_quantiles}{The number of quantiles to use in approximating CRPS when an exact solution is not available.}
}
\description{
These accuracy measures can be used to evaluate how accurately a forecast
distribution predicts a given actual value.
}
\section{Quantile/percentile score (pinball loss)}{


A quantile (or percentile) score evaluates how accurately a set of quantiles
(or percentiles) from the distribution match the given actual value. This
score uses a pinball loss function, and can be calculated via the average of
the score function given below:

The score function \eqn{s_p(q_p,y)} is given by \eqn{(1-p)(q_p-y)} if
\eqn{y < q_p}, and \eqn{p(y-q_p)} if \eqn{y \ge q_p}. Where \eqn{p} is the
quantile probability, \eqn{q_p = F^{-1}(p)} is the quantile with probability
\eqn{p}, and \eqn{y} is the actual value.

The resulting accuracy measure will average this score over all predicted
points at all desired quantiles (defined via the \code{probs} argument).

The percentile score is uses the same method with \code{probs} set to all
percentiles \code{probs = seq(0.01, 0.99, 0.01)}.
}

\section{Continuous ranked probability score (CRPS)}{


The continuous ranked probability score (CRPS) is the continuous analogue of
the pinball loss quantile score defined above. Its value is twice the
integral of the quantile score over all possible quantiles:

\deqn{
  CRPS(F,y) = 2 \int_0^1 s_p(q_p,y) dp
}{
  CRPS(F,y) = 2 integral_0^1 s_p(q_p,y) dp
}

It can be computed directly from the distribution via:

\deqn{
  CRPS(F,y) = \int_{-\infty}^\infty (F(x) - 1{y\leq x})^2 dx
}{
  CRPS(F,y) = integral_{-\infty}^\infty (F(x) - 1{y\leq x})^2 dx
}

For some forecast distribution \eqn{F} and actual value \eqn{y}.

Calculating the CRPS accuracy measure is computationally difficult for many
distributions, however it can be computed quickly and exactly for Normal and
emperical (sample) distributions. For other distributions the CRPS is
approximated using the quantile score of many quantiles (using the number of
quantiles specified in the \code{n_quantiles} argument).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mable.R
\name{as_mable}
\alias{as_mable}
\alias{as_mable.data.frame}
\title{Coerce a dataset to a mable}
\usage{
as_mable(x, ...)

\method{as_mable}{data.frame}(x, key = NULL, model = NULL, ...)
}
\arguments{
\item{x}{A dataset containing a list model column.}

\item{...}{Additional arguments passed to other methods.}

\item{key}{Structural variable(s) that identify each model.}

\item{model}{Identifiers for the columns containing model(s).}
}
\description{
Coerce a dataset to a mable
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vctrs-dable.R
\name{dable-vctrs}
\alias{dable-vctrs}
\alias{vec_ptype2.dcmp_ts}
\alias{vec_cast.dcmp_ts}
\title{Internal vctrs methods}
\usage{
\method{vec_ptype2}{dcmp_ts}(x, y, ...)

\method{vec_cast}{dcmp_ts}(x, to, ...)
}
\description{
These methods are the extensions that allow dable objects to
work with vctrs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vctrs-fable.R
\name{fable-vctrs}
\alias{fable-vctrs}
\alias{vec_ptype2.fbl_ts}
\alias{vec_cast.fbl_ts}
\title{Internal vctrs methods}
\usage{
\method{vec_ptype2}{fbl_ts}(x, y, ...)

\method{vec_cast}{fbl_ts}(x, to, ...)
}
\description{
These methods are the extensions that allow fable objects to
work with vctrs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{winkler_score}
\alias{winkler_score}
\alias{pinball_loss}
\alias{scaled_pinball_loss}
\alias{interval_accuracy_measures}
\title{Interval estimate accuracy measures}
\usage{
winkler_score(.dist, .actual, level = 95, na.rm = TRUE, ...)

pinball_loss(.dist, .actual, level = 95, na.rm = TRUE, ...)

scaled_pinball_loss(
  .dist,
  .actual,
  .train,
  level = 95,
  na.rm = TRUE,
  demean = FALSE,
  .period,
  d = .period == 1,
  D = .period > 1,
  ...
)

interval_accuracy_measures
}
\arguments{
\item{.dist}{The distribution of fitted values from the model, or forecasted values from the forecast.}

\item{.actual}{A vector of responses matching the fitted values
(for forecast accuracy, \code{new_data} must be provided).}

\item{level}{The level of the forecast interval.}

\item{na.rm}{Remove the missing values before calculating the accuracy measure}

\item{...}{Additional arguments for each measure.}

\item{.train}{A vector of responses used to train the model
(for forecast accuracy, the \code{orig_data} must be provided).}

\item{demean}{Should the response be demeaned (MASE)}

\item{.period}{The seasonal period of the data (defaulting to 'smallest' seasonal period).
from a model, or forecasted values from the forecast.}

\item{d}{Should the response model include a first difference?}

\item{D}{Should the response model include a seasonal difference?}
}
\description{
Interval estimate accuracy measures
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{MDA}
\alias{MDA}
\alias{MDV}
\alias{MDPV}
\alias{directional_accuracy_measures}
\title{Directional accuracy measures}
\usage{
MDA(.resid, .actual, na.rm = TRUE, reward = 1, penalty = 0, ...)

MDV(.resid, .actual, na.rm = TRUE, ...)

MDPV(.resid, .actual, na.rm = TRUE, ...)

directional_accuracy_measures
}
\arguments{
\item{.resid}{A vector of residuals from either the training (model accuracy)
or test (forecast accuracy) data.}

\item{.actual}{A vector of responses matching the fitted values
(for forecast accuracy, \code{new_data} must be provided).}

\item{na.rm}{Remove the missing values before calculating the accuracy measure}

\item{reward, penalty}{The weights given to correct and incorrect predicted
directions.}

\item{...}{Additional arguments for each measure.}
}
\description{
A collection of accuracy measures based on the accuracy of the prediction's
direction (say, increasing or decreasing).
}
\details{
\code{MDA()}: Mean Directional Accuracy
\code{MDV()}: Mean Directional Value
\code{MDPV()}: Mean Directional Percentage Value
}
\references{
Blaskowitz and H. Herwartz (2011) "On economic evaluation of directional forecasts". \emph{International Journal of Forecasting},
\bold{27}(4), 1058-1065.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/traverse.R
\name{traverse}
\alias{traverse}
\title{Recursively traverse an object}
\usage{
traverse(
  x,
  .f = list,
  .g = identity,
  .h = identity,
  base = function(.x) is_syntactic_literal(.x) || is_symbol(.x)
)
}
\arguments{
\item{x}{The object to traverse}

\item{.f}{A function for combining the recursed components}

\item{.g}{A function applied to the object before recursion}

\item{.h}{A function applied to the base case}

\item{base}{The base case for the recursion}
}
\description{
Recursively traverse an object
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.dcmp_ts}
\alias{autoplot.dcmp_ts}
\title{Decomposition plots}
\usage{
\method{autoplot}{dcmp_ts}(object, .vars = NULL, scale_bars = TRUE, level = c(80, 95), ...)
}
\arguments{
\item{object}{A dable.}

\item{.vars}{The column of the dable used to plot. By default, this will be the response variable of the decomposition.}

\item{scale_bars}{If \code{TRUE}, each facet will include a scale bar which represents the same units across each facet.}

\item{level}{If the decomposition contains distributions, which levels should be used to display intervals?}

\item{...}{Further arguments passed to \code{\link[ggplot2:geom_line]{ggplot2::geom_line()}}, which can be used to specify fixed aesthetics such as \code{colour = "red"} or \code{linewidth = 3}.}
}
\description{
Produces a faceted plot of the components used to build the response
variable of the dable. Useful for visualising how the components contribute
in a decomposition or model.
}
\examples{
\dontshow{if (requireNamespace("feasts", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(feasts)
library(tsibbledata)
aus_production \%>\% 
  model(STL(Beer)) \%>\%
  components() \%>\%  
  autoplot()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dable.R
\name{dable}
\alias{dable}
\title{Create a dable object}
\usage{
dable(..., response, method = NULL, seasons = list(), aliases = list())
}
\arguments{
\item{...}{Arguments passed to \code{\link[tsibble:tsibble]{tsibble::tsibble()}}.}

\item{response}{The name of the response variable column.}

\item{method}{The name of the decomposition method.}

\item{seasons}{A named list describing the structure of seasonal components
(such as \code{period}, and \code{base}).}

\item{aliases}{A named list of calls describing common aliases computed from
components.}
}
\description{
A dable (decomposition table) data class (\code{dcmp_ts}) which is a tsibble-like
data structure for representing decompositions. This data class is useful for
representing decompositions, as its print method describes how its columns
can be combined to produce the original data, and has a more appropriate
\code{autoplot()} method for displaying decompositions provided by ggtime.
Beyond this, a dable (\code{dcmp_ts}) behaves very similarly to a tsibble (\code{tbl_ts}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{agg_vec}
\alias{agg_vec}
\title{Create an aggregation vector}
\usage{
agg_vec(x = character(), aggregated = logical(vec_size(x)))
}
\arguments{
\item{x}{The vector of values.}

\item{aggregated}{A logical vector to identify which values are \verb{<aggregated>}.}
}
\description{
\lifecycle{maturing}
}
\details{
An aggregation vector extends usual vectors by adding \verb{<aggregated>} values.
These vectors are typically produced via the \code{\link[=aggregate_key]{aggregate_key()}} function,
however it can be useful to create them manually to produce more complicated
hierarchies (such as unbalanced hierarchies).
}
\examples{
agg_vec(
  x = c(NA, "A", "B"),
  aggregated = c(TRUE, FALSE, FALSE)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/components.R
\name{components.mdl_df}
\alias{components.mdl_df}
\alias{components.mdl_ts}
\title{Extract components from a fitted model}
\usage{
\method{components}{mdl_df}(object, ...)

\method{components}{mdl_ts}(object, ...)
}
\arguments{
\item{object}{A mable.}

\item{...}{Other arguments passed to methods.}
}
\description{
Allows you to extract elements of interest from the model which can be
useful in understanding how they contribute towards the overall fitted values.
}
\details{
A dable will be returned, which will allow you to easily plot the components
and see the way in which components are combined to give forecasts.

The components can also be visualised using the \code{\link[=autoplot]{autoplot()}} method provided
by the ggtime package.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

# Forecasting with an ETS(M,Ad,A) model to Australian beer production
aus_production \%>\%
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) \%>\% 
  components()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/outliers.R
\name{outliers}
\alias{outliers}
\alias{outliers.mdl_df}
\alias{outliers.mdl_ts}
\title{Identify outliers}
\usage{
outliers(object, ...)

\method{outliers}{mdl_df}(object, ...)

\method{outliers}{mdl_ts}(object, ...)
}
\arguments{
\item{object}{An object which can identify outliers.}

\item{...}{Arguments for further methods.}
}
\description{
Return a table of outlying observations using a fitted model.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/coherent.R
\name{coherent_cmat}
\alias{coherent_cmat}
\title{Zero-constraint matrix for coherent time series}
\usage{
coherent_cmat(data, sparse = FALSE)
}
\arguments{
\item{data}{A data object which contains linearly related coherent structures.}

\item{sparse}{If \code{TRUE}, a sparse matrix (class \code{"dgCMatrix"} from the
\pkg{Matrix} package) is returned. Defaults to \code{FALSE}.}
}
\value{
An \eqn{n_a \times n} matrix (dense or sparse) whose rows encode
each aggregation constraint as \eqn{\mathbf{C}\boldsymbol{y}_t =
  \boldsymbol{0}_{n_a}}.
}
\description{
Constructs the \eqn{n_a \times n} zero-constraint matrix \eqn{\mathbf{C}}
for a set of linearly related time series, where \eqn{n_a = n - n_b}
is the number of aggregated series.
}
\details{
Given the aggregation matrix \eqn{\mathbf{A}} (see \code{\link[=coherent_smat]{coherent_smat()}}), the
constraint matrix is defined as

\deqn{\mathbf{C} = [\mathbf{I}_{n_a} \;\; {-\mathbf{A}}],}

so that \eqn{\mathbf{C}\boldsymbol{y}_t = \boldsymbol{0}_{n_a}} for all
coherent vectors \eqn{\boldsymbol{y}_t}. This zero-constrained representation
yields the mapping matrix

\deqn{\mathbf{M} = \mathbf{I}_n -
  \mathbf{W}\mathbf{C}'(\mathbf{C}\mathbf{W}\mathbf{C}')^{-1}\mathbf{C},}

which requires inverting an \eqn{n_a \times n_a} matrix rather than the
\eqn{n_b \times n_b} matrix in the structural form, and is therefore more
efficient when \eqn{n_a < n_b}.
}
\references{
Hyndman, R. J., & Athanasopoulos, G. (2022). \emph{Notation for forecast
reconciliation}. \url{https://robjhyndman.com/hyndsight/reconciliation-notation.html}

Di Fonzo, T., & Girolimetto, D. (2021). Cross-temporal forecast
reconciliation: Optimal combination method and heuristic alternatives.
\emph{International Journal of Forecasting}. \doi{10.1016/j.ijforecast.2021.08.004}
}
\seealso{
\code{\link[=coherent_smat]{coherent_smat()}} for the corresponding structural matrix
\eqn{\mathbf{S}} and \code{\link[=aggregate_key]{aggregate_key()}} for computing cross-sectional
aggregations with tsibble data sets.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/temporal_aggregation.R
\name{aggregate_index}
\alias{aggregate_index}
\title{Expand a dataset to include temporal aggregates}
\usage{
aggregate_index(.data, .window, ..., .offset = "end", .bin_size = NULL)
}
\arguments{
\item{.data}{A tsibble.}

\item{.window}{Temporal aggregations to include. The default (NULL) will
automatically identify appropriate temporal aggregations. This can be
specified in several ways (see details).}

\item{...}{<\code{\link[rlang:args_data_masking]{data-masking}}> Name-value pairs of
summary functions. The name will be the name of the variable in the result.

The value can be:
\itemize{
\item A vector of length 1, e.g. \code{min(x)}, \code{n()}, or \code{sum(is.na(y))}.
\item A data frame with 1 row, to add multiple columns from a single expression.
}}

\item{.offset}{Offset the temporal aggregation windows to align with the start
or end of the data. If FALSE, no offset will be applied (giving common
breakpoints for temporal bins.)}

\item{.bin_size}{Temporary. Define the number of observations in each temporal bucket}
}
\description{
\lifecycle{experimental}
}
\details{
This feature is very experimental. It currently allows for temporal
aggregation of daily data as a proof of concept.

The aggregation \code{.window} can be specified in several ways:
\itemize{
\item A character string, containing one of "day", "week", "month", "quarter" or
"year". This can optionally be preceded by a (positive or negative) integer
and a space, or followed by "s".
\item A number, taken to be in days.
\item A \code{\link{difftime}} object.
}
}
\examples{
library(tsibble)
pedestrian \%>\%
  # Currently only supports daily data
  index_by(Date) \%>\% 
  dplyr::summarise(Count = sum(Count)) \%>\% 
  # Compute weekly aggregates
  fabletools:::aggregate_index("1 week", Count = sum(Count))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{response_vars}
\alias{response_vars}
\title{Return response variables}
\usage{
response_vars(x)
}
\arguments{
\item{x}{A dataset containing a response variable (such as a mable, fable, or dable).}
}
\description{
\code{response_vars()} returns a character vector of the response variables in the
object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform.R
\name{bias_adjust}
\alias{bias_adjust}
\title{Bias adjust back-transformation functions}
\usage{
bias_adjust(bt, sd)
}
\arguments{
\item{bt}{The back-transformation function}

\item{sd}{The forecast standard deviation}
}
\description{
To produce forecast means (instead of forecast medians) it is necessary to adjust the back-transformation function relative to the forecast variance.
}
\details{
More details about bias adjustment can be found in the transformations vignette: read the vignette:
\code{vignette("transformations", package = "fable")}
}
\examples{

adj_fn <- bias_adjust(function(x) exp(x), 1:10)
y <- rnorm(10)
exp(y)
adj_fn(y)

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{features_by_tag}
\alias{features_by_tag}
\title{Features by tag}
\description{
This documentation lists all available in currently loaded packages. This is
a useful reference for making a \code{\link[=feature_set]{feature_set()}} from particular tag(s).
}
\details{
\Sexpr[stage=render,results=rd]{fabletools:::rd_features_tag()}
}
\seealso{
\link{features_by_pkg}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mdl_ts.R
\name{model_sum}
\alias{model_sum}
\title{Provide a succinct summary of a model}
\usage{
model_sum(x)
}
\arguments{
\item{x}{The model to summarise}
}
\description{
Similarly to pillar's type_sum and obj_sum, model_sum is used to provide brief model summaries.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconciliation.R
\name{min_trace}
\alias{min_trace}
\title{Minimum trace forecast reconciliation}
\usage{
min_trace(
  models,
  method = c("wls_var", "ols", "wls_struct", "mint_cov", "mint_shrink"),
  sparse = NULL
)
}
\arguments{
\item{models}{A column of models in a mable.}

\item{method}{The reconciliation method to use.}

\item{sparse}{If TRUE, the reconciliation will be computed using sparse
matrix algebra? By default, sparse matrices will be used if the MatrixM
package is installed.}
}
\description{
Reconciles a hierarchy using the minimum trace combination method. The
response variable of the hierarchy must be aggregated using sums. The
forecasted time points must match for all series in the hierarchy (caution:
this is not yet tested for beyond the series length).
}
\references{
Wickramasuriya, S. L., Athanasopoulos, G., & Hyndman, R. J. (2019). Optimal forecast reconciliation for hierarchical and grouped time series through trace minimization. Journal of the American Statistical Association, 1-45. https://doi.org/10.1080/01621459.2018.1448825
}
\seealso{
\code{\link[=reconcile]{reconcile()}}, \code{\link[=aggregate_key]{aggregate_key()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/box_cox.R
\name{box_cox}
\alias{box_cox}
\alias{inv_box_cox}
\title{Box Cox Transformation}
\usage{
box_cox(x, lambda)

inv_box_cox(x, lambda)
}
\arguments{
\item{x}{a numeric vector.}

\item{lambda}{a numeric value for the transformation parameter.}
}
\value{
a transformed numeric vector of the same length as x.
}
\description{
\code{box_cox()} returns a transformation of the input variable using a Box-Cox
transformation. \code{inv_box_cox()} reverses the transformation.
}
\details{
The Box-Cox transformation is given by \deqn{f_\lambda(x) =\frac{x^\lambda -
1}{\lambda}}{f(x;lambda)=(x^lambda - 1)/lambda} if \eqn{\lambda\ne0}{lambda
is not equal to 0}. For \eqn{\lambda=0}{lambda=0},
\deqn{f_0(x)=\log(x)}{f(x;0)=log(x)}.
}
\examples{
library(tsibble)
library(dplyr)
airmiles \%>\% 
  as_tsibble() \%>\% 
  mutate(box_cox = box_cox(value, lambda = 0.3))

}
\references{
Box, G. E. P. and Cox, D. R. (1964) An analysis of
transformations. \emph{JRSS B} \bold{26} 211--246.
}
\author{
Rob J Hyndman & Mitchell O'Hara-Wild
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/refit.R
\name{refit.mdl_df}
\alias{refit.mdl_df}
\alias{refit.mdl_ts}
\title{Refit a mable to a new dataset}
\usage{
\method{refit}{mdl_df}(object, new_data, ...)

\method{refit}{mdl_ts}(object, new_data, ...)
}
\arguments{
\item{object}{A mable.}

\item{new_data}{A tsibble dataset used to refit the model.}

\item{...}{Additional optional arguments for refit methods.}
}
\description{
Applies a fitted model to a new dataset. For most methods this can be done
with or without re-estimation of the parameters.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)

fit <- as_tsibble(mdeaths) \%>\% 
  model(ETS(value ~ error("M") + trend("A") + season("A")))
fit \%>\% report()

fit \%>\% 
  refit(as_tsibble(fdeaths)) \%>\% 
  report(reinitialise = TRUE)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/xregs.R
\name{common_xregs}
\alias{common_xregs}
\alias{trend}
\alias{season}
\alias{fourier}
\title{Common exogenous regressors}
\usage{
common_xregs
}
\description{
These special functions provide interfaces to more complicated functions within
the model formulae interface.
}
\section{Specials}{


\subsection{trend}{
The \code{trend} special includes common linear trend regressors in the model. It also supports piecewise linear trend via the \code{knots} argument.
\preformatted{
trend(knots = NULL, origin = NULL)
}

\tabular{ll}{
\code{knots}    \tab A vector of times (same class as the data's time index) identifying the position of knots for a piecewise linear trend.\cr
\code{origin}   \tab An optional time value to act as the starting time for the trend.
}
}

\subsection{season}{
The \code{season} special includes seasonal dummy variables in the model.
\preformatted{
season(period = NULL)
}

\tabular{ll}{
\code{period}   \tab The periodic nature of the seasonality. This can be either a number indicating the number of observations in each seasonal period, or text to indicate the duration of the seasonal window (for example, annual seasonality would be "1 year").
}
}

\subsection{fourier}{
The \code{fourier} special includes seasonal fourier terms in the model. The maximum order of the fourier terms must be specified using \code{K}.
\preformatted{
fourier(period = NULL, K, origin = NULL)
}

\tabular{ll}{
\code{period}   \tab The periodic nature of the seasonality. This can be either a number indicating the number of observations in each seasonal period, or text to indicate the duration of the seasonal window (for example, annual seasonality would be "1 year"). \cr
\code{K}        \tab The maximum order of the fourier terms.\cr
\code{origin}   \tab An optional time value to act as the starting time for the fourier series.
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/residuals.R
\name{residuals.mdl_df}
\alias{residuals.mdl_df}
\alias{residuals.mdl_ts}
\title{Extract residuals values from models}
\usage{
\method{residuals}{mdl_df}(object, ...)

\method{residuals}{mdl_ts}(object, type = "innovation", ...)
}
\arguments{
\item{object}{A mable or time series model.}

\item{...}{Other arguments passed to the model method for \code{residuals()}}

\item{type}{The type of residuals to compute. If \code{type="response"}, residuals on the back-transformed data will be computed.}
}
\description{
Extracts the residuals from each of the models in a mable. A tsibble will
be returned containing these residuals.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{distribution_var}
\alias{distribution_var}
\title{Return distribution variable}
\usage{
distribution_var(x)
}
\arguments{
\item{x}{A dataset containing a distribution variable (such as a fable).}
}
\description{
\code{distribution_var()} returns a character vector of the distribution variable
in the data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_combination.R
\name{combination_model}
\alias{combination_model}
\title{Combination modelling}
\usage{
combination_model(..., cmbn_fn = combination_ensemble, cmbn_args = list())
}
\arguments{
\item{...}{Model definitions used in the combination.}

\item{cmbn_fn}{A function used to produce the combination.}

\item{cmbn_args}{Additional arguments passed to \code{cmbn_fn}.}
}
\description{
Combines multiple model definitions (passed via \code{...}) to produce a model
combination definition using some combination function (\code{cmbn_fn}). Currently
distributional forecasts are only supported for models producing normally
distributed forecasts.
}
\details{
A combination model can also be produced using mathematical operations.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibble)
library(tsibbledata)

# cmbn1 and cmbn2 are equivalent and equally weighted.
aus_production \%>\%
  model(
    cmbn1 = combination_model(SNAIVE(Beer), TSLM(Beer ~ trend() + season())),
    cmbn2 = (SNAIVE(Beer) + TSLM(Beer ~ trend() + season()))/2
  )

# An inverse variance weighted ensemble.
aus_production \%>\%
  model(
    cmbn1 = combination_model(
      SNAIVE(Beer), TSLM(Beer ~ trend() + season()), 
      cmbn_args = list(weights = "inv_var")
    )
  )
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform.R
\name{new_transformation}
\alias{new_transformation}
\alias{invert_transformation}
\title{Create a new modelling transformation}
\usage{
new_transformation(transformation, inverse)

invert_transformation(x, ...)
}
\arguments{
\item{transformation}{A function which transforms the data}

\item{inverse}{A function which is the inverse of a transformation}

\item{x}{A transformation (such as one created with \code{new_transformation}).}

\item{...}{Further arguments passed to other methods.}
}
\description{
Produces a new transformation for fable modelling functions which will be used to transform, back-transform, and adjust forecasts.
}
\details{
For more details about transformations, read the vignette:
\code{vignette("transformations", package = "fable")}
}
\examples{

scaled_logit <- function(x, lower=0, upper=1){
  log((x-lower)/(upper-x))
}
inv_scaled_logit <- function(x, lower=0, upper=1){
  (upper-lower)*exp(x)/(1+exp(x)) + lower
}
my_scaled_logit <- new_transformation(scaled_logit, inv_scaled_logit)

t_vals <- my_scaled_logit(1:10, 0, 100)
t_vals

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{feature_set}
\alias{feature_set}
\title{Create a feature set from tags}
\usage{
feature_set(pkgs = NULL, tags = NULL)
}
\arguments{
\item{pkgs}{The package(s) from which to search for features. If \code{NULL},
all registered features from currently loaded packages will be searched.}

\item{tags}{Tags used to identify similar groups of features. If \code{NULL},
all tags will be included.}
}
\description{
Construct a feature set from features available in currently loaded packages.
Lists of available features can be found in the following pages:
\itemize{
\item \link[=features_by_pkg]{Features by package}
\item \link[=features_by_tag]{Features by tag}
}
}
\section{Registering features}{

Features can be registered for use with the \code{feature_set()} function using
\code{\link[=register_feature]{register_feature()}}. This function allows you to register a feature along
with the tags associated with it. If the features are being registered from
within a package, this feature registration should happen at load time using
\verb{[.onLoad()]}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconciliation.R
\name{reconcile}
\alias{reconcile}
\alias{reconcile.mdl_df}
\title{Forecast reconciliation}
\usage{
reconcile(.data, ...)

\method{reconcile}{mdl_df}(.data, ...)
}
\arguments{
\item{.data}{A mable.}

\item{...}{Reconciliation methods applied to model columns within \code{.data}.}
}
\description{
This function allows you to specify the method used to reconcile forecasts
in accordance with its key structure.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
lung_deaths_agg <- as_tsibble(cbind(mdeaths, fdeaths)) \%>\%
  aggregate_key(key, value = sum(value))

lung_deaths_agg \%>\%
  model(lm = TSLM(value ~ trend() + season())) \%>\%
  reconcile(lm = min_trace(lm)) \%>\% 
  forecast()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specials.R
\name{new_specials}
\alias{new_specials}
\title{Create evaluation environment for specials}
\usage{
new_specials(..., .required_specials = NULL, .xreg_specials = NULL)
}
\arguments{
\item{...}{A named set of functions which used to parse formula inputs}

\item{.required_specials}{The names of specials which must be provided (and if not, are included with no inputs).}

\item{.xreg_specials}{The names of specials which will be only used as inputs to other specials (most commonly \code{xreg}).}
}
\description{
Allows extension packages to make use of the formula parsing of specials.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/features.R
\name{features}
\alias{features}
\alias{features_at}
\alias{features_all}
\alias{features_if}
\title{Extract features from a dataset}
\usage{
features(.tbl, .var, features, ...)

features_at(.tbl, .vars, features, ...)

features_all(.tbl, features, ...)

features_if(.tbl, .predicate, features, ...)
}
\arguments{
\item{.tbl}{A dataset}

\item{.var}{An expression that produces a vector from which the features are computed.}

\item{features}{A list of functions (or lambda expressions) for the features to compute. \code{\link[=feature_set]{feature_set()}} is a useful helper for building sets of features.}

\item{...}{Additional arguments to be passed to each feature. These arguments will only be passed to features which use it in their formal arguments (\code{\link[base:formals]{base::formals()}}), and not via their \code{...}. While passing \code{na.rm = TRUE} to \code{\link[stats:var]{stats::var()}} will work, it will not for \code{\link[base:mean]{base::mean()}} as its formals are \code{x} and \code{...}. To more precisely pass inputs to each function, you should use lambdas in the list of features (\code{~ mean(., na.rm = TRUE)}).}

\item{.vars}{A tidyselect compatible selection of the column(s) to compute features on.}

\item{.predicate}{A predicate function (or lambda expression) to be applied to the columns or a logical vector. The variables for which .predicate is or returns TRUE are selected.}
}
\description{
Create scalar valued summary features for a dataset from feature functions.
}
\details{
Lists of available features can be found in the following pages:
\itemize{
\item \link[=features_by_pkg]{Features by package}
\item \link[=features_by_tag]{Features by tag}
}
}
\examples{
# Provide a set of functions as a named list to features.
library(tsibble)
tourism \%>\% 
  features(Trips, features = list(mean = mean, sd = sd))

# Search and use useful features with `feature_set()`. 

\dontshow{if (requireNamespace("feasts", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(feasts)
\dontshow{\}) # examplesIf}
tourism \%>\% 
  features(Trips, features = feature_set(tags = "autocorrelation"))

# Best practice is to use anonymous functions for additional arguments
tourism \%>\% 
  features(Trips, list(~ quantile(., probs=seq(0,1,by=0.2))))


}
\seealso{
\code{\link[=feature_set]{feature_set()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{model}
\alias{model}
\alias{model.tbl_ts}
\title{Estimate models}
\usage{
model(.data, ...)

\method{model}{tbl_ts}(.data, ..., .safely = TRUE)
}
\arguments{
\item{.data}{A data structure suitable for the models (such as a \code{tsibble})}

\item{...}{Definitions for the models to be used. All models must share the
same response variable.}

\item{.safely}{If a model encounters an error, rather than aborting the process a \link[=null_model]{NULL model} will be returned instead. This allows for an error to occur when computing many models, without losing the results of the successful models.}
}
\description{
Trains specified model definition(s) to a dataset. This function will
estimate the a set of model definitions (passed via \code{...}) to each series
within \code{.data} (as identified by the key structure). The result will be a
mable (a model table), which neatly stores the estimated models in a tabular
structure. Rows of the data identify different series within the data, and
each model column contains all models from that model definition. Each cell
in the mable identifies a single model.
}
\section{Parallel}{


It is possible to estimate models in parallel using the
\href{https://cran.r-project.org/package=future}{future} package. By specifying a
\code{\link[future:plan]{future::plan()}} before estimating the models, they will be computed
according to that plan.
}

\section{Progress}{


Progress on model estimation can be obtained by wrapping the code with
\code{progressr::with_progress()}. Further customisation on how progress is
reported can be controlled using the \code{progressr} package.
}

\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

# Training an ETS(M,Ad,A) model to Australian beer production
aus_production \%>\%
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A")))

# Training a seasonal naive and ETS(A,A,A) model to the monthly 
# "Food retailing" turnover for selected Australian states.
library(dplyr)
progressr::with_progress(
aus_retail \%>\% 
  filter(
    Industry == "Food retailing",
    State \%in\% c("Victoria", "New South Wales", "Queensland")
  ) \%>\% 
  model(
    snaive = SNAIVE(Turnover),
    ets = ETS(log(Turnover) ~ error("A") + trend("A") + season("A")),
  )
)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconciliation.R
\name{middle_out}
\alias{middle_out}
\title{Middle out forecast reconciliation}
\usage{
middle_out(models, split = 1)
}
\arguments{
\item{models}{A column of models in a mable.}

\item{split}{The middle level of the hierarchy from which the bottom-up and
top-down approaches are used above and below respectively.}
}
\description{
\lifecycle{experimental}
}
\details{
Reconciles a hierarchy using the middle out reconciliation method. The
response variable of the hierarchy must be aggregated using sums. The
forecasted time points must match for all series in the hierarchy.
}
\seealso{
\code{\link[=reconcile]{reconcile()}}, \code{\link[=aggregate_key]{aggregate_key()}}
\href{https://otexts.com/fpp3/single-level.html#middle-out-approach}{\emph{Forecasting: Principles and Practice} - Middle-out approach}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forecast.R
\name{forecast.mdl_df}
\alias{forecast.mdl_df}
\alias{forecast.mdl_ts}
\title{Produce forecasts}
\usage{
\method{forecast}{mdl_df}(
  object,
  new_data = NULL,
  h = NULL,
  point_forecast = list(.mean = mean),
  ...
)

\method{forecast}{mdl_ts}(
  object,
  new_data = NULL,
  h = NULL,
  bias_adjust = NULL,
  simulate = FALSE,
  bootstrap = FALSE,
  times = 5000,
  point_forecast = list(.mean = mean),
  ...
)
}
\arguments{
\item{object}{The time series model used to produce the forecasts}

\item{new_data}{A \code{tsibble} containing future information used to forecast.}

\item{h}{The forecast horison (can be used instead of \code{new_data} for regular
time series with no exogenous regressors).}

\item{point_forecast}{The point forecast measure(s) which should be returned
in the resulting fable. Specified as a named list of functions which accept
a distribution and return a vector. To compute forecast medians, you can use
\code{list(.median = median)}.}

\item{...}{Additional arguments for forecast model methods.}

\item{bias_adjust}{Deprecated. Please use \code{point_forecast} to specify the
desired point forecast method.}

\item{simulate}{Should forecasts be based on simulated future paths instead
of analytical results.}

\item{bootstrap}{Should innovations from simulated forecasts be bootstrapped
from the model's fitted residuals. This allows the forecast distribution to
have a different underlying shape which could better represent the nature
of your data.}

\item{times}{The number of future paths for simulations if \code{simulate = TRUE}.}
}
\value{
A fable containing the following columns:
\itemize{
\item \code{.model}: The name of the model used to obtain the forecast. Taken from
the column names of models in the provided mable.
\item The forecast distribution. The name of this column will be the same as the
dependent variable in the model(s). If multiple dependent variables exist,
it will be named \code{.distribution}.
\item Point forecasts computed from the distribution using the functions in the
\code{point_forecast} argument.
\item All columns in \code{new_data}, excluding those whose names conflict with the
above.
}
}
\description{
The forecast function allows you to produce future predictions of a time series
from fitted models. If the response variable has been transformed in the
model formula, the transformation will be automatically back-transformed
(and bias adjusted if \code{bias_adjust} is \code{TRUE}). More details about
transformations in the fable framework can be found in
\code{vignette("transformations", package = "fable")}.
}
\details{
The forecasts returned contain both point forecasts and their distribution.
A specific forecast interval can be extracted from the distribution using the
\code{\link[=hilo]{hilo()}} function, and multiple intervals can be obtained using \code{\link[=report]{report()}}.
These intervals are stored in a single column using the \code{hilo} class, to
extract the numerical upper and lower bounds you can use \code{\link[=unpack_hilo]{unpack_hilo()}}.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE) && requireNamespace("urca", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibble)
library(tsibbledata)
library(dplyr)
library(tidyr)

# Forecasting with an ETS(M,Ad,A) model to Australian beer production
beer_fc <- aus_production \%>\%
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) \%>\% 
  forecast(h = "3 years")

# Compute 80\% and 95\% forecast intervals
beer_fc \%>\% 
  hilo(level = c(80, 95))

beer_fc \%>\% 
  autoplot(aus_production)

# Forecasting with a seasonal naive and linear model to the monthly 
# "Food retailing" turnover for each Australian state/territory.
library(dplyr)
aus_retail \%>\% 
  filter(Industry == "Food retailing") \%>\% 
  model(
    snaive = SNAIVE(Turnover),
    ets = TSLM(log(Turnover) ~ trend() + season()),
  ) \%>\% 
  forecast(h = "2 years 6 months") \%>\% 
  autoplot(filter(aus_retail, Month >= yearmonth("2000 Jan")), level = 90)
  
# Forecast GDP with a dynamic regression model on log(GDP) using population and
# an automatically chosen ARIMA error structure. Assume that population is fixed
# in the future.
aus_economy <- global_economy \%>\% 
  filter(Country == "Australia")
fit <- aus_economy \%>\% 
  model(lm = ARIMA(log(GDP) ~ Population))

future_aus <- new_data(aus_economy, n = 10) \%>\% 
  mutate(Population = last(aus_economy$Population))

fit \%>\% 
  forecast(new_data = future_aus) \%>\% 
  autoplot(aus_economy)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hypothesise.R
\name{hypothesize.mdl_df}
\alias{hypothesize.mdl_df}
\alias{hypothesize.mdl_ts}
\title{Run a hypothesis test from a mable}
\usage{
\method{hypothesize}{mdl_df}(x, ...)

\method{hypothesize}{mdl_ts}(x, tests = list(), ...)
}
\arguments{
\item{x}{A mable.}

\item{...}{Arguments for model methods.}

\item{tests}{a list of test functions to perform on the model}
}
\description{
This function will return the results of a hypothesis test for each model in
the mable.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

olympic_running \%>\%
  model(lm = TSLM(log(Time) ~ trend())) \%>\% 
  hypothesize()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.tbl_ts}
\alias{autoplot.tbl_ts}
\alias{autolayer.tbl_ts}
\title{Plot time series from a tsibble}
\usage{
\method{autoplot}{tbl_ts}(object, .vars = NULL, ...)

\method{autolayer}{tbl_ts}(object, .vars = NULL, ...)
}
\arguments{
\item{object}{A tsibble.}

\item{.vars}{A bare expression containing data you wish to plot. Multiple variables can be plotted using \code{\link[ggplot2:vars]{ggplot2::vars()}}.}

\item{...}{Further arguments passed to \code{\link[ggplot2:geom_line]{ggplot2::geom_line()}}, which can be used to specify fixed aesthetics such as \code{colour = "red"} or \code{linewidth = 3}.}
}
\description{
Produces a time series plot of one or more variables from a tsibble. If the
tsibble contains a multiple keys, separate time series will be identified by
colour.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)
library(tsibble)

tsibbledata::gafa_stock \%>\%
 autoplot(vars(Close, log(Close)))
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hilo.R
\name{unpack_hilo}
\alias{unpack_hilo}
\title{Unpack a hilo column}
\usage{
unpack_hilo(data, cols, names_sep = "_", names_repair = "check_unique")
}
\arguments{
\item{data}{A data frame.}

\item{cols}{Name of hilo columns to unpack.}

\item{names_sep}{If \code{NULL}, the default, the names will be left
as is. In \code{pack()}, inner names will come from the former outer names;
in \code{unpack()}, the new outer names will come from the inner names.

If a string, the inner and outer names will be used together. In
\code{unpack()}, the names of the new outer columns will be formed by pasting
together the outer and the inner column names, separated by \code{names_sep}. In
\code{pack()}, the new inner names will have the outer names + \code{names_sep}
automatically stripped. This makes \code{names_sep} roughly symmetric between
packing and unpacking.}

\item{names_repair}{Used to check that output data frame has valid
names. Must be one of the following options:
\itemize{
\item \verb{"minimal}": no name repair or checks, beyond basic existence,
\item \verb{"unique}": make sure names are unique and not empty,
\item \verb{"check_unique}": (the default), no name repair, but check they are unique,
\item \verb{"universal}": make the names unique and syntactic
\item a function: apply custom name repair.
\item \link[tidyr]{tidyr_legacy}: use the name repair from tidyr 0.8.
\item a formula: a purrr-style anonymous function (see \code{\link[rlang:as_function]{rlang::as_function()}})
}

See \code{\link[vctrs:vec_as_names]{vctrs::vec_as_names()}} for more details on these terms and the
strategies used to enforce them.}
}
\description{
\ifelse{html}{\href{https://lifecycle.r-lib.org/articles/stages.html#superseded}{\figure{lifecycle-superseded.svg}{options: alt='[Superseded]'}}}{\strong{[Superseded]}}

This function is superceded. It is recommended that you use the functionality
from the \href{https://pkg.mitchelloharawild.com/distributional/}{distributional}
package to extract elements from a \verb{<hilo>} object. For example, you can access
the lower bound with \verb{<hilo>$lower}.

Allows a hilo column to be unpacked into its component columns: "lower",
"upper", and "level".
}
\seealso{
\code{\link[tidyr:pack]{tidyr::unpack()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/broom.R
\name{glance.mdl_df}
\alias{glance.mdl_df}
\alias{glance.mdl_lst}
\alias{glance.mdl_ts}
\title{Glance a mable}
\usage{
\method{glance}{mdl_df}(x, ...)

\method{glance}{mdl_lst}(x, ...)

\method{glance}{mdl_ts}(x, ...)
}
\arguments{
\item{x}{A mable.}

\item{...}{Arguments for model methods.}
}
\description{
Uses the models within a mable to produce a one row summary of their fits.
This typically contains information about the residual variance,
information criterion, and other relevant summary statistics. Each model
will be represented with a row of output.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

olympic_running \%>\%
  model(lm = TSLM(log(Time) ~ trend())) \%>\% 
  glance()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/estimate.R
\name{estimate}
\alias{estimate}
\alias{estimate.tbl_ts}
\title{Estimate a model}
\usage{
estimate(.data, ...)

\method{estimate}{tbl_ts}(.data, .model, ...)
}
\arguments{
\item{.data}{A data structure suitable for the models (such as a \code{tsibble}).}

\item{...}{Further arguments passed to methods.}

\item{.model}{Definition for the model to be used.}
}
\description{
Estimate a model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/definitions.R
\name{new_model_class}
\alias{new_model_class}
\alias{new_model_definition}
\title{Create a new class of models}
\usage{
new_model_class(
  model = "Unknown model",
  train = function(.data, formula, specials, ...)
    abort("This model has not defined a training method."),
  specials = new_specials(),
  check = function(.data) {
 },
  prepare = function(...) {
 },
  ...,
  .env = caller_env(),
  .inherit = model_definition
)

new_model_definition(.class, formula, ..., .env = caller_env(n = 2))
}
\arguments{
\item{model}{The name of the model}

\item{train}{A function that trains the model to a dataset. \code{.data} is a tsibble
containing the data's index and response variables only. \code{formula} is the
user's provided formula. \code{specials} is the evaluated specials used in the formula.}

\item{specials}{Special functions produced using \code{\link[=new_specials]{new_specials()}}}

\item{check}{A function that is used to check the data for suitability with
the model. This can be used to check for missing values (both implicit and
explicit), regularity of observations, ordered time index, and univariate
responses.}

\item{prepare}{This allows you to modify the model class according to user
inputs. \code{...} is the arguments passed to \code{new_model_definition}, allowing
you to perform different checks or training procedures according to different
user inputs.}

\item{...}{Further arguments to \code{\link[R6:R6Class]{R6::R6Class()}}. This can be useful to set up
additional elements used in the other functions. For example, to use
\code{common_xregs}, an \code{origin} element in the model is used to store
the origin for \code{trend()} and \code{fourier()} specials. To use these specials, you
must add an \code{origin} element to the object (say with \code{origin = NULL}).}

\item{.env}{The environment from which functions should inherit from.}

\item{.inherit}{A model class to inherit from.}

\item{.class}{A model class (typically created with \code{\link[=new_model_class]{new_model_class()}}).}

\item{formula}{The user's model formula.}
}
\description{
Suitable for extension packages to create new models for fable.
}
\details{
This function produces a new R6 model definition. An understanding of R6 is
not required, however could be useful to provide more sophisticated model
interfaces. All functions have access to \code{self}, allowing the functions for
training the model and evaluating specials to access the model class itself.
This can be useful to obtain elements set in the \%TODO
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{accuracy.mdl_df}
\alias{accuracy.mdl_df}
\alias{accuracy.mdl_ts}
\alias{accuracy.fbl_ts}
\title{Evaluate accuracy of a forecast or model}
\usage{
\method{accuracy}{mdl_df}(object, measures = point_accuracy_measures, ...)

\method{accuracy}{mdl_ts}(object, measures = point_accuracy_measures, ...)

\method{accuracy}{fbl_ts}(object, data, measures = point_accuracy_measures, ..., by = NULL)
}
\arguments{
\item{object}{A model or forecast object}

\item{measures}{A list of accuracy measure functions to compute (such as \code{\link{point_accuracy_measures}}, \code{\link{interval_accuracy_measures}}, or \code{\link{distribution_accuracy_measures}})}

\item{...}{Additional arguments to be passed to measures that use it.}

\item{data}{A dataset containing the complete model dataset (both training and test data). The training portion of the data will be used in the computation of some accuracy measures, and the test data is used to compute the forecast errors.}

\item{by}{Variables over which the accuracy is computed (useful for computing across forecast horizons in cross-validation). If \code{by} is \code{NULL}, groups will be chosen automatically from the key structure.}
}
\description{
Summarise the performance of the model using accuracy measures. Accuracy
measures can be computed directly from models as the one-step-ahead fitted
residuals are available. When evaluating accuracy on forecasts, you will
need to provide a complete dataset that includes the future data and data
used to train the model.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibble)
library(tsibbledata)
library(dplyr)

fit <- aus_production \%>\%
  filter(Quarter < yearquarter("2006 Q1")) \%>\% 
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A")))

# In-sample training accuracy does not require extra data provided.
accuracy(fit)

# Out-of-sample forecast accuracy requires the future values to compare with.
# All available future data will be used, and a warning will be given if some
# data for the forecast window is unavailable.
fc <- fit \%>\% 
  forecast(h = "5 years")
fc \%>\% 
  accuracy(aus_production)
  
# It is also possible to compute interval and distributional measures of
# accuracy for models and forecasts which give forecast distributions.
fc \%>\% 
  accuracy(
    aus_production,
    measures = list(interval_accuracy_measures, distribution_accuracy_measures)
  )
\dontshow{\}) # examplesIf}
}
\seealso{
\href{https://otexts.com/fpp3/accuracy.html}{Evaluating forecast accuracy}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{parse_model}
\alias{parse_model}
\title{Parse the model specification for specials}
\usage{
parse_model(model)
}
\arguments{
\item{model}{A model definition}
}
\description{
Using a list of defined special functions, the user's formula specification and data
is parsed to extract important modelling components.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fabletools-package.R
\docType{package}
\name{fabletools-package}
\alias{fabletools}
\alias{fabletools-package}
\title{fabletools: Core Tools for Packages in the 'fable' Framework}
\description{
Provides tools, helpers and data structures for developing models and time series functions for 'fable' and extension packages. These tools support a consistent and tidy interface for time series modelling and analysis.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://fabletools.tidyverts.org/}
  \item \url{https://github.com/tidyverts/fabletools}
  \item Report bugs at \url{https://github.com/tidyverts/fabletools/issues}
}

}
\author{
\strong{Maintainer}: Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})

Authors:
\itemize{
  \item Mitchell O'Hara-Wild \email{mail@mitchelloharawild.com} (\href{https://orcid.org/0000-0001-6729-7695}{ORCID})
  \item Rob Hyndman
  \item Earo Wang (\href{https://orcid.org/0000-0001-6448-5260}{ORCID})
}

Other contributors:
\itemize{
  \item Di Cook [contributor]
  \item George Athanasopoulos [contributor]
  \item David Holt [contributor]
}

}
\keyword{package}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_combination.R
\name{combination_ensemble}
\alias{combination_ensemble}
\title{Ensemble combination}
\usage{
combination_ensemble(..., weights = c("equal", "inv_var"))
}
\arguments{
\item{...}{Estimated models used in the ensemble.}

\item{weights}{The method used to weight each model in the ensemble.}
}
\description{
Ensemble combination
}
\seealso{
\code{\link[=combination_weighted]{combination_weighted()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{ME}
\alias{ME}
\alias{MSE}
\alias{RMSE}
\alias{MAE}
\alias{MPE}
\alias{MAPE}
\alias{MASE}
\alias{RMSSE}
\alias{ACF1}
\alias{point_accuracy_measures}
\title{Point estimate accuracy measures}
\usage{
ME(.resid, na.rm = TRUE, ...)

MSE(.resid, na.rm = TRUE, ...)

RMSE(.resid, na.rm = TRUE, ...)

MAE(.resid, na.rm = TRUE, ...)

MPE(.resid, .actual, na.rm = TRUE, ...)

MAPE(.resid, .actual, na.rm = TRUE, ...)

MASE(
  .resid,
  .train,
  demean = FALSE,
  na.rm = TRUE,
  .period,
  d = .period == 1,
  D = .period > 1,
  ...
)

RMSSE(
  .resid,
  .train,
  demean = FALSE,
  na.rm = TRUE,
  .period,
  d = .period == 1,
  D = .period > 1,
  ...
)

ACF1(.resid, na.action = stats::na.pass, demean = TRUE, ...)

point_accuracy_measures
}
\arguments{
\item{.resid}{A vector of residuals from either the training (model accuracy)
or test (forecast accuracy) data.}

\item{na.rm}{Remove the missing values before calculating the accuracy measure}

\item{...}{Additional arguments for each measure.}

\item{.actual}{A vector of responses matching the fitted values
(for forecast accuracy, \code{new_data} must be provided).}

\item{.train}{A vector of responses used to train the model
(for forecast accuracy, the \code{orig_data} must be provided).}

\item{demean}{Should the response be demeaned (MASE)}

\item{.period}{The seasonal period of the data (defaulting to 'smallest' seasonal period).
from a model, or forecasted values from the forecast.}

\item{d}{Should the response model include a first difference?}

\item{D}{Should the response model include a seasonal difference?}

\item{na.action}{Function to handle missing values.}
}
\description{
Point estimate accuracy measures
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/parse.R
\name{parse_model_lhs}
\alias{parse_model_lhs}
\title{Parse the RHS of the model formula for transformations}
\usage{
parse_model_lhs(model)
}
\arguments{
\item{model}{A model definition}
}
\description{
Parse the RHS of the model formula for transformations
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate.R
\name{aggregation-vctrs}
\alias{aggregation-vctrs}
\alias{vec_ptype2.agg_vec}
\alias{vec_ptype2.agg_vec.agg_vec}
\alias{vec_ptype2.agg_vec.default}
\alias{vec_ptype2.agg_vec.character}
\alias{vec_ptype2.character.agg_vec}
\alias{vec_ptype_abbr.agg_vec}
\alias{vec_cast.agg_vec}
\alias{vec_cast.agg_vec.agg_vec}
\alias{vec_cast.agg_vec.default}
\alias{vec_cast.character.agg_vec}
\alias{vec_proxy_compare.agg_vec}
\title{Internal vctrs methods}
\usage{
\method{vec_ptype2}{agg_vec}(x, y, ...)

\method{vec_ptype2}{agg_vec.agg_vec}(x, y, ...)

\method{vec_ptype2}{agg_vec.default}(x, y, ...)

\method{vec_ptype2}{agg_vec.character}(x, y, ...)

\method{vec_ptype2}{character.agg_vec}(x, y, ...)

\method{vec_ptype_abbr}{agg_vec}(x, ...)

\method{vec_cast}{agg_vec}(x, to, ...)

\method{vec_cast}{agg_vec.agg_vec}(x, to, ...)

\method{vec_cast}{agg_vec.default}(x, to, ...)

\method{vec_cast}{character.agg_vec}(x, to, ...)

\method{vec_proxy_compare}{agg_vec}(x, ...)
}
\description{
These methods are the extensions that allow aggregation vectors to work with
vctrs.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/specials.R
\name{special_xreg}
\alias{special_xreg}
\title{Helper special for producing a model matrix of exogenous regressors}
\usage{
special_xreg(...)
}
\arguments{
\item{...}{Arguments for \code{fable_xreg_matrix} (see Details)}
}
\description{
Helper special for producing a model matrix of exogenous regressors
}
\details{
Currently the \code{fable_xreg_matrix} helper supports a single argument named
\code{default_intercept}. If this argument is TRUE (passed via \code{...} above), then
the intercept will be returned in the matrix if not specified (much like the
behaviour of \code{lm()}). If FALSE, then the intercept will only be included if
explicitly requested via \code{1} in the formula.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reconciliation.R
\name{bottom_up}
\alias{bottom_up}
\title{Bottom up forecast reconciliation}
\usage{
bottom_up(models)
}
\arguments{
\item{models}{A column of models in a mable.}
}
\description{
\lifecycle{experimental}
}
\details{
Reconciles a hierarchy using the bottom up reconciliation method. The
response variable of the hierarchy must be aggregated using sums. The
forecasted time points must match for all series in the hierarchy.
}
\seealso{
\code{\link[=reconcile]{reconcile()}}, \code{\link[=aggregate_key]{aggregate_key()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/broom.R
\name{tidy.mdl_df}
\alias{tidy.mdl_df}
\alias{coef.mdl_df}
\alias{coef.mdl_lst}
\alias{coef.mdl_ts}
\alias{tidy.mdl_lst}
\alias{tidy.mdl_ts}
\title{Extract model coefficients from a mable}
\usage{
\method{tidy}{mdl_df}(x, ...)

\method{coef}{mdl_df}(object, ...)

\method{coef}{mdl_lst}(object, ...)

\method{coef}{mdl_ts}(object, ...)

\method{tidy}{mdl_lst}(x, ...)

\method{tidy}{mdl_ts}(x, ...)
}
\arguments{
\item{x, object}{A mable.}

\item{...}{Arguments for model methods.}
}
\description{
This function will obtain the coefficients (and associated statistics) for
each model in the mable.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

olympic_running \%>\%
  model(lm = TSLM(log(Time) ~ trend())) \%>\% 
  tidy()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/interpolate.R
\name{interpolate.mdl_df}
\alias{interpolate.mdl_df}
\alias{interpolate.mdl_ts}
\title{Interpolate missing values}
\usage{
\method{interpolate}{mdl_df}(object, new_data, ...)

\method{interpolate}{mdl_ts}(object, new_data, ...)
}
\arguments{
\item{object}{A mable containing a single model column.}

\item{new_data}{A dataset with the same structure as the data used to fit the model.}

\item{...}{Other arguments passed to interpolate methods.}
}
\description{
Uses a fitted model to interpolate missing values from a dataset.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

# The fastest running times for the olympics are missing for years during 
# world wars as the olympics were not held.
olympic_running

olympic_running \%>\% 
  model(TSLM(Time ~ trend())) \%>\% 
  interpolate(olympic_running)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/broom.R
\name{augment.mdl_df}
\alias{augment.mdl_df}
\alias{augment.mdl_lst}
\alias{augment.mdl_ts}
\title{Augment a mable}
\usage{
\method{augment}{mdl_df}(x, ...)

\method{augment}{mdl_lst}(x, ...)

\method{augment}{mdl_ts}(x, type = NULL, ...)
}
\arguments{
\item{x}{A mable.}

\item{...}{Arguments for model methods.}

\item{type}{Deprecated.}
}
\description{
Uses a fitted model to augment the response variable with fitted values and
residuals. Response residuals (back-transformed) are stored in the \code{.resid}
column, while innovation residuals (transformed) are stored in the \code{.innov}
column.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

# Forecasting with an ETS(M,Ad,A) model to Australian beer production
aus_production \%>\%
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) \%>\% 
  augment()
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model.R
\name{model_rhs}
\alias{model_rhs}
\title{Extract the right hand side of a model}
\usage{
model_rhs(model)
}
\arguments{
\item{model}{A formula}
}
\description{
Extract the right hand side of a model
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fitted.R
\name{fitted.mdl_df}
\alias{fitted.mdl_df}
\alias{hfitted}
\alias{fitted.mdl_ts}
\title{Extract fitted values from models}
\usage{
\method{fitted}{mdl_df}(object, ...)

\method{fitted}{mdl_ts}(object, h = 1, ...)
}
\arguments{
\item{object}{A mable or time series model.}

\item{...}{Other arguments passed to the model method for \code{fitted()}}

\item{h}{The number of steps ahead that these fitted values are computed from.}
}
\description{
Extracts the fitted values from each of the models in a mable. A tsibble will
be returned containing these fitted values. Fitted values will be
automatically back-transformed if a transformation was specified.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/model_combination.R
\name{combination_weighted}
\alias{combination_weighted}
\title{Weighted combination}
\usage{
combination_weighted(..., weights = NULL)
}
\arguments{
\item{...}{Estimated models used in the ensemble.}

\item{weights}{The numeric weights applied to each model in \code{...}}
}
\description{
Weighted combination
}
\seealso{
\code{\link[=combination_ensemble]{combination_ensemble()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accessors.R
\name{mable_vars}
\alias{mable_vars}
\title{Return model column variables}
\usage{
mable_vars(x)
}
\arguments{
\item{x}{A dataset containing models (such as a mable).}
}
\description{
\code{mable_vars()} returns a character vector of the model variables in the
object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{autoplot.fbl_ts}
\alias{autoplot.fbl_ts}
\alias{autolayer.fbl_ts}
\title{Plot a set of forecasts}
\usage{
\method{autoplot}{fbl_ts}(object, data = NULL, level = c(80, 95), show_gap = TRUE, ...)

\method{autolayer}{fbl_ts}(
  object,
  data = NULL,
  level = c(80, 95),
  point_forecast = list(mean = mean),
  show_gap = TRUE,
  ...
)
}
\arguments{
\item{object}{A fable.}

\item{data}{A tsibble with the same key structure as the fable.}

\item{level}{The confidence level(s) for the plotted intervals.}

\item{show_gap}{Setting this to \code{FALSE} will connect the most recent value in \code{data} with the forecasts.}

\item{...}{Further arguments passed used to specify fixed aesthetics for the forecasts such as \code{colour = "red"} or \code{linewidth = 3}.}

\item{point_forecast}{The point forecast measure to be displayed in the plot.}
}
\description{
Produces a forecast plot from a fable. As the original data is not included
in the fable object, it will need to be specified via the \code{data} argument.
The \code{data} argument can be used to specify a shorter period of data, which is
useful to focus on the more recent observations.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE) && requireNamespace("tsibbledata", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(tsibbledata)

fc <- aus_production \%>\%
  model(ets = ETS(log(Beer) ~ error("M") + trend("Ad") + season("A"))) \%>\% 
  forecast(h = "3 years") 

fc \%>\% 
  autoplot(aus_production)
\dontshow{\}) # examplesIf}
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
aus_production \%>\% 
  autoplot(Beer) + 
  autolayer(fc)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reexports.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{\%>\%}
\alias{as_tsibble}
\alias{vars}
\alias{autoplot}
\alias{autolayer}
\alias{accuracy}
\alias{equation}
\alias{interpolate}
\alias{components}
\alias{augment}
\alias{glance}
\alias{tidy}
\alias{hypothesize}
\alias{generate}
\alias{refit}
\alias{forecast}
\alias{hilo}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{distributional}{\code{\link[distributional:hilo]{hilo()}}}

  \item{dplyr}{\code{\link[dplyr:\%>\%]{\%>\%}}, \code{\link[dplyr:vars]{vars()}}}

  \item{generics}{\code{\link[generics:accuracy]{accuracy()}}, \code{\link[generics:augment]{augment()}}, \code{\link[generics:components]{components()}}, \code{\link[generics:equation]{equation()}}, \code{\link[generics:forecast]{forecast()}}, \code{\link[generics:generate]{generate()}}, \code{\link[generics:glance]{glance()}}, \code{\link[generics:hypothesize]{hypothesize()}}, \code{\link[generics:interpolate]{interpolate()}}, \code{\link[generics:refit]{refit()}}, \code{\link[generics:tidy]{tidy()}}}

  \item{ggplot2}{\code{\link[ggplot2:autolayer]{autolayer()}}, \code{\link[ggplot2:autoplot]{autoplot()}}}

  \item{tsibble}{\code{\link[tsibble:as_tsibble]{as_tsibble()}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/report.R
\name{report}
\alias{report}
\title{Report information about an object}
\usage{
report(object, ...)
}
\arguments{
\item{object}{The object to report}

\item{...}{Additional options for the reporting function}
}
\description{
Displays the object in a suitable format for reporting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/frequency.R
\name{common_periods}
\alias{common_periods}
\alias{common_periods.default}
\alias{common_periods.tbl_ts}
\alias{common_periods.interval}
\alias{get_frequencies}
\alias{get_frequencies.numeric}
\alias{get_frequencies.NULL}
\alias{get_frequencies.character}
\alias{get_frequencies.Period}
\title{Extract frequencies for common seasonal periods}
\usage{
common_periods(x)

\method{common_periods}{default}(x)

\method{common_periods}{tbl_ts}(x)

\method{common_periods}{interval}(x)

get_frequencies(period, ...)

\method{get_frequencies}{numeric}(period, ...)

\method{get_frequencies}{`NULL`}(period, data, ..., .auto = c("smallest", "largest", "all"))

\method{get_frequencies}{character}(period, data, ...)

\method{get_frequencies}{Period}(period, data, ...)
}
\arguments{
\item{x}{An object containing temporal data (such as a \code{tsibble}, \code{interval}, \code{datetime} and others.)}

\item{period}{Specification of the time-series period}

\item{...}{Other arguments to be passed on to methods}

\item{data}{A tsibble}

\item{.auto}{The method used to automatically select the appropriate seasonal
periods}
}
\value{
A named vector of frequencies appropriate for the provided data.
}
\description{
Extract frequencies for common seasonal periods
}
\examples{
common_periods(tsibble::pedestrian)

}
\references{
\url{https://robjhyndman.com/hyndsight/seasonal-periods/}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/accuracy.R
\name{MAAPE}
\alias{MAAPE}
\title{Mean Arctangent Absolute Percentage Error}
\usage{
MAAPE(.resid, .actual, na.rm = TRUE, ...)
}
\arguments{
\item{.resid}{A vector of residuals from either the training (model accuracy)
or test (forecast accuracy) data.}

\item{.actual}{A vector of responses matching the fitted values
(for forecast accuracy, \code{new_data} must be provided).}

\item{na.rm}{Remove the missing values before calculating the accuracy measure}

\item{...}{Additional arguments for each measure.}
}
\description{
Mean Arctangent Absolute Percentage Error
}
\references{
Kim, Sungil and Heeyoung Kim (2016) "A new metric of absolute percentage error
for intermittent demand forecasts". \emph{International Journal of Forecasting},
\bold{32}(3), 669-679.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate.R
\name{generate.mdl_df}
\alias{generate.mdl_df}
\alias{generate.mdl_ts}
\title{Generate responses from a mable}
\usage{
\method{generate}{mdl_df}(x, new_data = NULL, h = NULL, times = 1, seed = NULL, ...)

\method{generate}{mdl_ts}(
  x,
  new_data = NULL,
  h = NULL,
  times = 1,
  seed = NULL,
  bootstrap = FALSE,
  bootstrap_block_size = 1,
  ...
)
}
\arguments{
\item{x}{A mable.}

\item{new_data}{The data to be generated (time index and exogenous regressors)}

\item{h}{The simulation horizon (can be used instead of \code{new_data} for regular
time series with no exogenous regressors).}

\item{times}{The number of replications.}

\item{seed}{The seed for the random generation from distributions.}

\item{...}{Additional arguments for individual simulation methods.}

\item{bootstrap}{If TRUE, then forecast distributions are computed using simulation with resampled errors.}

\item{bootstrap_block_size}{The bootstrap block size specifies the number of contiguous residuals to be taken in each bootstrap sample.}
}
\description{
Use a model's fitted distribution to simulate additional data with similar
behaviour to the response. This is a tidy implementation of
\code{\link[stats:simulate]{stats::simulate()}}.
}
\details{
Innovations are sampled by the model's assumed error distribution.
If \code{bootstrap} is \code{TRUE}, innovations will be sampled from the model's
residuals. If \code{new_data} contains the \code{.innov} column, those values will be
treated as innovations for the simulated paths.
}
\examples{
\dontshow{if (requireNamespace("fable", quietly = TRUE)) withAutoprint(\{ # examplesIf}
library(fable)
library(dplyr)
UKLungDeaths <- as_tsibble(cbind(mdeaths, fdeaths), pivot_longer = FALSE)
UKLungDeaths \%>\% 
  model(lm = TSLM(mdeaths ~ fourier("year", K = 4) + fdeaths)) \%>\% 
  generate(UKLungDeaths, times = 5)
\dontshow{\}) # examplesIf}
}
